
/*  Copyright (C) 2011 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <arpa/inet.h>
#include <config.h>
#include <fcntl.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>
#include <math.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include "zscanner/scanner.h"
#include "zscanner/functions.h"
#include "libknot/descriptor.h"

/*! \brief Mmap block size in bytes. This value is then adjusted to the
 *         multiple of memory pages which fit in.
 */
#define BLOCK_SIZE		30000000

/*! \brief Last artificial block which ensures final newline character. */
#define NEWLINE_BLOCK		"\n"

/*! \brief Shorthand for setting warning data. */
#define WARN(code) { s->error_code = code; }
/*! \brief Shorthand for setting error data. */
#define ERR(code) { s->error_code = code; s->stop = true; }

/*!
 * \brief Empty function which is called if no callback function is specified.
 */
static inline void noop(zs_scanner_t *s)
{
	(void)s;
}

/*!
 * \brief Writes record type number to r_data.
 *
 * \param type		Type number.
 * \param rdata_tail	Position where to write type number to.
 */
static inline void type_num(const uint16_t type, uint8_t **rdata_tail)
{
	*((uint16_t *)*rdata_tail) = htons(type);
	*rdata_tail += 2;
}

/*!
 * \brief Sets bit to bitmap window.
 *
 * \param type		Type number.
 * \param s		Scanner context.
 */
static inline void window_add_bit(const uint16_t type, zs_scanner_t *s) {
	uint8_t win      = type / 256;
	uint8_t bit_pos  = type % 256;
	uint8_t byte_pos = bit_pos / 8;

	((s->windows[win]).bitmap)[byte_pos] |= 128 >> (bit_pos % 8);

	if ((s->windows[win]).length < byte_pos + 1) {
		(s->windows[win]).length = byte_pos + 1;
	}

	if (s->last_window < win) {
		s->last_window = win;
	}
}

// Include scanner file (in Ragel).

static const int zone_scanner_start = 1077;
static const int zone_scanner_error = 0;





zs_scanner_t* zs_scanner_create(const char     *origin,
                                const uint16_t rclass,
                                const uint32_t ttl,
                                void (*process_record)(zs_scanner_t *),
                                void (*process_error)(zs_scanner_t *),
                                void *data)
{
	char settings[1024];

	if (origin == NULL) {
		return NULL;
	}

	zs_scanner_t *s = calloc(1, sizeof(zs_scanner_t));
	if (s == NULL) {
		return NULL;
	}

	// Nonzero initial scanner state.
	s->cs = zone_scanner_start;

	// Disable processing during parsing of settings.
	s->process_record = &noop;
	s->process_error = &noop;

	// Create ORIGIN directive and parse it using scanner to set up origin.
	const char *format;
	size_t origin_len = strlen(origin);
	if (origin_len > 0 && origin[origin_len - 1] != '.') {
		format = "$ORIGIN %s.\n";
	} else {
		format = "$ORIGIN %s\n";
	}
	int ret = snprintf(settings, sizeof(settings), format, origin);
	if (ret <= 0 || (size_t)ret >= sizeof(settings) ||
	    zs_scanner_parse(s, settings, settings + ret, true) != 0) {
		zs_scanner_free(s);
		return NULL;
	}

	// Set scanner defaults.
	s->default_class = rclass;
	s->default_ttl = ttl;
	s->process_record = process_record ? process_record : &noop;
	s->process_error = process_error ? process_error : &noop;
	s->data = data;
	s->path = strdup(".");
	s->line_counter = 1;

	return s;
}

void zs_scanner_free(zs_scanner_t *s)
{
	if (s != NULL) {
		free(s->path);
		free(s);
	}
}

static void parse_block(zs_scanner_t *s,
                        const char   *start,
                        const char   *end,
                        const bool   is_eof)
{
	// Necessary scanner variables.
	const char *p = start;
	const char *pe = end;
	char       *eof = NULL;
	int        stack[RAGEL_STACK_SIZE];

	// Auxiliary variables which are used in scanner body.
	struct in_addr  addr4;
	struct in6_addr addr6;
	uint32_t timestamp;
	int16_t  window;
	int      ret;

	// Next 2 variables are for better performance.
	// Restoring r_data pointer to next free space.
	uint8_t *rdata_tail = s->r_data + s->r_data_tail;
	// Initialization of the last r_data byte.
	uint8_t *rdata_stop = s->r_data + MAX_RDATA_LENGTH - 1;

	// Restoring scanner states.
	int cs  = s->cs;
	int top = s->top;
	memcpy(stack, s->stack, sizeof(stack));

	// End of file check.
	if (is_eof) {
		eof = (char *)pe;
	}

	// Writing scanner body (in C).
	
	{
	short _widec;
	if ( p == pe )
		goto _test_eof;
	goto _resume;

_again:
	switch ( cs ) {
		case 1077: goto st1077;
		case 0: goto st0;
		case 1: goto st1;
		case 2: goto st2;
		case 3: goto st3;
		case 4: goto st4;
		case 5: goto st5;
		case 6: goto st6;
		case 7: goto st7;
		case 8: goto st8;
		case 9: goto st9;
		case 10: goto st10;
		case 11: goto st11;
		case 12: goto st12;
		case 1078: goto st1078;
		case 13: goto st13;
		case 14: goto st14;
		case 15: goto st15;
		case 16: goto st16;
		case 17: goto st17;
		case 18: goto st18;
		case 19: goto st19;
		case 20: goto st20;
		case 21: goto st21;
		case 22: goto st22;
		case 23: goto st23;
		case 24: goto st24;
		case 25: goto st25;
		case 26: goto st26;
		case 27: goto st27;
		case 28: goto st28;
		case 29: goto st29;
		case 30: goto st30;
		case 31: goto st31;
		case 32: goto st32;
		case 33: goto st33;
		case 34: goto st34;
		case 35: goto st35;
		case 36: goto st36;
		case 37: goto st37;
		case 38: goto st38;
		case 39: goto st39;
		case 40: goto st40;
		case 41: goto st41;
		case 42: goto st42;
		case 43: goto st43;
		case 44: goto st44;
		case 45: goto st45;
		case 46: goto st46;
		case 47: goto st47;
		case 48: goto st48;
		case 49: goto st49;
		case 50: goto st50;
		case 51: goto st51;
		case 52: goto st52;
		case 53: goto st53;
		case 54: goto st54;
		case 55: goto st55;
		case 56: goto st56;
		case 57: goto st57;
		case 58: goto st58;
		case 59: goto st59;
		case 60: goto st60;
		case 61: goto st61;
		case 62: goto st62;
		case 63: goto st63;
		case 64: goto st64;
		case 65: goto st65;
		case 66: goto st66;
		case 67: goto st67;
		case 68: goto st68;
		case 69: goto st69;
		case 70: goto st70;
		case 71: goto st71;
		case 72: goto st72;
		case 73: goto st73;
		case 74: goto st74;
		case 75: goto st75;
		case 76: goto st76;
		case 77: goto st77;
		case 78: goto st78;
		case 79: goto st79;
		case 80: goto st80;
		case 81: goto st81;
		case 82: goto st82;
		case 83: goto st83;
		case 84: goto st84;
		case 85: goto st85;
		case 86: goto st86;
		case 87: goto st87;
		case 88: goto st88;
		case 89: goto st89;
		case 90: goto st90;
		case 91: goto st91;
		case 92: goto st92;
		case 93: goto st93;
		case 94: goto st94;
		case 95: goto st95;
		case 96: goto st96;
		case 97: goto st97;
		case 98: goto st98;
		case 99: goto st99;
		case 100: goto st100;
		case 101: goto st101;
		case 102: goto st102;
		case 103: goto st103;
		case 104: goto st104;
		case 105: goto st105;
		case 106: goto st106;
		case 107: goto st107;
		case 108: goto st108;
		case 109: goto st109;
		case 110: goto st110;
		case 111: goto st111;
		case 112: goto st112;
		case 113: goto st113;
		case 114: goto st114;
		case 115: goto st115;
		case 116: goto st116;
		case 117: goto st117;
		case 118: goto st118;
		case 119: goto st119;
		case 120: goto st120;
		case 121: goto st121;
		case 122: goto st122;
		case 123: goto st123;
		case 124: goto st124;
		case 125: goto st125;
		case 126: goto st126;
		case 127: goto st127;
		case 128: goto st128;
		case 129: goto st129;
		case 130: goto st130;
		case 131: goto st131;
		case 132: goto st132;
		case 133: goto st133;
		case 134: goto st134;
		case 1079: goto st1079;
		case 135: goto st135;
		case 136: goto st136;
		case 137: goto st137;
		case 138: goto st138;
		case 139: goto st139;
		case 140: goto st140;
		case 141: goto st141;
		case 142: goto st142;
		case 143: goto st143;
		case 144: goto st144;
		case 1080: goto st1080;
		case 145: goto st145;
		case 146: goto st146;
		case 147: goto st147;
		case 148: goto st148;
		case 149: goto st149;
		case 150: goto st150;
		case 151: goto st151;
		case 152: goto st152;
		case 1081: goto st1081;
		case 153: goto st153;
		case 154: goto st154;
		case 155: goto st155;
		case 1082: goto st1082;
		case 156: goto st156;
		case 157: goto st157;
		case 158: goto st158;
		case 159: goto st159;
		case 160: goto st160;
		case 161: goto st161;
		case 162: goto st162;
		case 163: goto st163;
		case 164: goto st164;
		case 165: goto st165;
		case 166: goto st166;
		case 1083: goto st1083;
		case 167: goto st167;
		case 168: goto st168;
		case 169: goto st169;
		case 170: goto st170;
		case 1084: goto st1084;
		case 171: goto st171;
		case 172: goto st172;
		case 173: goto st173;
		case 174: goto st174;
		case 175: goto st175;
		case 176: goto st176;
		case 177: goto st177;
		case 178: goto st178;
		case 179: goto st179;
		case 180: goto st180;
		case 181: goto st181;
		case 182: goto st182;
		case 183: goto st183;
		case 184: goto st184;
		case 185: goto st185;
		case 1085: goto st1085;
		case 186: goto st186;
		case 187: goto st187;
		case 188: goto st188;
		case 189: goto st189;
		case 190: goto st190;
		case 191: goto st191;
		case 192: goto st192;
		case 193: goto st193;
		case 194: goto st194;
		case 195: goto st195;
		case 196: goto st196;
		case 197: goto st197;
		case 198: goto st198;
		case 199: goto st199;
		case 200: goto st200;
		case 201: goto st201;
		case 1086: goto st1086;
		case 202: goto st202;
		case 203: goto st203;
		case 204: goto st204;
		case 205: goto st205;
		case 206: goto st206;
		case 207: goto st207;
		case 208: goto st208;
		case 209: goto st209;
		case 210: goto st210;
		case 211: goto st211;
		case 212: goto st212;
		case 213: goto st213;
		case 214: goto st214;
		case 215: goto st215;
		case 216: goto st216;
		case 217: goto st217;
		case 218: goto st218;
		case 219: goto st219;
		case 220: goto st220;
		case 221: goto st221;
		case 222: goto st222;
		case 223: goto st223;
		case 224: goto st224;
		case 225: goto st225;
		case 226: goto st226;
		case 227: goto st227;
		case 228: goto st228;
		case 229: goto st229;
		case 230: goto st230;
		case 231: goto st231;
		case 232: goto st232;
		case 233: goto st233;
		case 234: goto st234;
		case 235: goto st235;
		case 236: goto st236;
		case 237: goto st237;
		case 238: goto st238;
		case 239: goto st239;
		case 240: goto st240;
		case 241: goto st241;
		case 242: goto st242;
		case 243: goto st243;
		case 244: goto st244;
		case 245: goto st245;
		case 246: goto st246;
		case 247: goto st247;
		case 248: goto st248;
		case 249: goto st249;
		case 250: goto st250;
		case 251: goto st251;
		case 252: goto st252;
		case 253: goto st253;
		case 254: goto st254;
		case 1087: goto st1087;
		case 255: goto st255;
		case 256: goto st256;
		case 1088: goto st1088;
		case 257: goto st257;
		case 258: goto st258;
		case 259: goto st259;
		case 260: goto st260;
		case 261: goto st261;
		case 262: goto st262;
		case 263: goto st263;
		case 264: goto st264;
		case 265: goto st265;
		case 1089: goto st1089;
		case 1090: goto st1090;
		case 266: goto st266;
		case 267: goto st267;
		case 268: goto st268;
		case 269: goto st269;
		case 270: goto st270;
		case 271: goto st271;
		case 272: goto st272;
		case 273: goto st273;
		case 274: goto st274;
		case 275: goto st275;
		case 276: goto st276;
		case 277: goto st277;
		case 278: goto st278;
		case 279: goto st279;
		case 1091: goto st1091;
		case 280: goto st280;
		case 281: goto st281;
		case 282: goto st282;
		case 283: goto st283;
		case 284: goto st284;
		case 285: goto st285;
		case 286: goto st286;
		case 287: goto st287;
		case 288: goto st288;
		case 289: goto st289;
		case 1092: goto st1092;
		case 290: goto st290;
		case 291: goto st291;
		case 292: goto st292;
		case 293: goto st293;
		case 294: goto st294;
		case 295: goto st295;
		case 296: goto st296;
		case 297: goto st297;
		case 298: goto st298;
		case 299: goto st299;
		case 300: goto st300;
		case 301: goto st301;
		case 302: goto st302;
		case 303: goto st303;
		case 1093: goto st1093;
		case 304: goto st304;
		case 305: goto st305;
		case 306: goto st306;
		case 307: goto st307;
		case 308: goto st308;
		case 309: goto st309;
		case 310: goto st310;
		case 1094: goto st1094;
		case 311: goto st311;
		case 312: goto st312;
		case 313: goto st313;
		case 314: goto st314;
		case 315: goto st315;
		case 316: goto st316;
		case 317: goto st317;
		case 318: goto st318;
		case 319: goto st319;
		case 1095: goto st1095;
		case 1096: goto st1096;
		case 1097: goto st1097;
		case 320: goto st320;
		case 321: goto st321;
		case 322: goto st322;
		case 323: goto st323;
		case 324: goto st324;
		case 325: goto st325;
		case 326: goto st326;
		case 327: goto st327;
		case 1098: goto st1098;
		case 1099: goto st1099;
		case 328: goto st328;
		case 329: goto st329;
		case 330: goto st330;
		case 331: goto st331;
		case 332: goto st332;
		case 333: goto st333;
		case 334: goto st334;
		case 1100: goto st1100;
		case 335: goto st335;
		case 336: goto st336;
		case 337: goto st337;
		case 338: goto st338;
		case 339: goto st339;
		case 340: goto st340;
		case 341: goto st341;
		case 342: goto st342;
		case 343: goto st343;
		case 344: goto st344;
		case 345: goto st345;
		case 346: goto st346;
		case 347: goto st347;
		case 348: goto st348;
		case 349: goto st349;
		case 350: goto st350;
		case 351: goto st351;
		case 352: goto st352;
		case 353: goto st353;
		case 354: goto st354;
		case 355: goto st355;
		case 356: goto st356;
		case 357: goto st357;
		case 358: goto st358;
		case 359: goto st359;
		case 360: goto st360;
		case 361: goto st361;
		case 362: goto st362;
		case 363: goto st363;
		case 364: goto st364;
		case 365: goto st365;
		case 366: goto st366;
		case 367: goto st367;
		case 368: goto st368;
		case 369: goto st369;
		case 370: goto st370;
		case 371: goto st371;
		case 372: goto st372;
		case 373: goto st373;
		case 374: goto st374;
		case 375: goto st375;
		case 376: goto st376;
		case 377: goto st377;
		case 378: goto st378;
		case 379: goto st379;
		case 380: goto st380;
		case 381: goto st381;
		case 382: goto st382;
		case 383: goto st383;
		case 384: goto st384;
		case 385: goto st385;
		case 386: goto st386;
		case 387: goto st387;
		case 388: goto st388;
		case 389: goto st389;
		case 390: goto st390;
		case 391: goto st391;
		case 392: goto st392;
		case 393: goto st393;
		case 394: goto st394;
		case 395: goto st395;
		case 396: goto st396;
		case 397: goto st397;
		case 398: goto st398;
		case 399: goto st399;
		case 400: goto st400;
		case 401: goto st401;
		case 402: goto st402;
		case 403: goto st403;
		case 404: goto st404;
		case 405: goto st405;
		case 406: goto st406;
		case 407: goto st407;
		case 408: goto st408;
		case 409: goto st409;
		case 410: goto st410;
		case 411: goto st411;
		case 412: goto st412;
		case 413: goto st413;
		case 414: goto st414;
		case 415: goto st415;
		case 416: goto st416;
		case 417: goto st417;
		case 418: goto st418;
		case 419: goto st419;
		case 420: goto st420;
		case 421: goto st421;
		case 422: goto st422;
		case 423: goto st423;
		case 424: goto st424;
		case 425: goto st425;
		case 426: goto st426;
		case 427: goto st427;
		case 428: goto st428;
		case 429: goto st429;
		case 430: goto st430;
		case 431: goto st431;
		case 432: goto st432;
		case 433: goto st433;
		case 434: goto st434;
		case 435: goto st435;
		case 436: goto st436;
		case 437: goto st437;
		case 438: goto st438;
		case 439: goto st439;
		case 440: goto st440;
		case 441: goto st441;
		case 442: goto st442;
		case 443: goto st443;
		case 444: goto st444;
		case 445: goto st445;
		case 446: goto st446;
		case 447: goto st447;
		case 448: goto st448;
		case 449: goto st449;
		case 450: goto st450;
		case 451: goto st451;
		case 452: goto st452;
		case 453: goto st453;
		case 454: goto st454;
		case 1101: goto st1101;
		case 1102: goto st1102;
		case 1103: goto st1103;
		case 455: goto st455;
		case 456: goto st456;
		case 457: goto st457;
		case 458: goto st458;
		case 459: goto st459;
		case 1104: goto st1104;
		case 460: goto st460;
		case 461: goto st461;
		case 462: goto st462;
		case 463: goto st463;
		case 1105: goto st1105;
		case 1106: goto st1106;
		case 1107: goto st1107;
		case 464: goto st464;
		case 465: goto st465;
		case 1108: goto st1108;
		case 466: goto st466;
		case 467: goto st467;
		case 468: goto st468;
		case 1109: goto st1109;
		case 469: goto st469;
		case 470: goto st470;
		case 471: goto st471;
		case 472: goto st472;
		case 473: goto st473;
		case 474: goto st474;
		case 475: goto st475;
		case 476: goto st476;
		case 477: goto st477;
		case 478: goto st478;
		case 479: goto st479;
		case 480: goto st480;
		case 481: goto st481;
		case 482: goto st482;
		case 483: goto st483;
		case 484: goto st484;
		case 485: goto st485;
		case 486: goto st486;
		case 487: goto st487;
		case 488: goto st488;
		case 489: goto st489;
		case 490: goto st490;
		case 491: goto st491;
		case 492: goto st492;
		case 493: goto st493;
		case 494: goto st494;
		case 495: goto st495;
		case 496: goto st496;
		case 497: goto st497;
		case 498: goto st498;
		case 499: goto st499;
		case 500: goto st500;
		case 501: goto st501;
		case 502: goto st502;
		case 503: goto st503;
		case 504: goto st504;
		case 505: goto st505;
		case 506: goto st506;
		case 507: goto st507;
		case 508: goto st508;
		case 509: goto st509;
		case 510: goto st510;
		case 511: goto st511;
		case 512: goto st512;
		case 513: goto st513;
		case 514: goto st514;
		case 515: goto st515;
		case 516: goto st516;
		case 517: goto st517;
		case 518: goto st518;
		case 519: goto st519;
		case 520: goto st520;
		case 521: goto st521;
		case 522: goto st522;
		case 523: goto st523;
		case 524: goto st524;
		case 525: goto st525;
		case 526: goto st526;
		case 527: goto st527;
		case 528: goto st528;
		case 529: goto st529;
		case 530: goto st530;
		case 531: goto st531;
		case 532: goto st532;
		case 533: goto st533;
		case 534: goto st534;
		case 535: goto st535;
		case 536: goto st536;
		case 537: goto st537;
		case 538: goto st538;
		case 539: goto st539;
		case 540: goto st540;
		case 541: goto st541;
		case 542: goto st542;
		case 543: goto st543;
		case 544: goto st544;
		case 545: goto st545;
		case 546: goto st546;
		case 547: goto st547;
		case 548: goto st548;
		case 549: goto st549;
		case 550: goto st550;
		case 551: goto st551;
		case 552: goto st552;
		case 553: goto st553;
		case 554: goto st554;
		case 555: goto st555;
		case 556: goto st556;
		case 557: goto st557;
		case 558: goto st558;
		case 559: goto st559;
		case 560: goto st560;
		case 561: goto st561;
		case 562: goto st562;
		case 563: goto st563;
		case 1110: goto st1110;
		case 564: goto st564;
		case 565: goto st565;
		case 566: goto st566;
		case 567: goto st567;
		case 568: goto st568;
		case 569: goto st569;
		case 570: goto st570;
		case 571: goto st571;
		case 572: goto st572;
		case 573: goto st573;
		case 574: goto st574;
		case 575: goto st575;
		case 576: goto st576;
		case 577: goto st577;
		case 578: goto st578;
		case 579: goto st579;
		case 580: goto st580;
		case 581: goto st581;
		case 582: goto st582;
		case 583: goto st583;
		case 584: goto st584;
		case 585: goto st585;
		case 586: goto st586;
		case 587: goto st587;
		case 588: goto st588;
		case 589: goto st589;
		case 590: goto st590;
		case 591: goto st591;
		case 592: goto st592;
		case 593: goto st593;
		case 594: goto st594;
		case 595: goto st595;
		case 596: goto st596;
		case 597: goto st597;
		case 598: goto st598;
		case 599: goto st599;
		case 600: goto st600;
		case 601: goto st601;
		case 602: goto st602;
		case 603: goto st603;
		case 604: goto st604;
		case 1111: goto st1111;
		case 605: goto st605;
		case 606: goto st606;
		case 1112: goto st1112;
		case 607: goto st607;
		case 608: goto st608;
		case 609: goto st609;
		case 610: goto st610;
		case 611: goto st611;
		case 612: goto st612;
		case 613: goto st613;
		case 614: goto st614;
		case 615: goto st615;
		case 616: goto st616;
		case 617: goto st617;
		case 618: goto st618;
		case 619: goto st619;
		case 620: goto st620;
		case 1113: goto st1113;
		case 621: goto st621;
		case 622: goto st622;
		case 623: goto st623;
		case 624: goto st624;
		case 625: goto st625;
		case 626: goto st626;
		case 627: goto st627;
		case 628: goto st628;
		case 629: goto st629;
		case 630: goto st630;
		case 631: goto st631;
		case 632: goto st632;
		case 633: goto st633;
		case 634: goto st634;
		case 635: goto st635;
		case 636: goto st636;
		case 637: goto st637;
		case 638: goto st638;
		case 639: goto st639;
		case 640: goto st640;
		case 641: goto st641;
		case 642: goto st642;
		case 1114: goto st1114;
		case 643: goto st643;
		case 644: goto st644;
		case 645: goto st645;
		case 646: goto st646;
		case 647: goto st647;
		case 1115: goto st1115;
		case 648: goto st648;
		case 649: goto st649;
		case 650: goto st650;
		case 651: goto st651;
		case 652: goto st652;
		case 1116: goto st1116;
		case 653: goto st653;
		case 654: goto st654;
		case 655: goto st655;
		case 656: goto st656;
		case 657: goto st657;
		case 1117: goto st1117;
		case 1118: goto st1118;
		case 1119: goto st1119;
		case 658: goto st658;
		case 659: goto st659;
		case 1120: goto st1120;
		case 660: goto st660;
		case 661: goto st661;
		case 662: goto st662;
		case 663: goto st663;
		case 664: goto st664;
		case 665: goto st665;
		case 666: goto st666;
		case 667: goto st667;
		case 668: goto st668;
		case 669: goto st669;
		case 670: goto st670;
		case 671: goto st671;
		case 672: goto st672;
		case 673: goto st673;
		case 674: goto st674;
		case 675: goto st675;
		case 676: goto st676;
		case 677: goto st677;
		case 678: goto st678;
		case 679: goto st679;
		case 680: goto st680;
		case 681: goto st681;
		case 682: goto st682;
		case 683: goto st683;
		case 684: goto st684;
		case 685: goto st685;
		case 686: goto st686;
		case 1121: goto st1121;
		case 1122: goto st1122;
		case 1123: goto st1123;
		case 687: goto st687;
		case 688: goto st688;
		case 689: goto st689;
		case 1124: goto st1124;
		case 1125: goto st1125;
		case 690: goto st690;
		case 691: goto st691;
		case 692: goto st692;
		case 693: goto st693;
		case 1126: goto st1126;
		case 1127: goto st1127;
		case 694: goto st694;
		case 695: goto st695;
		case 696: goto st696;
		case 697: goto st697;
		case 1128: goto st1128;
		case 1129: goto st1129;
		case 698: goto st698;
		case 699: goto st699;
		case 700: goto st700;
		case 701: goto st701;
		case 702: goto st702;
		case 703: goto st703;
		case 704: goto st704;
		case 705: goto st705;
		case 706: goto st706;
		case 707: goto st707;
		case 708: goto st708;
		case 709: goto st709;
		case 710: goto st710;
		case 711: goto st711;
		case 712: goto st712;
		case 713: goto st713;
		case 714: goto st714;
		case 715: goto st715;
		case 716: goto st716;
		case 717: goto st717;
		case 718: goto st718;
		case 719: goto st719;
		case 720: goto st720;
		case 721: goto st721;
		case 722: goto st722;
		case 1130: goto st1130;
		case 723: goto st723;
		case 724: goto st724;
		case 725: goto st725;
		case 726: goto st726;
		case 727: goto st727;
		case 728: goto st728;
		case 729: goto st729;
		case 730: goto st730;
		case 731: goto st731;
		case 732: goto st732;
		case 733: goto st733;
		case 734: goto st734;
		case 735: goto st735;
		case 736: goto st736;
		case 737: goto st737;
		case 1131: goto st1131;
		case 738: goto st738;
		case 739: goto st739;
		case 740: goto st740;
		case 741: goto st741;
		case 742: goto st742;
		case 743: goto st743;
		case 744: goto st744;
		case 745: goto st745;
		case 746: goto st746;
		case 747: goto st747;
		case 748: goto st748;
		case 749: goto st749;
		case 750: goto st750;
		case 1132: goto st1132;
		case 751: goto st751;
		case 752: goto st752;
		case 753: goto st753;
		case 754: goto st754;
		case 755: goto st755;
		case 756: goto st756;
		case 757: goto st757;
		case 758: goto st758;
		case 759: goto st759;
		case 760: goto st760;
		case 761: goto st761;
		case 1133: goto st1133;
		case 1134: goto st1134;
		case 762: goto st762;
		case 763: goto st763;
		case 764: goto st764;
		case 1135: goto st1135;
		case 765: goto st765;
		case 766: goto st766;
		case 767: goto st767;
		case 768: goto st768;
		case 769: goto st769;
		case 770: goto st770;
		case 771: goto st771;
		case 772: goto st772;
		case 773: goto st773;
		case 774: goto st774;
		case 1136: goto st1136;
		case 1137: goto st1137;
		case 1138: goto st1138;
		case 775: goto st775;
		case 776: goto st776;
		case 777: goto st777;
		case 778: goto st778;
		case 779: goto st779;
		case 780: goto st780;
		case 781: goto st781;
		case 782: goto st782;
		case 783: goto st783;
		case 784: goto st784;
		case 785: goto st785;
		case 1139: goto st1139;
		case 1140: goto st1140;
		case 1141: goto st1141;
		case 786: goto st786;
		case 787: goto st787;
		case 788: goto st788;
		case 789: goto st789;
		case 790: goto st790;
		case 791: goto st791;
		case 792: goto st792;
		case 793: goto st793;
		case 794: goto st794;
		case 795: goto st795;
		case 796: goto st796;
		case 797: goto st797;
		case 1142: goto st1142;
		case 798: goto st798;
		case 799: goto st799;
		case 800: goto st800;
		case 1143: goto st1143;
		case 1144: goto st1144;
		case 801: goto st801;
		case 1145: goto st1145;
		case 1146: goto st1146;
		case 802: goto st802;
		case 1147: goto st1147;
		case 1148: goto st1148;
		case 803: goto st803;
		case 804: goto st804;
		case 805: goto st805;
		case 806: goto st806;
		case 807: goto st807;
		case 808: goto st808;
		case 809: goto st809;
		case 810: goto st810;
		case 811: goto st811;
		case 812: goto st812;
		case 813: goto st813;
		case 814: goto st814;
		case 815: goto st815;
		case 816: goto st816;
		case 817: goto st817;
		case 818: goto st818;
		case 819: goto st819;
		case 820: goto st820;
		case 821: goto st821;
		case 822: goto st822;
		case 823: goto st823;
		case 824: goto st824;
		case 825: goto st825;
		case 826: goto st826;
		case 827: goto st827;
		case 828: goto st828;
		case 829: goto st829;
		case 830: goto st830;
		case 831: goto st831;
		case 832: goto st832;
		case 833: goto st833;
		case 834: goto st834;
		case 835: goto st835;
		case 836: goto st836;
		case 837: goto st837;
		case 838: goto st838;
		case 839: goto st839;
		case 840: goto st840;
		case 841: goto st841;
		case 842: goto st842;
		case 843: goto st843;
		case 844: goto st844;
		case 1149: goto st1149;
		case 845: goto st845;
		case 846: goto st846;
		case 847: goto st847;
		case 848: goto st848;
		case 849: goto st849;
		case 850: goto st850;
		case 851: goto st851;
		case 852: goto st852;
		case 853: goto st853;
		case 854: goto st854;
		case 855: goto st855;
		case 856: goto st856;
		case 857: goto st857;
		case 858: goto st858;
		case 859: goto st859;
		case 860: goto st860;
		case 861: goto st861;
		case 862: goto st862;
		case 863: goto st863;
		case 864: goto st864;
		case 865: goto st865;
		case 866: goto st866;
		case 867: goto st867;
		case 868: goto st868;
		case 869: goto st869;
		case 870: goto st870;
		case 871: goto st871;
		case 872: goto st872;
		case 873: goto st873;
		case 874: goto st874;
		case 875: goto st875;
		case 876: goto st876;
		case 877: goto st877;
		case 878: goto st878;
		case 879: goto st879;
		case 880: goto st880;
		case 881: goto st881;
		case 882: goto st882;
		case 883: goto st883;
		case 884: goto st884;
		case 885: goto st885;
		case 886: goto st886;
		case 887: goto st887;
		case 888: goto st888;
		case 889: goto st889;
		case 890: goto st890;
		case 891: goto st891;
		case 892: goto st892;
		case 893: goto st893;
		case 894: goto st894;
		case 895: goto st895;
		case 896: goto st896;
		case 897: goto st897;
		case 898: goto st898;
		case 899: goto st899;
		case 900: goto st900;
		case 901: goto st901;
		case 902: goto st902;
		case 903: goto st903;
		case 904: goto st904;
		case 905: goto st905;
		case 906: goto st906;
		case 907: goto st907;
		case 908: goto st908;
		case 909: goto st909;
		case 910: goto st910;
		case 911: goto st911;
		case 912: goto st912;
		case 913: goto st913;
		case 914: goto st914;
		case 915: goto st915;
		case 916: goto st916;
		case 917: goto st917;
		case 918: goto st918;
		case 919: goto st919;
		case 920: goto st920;
		case 921: goto st921;
		case 922: goto st922;
		case 923: goto st923;
		case 924: goto st924;
		case 925: goto st925;
		case 926: goto st926;
		case 927: goto st927;
		case 928: goto st928;
		case 929: goto st929;
		case 930: goto st930;
		case 931: goto st931;
		case 932: goto st932;
		case 933: goto st933;
		case 934: goto st934;
		case 935: goto st935;
		case 936: goto st936;
		case 937: goto st937;
		case 938: goto st938;
		case 939: goto st939;
		case 940: goto st940;
		case 941: goto st941;
		case 942: goto st942;
		case 943: goto st943;
		case 944: goto st944;
		case 945: goto st945;
		case 946: goto st946;
		case 947: goto st947;
		case 948: goto st948;
		case 949: goto st949;
		case 950: goto st950;
		case 951: goto st951;
		case 952: goto st952;
		case 953: goto st953;
		case 954: goto st954;
		case 955: goto st955;
		case 956: goto st956;
		case 957: goto st957;
		case 958: goto st958;
		case 959: goto st959;
		case 960: goto st960;
		case 961: goto st961;
		case 962: goto st962;
		case 963: goto st963;
		case 964: goto st964;
		case 965: goto st965;
		case 966: goto st966;
		case 967: goto st967;
		case 968: goto st968;
		case 969: goto st969;
		case 970: goto st970;
		case 971: goto st971;
		case 972: goto st972;
		case 973: goto st973;
		case 974: goto st974;
		case 975: goto st975;
		case 976: goto st976;
		case 977: goto st977;
		case 978: goto st978;
		case 1150: goto st1150;
		case 979: goto st979;
		case 980: goto st980;
		case 981: goto st981;
		case 982: goto st982;
		case 983: goto st983;
		case 984: goto st984;
		case 985: goto st985;
		case 986: goto st986;
		case 1151: goto st1151;
		case 987: goto st987;
		case 988: goto st988;
		case 989: goto st989;
		case 990: goto st990;
		case 991: goto st991;
		case 1152: goto st1152;
		case 992: goto st992;
		case 993: goto st993;
		case 994: goto st994;
		case 995: goto st995;
		case 996: goto st996;
		case 997: goto st997;
		case 998: goto st998;
		case 999: goto st999;
		case 1000: goto st1000;
		case 1001: goto st1001;
		case 1002: goto st1002;
		case 1003: goto st1003;
		case 1004: goto st1004;
		case 1005: goto st1005;
		case 1006: goto st1006;
		case 1007: goto st1007;
		case 1008: goto st1008;
		case 1009: goto st1009;
		case 1153: goto st1153;
		case 1010: goto st1010;
		case 1011: goto st1011;
		case 1012: goto st1012;
		case 1013: goto st1013;
		case 1014: goto st1014;
		case 1015: goto st1015;
		case 1016: goto st1016;
		case 1017: goto st1017;
		case 1018: goto st1018;
		case 1019: goto st1019;
		case 1020: goto st1020;
		case 1021: goto st1021;
		case 1022: goto st1022;
		case 1023: goto st1023;
		case 1024: goto st1024;
		case 1025: goto st1025;
		case 1026: goto st1026;
		case 1027: goto st1027;
		case 1028: goto st1028;
		case 1154: goto st1154;
		case 1029: goto st1029;
		case 1030: goto st1030;
		case 1031: goto st1031;
		case 1032: goto st1032;
		case 1033: goto st1033;
		case 1034: goto st1034;
		case 1035: goto st1035;
		case 1036: goto st1036;
		case 1037: goto st1037;
		case 1038: goto st1038;
		case 1039: goto st1039;
		case 1040: goto st1040;
		case 1041: goto st1041;
		case 1042: goto st1042;
		case 1043: goto st1043;
		case 1155: goto st1155;
		case 1156: goto st1156;
		case 1157: goto st1157;
		case 1044: goto st1044;
		case 1045: goto st1045;
		case 1046: goto st1046;
		case 1047: goto st1047;
		case 1048: goto st1048;
		case 1049: goto st1049;
		case 1050: goto st1050;
		case 1158: goto st1158;
		case 1051: goto st1051;
		case 1052: goto st1052;
		case 1053: goto st1053;
		case 1054: goto st1054;
		case 1055: goto st1055;
		case 1056: goto st1056;
		case 1057: goto st1057;
		case 1058: goto st1058;
		case 1059: goto st1059;
		case 1060: goto st1060;
		case 1061: goto st1061;
		case 1062: goto st1062;
		case 1063: goto st1063;
		case 1159: goto st1159;
		case 1064: goto st1064;
		case 1065: goto st1065;
		case 1066: goto st1066;
		case 1067: goto st1067;
		case 1068: goto st1068;
		case 1069: goto st1069;
		case 1070: goto st1070;
		case 1160: goto st1160;
		case 1071: goto st1071;
		case 1072: goto st1072;
		case 1073: goto st1073;
		case 1074: goto st1074;
		case 1075: goto st1075;
		case 1076: goto st1076;
		case 1161: goto st1161;
	default: break;
	}

	if ( ++p == pe )
		goto _test_eof;
_resume:
	switch ( cs )
	{
tr19:
	{
		s->line_counter++;
	}
	goto st1077;
tr81:
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);

		// Stop scanner if required.
		if (s->stop == true) {
			return;
		}
	}
	{
		s->line_counter++;
	}
	goto st1077;
tr3259:
	{
		s->stop = false;
	}
	{
		s->line_counter++;
	}
	goto st1077;
st1077:
	if ( ++p == pe )
		goto _test_eof1077;
case 1077:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3200;
		case 32: goto tr3200;
		case 36: goto st145;
		case 40: goto tr3202;
		case 41: goto tr3203;
		case 42: goto tr3204;
		case 92: goto tr3204;
		case 95: goto tr3204;
		case 778: goto tr19;
		case 827: goto st144;
		case 1034: goto tr3205;
		case 1083: goto tr3206;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr3204;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr3204;
	} else
		goto tr3204;
	goto tr3199;
tr0:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	goto st0;
tr23:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	goto st0;
tr35:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	goto st0;
tr55:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr69:
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr77:
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	goto st0;
tr83:
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	goto st0;
tr95:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	goto st0;
tr119:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	goto st0;
tr125:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	goto st0;
tr547:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	goto st0;
tr549:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr614:
	{
		ERR(ZS_BAD_DIRECTIVE);
		p--; {goto st253;}
	}
	goto st0;
tr625:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st253;}
	}
	goto st0;
tr637:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	goto st0;
tr661:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr676:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	goto st0;
tr704:
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	{
		ERR(ZS_BAD_DIRECTIVE);
		p--; {goto st253;}
	}
	goto st0;
tr714:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	goto st0;
tr720:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	goto st0;
tr731:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	goto st0;
tr744:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr797:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	goto st0;
tr802:
	{
		WARN(ZS_BAD_DNAME_CHAR);
		p--; {goto st253;}
	}
	goto st0;
tr815:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_DNAME_CHAR);
		p--; {goto st253;}
	}
	goto st0;
tr823:
	{
		WARN(ZS_BAD_TEXT_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_TEXT);
		p--; {goto st253;}
	}
	goto st0;
tr830:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_TEXT_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_TEXT);
		p--; {goto st253;}
	}
	goto st0;
tr840:
	{
		WARN(ZS_BAD_TEXT);
		p--; {goto st253;}
	}
	goto st0;
tr847:
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st253;}
	}
	goto st0;
tr853:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st253;}
	}
	goto st0;
tr855:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st253;}
	}
	goto st0;
tr867:
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st253;}
	}
	goto st0;
tr869:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st253;}
	}
	goto st0;
tr882:
	{
		ERR(ZS_BAD_ORIGIN);
		p--; {goto st253;}
	}
	goto st0;
tr894:
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	{
		ERR(ZS_BAD_ORIGIN);
		p--; {goto st253;}
	}
	goto st0;
tr908:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		ERR(ZS_BAD_ORIGIN);
		p--; {goto st253;}
	}
	goto st0;
tr920:
	{
		ERR(ZS_BAD_INCLUDE_FILENAME);
		p--; {goto st253;}
	}
	goto st0;
tr921:
	{
		ERR(ZS_BAD_INCLUDE_FILENAME);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	goto st0;
tr928:
	{
		ERR(ZS_BAD_INCLUDE_ORIGIN);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	goto st0;
tr938:
	{
		ERR(ZS_BAD_INCLUDE_ORIGIN);
		p--; {goto st253;}
	}
	goto st0;
tr953:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		ERR(ZS_BAD_INCLUDE_ORIGIN);
		p--; {goto st253;}
	}
	goto st0;
tr965:
	{
		WARN(ZS_BAD_BASE64_CHAR);
		p--; {goto st253;}
	}
	goto st0;
tr979:
	{
		WARN(ZS_BAD_BITMAP);
		p--; {goto st253;}
	}
	goto st0;
tr1449:
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr1455:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr1463:
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr1513:
	{
		WARN(ZS_BAD_ALGORITHM);
		p--; {goto st253;}
	}
	goto st0;
tr1624:
	{
		WARN(ZS_BAD_CERT_TYPE);
		p--; {goto st253;}
	}
	goto st0;
tr1677:
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr1700:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr1719:
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr1867:
	{
		s->long_string = false;
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr1888:
	{
		WARN(ZS_BAD_LOC_DATA);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr2245:
	{
		WARN(ZS_BAD_APL);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr2260:
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_APL);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr2303:
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr2360:
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr2375:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr2389:
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_GATEWAY_KEY);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr2412:
	{
		WARN(ZS_BAD_GATEWAY_KEY);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr2426:
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr2433:
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_GATEWAY_KEY);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr2538:
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr2591:
	{
		WARN(ZS_BAD_TIMESTAMP_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr2947:
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr3035:
	{
		WARN(ZS_BAD_BASE32HEX_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr3178:
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_CHAR_COLON);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr3184:
	{
		WARN(ZS_BAD_CHAR_COLON);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr3188:
	{
		WARN(ZS_BAD_CHAR_DASH);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	goto st0;
tr3199:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	goto st0;
tr3217:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	goto st0;
tr3237:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	goto st0;
tr3253:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st253;}
	}
	{
		ERR(ZS_BAD_DIRECTIVE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	goto st0;
tr3276:
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	goto st0;
st0:
cs = 0;
	goto _out;
tr2:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1;
tr3:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1;
tr3200:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st1;
tr3202:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1;
tr3203:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1;
tr3263:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st1;
tr3264:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st1;
tr3254:
	{
		s->stop = false;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st1;
tr3256:
	{
		s->stop = false;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1;
tr3257:
	{
		s->stop = false;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1;
st1:
	if ( ++p == pe )
		goto _test_eof1;
case 1:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1;
		case 32: goto st1;
		case 40: goto tr2;
		case 41: goto tr3;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 80: goto tr15;
		case 82: goto tr16;
		case 83: goto tr17;
		case 84: goto tr18;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 112: goto tr15;
		case 114: goto tr16;
		case 115: goto tr17;
		case 116: goto tr18;
		case 778: goto tr19;
		case 827: goto st144;
		case 1034: goto tr21;
		case 1083: goto st249;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr0;
tr4:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st2;
tr27:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st2;
st2:
	if ( ++p == pe )
		goto _test_eof2;
case 2:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr24;
		case 32: goto tr24;
		case 40: goto tr25;
		case 41: goto tr26;
		case 68: goto tr28;
		case 72: goto tr29;
		case 77: goto tr30;
		case 83: goto st159;
		case 87: goto tr32;
		case 100: goto tr28;
		case 104: goto tr29;
		case 109: goto tr30;
		case 115: goto st159;
		case 119: goto tr32;
		case 1034: goto tr33;
		case 1083: goto tr34;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr27;
	goto tr23;
tr37:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st3;
tr38:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st3;
tr53:
	{
		s->line_counter++;
	}
	goto st3;
tr24:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st3;
tr25:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st3;
tr26:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st3;
tr33:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st3;
tr655:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st3;
tr656:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st3;
tr657:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st3;
tr659:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st3;
st3:
	if ( ++p == pe )
		goto _test_eof3;
case 3:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st3;
		case 32: goto st3;
		case 40: goto tr37;
		case 41: goto tr38;
		case 65: goto tr39;
		case 67: goto tr40;
		case 68: goto tr41;
		case 69: goto tr42;
		case 72: goto tr43;
		case 73: goto tr44;
		case 75: goto tr45;
		case 76: goto tr46;
		case 77: goto tr47;
		case 78: goto tr48;
		case 80: goto tr49;
		case 82: goto tr50;
		case 83: goto tr51;
		case 84: goto tr52;
		case 97: goto tr39;
		case 99: goto tr40;
		case 100: goto tr41;
		case 101: goto tr42;
		case 104: goto tr43;
		case 105: goto tr44;
		case 107: goto tr45;
		case 108: goto tr46;
		case 109: goto tr47;
		case 110: goto tr48;
		case 112: goto tr49;
		case 114: goto tr50;
		case 115: goto tr51;
		case 116: goto tr52;
		case 1034: goto tr53;
		case 1083: goto st162;
	}
	goto tr35;
tr5:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st4;
tr39:
	{
		s->r_class = s->default_class;
	}
	goto st4;
tr574:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st4;
st4:
	if ( ++p == pe )
		goto _test_eof4;
case 4:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr56;
		case 32: goto tr56;
		case 40: goto tr57;
		case 41: goto tr58;
		case 65: goto st220;
		case 70: goto st223;
		case 80: goto st227;
		case 97: goto st220;
		case 102: goto st223;
		case 112: goto st227;
		case 2058: goto tr62;
		case 2107: goto tr63;
		case 2314: goto tr64;
		case 2363: goto tr64;
		case 2570: goto tr65;
		case 2619: goto tr66;
	}
	goto tr55;
tr70:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr71:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr73:
	{
		s->line_counter++;
	}
	goto st5;
tr56:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr57:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr58:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr62:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr140:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr141:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr142:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr143:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr151:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr152:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr153:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr154:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr161:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr162:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr163:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr164:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr172:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr173:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr174:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr175:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr186:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr187:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr188:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr189:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr198:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr199:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr200:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr201:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr209:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr210:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr211:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr212:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr217:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr218:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr219:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr220:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr230:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr231:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr232:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr233:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr239:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr240:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr241:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr242:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr251:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr252:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr253:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr254:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr291:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr292:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr293:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr294:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr302:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr303:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr304:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr305:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr310:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr311:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr312:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr313:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr323:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr324:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr325:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr326:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr332:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr333:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr334:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr335:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr341:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr342:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr343:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr344:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr349:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr350:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr351:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr352:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr362:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr363:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr364:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr365:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr370:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr371:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr372:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr373:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr384:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr385:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr386:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr387:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr393:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr394:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr395:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr396:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr401:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr402:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr403:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr405:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr411:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr412:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr413:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr415:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr420:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr421:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr422:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr424:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr433:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr434:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr435:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr436:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr443:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr444:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr445:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr446:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr454:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr455:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr456:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr457:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr465:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr466:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr467:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr468:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr473:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr474:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr475:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr476:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr486:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr487:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr488:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr489:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr495:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr496:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr497:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr498:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr504:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr505:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr506:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr507:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr515:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr516:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr517:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr518:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr528:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr529:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr530:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr531:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr537:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr538:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr539:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr540:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr550:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr551:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr552:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr554:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr757:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr758:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr759:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr760:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr768:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr769:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr770:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr771:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
tr777:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	goto st5;
tr778:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st5;
tr779:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st5;
tr780:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	goto st5;
st5:
	if ( ++p == pe )
		goto _test_eof5;
case 5:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st5;
		case 32: goto st5;
		case 40: goto tr70;
		case 41: goto tr71;
		case 92: goto st9;
		case 2058: goto tr73;
		case 2107: goto st10;
		case 2314: goto tr67;
		case 2363: goto tr67;
		case 2570: goto tr75;
		case 2619: goto tr76;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr67;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr67;
	} else
		goto tr67;
	goto tr69;
tr64:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr67:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr84:
	{ p--; }
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr85:
	{
		switch (s->r_type) {
		// Next types must not have empty rdata.
		case KNOT_RRTYPE_A:
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
		case KNOT_RRTYPE_SOA:
		case KNOT_RRTYPE_HINFO:
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
		case KNOT_RRTYPE_RP:
		case KNOT_RRTYPE_AAAA:
		case KNOT_RRTYPE_LOC:
		case KNOT_RRTYPE_SRV:
		case KNOT_RRTYPE_NAPTR:
		case KNOT_RRTYPE_CERT:
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_SSHFP:
		case KNOT_RRTYPE_IPSECKEY:
		case KNOT_RRTYPE_RRSIG:
		case KNOT_RRTYPE_NSEC:
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_DHCID:
		case KNOT_RRTYPE_NSEC3:
		case KNOT_RRTYPE_NSEC3PARAM:
		case KNOT_RRTYPE_TLSA:
		case KNOT_RRTYPE_CDS:
		case KNOT_RRTYPE_CDNSKEY:
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L32:
		case KNOT_RRTYPE_L64:
		case KNOT_RRTYPE_LP:
		case KNOT_RRTYPE_EUI48:
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st448;}
		// Next types can have empty rdata.
		case KNOT_RRTYPE_APL:
		default:
			{stack[top++] = 6; goto st457;}
		}
	}
	goto st6;
tr145:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr156:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr166:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr177:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr191:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr203:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr214:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr222:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr235:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr244:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr256:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr296:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr307:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr315:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr328:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr337:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr346:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr354:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr367:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr375:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr389:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr398:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr407:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr417:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr426:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr438:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr448:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr459:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr470:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr478:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr491:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr500:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr509:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr520:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr533:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr542:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr556:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr762:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr773:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
tr782:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 6; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 6; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 6; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 6; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 6; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 6; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 6; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 6; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 6; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 6; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 6; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 6; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 6; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 6; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 6; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 6; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 6; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 6; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 6; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 6; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 6; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 6; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 6; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 6; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 6; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 6; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 6; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st6;
st6:
	if ( ++p == pe )
		goto _test_eof6;
case 6:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st7;
		case 32: goto st7;
		case 40: goto tr79;
		case 41: goto tr80;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr81;
		case 1083: goto st8;
	}
	goto tr77;
tr79:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st7;
tr80:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st7;
st7:
	if ( ++p == pe )
		goto _test_eof7;
case 7:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st7;
		case 32: goto st7;
		case 40: goto tr79;
		case 41: goto tr80;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr81;
		case 1083: goto st8;
	}
	goto tr83;
st8:
	if ( ++p == pe )
		goto _test_eof8;
case 8:
	if ( (*p) == 10 )
		goto tr81;
	goto st8;
st9:
	if ( ++p == pe )
		goto _test_eof9;
case 9:
	if ( (*p) == 35 )
		goto tr85;
	goto tr84;
tr63:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr144:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr155:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr165:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr176:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr190:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr202:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr213:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr221:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr234:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr243:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr255:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr295:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr306:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr314:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr327:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr336:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr345:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr353:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr366:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr374:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr388:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr397:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr406:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr416:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr425:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr437:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr447:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr458:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr469:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr477:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr490:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr499:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr508:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr519:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr532:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr541:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr555:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr761:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr772:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
tr781:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	goto st10;
st10:
	if ( ++p == pe )
		goto _test_eof10;
case 10:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr73;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st10;
	goto tr69;
tr87:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st11;
tr88:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st11;
tr65:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr75:
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr146:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr157:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr167:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr178:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr192:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr204:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr215:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr223:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr236:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr245:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr257:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr297:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr308:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr316:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr329:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr338:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr347:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr355:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr368:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr376:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr390:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr399:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr408:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr418:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr427:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr439:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr449:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr460:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr471:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr479:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr492:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr501:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr510:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr521:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr534:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr543:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr557:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr763:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr774:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
tr783:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 11; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 11; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 11; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 11; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 11; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 11; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 11; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 11; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 11; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 11; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 11; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 11; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 11; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 11; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 11; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 11; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 11; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 11; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 11; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 11; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 11; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 11; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 11; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 11; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 11; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 11; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 11; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st11;
st11:
	if ( ++p == pe )
		goto _test_eof11;
case 11:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st11;
		case 32: goto st11;
		case 40: goto tr87;
		case 41: goto tr88;
		case 92: goto st9;
		case 1802: goto tr81;
		case 1851: goto st8;
		case 2058: goto tr81;
		case 2107: goto st12;
		case 2314: goto tr90;
		case 2363: goto tr91;
		case 2570: goto tr81;
		case 2619: goto tr92;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr67;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr67;
	} else
		goto tr67;
	goto tr77;
tr681:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st12;
st12:
	if ( ++p == pe )
		goto _test_eof12;
case 12:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr81;
		case 1034: goto tr93;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto st12;
	} else if ( _widec >= 640 )
		goto st8;
	goto tr77;
tr93:
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);

		// Stop scanner if required.
		if (s->stop == true) {
			return;
		}
	}
	goto st1078;
tr680:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);

		// Stop scanner if required.
		if (s->stop == true) {
			return;
		}
	}
	goto st1078;
st1078:
	if ( ++p == pe )
		goto _test_eof1078;
case 1078:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr3207;
		case 32: goto tr3207;
		case 36: goto tr3208;
		case 40: goto tr3209;
		case 41: goto tr3210;
		case 42: goto tr3211;
		case 58: goto tr67;
		case 92: goto tr3212;
		case 95: goto tr3211;
		case 1802: goto tr19;
		case 1851: goto st144;
		case 2058: goto tr3213;
		case 2107: goto tr3214;
		case 2314: goto tr115;
		case 2363: goto tr116;
		case 2570: goto tr3215;
		case 2619: goto tr3216;
	}
	if ( _widec < 60 ) {
		if ( _widec < 11 ) {
			if ( _widec <= 8 )
				goto tr67;
		} else if ( _widec > 44 ) {
			if ( 45 <= _widec && _widec <= 57 )
				goto tr3211;
		} else
			goto tr67;
	} else if ( _widec > 63 ) {
		if ( _widec < 91 ) {
			if ( 64 <= _widec && _widec <= 90 )
				goto tr3211;
		} else if ( _widec > 96 ) {
			if ( _widec > 122 ) {
				if ( 123 <= _widec )
					goto tr67;
			} else if ( _widec >= 97 )
				goto tr3211;
		} else
			goto tr67;
	} else
		goto tr67;
	goto tr714;
tr96:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st13;
tr97:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st13;
tr3207:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st13;
tr3209:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st13;
tr3210:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st13;
st13:
	if ( ++p == pe )
		goto _test_eof13;
case 13:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st13;
		case 32: goto st13;
		case 40: goto tr96;
		case 41: goto tr97;
		case 58: goto tr67;
		case 65: goto tr99;
		case 67: goto tr100;
		case 68: goto tr101;
		case 69: goto tr102;
		case 72: goto tr103;
		case 73: goto tr104;
		case 75: goto tr105;
		case 76: goto tr106;
		case 77: goto tr107;
		case 78: goto tr108;
		case 80: goto tr109;
		case 82: goto tr110;
		case 83: goto tr111;
		case 84: goto tr112;
		case 92: goto st9;
		case 97: goto tr99;
		case 99: goto tr100;
		case 100: goto tr101;
		case 101: goto tr102;
		case 104: goto tr103;
		case 105: goto tr104;
		case 107: goto tr105;
		case 108: goto tr106;
		case 109: goto tr107;
		case 110: goto tr108;
		case 112: goto tr109;
		case 114: goto tr110;
		case 115: goto tr111;
		case 116: goto tr112;
		case 1802: goto tr19;
		case 1851: goto st144;
		case 2058: goto tr113;
		case 2107: goto st229;
		case 2314: goto tr115;
		case 2363: goto tr116;
		case 2570: goto tr117;
		case 2619: goto tr118;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr67;
	} else if ( _widec > 47 ) {
		if ( _widec > 57 ) {
			if ( 60 <= _widec )
				goto tr67;
		} else if ( _widec >= 48 )
			goto tr98;
	} else
		goto tr67;
	goto tr95;
tr98:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 14; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 14; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 14; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 14; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 14; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 14; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 14; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 14; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 14; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 14; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 14; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 14; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 14; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 14; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 14; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 14; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 14; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 14; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 14; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 14; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 14; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 14; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 14; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 14; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 14; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 14; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 14; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st14;
st14:
	if ( ++p == pe )
		goto _test_eof14;
case 14:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr120;
		case 32: goto tr120;
		case 40: goto tr121;
		case 41: goto tr122;
		case 68: goto tr28;
		case 72: goto tr29;
		case 77: goto tr30;
		case 83: goto st159;
		case 87: goto tr32;
		case 100: goto tr28;
		case 104: goto tr29;
		case 109: goto tr30;
		case 115: goto st159;
		case 119: goto tr32;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr123;
		case 1083: goto tr124;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr27;
	goto tr119;
tr127:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st15;
tr128:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st15;
tr120:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st15;
tr121:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st15;
tr122:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st15;
st15:
	if ( ++p == pe )
		goto _test_eof15;
case 15:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st15;
		case 32: goto st15;
		case 40: goto tr127;
		case 41: goto tr128;
		case 65: goto tr39;
		case 67: goto tr40;
		case 68: goto tr41;
		case 69: goto tr42;
		case 72: goto tr43;
		case 73: goto tr44;
		case 75: goto tr45;
		case 76: goto tr46;
		case 77: goto tr47;
		case 78: goto tr48;
		case 80: goto tr49;
		case 82: goto tr50;
		case 83: goto tr51;
		case 84: goto tr52;
		case 97: goto tr39;
		case 99: goto tr40;
		case 100: goto tr41;
		case 101: goto tr42;
		case 104: goto tr43;
		case 105: goto tr44;
		case 107: goto tr45;
		case 108: goto tr46;
		case 109: goto tr47;
		case 110: goto tr48;
		case 112: goto tr49;
		case 114: goto tr50;
		case 115: goto tr51;
		case 116: goto tr52;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr129;
		case 1083: goto st252;
	}
	goto tr125;
tr6:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st16;
tr40:
	{
		s->r_class = s->default_class;
	}
	goto st16;
tr575:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st16;
st16:
	if ( ++p == pe )
		goto _test_eof16;
case 16:
	switch( (*p) ) {
		case 68: goto st17;
		case 69: goto st25;
		case 78: goto st28;
		case 100: goto st17;
		case 101: goto st25;
		case 110: goto st28;
	}
	goto tr35;
st17:
	if ( ++p == pe )
		goto _test_eof17;
case 17:
	switch( (*p) ) {
		case 78: goto st18;
		case 83: goto st24;
		case 110: goto st18;
		case 115: goto st24;
	}
	goto tr35;
st18:
	if ( ++p == pe )
		goto _test_eof18;
case 18:
	switch( (*p) ) {
		case 83: goto st19;
		case 115: goto st19;
	}
	goto tr35;
st19:
	if ( ++p == pe )
		goto _test_eof19;
case 19:
	switch( (*p) ) {
		case 75: goto st20;
		case 107: goto st20;
	}
	goto tr35;
st20:
	if ( ++p == pe )
		goto _test_eof20;
case 20:
	switch( (*p) ) {
		case 69: goto st21;
		case 101: goto st21;
	}
	goto tr35;
st21:
	if ( ++p == pe )
		goto _test_eof21;
case 21:
	switch( (*p) ) {
		case 89: goto st22;
		case 121: goto st22;
	}
	goto tr35;
st22:
	if ( ++p == pe )
		goto _test_eof22;
case 22:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr140;
		case 32: goto tr140;
		case 40: goto tr141;
		case 41: goto tr142;
		case 2058: goto tr143;
		case 2107: goto tr144;
		case 2314: goto tr145;
		case 2363: goto tr145;
		case 2570: goto tr146;
		case 2619: goto tr147;
	}
	goto tr55;
tr149:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st23;
tr150:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st23;
tr66:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr76:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr147:
	{ s->r_type = KNOT_RRTYPE_CDNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr158:
	{ s->r_type = KNOT_RRTYPE_CDS; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr168:
	{ s->r_type = KNOT_RRTYPE_CERT; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr179:
	{ s->r_type = KNOT_RRTYPE_CNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr193:
	{ s->r_type = KNOT_RRTYPE_DHCID; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr205:
	{ s->r_type = KNOT_RRTYPE_DNAME; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr216:
	{ s->r_type = KNOT_RRTYPE_DNSKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr224:
	{ s->r_type = KNOT_RRTYPE_DS; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr237:
	{ s->r_type = KNOT_RRTYPE_EUI48; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr246:
	{ s->r_type = KNOT_RRTYPE_EUI64; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr258:
	{ s->r_type = KNOT_RRTYPE_HINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr298:
	{ s->r_type = KNOT_RRTYPE_IPSECKEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr309:
	{ s->r_type = KNOT_RRTYPE_KEY; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr317:
	{ s->r_type = KNOT_RRTYPE_KX; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr330:
	{ s->r_type = KNOT_RRTYPE_L32; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr339:
	{ s->r_type = KNOT_RRTYPE_L64; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr348:
	{ s->r_type = KNOT_RRTYPE_LOC; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr356:
	{ s->r_type = KNOT_RRTYPE_LP; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr369:
	{ s->r_type = KNOT_RRTYPE_MINFO; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr377:
	{ s->r_type = KNOT_RRTYPE_MX; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr391:
	{ s->r_type = KNOT_RRTYPE_NAPTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr400:
	{ s->r_type = KNOT_RRTYPE_NID; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr409:
	{ s->r_type = KNOT_RRTYPE_NS; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr419:
	{ s->r_type = KNOT_RRTYPE_NSEC; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr428:
	{ s->r_type = KNOT_RRTYPE_NSEC3; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr440:
	{ s->r_type = KNOT_RRTYPE_NSEC3PARAM; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr450:
	{ s->r_type = KNOT_RRTYPE_PTR; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr461:
	{ s->r_type = KNOT_RRTYPE_RP; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr472:
	{ s->r_type = KNOT_RRTYPE_RRSIG; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr480:
	{ s->r_type = KNOT_RRTYPE_RT; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr493:
	{ s->r_type = KNOT_RRTYPE_SOA; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr502:
	{ s->r_type = KNOT_RRTYPE_SPF; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr511:
	{ s->r_type = KNOT_RRTYPE_SRV; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr522:
	{ s->r_type = KNOT_RRTYPE_SSHFP; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr535:
	{ s->r_type = KNOT_RRTYPE_TLSA; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr544:
	{ s->r_type = KNOT_RRTYPE_TXT; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr558:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_type = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr764:
	{ s->r_type = KNOT_RRTYPE_AAAA; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr775:
	{ s->r_type = KNOT_RRTYPE_AFSDB; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
tr784:
	{ s->r_type = KNOT_RRTYPE_APL; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 23; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 23; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 23; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 23; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 23; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 23; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 23; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 23; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 23; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 23; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 23; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 23; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 23; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 23; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 23; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 23; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 23; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 23; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 23; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 23; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 23; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 23; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 23; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 23; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 23; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 23; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 23; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st23;
st23:
	if ( ++p == pe )
		goto _test_eof23;
case 23:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto st7;
		case 778: goto tr81;
		case 800: goto st7;
		case 808: goto tr79;
		case 809: goto tr80;
		case 827: goto st8;
		case 1033: goto st23;
		case 1034: goto tr81;
		case 1056: goto st23;
		case 1064: goto tr149;
		case 1065: goto tr150;
		case 1083: goto st12;
	}
	if ( 896 <= _widec && _widec <= 1151 )
		goto st10;
	goto tr77;
st24:
	if ( ++p == pe )
		goto _test_eof24;
case 24:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr151;
		case 32: goto tr151;
		case 40: goto tr152;
		case 41: goto tr153;
		case 2058: goto tr154;
		case 2107: goto tr155;
		case 2314: goto tr156;
		case 2363: goto tr156;
		case 2570: goto tr157;
		case 2619: goto tr158;
	}
	goto tr55;
st25:
	if ( ++p == pe )
		goto _test_eof25;
case 25:
	switch( (*p) ) {
		case 82: goto st26;
		case 114: goto st26;
	}
	goto tr35;
st26:
	if ( ++p == pe )
		goto _test_eof26;
case 26:
	switch( (*p) ) {
		case 84: goto st27;
		case 116: goto st27;
	}
	goto tr35;
st27:
	if ( ++p == pe )
		goto _test_eof27;
case 27:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr161;
		case 32: goto tr161;
		case 40: goto tr162;
		case 41: goto tr163;
		case 2058: goto tr164;
		case 2107: goto tr165;
		case 2314: goto tr166;
		case 2363: goto tr166;
		case 2570: goto tr167;
		case 2619: goto tr168;
	}
	goto tr55;
st28:
	if ( ++p == pe )
		goto _test_eof28;
case 28:
	switch( (*p) ) {
		case 65: goto st29;
		case 97: goto st29;
	}
	goto tr35;
st29:
	if ( ++p == pe )
		goto _test_eof29;
case 29:
	switch( (*p) ) {
		case 77: goto st30;
		case 109: goto st30;
	}
	goto tr35;
st30:
	if ( ++p == pe )
		goto _test_eof30;
case 30:
	switch( (*p) ) {
		case 69: goto st31;
		case 101: goto st31;
	}
	goto tr35;
st31:
	if ( ++p == pe )
		goto _test_eof31;
case 31:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr172;
		case 32: goto tr172;
		case 40: goto tr173;
		case 41: goto tr174;
		case 2058: goto tr175;
		case 2107: goto tr176;
		case 2314: goto tr177;
		case 2363: goto tr177;
		case 2570: goto tr178;
		case 2619: goto tr179;
	}
	goto tr55;
tr7:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st32;
tr41:
	{
		s->r_class = s->default_class;
	}
	goto st32;
tr576:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st32;
st32:
	if ( ++p == pe )
		goto _test_eof32;
case 32:
	switch( (*p) ) {
		case 72: goto st33;
		case 78: goto st37;
		case 83: goto st45;
		case 104: goto st33;
		case 110: goto st37;
		case 115: goto st45;
	}
	goto tr35;
st33:
	if ( ++p == pe )
		goto _test_eof33;
case 33:
	switch( (*p) ) {
		case 67: goto st34;
		case 99: goto st34;
	}
	goto tr35;
st34:
	if ( ++p == pe )
		goto _test_eof34;
case 34:
	switch( (*p) ) {
		case 73: goto st35;
		case 105: goto st35;
	}
	goto tr35;
st35:
	if ( ++p == pe )
		goto _test_eof35;
case 35:
	switch( (*p) ) {
		case 68: goto st36;
		case 100: goto st36;
	}
	goto tr35;
st36:
	if ( ++p == pe )
		goto _test_eof36;
case 36:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr186;
		case 32: goto tr186;
		case 40: goto tr187;
		case 41: goto tr188;
		case 2058: goto tr189;
		case 2107: goto tr190;
		case 2314: goto tr191;
		case 2363: goto tr191;
		case 2570: goto tr192;
		case 2619: goto tr193;
	}
	goto tr55;
st37:
	if ( ++p == pe )
		goto _test_eof37;
case 37:
	switch( (*p) ) {
		case 65: goto st38;
		case 83: goto st41;
		case 97: goto st38;
		case 115: goto st41;
	}
	goto tr35;
st38:
	if ( ++p == pe )
		goto _test_eof38;
case 38:
	switch( (*p) ) {
		case 77: goto st39;
		case 109: goto st39;
	}
	goto tr35;
st39:
	if ( ++p == pe )
		goto _test_eof39;
case 39:
	switch( (*p) ) {
		case 69: goto st40;
		case 101: goto st40;
	}
	goto tr35;
st40:
	if ( ++p == pe )
		goto _test_eof40;
case 40:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr198;
		case 32: goto tr198;
		case 40: goto tr199;
		case 41: goto tr200;
		case 2058: goto tr201;
		case 2107: goto tr202;
		case 2314: goto tr203;
		case 2363: goto tr203;
		case 2570: goto tr204;
		case 2619: goto tr205;
	}
	goto tr55;
st41:
	if ( ++p == pe )
		goto _test_eof41;
case 41:
	switch( (*p) ) {
		case 75: goto st42;
		case 107: goto st42;
	}
	goto tr35;
st42:
	if ( ++p == pe )
		goto _test_eof42;
case 42:
	switch( (*p) ) {
		case 69: goto st43;
		case 101: goto st43;
	}
	goto tr35;
st43:
	if ( ++p == pe )
		goto _test_eof43;
case 43:
	switch( (*p) ) {
		case 89: goto st44;
		case 121: goto st44;
	}
	goto tr35;
st44:
	if ( ++p == pe )
		goto _test_eof44;
case 44:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr209;
		case 32: goto tr209;
		case 40: goto tr210;
		case 41: goto tr211;
		case 2058: goto tr212;
		case 2107: goto tr213;
		case 2314: goto tr214;
		case 2363: goto tr214;
		case 2570: goto tr215;
		case 2619: goto tr216;
	}
	goto tr55;
st45:
	if ( ++p == pe )
		goto _test_eof45;
case 45:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr217;
		case 32: goto tr217;
		case 40: goto tr218;
		case 41: goto tr219;
		case 2058: goto tr220;
		case 2107: goto tr221;
		case 2314: goto tr222;
		case 2363: goto tr222;
		case 2570: goto tr223;
		case 2619: goto tr224;
	}
	goto tr55;
tr8:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st46;
tr42:
	{
		s->r_class = s->default_class;
	}
	goto st46;
tr577:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st46;
st46:
	if ( ++p == pe )
		goto _test_eof46;
case 46:
	switch( (*p) ) {
		case 85: goto st47;
		case 117: goto st47;
	}
	goto tr35;
st47:
	if ( ++p == pe )
		goto _test_eof47;
case 47:
	switch( (*p) ) {
		case 73: goto st48;
		case 105: goto st48;
	}
	goto tr35;
st48:
	if ( ++p == pe )
		goto _test_eof48;
case 48:
	switch( (*p) ) {
		case 52: goto st49;
		case 54: goto st51;
	}
	goto tr35;
st49:
	if ( ++p == pe )
		goto _test_eof49;
case 49:
	if ( (*p) == 56 )
		goto st50;
	goto tr35;
st50:
	if ( ++p == pe )
		goto _test_eof50;
case 50:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr230;
		case 32: goto tr230;
		case 40: goto tr231;
		case 41: goto tr232;
		case 2058: goto tr233;
		case 2107: goto tr234;
		case 2314: goto tr235;
		case 2363: goto tr235;
		case 2570: goto tr236;
		case 2619: goto tr237;
	}
	goto tr55;
st51:
	if ( ++p == pe )
		goto _test_eof51;
case 51:
	if ( (*p) == 52 )
		goto st52;
	goto tr35;
st52:
	if ( ++p == pe )
		goto _test_eof52;
case 52:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr239;
		case 32: goto tr239;
		case 40: goto tr240;
		case 41: goto tr241;
		case 2058: goto tr242;
		case 2107: goto tr243;
		case 2314: goto tr244;
		case 2363: goto tr244;
		case 2570: goto tr245;
		case 2619: goto tr246;
	}
	goto tr55;
tr9:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st53;
tr43:
	{
		s->r_class = s->default_class;
	}
	goto st53;
tr578:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st53;
st53:
	if ( ++p == pe )
		goto _test_eof53;
case 53:
	switch( (*p) ) {
		case 73: goto st54;
		case 105: goto st54;
	}
	goto tr35;
st54:
	if ( ++p == pe )
		goto _test_eof54;
case 54:
	switch( (*p) ) {
		case 78: goto st55;
		case 110: goto st55;
	}
	goto tr35;
st55:
	if ( ++p == pe )
		goto _test_eof55;
case 55:
	switch( (*p) ) {
		case 70: goto st56;
		case 102: goto st56;
	}
	goto tr35;
st56:
	if ( ++p == pe )
		goto _test_eof56;
case 56:
	switch( (*p) ) {
		case 79: goto st57;
		case 111: goto st57;
	}
	goto tr35;
st57:
	if ( ++p == pe )
		goto _test_eof57;
case 57:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr251;
		case 32: goto tr251;
		case 40: goto tr252;
		case 41: goto tr253;
		case 2058: goto tr254;
		case 2107: goto tr255;
		case 2314: goto tr256;
		case 2363: goto tr256;
		case 2570: goto tr257;
		case 2619: goto tr258;
	}
	goto tr55;
tr44:
	{
		s->r_class = s->default_class;
	}
	goto st58;
st58:
	if ( ++p == pe )
		goto _test_eof58;
case 58:
	switch( (*p) ) {
		case 78: goto st59;
		case 80: goto st62;
		case 110: goto st59;
		case 112: goto st62;
	}
	goto tr35;
st59:
	if ( ++p == pe )
		goto _test_eof59;
case 59:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr261;
		case 32: goto tr261;
		case 40: goto tr262;
		case 41: goto tr263;
		case 1034: goto tr264;
		case 1083: goto tr265;
	}
	goto tr35;
tr267:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st60;
tr268:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st60;
tr283:
	{
		s->line_counter++;
	}
	goto st60;
tr590:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st60;
tr591:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st60;
tr592:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st60;
tr599:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st60;
tr261:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	goto st60;
tr262:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st60;
tr263:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st60;
tr264:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	{
		s->line_counter++;
	}
	goto st60;
tr608:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st60;
tr609:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st60;
tr610:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st60;
tr612:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st60;
st60:
	if ( ++p == pe )
		goto _test_eof60;
case 60:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st60;
		case 32: goto st60;
		case 40: goto tr267;
		case 41: goto tr268;
		case 65: goto st4;
		case 67: goto st16;
		case 68: goto st32;
		case 69: goto st46;
		case 72: goto st53;
		case 73: goto st61;
		case 75: goto st69;
		case 76: goto st73;
		case 77: goto st81;
		case 78: goto st87;
		case 80: goto st103;
		case 82: goto st106;
		case 83: goto st113;
		case 84: goto st124;
		case 97: goto st4;
		case 99: goto st16;
		case 100: goto st32;
		case 101: goto st46;
		case 104: goto st53;
		case 105: goto st61;
		case 107: goto st69;
		case 108: goto st73;
		case 109: goto st81;
		case 110: goto st87;
		case 112: goto st103;
		case 114: goto st106;
		case 115: goto st113;
		case 116: goto st124;
		case 1034: goto tr283;
		case 1083: goto st134;
	}
	goto tr35;
tr579:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st61;
st61:
	if ( ++p == pe )
		goto _test_eof61;
case 61:
	switch( (*p) ) {
		case 80: goto st62;
		case 112: goto st62;
	}
	goto tr35;
st62:
	if ( ++p == pe )
		goto _test_eof62;
case 62:
	switch( (*p) ) {
		case 83: goto st63;
		case 115: goto st63;
	}
	goto tr35;
st63:
	if ( ++p == pe )
		goto _test_eof63;
case 63:
	switch( (*p) ) {
		case 69: goto st64;
		case 101: goto st64;
	}
	goto tr35;
st64:
	if ( ++p == pe )
		goto _test_eof64;
case 64:
	switch( (*p) ) {
		case 67: goto st65;
		case 99: goto st65;
	}
	goto tr35;
st65:
	if ( ++p == pe )
		goto _test_eof65;
case 65:
	switch( (*p) ) {
		case 75: goto st66;
		case 107: goto st66;
	}
	goto tr35;
st66:
	if ( ++p == pe )
		goto _test_eof66;
case 66:
	switch( (*p) ) {
		case 69: goto st67;
		case 101: goto st67;
	}
	goto tr35;
st67:
	if ( ++p == pe )
		goto _test_eof67;
case 67:
	switch( (*p) ) {
		case 89: goto st68;
		case 121: goto st68;
	}
	goto tr35;
st68:
	if ( ++p == pe )
		goto _test_eof68;
case 68:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr291;
		case 32: goto tr291;
		case 40: goto tr292;
		case 41: goto tr293;
		case 2058: goto tr294;
		case 2107: goto tr295;
		case 2314: goto tr296;
		case 2363: goto tr296;
		case 2570: goto tr297;
		case 2619: goto tr298;
	}
	goto tr55;
tr11:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st69;
tr45:
	{
		s->r_class = s->default_class;
	}
	goto st69;
tr580:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st69;
st69:
	if ( ++p == pe )
		goto _test_eof69;
case 69:
	switch( (*p) ) {
		case 69: goto st70;
		case 88: goto st72;
		case 101: goto st70;
		case 120: goto st72;
	}
	goto tr35;
st70:
	if ( ++p == pe )
		goto _test_eof70;
case 70:
	switch( (*p) ) {
		case 89: goto st71;
		case 121: goto st71;
	}
	goto tr35;
st71:
	if ( ++p == pe )
		goto _test_eof71;
case 71:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr302;
		case 32: goto tr302;
		case 40: goto tr303;
		case 41: goto tr304;
		case 2058: goto tr305;
		case 2107: goto tr306;
		case 2314: goto tr307;
		case 2363: goto tr307;
		case 2570: goto tr308;
		case 2619: goto tr309;
	}
	goto tr55;
st72:
	if ( ++p == pe )
		goto _test_eof72;
case 72:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr310;
		case 32: goto tr310;
		case 40: goto tr311;
		case 41: goto tr312;
		case 2058: goto tr313;
		case 2107: goto tr314;
		case 2314: goto tr315;
		case 2363: goto tr315;
		case 2570: goto tr316;
		case 2619: goto tr317;
	}
	goto tr55;
tr12:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st73;
tr46:
	{
		s->r_class = s->default_class;
	}
	goto st73;
tr581:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st73;
st73:
	if ( ++p == pe )
		goto _test_eof73;
case 73:
	switch( (*p) ) {
		case 51: goto st74;
		case 54: goto st76;
		case 79: goto st78;
		case 80: goto st80;
		case 111: goto st78;
		case 112: goto st80;
	}
	goto tr35;
st74:
	if ( ++p == pe )
		goto _test_eof74;
case 74:
	if ( (*p) == 50 )
		goto st75;
	goto tr35;
st75:
	if ( ++p == pe )
		goto _test_eof75;
case 75:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr323;
		case 32: goto tr323;
		case 40: goto tr324;
		case 41: goto tr325;
		case 2058: goto tr326;
		case 2107: goto tr327;
		case 2314: goto tr328;
		case 2363: goto tr328;
		case 2570: goto tr329;
		case 2619: goto tr330;
	}
	goto tr55;
st76:
	if ( ++p == pe )
		goto _test_eof76;
case 76:
	if ( (*p) == 52 )
		goto st77;
	goto tr35;
st77:
	if ( ++p == pe )
		goto _test_eof77;
case 77:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr332;
		case 32: goto tr332;
		case 40: goto tr333;
		case 41: goto tr334;
		case 2058: goto tr335;
		case 2107: goto tr336;
		case 2314: goto tr337;
		case 2363: goto tr337;
		case 2570: goto tr338;
		case 2619: goto tr339;
	}
	goto tr55;
st78:
	if ( ++p == pe )
		goto _test_eof78;
case 78:
	switch( (*p) ) {
		case 67: goto st79;
		case 99: goto st79;
	}
	goto tr35;
st79:
	if ( ++p == pe )
		goto _test_eof79;
case 79:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr341;
		case 32: goto tr341;
		case 40: goto tr342;
		case 41: goto tr343;
		case 2058: goto tr344;
		case 2107: goto tr345;
		case 2314: goto tr346;
		case 2363: goto tr346;
		case 2570: goto tr347;
		case 2619: goto tr348;
	}
	goto tr55;
st80:
	if ( ++p == pe )
		goto _test_eof80;
case 80:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr349;
		case 32: goto tr349;
		case 40: goto tr350;
		case 41: goto tr351;
		case 2058: goto tr352;
		case 2107: goto tr353;
		case 2314: goto tr354;
		case 2363: goto tr354;
		case 2570: goto tr355;
		case 2619: goto tr356;
	}
	goto tr55;
tr13:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st81;
tr47:
	{
		s->r_class = s->default_class;
	}
	goto st81;
tr582:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st81;
st81:
	if ( ++p == pe )
		goto _test_eof81;
case 81:
	switch( (*p) ) {
		case 73: goto st82;
		case 88: goto st86;
		case 105: goto st82;
		case 120: goto st86;
	}
	goto tr35;
st82:
	if ( ++p == pe )
		goto _test_eof82;
case 82:
	switch( (*p) ) {
		case 78: goto st83;
		case 110: goto st83;
	}
	goto tr35;
st83:
	if ( ++p == pe )
		goto _test_eof83;
case 83:
	switch( (*p) ) {
		case 70: goto st84;
		case 102: goto st84;
	}
	goto tr35;
st84:
	if ( ++p == pe )
		goto _test_eof84;
case 84:
	switch( (*p) ) {
		case 79: goto st85;
		case 111: goto st85;
	}
	goto tr35;
st85:
	if ( ++p == pe )
		goto _test_eof85;
case 85:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr362;
		case 32: goto tr362;
		case 40: goto tr363;
		case 41: goto tr364;
		case 2058: goto tr365;
		case 2107: goto tr366;
		case 2314: goto tr367;
		case 2363: goto tr367;
		case 2570: goto tr368;
		case 2619: goto tr369;
	}
	goto tr55;
st86:
	if ( ++p == pe )
		goto _test_eof86;
case 86:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr370;
		case 32: goto tr370;
		case 40: goto tr371;
		case 41: goto tr372;
		case 2058: goto tr373;
		case 2107: goto tr374;
		case 2314: goto tr375;
		case 2363: goto tr375;
		case 2570: goto tr376;
		case 2619: goto tr377;
	}
	goto tr55;
tr14:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st87;
tr48:
	{
		s->r_class = s->default_class;
	}
	goto st87;
tr583:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st87;
st87:
	if ( ++p == pe )
		goto _test_eof87;
case 87:
	switch( (*p) ) {
		case 65: goto st88;
		case 73: goto st92;
		case 83: goto st94;
		case 97: goto st88;
		case 105: goto st92;
		case 115: goto st94;
	}
	goto tr35;
st88:
	if ( ++p == pe )
		goto _test_eof88;
case 88:
	switch( (*p) ) {
		case 80: goto st89;
		case 112: goto st89;
	}
	goto tr35;
st89:
	if ( ++p == pe )
		goto _test_eof89;
case 89:
	switch( (*p) ) {
		case 84: goto st90;
		case 116: goto st90;
	}
	goto tr35;
st90:
	if ( ++p == pe )
		goto _test_eof90;
case 90:
	switch( (*p) ) {
		case 82: goto st91;
		case 114: goto st91;
	}
	goto tr35;
st91:
	if ( ++p == pe )
		goto _test_eof91;
case 91:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr384;
		case 32: goto tr384;
		case 40: goto tr385;
		case 41: goto tr386;
		case 2058: goto tr387;
		case 2107: goto tr388;
		case 2314: goto tr389;
		case 2363: goto tr389;
		case 2570: goto tr390;
		case 2619: goto tr391;
	}
	goto tr55;
st92:
	if ( ++p == pe )
		goto _test_eof92;
case 92:
	switch( (*p) ) {
		case 68: goto st93;
		case 100: goto st93;
	}
	goto tr35;
st93:
	if ( ++p == pe )
		goto _test_eof93;
case 93:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr393;
		case 32: goto tr393;
		case 40: goto tr394;
		case 41: goto tr395;
		case 2058: goto tr396;
		case 2107: goto tr397;
		case 2314: goto tr398;
		case 2363: goto tr398;
		case 2570: goto tr399;
		case 2619: goto tr400;
	}
	goto tr55;
st94:
	if ( ++p == pe )
		goto _test_eof94;
case 94:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr401;
		case 32: goto tr401;
		case 40: goto tr402;
		case 41: goto tr403;
		case 69: goto st95;
		case 101: goto st95;
		case 2058: goto tr405;
		case 2107: goto tr406;
		case 2314: goto tr407;
		case 2363: goto tr407;
		case 2570: goto tr408;
		case 2619: goto tr409;
	}
	goto tr55;
st95:
	if ( ++p == pe )
		goto _test_eof95;
case 95:
	switch( (*p) ) {
		case 67: goto st96;
		case 99: goto st96;
	}
	goto tr35;
st96:
	if ( ++p == pe )
		goto _test_eof96;
case 96:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr411;
		case 32: goto tr411;
		case 40: goto tr412;
		case 41: goto tr413;
		case 51: goto st97;
		case 2058: goto tr415;
		case 2107: goto tr416;
		case 2314: goto tr417;
		case 2363: goto tr417;
		case 2570: goto tr418;
		case 2619: goto tr419;
	}
	goto tr55;
st97:
	if ( ++p == pe )
		goto _test_eof97;
case 97:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr420;
		case 32: goto tr420;
		case 40: goto tr421;
		case 41: goto tr422;
		case 80: goto st98;
		case 112: goto st98;
		case 2058: goto tr424;
		case 2107: goto tr425;
		case 2314: goto tr426;
		case 2363: goto tr426;
		case 2570: goto tr427;
		case 2619: goto tr428;
	}
	goto tr55;
st98:
	if ( ++p == pe )
		goto _test_eof98;
case 98:
	switch( (*p) ) {
		case 65: goto st99;
		case 97: goto st99;
	}
	goto tr35;
st99:
	if ( ++p == pe )
		goto _test_eof99;
case 99:
	switch( (*p) ) {
		case 82: goto st100;
		case 114: goto st100;
	}
	goto tr35;
st100:
	if ( ++p == pe )
		goto _test_eof100;
case 100:
	switch( (*p) ) {
		case 65: goto st101;
		case 97: goto st101;
	}
	goto tr35;
st101:
	if ( ++p == pe )
		goto _test_eof101;
case 101:
	switch( (*p) ) {
		case 77: goto st102;
		case 109: goto st102;
	}
	goto tr35;
st102:
	if ( ++p == pe )
		goto _test_eof102;
case 102:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr433;
		case 32: goto tr433;
		case 40: goto tr434;
		case 41: goto tr435;
		case 2058: goto tr436;
		case 2107: goto tr437;
		case 2314: goto tr438;
		case 2363: goto tr438;
		case 2570: goto tr439;
		case 2619: goto tr440;
	}
	goto tr55;
tr15:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st103;
tr49:
	{
		s->r_class = s->default_class;
	}
	goto st103;
tr584:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st103;
st103:
	if ( ++p == pe )
		goto _test_eof103;
case 103:
	switch( (*p) ) {
		case 84: goto st104;
		case 116: goto st104;
	}
	goto tr35;
st104:
	if ( ++p == pe )
		goto _test_eof104;
case 104:
	switch( (*p) ) {
		case 82: goto st105;
		case 114: goto st105;
	}
	goto tr35;
st105:
	if ( ++p == pe )
		goto _test_eof105;
case 105:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr443;
		case 32: goto tr443;
		case 40: goto tr444;
		case 41: goto tr445;
		case 2058: goto tr446;
		case 2107: goto tr447;
		case 2314: goto tr448;
		case 2363: goto tr448;
		case 2570: goto tr449;
		case 2619: goto tr450;
	}
	goto tr55;
tr16:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st106;
tr50:
	{
		s->r_class = s->default_class;
	}
	goto st106;
tr585:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st106;
st106:
	if ( ++p == pe )
		goto _test_eof106;
case 106:
	switch( (*p) ) {
		case 80: goto st107;
		case 82: goto st108;
		case 84: goto st112;
		case 112: goto st107;
		case 114: goto st108;
		case 116: goto st112;
	}
	goto tr35;
st107:
	if ( ++p == pe )
		goto _test_eof107;
case 107:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr454;
		case 32: goto tr454;
		case 40: goto tr455;
		case 41: goto tr456;
		case 2058: goto tr457;
		case 2107: goto tr458;
		case 2314: goto tr459;
		case 2363: goto tr459;
		case 2570: goto tr460;
		case 2619: goto tr461;
	}
	goto tr55;
st108:
	if ( ++p == pe )
		goto _test_eof108;
case 108:
	switch( (*p) ) {
		case 83: goto st109;
		case 115: goto st109;
	}
	goto tr35;
st109:
	if ( ++p == pe )
		goto _test_eof109;
case 109:
	switch( (*p) ) {
		case 73: goto st110;
		case 105: goto st110;
	}
	goto tr35;
st110:
	if ( ++p == pe )
		goto _test_eof110;
case 110:
	switch( (*p) ) {
		case 71: goto st111;
		case 103: goto st111;
	}
	goto tr35;
st111:
	if ( ++p == pe )
		goto _test_eof111;
case 111:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr465;
		case 32: goto tr465;
		case 40: goto tr466;
		case 41: goto tr467;
		case 2058: goto tr468;
		case 2107: goto tr469;
		case 2314: goto tr470;
		case 2363: goto tr470;
		case 2570: goto tr471;
		case 2619: goto tr472;
	}
	goto tr55;
st112:
	if ( ++p == pe )
		goto _test_eof112;
case 112:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr473;
		case 32: goto tr473;
		case 40: goto tr474;
		case 41: goto tr475;
		case 2058: goto tr476;
		case 2107: goto tr477;
		case 2314: goto tr478;
		case 2363: goto tr478;
		case 2570: goto tr479;
		case 2619: goto tr480;
	}
	goto tr55;
tr17:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st113;
tr51:
	{
		s->r_class = s->default_class;
	}
	goto st113;
tr586:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st113;
st113:
	if ( ++p == pe )
		goto _test_eof113;
case 113:
	switch( (*p) ) {
		case 79: goto st114;
		case 80: goto st116;
		case 82: goto st118;
		case 83: goto st120;
		case 111: goto st114;
		case 112: goto st116;
		case 114: goto st118;
		case 115: goto st120;
	}
	goto tr35;
st114:
	if ( ++p == pe )
		goto _test_eof114;
case 114:
	switch( (*p) ) {
		case 65: goto st115;
		case 97: goto st115;
	}
	goto tr35;
st115:
	if ( ++p == pe )
		goto _test_eof115;
case 115:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr486;
		case 32: goto tr486;
		case 40: goto tr487;
		case 41: goto tr488;
		case 2058: goto tr489;
		case 2107: goto tr490;
		case 2314: goto tr491;
		case 2363: goto tr491;
		case 2570: goto tr492;
		case 2619: goto tr493;
	}
	goto tr55;
st116:
	if ( ++p == pe )
		goto _test_eof116;
case 116:
	switch( (*p) ) {
		case 70: goto st117;
		case 102: goto st117;
	}
	goto tr35;
st117:
	if ( ++p == pe )
		goto _test_eof117;
case 117:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr495;
		case 32: goto tr495;
		case 40: goto tr496;
		case 41: goto tr497;
		case 2058: goto tr498;
		case 2107: goto tr499;
		case 2314: goto tr500;
		case 2363: goto tr500;
		case 2570: goto tr501;
		case 2619: goto tr502;
	}
	goto tr55;
st118:
	if ( ++p == pe )
		goto _test_eof118;
case 118:
	switch( (*p) ) {
		case 86: goto st119;
		case 118: goto st119;
	}
	goto tr35;
st119:
	if ( ++p == pe )
		goto _test_eof119;
case 119:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr504;
		case 32: goto tr504;
		case 40: goto tr505;
		case 41: goto tr506;
		case 2058: goto tr507;
		case 2107: goto tr508;
		case 2314: goto tr509;
		case 2363: goto tr509;
		case 2570: goto tr510;
		case 2619: goto tr511;
	}
	goto tr55;
st120:
	if ( ++p == pe )
		goto _test_eof120;
case 120:
	switch( (*p) ) {
		case 72: goto st121;
		case 104: goto st121;
	}
	goto tr35;
st121:
	if ( ++p == pe )
		goto _test_eof121;
case 121:
	switch( (*p) ) {
		case 70: goto st122;
		case 102: goto st122;
	}
	goto tr35;
st122:
	if ( ++p == pe )
		goto _test_eof122;
case 122:
	switch( (*p) ) {
		case 80: goto st123;
		case 112: goto st123;
	}
	goto tr35;
st123:
	if ( ++p == pe )
		goto _test_eof123;
case 123:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr515;
		case 32: goto tr515;
		case 40: goto tr516;
		case 41: goto tr517;
		case 2058: goto tr518;
		case 2107: goto tr519;
		case 2314: goto tr520;
		case 2363: goto tr520;
		case 2570: goto tr521;
		case 2619: goto tr522;
	}
	goto tr55;
tr18:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st124;
tr52:
	{
		s->r_class = s->default_class;
	}
	goto st124;
tr587:
	{
		s->r_ttl = s->default_ttl;
	}
	goto st124;
st124:
	if ( ++p == pe )
		goto _test_eof124;
case 124:
	switch( (*p) ) {
		case 76: goto st125;
		case 88: goto st128;
		case 89: goto st130;
		case 108: goto st125;
		case 120: goto st128;
		case 121: goto st130;
	}
	goto tr35;
st125:
	if ( ++p == pe )
		goto _test_eof125;
case 125:
	switch( (*p) ) {
		case 83: goto st126;
		case 115: goto st126;
	}
	goto tr35;
st126:
	if ( ++p == pe )
		goto _test_eof126;
case 126:
	switch( (*p) ) {
		case 65: goto st127;
		case 97: goto st127;
	}
	goto tr35;
st127:
	if ( ++p == pe )
		goto _test_eof127;
case 127:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr528;
		case 32: goto tr528;
		case 40: goto tr529;
		case 41: goto tr530;
		case 2058: goto tr531;
		case 2107: goto tr532;
		case 2314: goto tr533;
		case 2363: goto tr533;
		case 2570: goto tr534;
		case 2619: goto tr535;
	}
	goto tr55;
st128:
	if ( ++p == pe )
		goto _test_eof128;
case 128:
	switch( (*p) ) {
		case 84: goto st129;
		case 116: goto st129;
	}
	goto tr35;
st129:
	if ( ++p == pe )
		goto _test_eof129;
case 129:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr537;
		case 32: goto tr537;
		case 40: goto tr538;
		case 41: goto tr539;
		case 2058: goto tr540;
		case 2107: goto tr541;
		case 2314: goto tr542;
		case 2363: goto tr542;
		case 2570: goto tr543;
		case 2619: goto tr544;
	}
	goto tr55;
st130:
	if ( ++p == pe )
		goto _test_eof130;
case 130:
	switch( (*p) ) {
		case 80: goto st131;
		case 112: goto st131;
	}
	goto tr35;
st131:
	if ( ++p == pe )
		goto _test_eof131;
case 131:
	switch( (*p) ) {
		case 69: goto st132;
		case 101: goto st132;
	}
	goto tr35;
st132:
	if ( ++p == pe )
		goto _test_eof132;
case 132:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr548;
	goto tr547;
tr548:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st133;
tr553:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st133;
st133:
	if ( ++p == pe )
		goto _test_eof133;
case 133:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr550;
		case 32: goto tr550;
		case 40: goto tr551;
		case 41: goto tr552;
		case 2058: goto tr554;
		case 2107: goto tr555;
		case 2314: goto tr556;
		case 2363: goto tr556;
		case 2570: goto tr557;
		case 2619: goto tr558;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr553;
	goto tr549;
tr600:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st134;
tr265:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	goto st134;
tr613:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st134;
st134:
	if ( ++p == pe )
		goto _test_eof134;
case 134:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr283;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st134;
	goto tr35;
tr129:
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);

		// Stop scanner if required.
		if (s->stop == true) {
			return;
		}
	}
	goto st1079;
tr123:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);

		// Stop scanner if required.
		if (s->stop == true) {
			return;
		}
	}
	goto st1079;
st1079:
	if ( ++p == pe )
		goto _test_eof1079;
case 1079:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3218;
		case 32: goto tr3218;
		case 36: goto st145;
		case 40: goto tr3219;
		case 41: goto tr3220;
		case 42: goto tr3204;
		case 65: goto tr3221;
		case 67: goto tr3222;
		case 68: goto tr3223;
		case 69: goto tr3224;
		case 72: goto tr3225;
		case 73: goto tr3226;
		case 75: goto tr3227;
		case 76: goto tr3228;
		case 77: goto tr3229;
		case 78: goto tr3230;
		case 80: goto tr3231;
		case 82: goto tr3232;
		case 83: goto tr3233;
		case 84: goto tr3234;
		case 92: goto tr3204;
		case 95: goto tr3204;
		case 97: goto tr3221;
		case 99: goto tr3222;
		case 100: goto tr3223;
		case 101: goto tr3224;
		case 104: goto tr3225;
		case 105: goto tr3226;
		case 107: goto tr3227;
		case 108: goto tr3228;
		case 109: goto tr3229;
		case 110: goto tr3230;
		case 112: goto tr3231;
		case 114: goto tr3232;
		case 115: goto tr3233;
		case 116: goto tr3234;
		case 778: goto tr19;
		case 827: goto st144;
		case 1034: goto tr3235;
		case 1083: goto tr3236;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr3204;
	} else if ( _widec > 90 ) {
		if ( 98 <= _widec && _widec <= 122 )
			goto tr3204;
	} else
		goto tr3204;
	goto tr3217;
tr560:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st135;
tr561:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st135;
tr3218:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st135;
tr3219:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st135;
tr3220:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st135;
st135:
	if ( ++p == pe )
		goto _test_eof135;
case 135:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st135;
		case 32: goto st135;
		case 40: goto tr560;
		case 41: goto tr561;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 80: goto tr15;
		case 82: goto tr16;
		case 83: goto tr17;
		case 84: goto tr18;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 112: goto tr15;
		case 114: goto tr16;
		case 115: goto tr17;
		case 116: goto tr18;
		case 778: goto tr19;
		case 827: goto st144;
		case 1034: goto tr562;
		case 1083: goto st250;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr0;
tr10:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	goto st136;
st136:
	if ( ++p == pe )
		goto _test_eof136;
case 136:
	switch( (*p) ) {
		case 78: goto st137;
		case 80: goto st62;
		case 110: goto st137;
		case 112: goto st62;
	}
	goto tr35;
st137:
	if ( ++p == pe )
		goto _test_eof137;
case 137:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr565;
		case 32: goto tr565;
		case 40: goto tr566;
		case 41: goto tr567;
		case 1034: goto tr568;
		case 1083: goto tr569;
	}
	goto tr35;
tr571:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st138;
tr572:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st138;
tr588:
	{
		s->line_counter++;
	}
	goto st138;
tr565:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	goto st138;
tr566:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st138;
tr567:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st138;
tr568:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	{
		s->line_counter++;
	}
	goto st138;
st138:
	if ( ++p == pe )
		goto _test_eof138;
case 138:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st138;
		case 32: goto st138;
		case 40: goto tr571;
		case 41: goto tr572;
		case 65: goto tr574;
		case 67: goto tr575;
		case 68: goto tr576;
		case 69: goto tr577;
		case 72: goto tr578;
		case 73: goto tr579;
		case 75: goto tr580;
		case 76: goto tr581;
		case 77: goto tr582;
		case 78: goto tr583;
		case 80: goto tr584;
		case 82: goto tr585;
		case 83: goto tr586;
		case 84: goto tr587;
		case 97: goto tr574;
		case 99: goto tr575;
		case 100: goto tr576;
		case 101: goto tr577;
		case 104: goto tr578;
		case 105: goto tr579;
		case 107: goto tr580;
		case 108: goto tr581;
		case 109: goto tr582;
		case 110: goto tr583;
		case 112: goto tr584;
		case 114: goto tr585;
		case 115: goto tr586;
		case 116: goto tr587;
		case 1034: goto tr588;
		case 1083: goto st143;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr573;
	goto tr547;
tr573:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st139;
tr593:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st139;
st139:
	if ( ++p == pe )
		goto _test_eof139;
case 139:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr590;
		case 32: goto tr590;
		case 40: goto tr591;
		case 41: goto tr592;
		case 68: goto tr594;
		case 72: goto tr595;
		case 77: goto tr596;
		case 83: goto st140;
		case 87: goto tr598;
		case 100: goto tr594;
		case 104: goto tr595;
		case 109: goto tr596;
		case 115: goto st140;
		case 119: goto tr598;
		case 1034: goto tr599;
		case 1083: goto tr600;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr593;
	goto tr23;
tr594:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st140;
tr595:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st140;
tr596:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st140;
tr598:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st140;
st140:
	if ( ++p == pe )
		goto _test_eof140;
case 140:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr590;
		case 32: goto tr590;
		case 40: goto tr591;
		case 41: goto tr592;
		case 1034: goto tr599;
		case 1083: goto tr600;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr601;
	goto tr23;
tr602:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st141;
tr601:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st141;
tr611:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st141;
st141:
	if ( ++p == pe )
		goto _test_eof141;
case 141:
	switch( (*p) ) {
		case 68: goto tr603;
		case 72: goto tr604;
		case 77: goto tr605;
		case 83: goto st142;
		case 87: goto tr607;
		case 100: goto tr603;
		case 104: goto tr604;
		case 109: goto tr605;
		case 115: goto st142;
		case 119: goto tr607;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr602;
	goto tr23;
tr603:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st142;
tr604:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st142;
tr605:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st142;
tr607:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st142;
st142:
	if ( ++p == pe )
		goto _test_eof142;
case 142:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr608;
		case 32: goto tr608;
		case 40: goto tr609;
		case 41: goto tr610;
		case 1034: goto tr612;
		case 1083: goto tr613;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr611;
	goto tr23;
tr569:
	{
		s->r_class = KNOT_CLASS_IN;
	}
	goto st143;
st143:
	if ( ++p == pe )
		goto _test_eof143;
case 143:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr588;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st143;
	goto tr35;
tr3260:
	{
		s->stop = false;
	}
	goto st144;
st144:
	if ( ++p == pe )
		goto _test_eof144;
case 144:
	if ( (*p) == 10 )
		goto tr19;
	goto st144;
tr562:
	{
		s->line_counter++;
	}
	goto st1080;
tr729:
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);

		// Stop scanner if required.
		if (s->stop == true) {
			return;
		}
	}
	goto st1080;
tr724:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);

		// Stop scanner if required.
		if (s->stop == true) {
			return;
		}
	}
	goto st1080;
tr3235:
	{
		s->line_counter++;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st1080;
st1080:
	if ( ++p == pe )
		goto _test_eof1080;
case 1080:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3218;
		case 32: goto tr3218;
		case 36: goto st145;
		case 40: goto tr3219;
		case 41: goto tr3220;
		case 42: goto tr3204;
		case 65: goto tr3239;
		case 67: goto tr3240;
		case 68: goto tr3241;
		case 69: goto tr3242;
		case 72: goto tr3243;
		case 73: goto tr3244;
		case 75: goto tr3245;
		case 76: goto tr3246;
		case 77: goto tr3247;
		case 78: goto tr3248;
		case 80: goto tr3249;
		case 82: goto tr3250;
		case 83: goto tr3251;
		case 84: goto tr3252;
		case 92: goto tr3204;
		case 95: goto tr3204;
		case 97: goto tr3239;
		case 99: goto tr3240;
		case 100: goto tr3241;
		case 101: goto tr3242;
		case 104: goto tr3243;
		case 105: goto tr3244;
		case 107: goto tr3245;
		case 108: goto tr3246;
		case 109: goto tr3247;
		case 110: goto tr3248;
		case 112: goto tr3249;
		case 114: goto tr3250;
		case 115: goto tr3251;
		case 116: goto tr3252;
		case 778: goto tr19;
		case 827: goto st144;
		case 1034: goto tr3235;
		case 1083: goto tr3236;
	}
	if ( _widec < 48 ) {
		if ( 45 <= _widec && _widec <= 47 )
			goto tr3204;
	} else if ( _widec > 57 ) {
		if ( _widec > 90 ) {
			if ( 98 <= _widec && _widec <= 122 )
				goto tr3204;
		} else if ( _widec >= 64 )
			goto tr3204;
	} else
		goto tr3238;
	goto tr3237;
tr3255:
	{
		s->stop = false;
	}
	goto st145;
st145:
	if ( ++p == pe )
		goto _test_eof145;
case 145:
	switch( (*p) ) {
		case 73: goto tr615;
		case 79: goto tr616;
		case 84: goto tr617;
		case 105: goto tr615;
		case 111: goto tr616;
		case 116: goto tr617;
	}
	goto tr614;
tr615:
	{
		s->stop = true;
	}
	goto st146;
st146:
	if ( ++p == pe )
		goto _test_eof146;
case 146:
	switch( (*p) ) {
		case 78: goto st147;
		case 110: goto st147;
	}
	goto tr614;
st147:
	if ( ++p == pe )
		goto _test_eof147;
case 147:
	switch( (*p) ) {
		case 67: goto st148;
		case 99: goto st148;
	}
	goto tr614;
st148:
	if ( ++p == pe )
		goto _test_eof148;
case 148:
	switch( (*p) ) {
		case 76: goto st149;
		case 108: goto st149;
	}
	goto tr614;
st149:
	if ( ++p == pe )
		goto _test_eof149;
case 149:
	switch( (*p) ) {
		case 85: goto st150;
		case 117: goto st150;
	}
	goto tr614;
st150:
	if ( ++p == pe )
		goto _test_eof150;
case 150:
	switch( (*p) ) {
		case 68: goto st151;
		case 100: goto st151;
	}
	goto tr614;
st151:
	if ( ++p == pe )
		goto _test_eof151;
case 151:
	switch( (*p) ) {
		case 69: goto st152;
		case 101: goto st152;
	}
	goto tr614;
st152:
	if ( ++p == pe )
		goto _test_eof152;
case 152:
	switch( (*p) ) {
		case 32: goto tr624;
		case 59: goto tr624;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr624;
	} else if ( (*p) >= 9 )
		goto tr624;
	goto tr614;
tr624:
	{ p--; {stack[top++] = 1081; goto st297;} }
	goto st1081;
tr710:
	{ p--; {stack[top++] = 1081; goto st285;} }
	goto st1081;
tr713:
	{ p--; {stack[top++] = 1081; goto st276;} }
	goto st1081;
st1081:
	if ( ++p == pe )
		goto _test_eof1081;
case 1081:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3254;
		case 32: goto tr3254;
		case 36: goto tr3255;
		case 40: goto tr3256;
		case 41: goto tr3257;
		case 42: goto tr3258;
		case 92: goto tr3258;
		case 95: goto tr3258;
		case 778: goto tr3259;
		case 827: goto tr3260;
		case 1034: goto tr3261;
		case 1083: goto tr3262;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr3258;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr3258;
	} else
		goto tr3258;
	goto tr3253;
tr3204:
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 153; goto st255;} }
	goto st153;
tr3258:
	{
		s->stop = false;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 153; goto st255;} }
	goto st153;
st153:
	if ( ++p == pe )
		goto _test_eof153;
case 153:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr626;
		case 32: goto tr626;
		case 40: goto tr627;
		case 41: goto tr628;
		case 1034: goto tr629;
		case 1083: goto tr630;
	}
	goto tr625;
tr632:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st154;
tr633:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st154;
tr634:
	{
		s->line_counter++;
	}
	goto st154;
tr626:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	goto st154;
tr627:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st154;
tr628:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st154;
tr629:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st154;
st154:
	if ( ++p == pe )
		goto _test_eof154;
case 154:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st154;
		case 32: goto st154;
		case 40: goto tr632;
		case 41: goto tr633;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 80: goto tr15;
		case 82: goto tr16;
		case 83: goto tr17;
		case 84: goto tr18;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 112: goto tr15;
		case 114: goto tr16;
		case 115: goto tr17;
		case 116: goto tr18;
		case 1034: goto tr634;
		case 1083: goto st155;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr547;
tr630:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	goto st155;
st155:
	if ( ++p == pe )
		goto _test_eof155;
case 155:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr634;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st155;
	goto st0;
tr21:
	{
		s->line_counter++;
	}
	goto st1082;
tr693:
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);

		// Stop scanner if required.
		if (s->stop == true) {
			return;
		}
	}
	goto st1082;
tr718:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);

		// Stop scanner if required.
		if (s->stop == true) {
			return;
		}
	}
	goto st1082;
tr3205:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1082;
tr3265:
	{
		s->line_counter++;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st1082;
tr3261:
	{
		s->stop = false;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1082;
tr3269:
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);

		// Stop scanner if required.
		if (s->stop == true) {
			return;
		}
	}
	{
		s->line_counter++;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st1082;
tr3292:
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);

		// Stop scanner if required.
		if (s->stop == true) {
			return;
		}
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st1082;
st1082:
	if ( ++p == pe )
		goto _test_eof1082;
case 1082:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3200;
		case 32: goto tr3200;
		case 36: goto st145;
		case 40: goto tr3263;
		case 41: goto tr3264;
		case 42: goto tr3204;
		case 65: goto tr3239;
		case 67: goto tr3240;
		case 68: goto tr3241;
		case 69: goto tr3242;
		case 72: goto tr3243;
		case 73: goto tr3244;
		case 75: goto tr3245;
		case 76: goto tr3246;
		case 77: goto tr3247;
		case 78: goto tr3248;
		case 80: goto tr3249;
		case 82: goto tr3250;
		case 83: goto tr3251;
		case 84: goto tr3252;
		case 92: goto tr3204;
		case 95: goto tr3204;
		case 97: goto tr3239;
		case 99: goto tr3240;
		case 100: goto tr3241;
		case 101: goto tr3242;
		case 104: goto tr3243;
		case 105: goto tr3244;
		case 107: goto tr3245;
		case 108: goto tr3246;
		case 109: goto tr3247;
		case 110: goto tr3248;
		case 112: goto tr3249;
		case 114: goto tr3250;
		case 115: goto tr3251;
		case 116: goto tr3252;
		case 778: goto tr19;
		case 827: goto st144;
		case 1034: goto tr3265;
		case 1083: goto tr3206;
	}
	if ( _widec < 48 ) {
		if ( 45 <= _widec && _widec <= 47 )
			goto tr3204;
	} else if ( _widec > 57 ) {
		if ( _widec > 90 ) {
			if ( 98 <= _widec && _widec <= 122 )
				goto tr3204;
		} else if ( _widec >= 64 )
			goto tr3204;
	} else
		goto tr3238;
	goto tr3237;
tr3238:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 156; goto st255;} }
	goto st156;
st156:
	if ( ++p == pe )
		goto _test_eof156;
case 156:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr638;
		case 32: goto tr638;
		case 40: goto tr639;
		case 41: goto tr640;
		case 68: goto tr28;
		case 72: goto tr29;
		case 77: goto tr30;
		case 83: goto st159;
		case 87: goto tr32;
		case 100: goto tr28;
		case 104: goto tr29;
		case 109: goto tr30;
		case 115: goto st159;
		case 119: goto tr32;
		case 1034: goto tr641;
		case 1083: goto tr642;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr27;
	goto tr637;
tr644:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st157;
tr645:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st157;
tr646:
	{
		s->line_counter++;
	}
	goto st157;
tr638:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st157;
tr639:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st157;
tr640:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st157;
tr641:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st157;
st157:
	if ( ++p == pe )
		goto _test_eof157;
case 157:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st157;
		case 32: goto st157;
		case 40: goto tr644;
		case 41: goto tr645;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 80: goto tr15;
		case 82: goto tr16;
		case 83: goto tr17;
		case 84: goto tr18;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 112: goto tr15;
		case 114: goto tr16;
		case 115: goto tr17;
		case 116: goto tr18;
		case 1034: goto tr646;
		case 1083: goto st158;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr547;
tr642:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st158;
st158:
	if ( ++p == pe )
		goto _test_eof158;
case 158:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr646;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st158;
	goto tr35;
tr28:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st159;
tr29:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st159;
tr30:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st159;
tr32:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st159;
st159:
	if ( ++p == pe )
		goto _test_eof159;
case 159:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr24;
		case 32: goto tr24;
		case 40: goto tr25;
		case 41: goto tr26;
		case 1034: goto tr33;
		case 1083: goto tr34;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr648;
	goto tr23;
tr649:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st160;
tr648:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st160;
tr658:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st160;
st160:
	if ( ++p == pe )
		goto _test_eof160;
case 160:
	switch( (*p) ) {
		case 68: goto tr650;
		case 72: goto tr651;
		case 77: goto tr652;
		case 83: goto st161;
		case 87: goto tr654;
		case 100: goto tr650;
		case 104: goto tr651;
		case 109: goto tr652;
		case 115: goto st161;
		case 119: goto tr654;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr649;
	goto tr23;
tr650:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st161;
tr651:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st161;
tr652:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st161;
tr654:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st161;
st161:
	if ( ++p == pe )
		goto _test_eof161;
case 161:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr655;
		case 32: goto tr655;
		case 40: goto tr656;
		case 41: goto tr657;
		case 1034: goto tr659;
		case 1083: goto tr660;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr658;
	goto tr23;
tr34:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st162;
tr660:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st162;
st162:
	if ( ++p == pe )
		goto _test_eof162;
case 162:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr53;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st162;
	goto tr35;
tr3221:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 163; goto st255;} }
	goto st163;
tr3239:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 163; goto st255;} }
	goto st163;
st163:
	if ( ++p == pe )
		goto _test_eof163;
case 163:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr662;
		case 32: goto tr662;
		case 40: goto tr663;
		case 41: goto tr664;
		case 65: goto st220;
		case 70: goto st223;
		case 80: goto st227;
		case 97: goto st220;
		case 102: goto st223;
		case 112: goto st227;
		case 2058: goto tr665;
		case 2107: goto tr666;
		case 2314: goto tr64;
		case 2363: goto tr64;
		case 2570: goto tr667;
		case 2619: goto tr668;
	}
	goto tr661;
tr670:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st164;
tr671:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st164;
tr672:
	{
		s->line_counter++;
	}
	goto st164;
tr662:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st164;
tr663:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st164;
tr664:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st164;
tr665:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st164;
st164:
	if ( ++p == pe )
		goto _test_eof164;
case 164:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st164;
		case 32: goto st164;
		case 40: goto tr670;
		case 41: goto tr671;
		case 58: goto tr67;
		case 65: goto tr99;
		case 67: goto tr100;
		case 68: goto tr101;
		case 69: goto tr102;
		case 72: goto tr103;
		case 73: goto tr104;
		case 75: goto tr105;
		case 76: goto tr106;
		case 77: goto tr107;
		case 78: goto tr108;
		case 80: goto tr109;
		case 82: goto tr110;
		case 83: goto tr111;
		case 84: goto tr112;
		case 92: goto st9;
		case 97: goto tr99;
		case 99: goto tr100;
		case 100: goto tr101;
		case 101: goto tr102;
		case 104: goto tr103;
		case 105: goto tr104;
		case 107: goto tr105;
		case 108: goto tr106;
		case 109: goto tr107;
		case 110: goto tr108;
		case 112: goto tr109;
		case 114: goto tr110;
		case 115: goto tr111;
		case 116: goto tr112;
		case 2058: goto tr672;
		case 2107: goto st234;
		case 2314: goto tr67;
		case 2363: goto tr67;
		case 2570: goto tr674;
		case 2619: goto tr675;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr67;
	} else if ( _widec > 47 ) {
		if ( _widec > 57 ) {
			if ( 60 <= _widec )
				goto tr67;
		} else if ( _widec >= 48 )
			goto tr98;
	} else
		goto tr67;
	goto tr549;
tr99:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 165; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 165; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 165; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 165; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 165; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 165; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 165; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 165; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 165; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 165; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 165; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 165; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 165; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 165; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 165; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 165; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 165; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 165; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 165; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 165; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 165; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 165; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 165; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 165; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 165; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 165; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 165; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st165;
st165:
	if ( ++p == pe )
		goto _test_eof165;
case 165:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr677;
		case 32: goto tr677;
		case 40: goto tr678;
		case 41: goto tr679;
		case 65: goto st220;
		case 70: goto st223;
		case 80: goto st227;
		case 97: goto st220;
		case 102: goto st223;
		case 112: goto st227;
		case 1802: goto tr81;
		case 1851: goto st8;
		case 2058: goto tr680;
		case 2107: goto tr681;
		case 2314: goto tr682;
		case 2363: goto tr683;
		case 2570: goto tr684;
		case 2619: goto tr685;
	}
	goto tr676;
tr687:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st166;
tr688:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st166;
tr677:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st166;
tr678:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st166;
tr679:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st166;
st166:
	if ( ++p == pe )
		goto _test_eof166;
case 166:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st166;
		case 32: goto st166;
		case 40: goto tr687;
		case 41: goto tr688;
		case 92: goto st9;
		case 1802: goto tr81;
		case 1851: goto st8;
		case 2058: goto tr93;
		case 2107: goto st12;
		case 2314: goto tr90;
		case 2363: goto tr91;
		case 2570: goto tr689;
		case 2619: goto tr92;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr67;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr67;
	} else
		goto tr67;
	goto tr77;
tr90:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1083; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1083; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1083; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1083; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1083; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1083; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1083; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1083; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1083; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1083; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1083; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1083; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1083; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1083; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1083; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1083; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1083; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1083; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1083; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1083; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1083; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1083; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1083; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1083; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1083; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1083; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1083; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);

		// Stop scanner if required.
		if (s->stop == true) {
			return;
		}
	}
	{
		s->line_counter++;
	}
	goto st1083;
tr115:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1083; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1083; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1083; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1083; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1083; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1083; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1083; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1083; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1083; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1083; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1083; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1083; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1083; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1083; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1083; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1083; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1083; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1083; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1083; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1083; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1083; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1083; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1083; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1083; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1083; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1083; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1083; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1083;
tr682:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1083; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1083; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1083; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1083; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1083; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1083; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1083; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1083; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1083; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1083; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1083; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1083; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1083; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1083; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1083; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1083; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1083; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1083; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1083; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1083; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1083; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1083; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1083; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1083; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1083; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1083; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1083; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);

		// Stop scanner if required.
		if (s->stop == true) {
			return;
		}
	}
	{
		s->line_counter++;
	}
	goto st1083;
st1083:
	if ( ++p == pe )
		goto _test_eof1083;
case 1083:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3266;
		case 32: goto tr3266;
		case 36: goto st145;
		case 40: goto tr3267;
		case 41: goto tr3268;
		case 42: goto tr3204;
		case 92: goto tr3204;
		case 95: goto tr3204;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr3269;
		case 1083: goto tr3270;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr3204;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr3204;
	} else
		goto tr3204;
	goto tr714;
tr691:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st167;
tr692:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st167;
tr715:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	goto st167;
tr716:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st167;
tr717:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st167;
tr3266:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st167;
tr3267:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st167;
tr3268:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st167;
st167:
	if ( ++p == pe )
		goto _test_eof167;
case 167:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st167;
		case 32: goto st167;
		case 40: goto tr691;
		case 41: goto tr692;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 80: goto tr15;
		case 82: goto tr16;
		case 83: goto tr17;
		case 84: goto tr18;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 112: goto tr15;
		case 114: goto tr16;
		case 115: goto tr17;
		case 116: goto tr18;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr693;
		case 1083: goto st168;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr0;
tr719:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	goto st168;
tr3270:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st168;
st168:
	if ( ++p == pe )
		goto _test_eof168;
case 168:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr81;
		case 1034: goto tr693;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto st168;
	} else if ( _widec >= 640 )
		goto st8;
	goto tr83;
tr683:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 169; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 169; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 169; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 169; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 169; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 169; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 169; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 169; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 169; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 169; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 169; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 169; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 169; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 169; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 169; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 169; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 169; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 169; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 169; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 169; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 169; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 169; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 169; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 169; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 169; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 169; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 169; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st169;
tr91:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 169; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 169; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 169; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 169; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 169; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 169; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 169; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 169; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 169; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 169; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 169; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 169; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 169; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 169; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 169; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 169; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 169; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 169; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 169; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 169; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 169; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 169; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 169; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 169; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 169; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 169; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 169; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st169;
st169:
	if ( ++p == pe )
		goto _test_eof169;
case 169:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st170;
		case 32: goto st170;
		case 40: goto tr696;
		case 41: goto tr697;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr81;
		case 1083: goto st8;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st8;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto st8;
	} else
		goto st8;
	goto tr77;
tr696:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st170;
tr697:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st170;
st170:
	if ( ++p == pe )
		goto _test_eof170;
case 170:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st170;
		case 32: goto st170;
		case 40: goto tr696;
		case 41: goto tr697;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr81;
		case 1083: goto st8;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st8;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto st8;
	} else
		goto st8;
	goto tr83;
tr684:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1084; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1084; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1084; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1084; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1084; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1084; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1084; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1084; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1084; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1084; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1084; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1084; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1084; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1084; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1084; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1084; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1084; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1084; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1084; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1084; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1084; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1084; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1084; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1084; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1084; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1084; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1084; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);

		// Stop scanner if required.
		if (s->stop == true) {
			return;
		}
	}
	goto st1084;
tr689:
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1084; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1084; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1084; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1084; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1084; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1084; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1084; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1084; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1084; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1084; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1084; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1084; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1084; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1084; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1084; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1084; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1084; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1084; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1084; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1084; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1084; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1084; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1084; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1084; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1084; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1084; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1084; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);

		// Stop scanner if required.
		if (s->stop == true) {
			return;
		}
	}
	goto st1084;
st1084:
	if ( ++p == pe )
		goto _test_eof1084;
case 1084:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr3271;
		case 32: goto tr3271;
		case 36: goto tr3208;
		case 40: goto tr3272;
		case 41: goto tr3273;
		case 42: goto tr3211;
		case 58: goto tr67;
		case 92: goto tr3212;
		case 95: goto tr3211;
		case 1802: goto tr81;
		case 1851: goto st8;
		case 2058: goto tr3269;
		case 2107: goto tr3274;
		case 2314: goto tr90;
		case 2363: goto tr91;
		case 2570: goto tr3269;
		case 2619: goto tr3275;
	}
	if ( _widec < 60 ) {
		if ( _widec < 11 ) {
			if ( _widec <= 8 )
				goto tr67;
		} else if ( _widec > 44 ) {
			if ( 45 <= _widec && _widec <= 57 )
				goto tr3211;
		} else
			goto tr67;
	} else if ( _widec > 63 ) {
		if ( _widec < 91 ) {
			if ( 64 <= _widec && _widec <= 90 )
				goto tr3211;
		} else if ( _widec > 96 ) {
			if ( _widec > 122 ) {
				if ( 123 <= _widec )
					goto tr67;
			} else if ( _widec >= 97 )
				goto tr3211;
		} else
			goto tr67;
	} else
		goto tr67;
	goto tr714;
tr699:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st171;
tr700:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st171;
tr674:
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 171; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 171; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 171; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 171; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 171; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 171; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 171; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 171; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 171; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 171; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 171; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 171; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 171; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 171; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 171; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 171; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 171; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 171; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 171; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 171; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 171; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 171; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 171; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 171; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 171; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 171; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 171; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st171;
tr667:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 171; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 171; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 171; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 171; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 171; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 171; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 171; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 171; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 171; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 171; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 171; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 171; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 171; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 171; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 171; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 171; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 171; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 171; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 171; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 171; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 171; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 171; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 171; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 171; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 171; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 171; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 171; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st171;
tr3271:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st171;
tr3272:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st171;
tr3273:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st171;
st171:
	if ( ++p == pe )
		goto _test_eof171;
case 171:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st171;
		case 32: goto st171;
		case 40: goto tr699;
		case 41: goto tr700;
		case 58: goto tr67;
		case 65: goto tr99;
		case 67: goto tr100;
		case 68: goto tr101;
		case 69: goto tr102;
		case 72: goto tr103;
		case 73: goto tr104;
		case 75: goto tr105;
		case 76: goto tr106;
		case 77: goto tr107;
		case 78: goto tr108;
		case 80: goto tr109;
		case 82: goto tr110;
		case 83: goto tr111;
		case 84: goto tr112;
		case 92: goto st9;
		case 97: goto tr99;
		case 99: goto tr100;
		case 100: goto tr101;
		case 101: goto tr102;
		case 104: goto tr103;
		case 105: goto tr104;
		case 107: goto tr105;
		case 108: goto tr106;
		case 109: goto tr107;
		case 110: goto tr108;
		case 112: goto tr109;
		case 114: goto tr110;
		case 115: goto tr111;
		case 116: goto tr112;
		case 1802: goto tr81;
		case 1851: goto st8;
		case 2058: goto tr693;
		case 2107: goto st185;
		case 2314: goto tr90;
		case 2363: goto tr91;
		case 2570: goto tr693;
		case 2619: goto tr702;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr67;
	} else if ( _widec > 47 ) {
		if ( _widec > 57 ) {
			if ( 60 <= _widec )
				goto tr67;
		} else if ( _widec >= 48 )
			goto tr98;
	} else
		goto tr67;
	goto tr95;
tr100:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 172; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 172; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 172; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 172; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 172; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 172; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 172; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 172; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 172; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 172; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 172; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 172; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 172; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 172; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 172; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 172; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 172; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 172; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 172; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 172; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 172; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 172; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 172; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 172; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 172; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 172; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 172; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st172;
st172:
	if ( ++p == pe )
		goto _test_eof172;
case 172:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st7;
		case 32: goto st7;
		case 40: goto tr79;
		case 41: goto tr80;
		case 68: goto st17;
		case 69: goto st25;
		case 78: goto st28;
		case 100: goto st17;
		case 101: goto st25;
		case 110: goto st28;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr81;
		case 1083: goto st8;
	}
	goto tr676;
tr101:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 173; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 173; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 173; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 173; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 173; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 173; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 173; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 173; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 173; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 173; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 173; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 173; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 173; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 173; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 173; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 173; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 173; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 173; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 173; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 173; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 173; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 173; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 173; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 173; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 173; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 173; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 173; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st173;
st173:
	if ( ++p == pe )
		goto _test_eof173;
case 173:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st7;
		case 32: goto st7;
		case 40: goto tr79;
		case 41: goto tr80;
		case 72: goto st33;
		case 78: goto st37;
		case 83: goto st45;
		case 104: goto st33;
		case 110: goto st37;
		case 115: goto st45;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr81;
		case 1083: goto st8;
	}
	goto tr676;
tr102:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 174; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 174; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 174; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 174; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 174; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 174; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 174; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 174; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 174; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 174; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 174; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 174; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 174; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 174; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 174; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 174; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 174; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 174; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 174; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 174; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 174; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 174; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 174; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 174; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 174; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 174; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 174; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st174;
st174:
	if ( ++p == pe )
		goto _test_eof174;
case 174:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st7;
		case 32: goto st7;
		case 40: goto tr79;
		case 41: goto tr80;
		case 85: goto st47;
		case 117: goto st47;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr81;
		case 1083: goto st8;
	}
	goto tr676;
tr103:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 175; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 175; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 175; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 175; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 175; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 175; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 175; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 175; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 175; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 175; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 175; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 175; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 175; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 175; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 175; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 175; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 175; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 175; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 175; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 175; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 175; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 175; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 175; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 175; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 175; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 175; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 175; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st175;
st175:
	if ( ++p == pe )
		goto _test_eof175;
case 175:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st7;
		case 32: goto st7;
		case 40: goto tr79;
		case 41: goto tr80;
		case 73: goto st54;
		case 105: goto st54;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr81;
		case 1083: goto st8;
	}
	goto tr676;
tr104:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 176; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 176; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 176; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 176; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 176; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 176; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 176; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 176; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 176; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 176; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 176; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 176; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 176; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 176; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 176; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 176; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 176; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 176; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 176; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 176; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 176; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 176; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 176; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 176; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 176; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 176; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 176; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st176;
st176:
	if ( ++p == pe )
		goto _test_eof176;
case 176:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st7;
		case 32: goto st7;
		case 40: goto tr79;
		case 41: goto tr80;
		case 78: goto st137;
		case 80: goto st62;
		case 110: goto st137;
		case 112: goto st62;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr81;
		case 1083: goto st8;
	}
	goto tr676;
tr105:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 177; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 177; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 177; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 177; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 177; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 177; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 177; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 177; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 177; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 177; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 177; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 177; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 177; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 177; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 177; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 177; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 177; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 177; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 177; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 177; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 177; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 177; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 177; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 177; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 177; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 177; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 177; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st177;
st177:
	if ( ++p == pe )
		goto _test_eof177;
case 177:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st7;
		case 32: goto st7;
		case 40: goto tr79;
		case 41: goto tr80;
		case 69: goto st70;
		case 88: goto st72;
		case 101: goto st70;
		case 120: goto st72;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr81;
		case 1083: goto st8;
	}
	goto tr676;
tr106:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 178; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 178; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 178; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 178; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 178; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 178; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 178; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 178; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 178; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 178; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 178; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 178; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 178; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 178; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 178; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 178; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 178; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 178; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 178; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 178; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 178; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 178; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 178; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 178; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 178; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 178; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 178; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st178;
st178:
	if ( ++p == pe )
		goto _test_eof178;
case 178:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st7;
		case 32: goto st7;
		case 40: goto tr79;
		case 41: goto tr80;
		case 51: goto st74;
		case 54: goto st76;
		case 79: goto st78;
		case 80: goto st80;
		case 111: goto st78;
		case 112: goto st80;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr81;
		case 1083: goto st8;
	}
	goto tr676;
tr107:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 179; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 179; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 179; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 179; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 179; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 179; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 179; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 179; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 179; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 179; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 179; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 179; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 179; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 179; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 179; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 179; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 179; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 179; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 179; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 179; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 179; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 179; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 179; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 179; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 179; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 179; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 179; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st179;
st179:
	if ( ++p == pe )
		goto _test_eof179;
case 179:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st7;
		case 32: goto st7;
		case 40: goto tr79;
		case 41: goto tr80;
		case 73: goto st82;
		case 88: goto st86;
		case 105: goto st82;
		case 120: goto st86;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr81;
		case 1083: goto st8;
	}
	goto tr676;
tr108:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 180; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 180; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 180; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 180; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 180; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 180; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 180; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 180; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 180; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 180; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 180; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 180; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 180; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 180; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 180; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 180; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 180; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 180; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 180; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 180; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 180; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 180; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 180; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 180; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 180; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 180; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 180; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st180;
st180:
	if ( ++p == pe )
		goto _test_eof180;
case 180:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st7;
		case 32: goto st7;
		case 40: goto tr79;
		case 41: goto tr80;
		case 65: goto st88;
		case 73: goto st92;
		case 83: goto st94;
		case 97: goto st88;
		case 105: goto st92;
		case 115: goto st94;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr81;
		case 1083: goto st8;
	}
	goto tr676;
tr109:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 181; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 181; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 181; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 181; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 181; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 181; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 181; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 181; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 181; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 181; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 181; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 181; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 181; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 181; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 181; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 181; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 181; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 181; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 181; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 181; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 181; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 181; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 181; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 181; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 181; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 181; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 181; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st181;
st181:
	if ( ++p == pe )
		goto _test_eof181;
case 181:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st7;
		case 32: goto st7;
		case 40: goto tr79;
		case 41: goto tr80;
		case 84: goto st104;
		case 116: goto st104;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr81;
		case 1083: goto st8;
	}
	goto tr676;
tr110:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 182; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 182; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 182; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 182; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 182; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 182; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 182; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 182; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 182; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 182; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 182; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 182; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 182; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 182; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 182; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 182; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 182; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 182; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 182; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 182; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 182; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 182; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 182; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 182; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 182; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 182; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 182; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st182;
st182:
	if ( ++p == pe )
		goto _test_eof182;
case 182:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st7;
		case 32: goto st7;
		case 40: goto tr79;
		case 41: goto tr80;
		case 80: goto st107;
		case 82: goto st108;
		case 84: goto st112;
		case 112: goto st107;
		case 114: goto st108;
		case 116: goto st112;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr81;
		case 1083: goto st8;
	}
	goto tr676;
tr111:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 183; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 183; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 183; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 183; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 183; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 183; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 183; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 183; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 183; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 183; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 183; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 183; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 183; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 183; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 183; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 183; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 183; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 183; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 183; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 183; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 183; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 183; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 183; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 183; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 183; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 183; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 183; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st183;
st183:
	if ( ++p == pe )
		goto _test_eof183;
case 183:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st7;
		case 32: goto st7;
		case 40: goto tr79;
		case 41: goto tr80;
		case 79: goto st114;
		case 80: goto st116;
		case 82: goto st118;
		case 83: goto st120;
		case 111: goto st114;
		case 112: goto st116;
		case 114: goto st118;
		case 115: goto st120;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr81;
		case 1083: goto st8;
	}
	goto tr676;
tr112:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 184; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 184; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 184; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 184; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 184; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 184; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 184; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 184; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 184; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 184; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 184; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 184; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 184; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 184; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 184; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 184; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 184; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 184; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 184; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 184; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 184; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 184; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 184; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 184; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 184; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 184; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 184; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st184;
st184:
	if ( ++p == pe )
		goto _test_eof184;
case 184:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st7;
		case 32: goto st7;
		case 40: goto tr79;
		case 41: goto tr80;
		case 76: goto st125;
		case 88: goto st128;
		case 89: goto st130;
		case 108: goto st125;
		case 120: goto st128;
		case 121: goto st130;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr81;
		case 1083: goto st8;
	}
	goto tr676;
tr736:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st185;
tr3274:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st185;
st185:
	if ( ++p == pe )
		goto _test_eof185;
case 185:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr81;
		case 1034: goto tr703;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto st185;
	} else if ( _widec >= 640 )
		goto st8;
	goto tr77;
tr113:
	{
		s->line_counter++;
	}
	goto st1085;
tr703:
	{
		s->line_counter++;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);

		// Stop scanner if required.
		if (s->stop == true) {
			return;
		}
	}
	goto st1085;
tr735:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);

		// Stop scanner if required.
		if (s->stop == true) {
			return;
		}
	}
	goto st1085;
tr3213:
	{
		s->line_counter++;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st1085;
st1085:
	if ( ++p == pe )
		goto _test_eof1085;
case 1085:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr3207;
		case 32: goto tr3207;
		case 36: goto tr3208;
		case 40: goto tr3209;
		case 41: goto tr3210;
		case 42: goto tr3211;
		case 58: goto tr67;
		case 65: goto tr3278;
		case 67: goto tr3279;
		case 68: goto tr3280;
		case 69: goto tr3281;
		case 72: goto tr3282;
		case 73: goto tr3283;
		case 75: goto tr3284;
		case 76: goto tr3285;
		case 77: goto tr3286;
		case 78: goto tr3287;
		case 80: goto tr3288;
		case 82: goto tr3289;
		case 83: goto tr3290;
		case 84: goto tr3291;
		case 92: goto tr3212;
		case 95: goto tr3211;
		case 97: goto tr3278;
		case 99: goto tr3279;
		case 100: goto tr3280;
		case 101: goto tr3281;
		case 104: goto tr3282;
		case 105: goto tr3283;
		case 107: goto tr3284;
		case 108: goto tr3285;
		case 109: goto tr3286;
		case 110: goto tr3287;
		case 112: goto tr3288;
		case 114: goto tr3289;
		case 115: goto tr3290;
		case 116: goto tr3291;
		case 1802: goto tr19;
		case 1851: goto st144;
		case 2058: goto tr3213;
		case 2107: goto tr3214;
		case 2314: goto tr115;
		case 2363: goto tr116;
		case 2570: goto tr3215;
		case 2619: goto tr3216;
	}
	if ( _widec < 60 ) {
		if ( _widec < 11 ) {
			if ( _widec <= 8 )
				goto tr67;
		} else if ( _widec > 44 ) {
			if ( _widec > 47 ) {
				if ( 48 <= _widec && _widec <= 57 )
					goto tr3277;
			} else if ( _widec >= 45 )
				goto tr3211;
		} else
			goto tr67;
	} else if ( _widec > 63 ) {
		if ( _widec < 91 ) {
			if ( 64 <= _widec && _widec <= 90 )
				goto tr3211;
		} else if ( _widec > 96 ) {
			if ( _widec > 122 ) {
				if ( 123 <= _widec )
					goto tr67;
			} else if ( _widec >= 98 )
				goto tr3211;
		} else
			goto tr67;
	} else
		goto tr67;
	goto tr3276;
tr3208:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 186; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 186; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 186; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 186; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 186; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 186; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 186; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 186; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 186; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 186; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 186; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 186; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 186; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 186; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 186; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 186; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 186; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 186; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 186; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 186; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 186; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 186; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 186; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 186; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 186; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 186; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 186; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st186;
st186:
	if ( ++p == pe )
		goto _test_eof186;
case 186:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st7;
		case 32: goto st7;
		case 40: goto tr79;
		case 41: goto tr80;
		case 73: goto tr615;
		case 79: goto tr616;
		case 84: goto tr617;
		case 105: goto tr615;
		case 111: goto tr616;
		case 116: goto tr617;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr81;
		case 1083: goto st8;
	}
	goto tr704;
tr616:
	{
		s->stop = true;
	}
	goto st187;
st187:
	if ( ++p == pe )
		goto _test_eof187;
case 187:
	switch( (*p) ) {
		case 82: goto st188;
		case 114: goto st188;
	}
	goto tr614;
st188:
	if ( ++p == pe )
		goto _test_eof188;
case 188:
	switch( (*p) ) {
		case 73: goto st189;
		case 105: goto st189;
	}
	goto tr614;
st189:
	if ( ++p == pe )
		goto _test_eof189;
case 189:
	switch( (*p) ) {
		case 71: goto st190;
		case 103: goto st190;
	}
	goto tr614;
st190:
	if ( ++p == pe )
		goto _test_eof190;
case 190:
	switch( (*p) ) {
		case 73: goto st191;
		case 105: goto st191;
	}
	goto tr614;
st191:
	if ( ++p == pe )
		goto _test_eof191;
case 191:
	switch( (*p) ) {
		case 78: goto st192;
		case 110: goto st192;
	}
	goto tr614;
st192:
	if ( ++p == pe )
		goto _test_eof192;
case 192:
	switch( (*p) ) {
		case 32: goto tr710;
		case 59: goto tr710;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr710;
	} else if ( (*p) >= 9 )
		goto tr710;
	goto tr614;
tr617:
	{
		s->stop = true;
	}
	goto st193;
st193:
	if ( ++p == pe )
		goto _test_eof193;
case 193:
	switch( (*p) ) {
		case 84: goto st194;
		case 116: goto st194;
	}
	goto tr614;
st194:
	if ( ++p == pe )
		goto _test_eof194;
case 194:
	switch( (*p) ) {
		case 76: goto st195;
		case 108: goto st195;
	}
	goto tr614;
st195:
	if ( ++p == pe )
		goto _test_eof195;
case 195:
	switch( (*p) ) {
		case 32: goto tr713;
		case 59: goto tr713;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr713;
	} else if ( (*p) >= 9 )
		goto tr713;
	goto tr614;
tr3211:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 196; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 196; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 196; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 196; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 196; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 196; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 196; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 196; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 196; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 196; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 196; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 196; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 196; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 196; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 196; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 196; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 196; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 196; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 196; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 196; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 196; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 196; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 196; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 196; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 196; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 196; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 196; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 196; goto st255;} }
	goto st196;
st196:
	if ( ++p == pe )
		goto _test_eof196;
case 196:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr715;
		case 32: goto tr715;
		case 40: goto tr716;
		case 41: goto tr717;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr718;
		case 1083: goto tr719;
	}
	goto tr714;
tr3277:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 197; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 197; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 197; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 197; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 197; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 197; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 197; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 197; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 197; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 197; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 197; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 197; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 197; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 197; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 197; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 197; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 197; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 197; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 197; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 197; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 197; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 197; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 197; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 197; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 197; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 197; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 197; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 197; goto st255;} }
	goto st197;
st197:
	if ( ++p == pe )
		goto _test_eof197;
case 197:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr721;
		case 32: goto tr721;
		case 40: goto tr722;
		case 41: goto tr723;
		case 68: goto tr28;
		case 72: goto tr29;
		case 77: goto tr30;
		case 83: goto st159;
		case 87: goto tr32;
		case 100: goto tr28;
		case 104: goto tr29;
		case 109: goto tr30;
		case 115: goto st159;
		case 119: goto tr32;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr724;
		case 1083: goto tr725;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr27;
	goto tr720;
tr727:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st198;
tr728:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st198;
tr721:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st198;
tr722:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st198;
tr723:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st198;
st198:
	if ( ++p == pe )
		goto _test_eof198;
case 198:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st198;
		case 32: goto st198;
		case 40: goto tr727;
		case 41: goto tr728;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 80: goto tr15;
		case 82: goto tr16;
		case 83: goto tr17;
		case 84: goto tr18;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 112: goto tr15;
		case 114: goto tr16;
		case 115: goto tr17;
		case 116: goto tr18;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr729;
		case 1083: goto st199;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr0;
tr725:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st199;
st199:
	if ( ++p == pe )
		goto _test_eof199;
case 199:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr81;
		case 1034: goto tr729;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto st199;
	} else if ( _widec >= 640 )
		goto st8;
	goto tr125;
tr3278:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 200; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 200; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 200; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 200; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 200; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 200; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 200; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 200; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 200; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 200; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 200; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 200; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 200; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 200; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 200; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 200; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 200; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 200; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 200; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 200; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 200; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 200; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 200; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 200; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 200; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 200; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 200; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 200; goto st255;} }
	goto st200;
st200:
	if ( ++p == pe )
		goto _test_eof200;
case 200:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr732;
		case 32: goto tr732;
		case 40: goto tr733;
		case 41: goto tr734;
		case 65: goto st220;
		case 70: goto st223;
		case 80: goto st227;
		case 97: goto st220;
		case 102: goto st223;
		case 112: goto st227;
		case 1802: goto tr81;
		case 1851: goto st8;
		case 2058: goto tr735;
		case 2107: goto tr736;
		case 2314: goto tr682;
		case 2363: goto tr683;
		case 2570: goto tr737;
		case 2619: goto tr738;
	}
	goto tr731;
tr740:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st201;
tr741:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st201;
tr732:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st201;
tr733:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st201;
tr734:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st201;
st201:
	if ( ++p == pe )
		goto _test_eof201;
case 201:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st201;
		case 32: goto st201;
		case 40: goto tr740;
		case 41: goto tr741;
		case 58: goto tr67;
		case 65: goto tr99;
		case 67: goto tr100;
		case 68: goto tr101;
		case 69: goto tr102;
		case 72: goto tr103;
		case 73: goto tr104;
		case 75: goto tr105;
		case 76: goto tr106;
		case 77: goto tr107;
		case 78: goto tr108;
		case 80: goto tr109;
		case 82: goto tr110;
		case 83: goto tr111;
		case 84: goto tr112;
		case 92: goto st9;
		case 97: goto tr99;
		case 99: goto tr100;
		case 100: goto tr101;
		case 101: goto tr102;
		case 104: goto tr103;
		case 105: goto tr104;
		case 107: goto tr105;
		case 108: goto tr106;
		case 109: goto tr107;
		case 110: goto tr108;
		case 112: goto tr109;
		case 114: goto tr110;
		case 115: goto tr111;
		case 116: goto tr112;
		case 1802: goto tr81;
		case 1851: goto st8;
		case 2058: goto tr703;
		case 2107: goto st185;
		case 2314: goto tr90;
		case 2363: goto tr91;
		case 2570: goto tr742;
		case 2619: goto tr702;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr67;
	} else if ( _widec > 47 ) {
		if ( _widec > 57 ) {
			if ( 60 <= _widec )
				goto tr67;
		} else if ( _widec >= 48 )
			goto tr98;
	} else
		goto tr67;
	goto tr95;
tr117:
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1086; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1086; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1086; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1086; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1086; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1086; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1086; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1086; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1086; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1086; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1086; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1086; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1086; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1086; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1086; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1086; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1086; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1086; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1086; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1086; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1086; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1086; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1086; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1086; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1086; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1086; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1086; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st1086;
tr742:
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1086; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1086; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1086; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1086; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1086; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1086; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1086; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1086; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1086; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1086; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1086; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1086; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1086; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1086; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1086; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1086; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1086; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1086; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1086; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1086; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1086; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1086; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1086; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1086; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1086; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1086; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1086; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);

		// Stop scanner if required.
		if (s->stop == true) {
			return;
		}
	}
	goto st1086;
tr737:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1086; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1086; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1086; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1086; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1086; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1086; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1086; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1086; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1086; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1086; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1086; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1086; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1086; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1086; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1086; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1086; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1086; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1086; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1086; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1086; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1086; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1086; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1086; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1086; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1086; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1086; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1086; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		if (rdata_tail - s->r_data > UINT16_MAX) {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
		s->r_data_length = rdata_tail - s->r_data;

		s->process_record(s);

		// Stop scanner if required.
		if (s->stop == true) {
			return;
		}
	}
	goto st1086;
tr3215:
	{
		s->line_counter++;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 1086; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 1086; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 1086; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 1086; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 1086; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 1086; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 1086; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 1086; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 1086; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 1086; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 1086; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 1086; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 1086; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 1086; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 1086; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 1086; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 1086; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 1086; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 1086; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 1086; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 1086; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 1086; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 1086; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 1086; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 1086; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 1086; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 1086; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st1086;
st1086:
	if ( ++p == pe )
		goto _test_eof1086;
case 1086:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr3271;
		case 32: goto tr3271;
		case 36: goto tr3208;
		case 40: goto tr3272;
		case 41: goto tr3273;
		case 42: goto tr3211;
		case 58: goto tr67;
		case 65: goto tr3278;
		case 67: goto tr3279;
		case 68: goto tr3280;
		case 69: goto tr3281;
		case 72: goto tr3282;
		case 73: goto tr3283;
		case 75: goto tr3284;
		case 76: goto tr3285;
		case 77: goto tr3286;
		case 78: goto tr3287;
		case 80: goto tr3288;
		case 82: goto tr3289;
		case 83: goto tr3290;
		case 84: goto tr3291;
		case 92: goto tr3212;
		case 95: goto tr3211;
		case 97: goto tr3278;
		case 99: goto tr3279;
		case 100: goto tr3280;
		case 101: goto tr3281;
		case 104: goto tr3282;
		case 105: goto tr3283;
		case 107: goto tr3284;
		case 108: goto tr3285;
		case 109: goto tr3286;
		case 110: goto tr3287;
		case 112: goto tr3288;
		case 114: goto tr3289;
		case 115: goto tr3290;
		case 116: goto tr3291;
		case 1802: goto tr81;
		case 1851: goto st8;
		case 2058: goto tr3292;
		case 2107: goto tr3274;
		case 2314: goto tr90;
		case 2363: goto tr91;
		case 2570: goto tr3292;
		case 2619: goto tr3275;
	}
	if ( _widec < 60 ) {
		if ( _widec < 11 ) {
			if ( _widec <= 8 )
				goto tr67;
		} else if ( _widec > 44 ) {
			if ( _widec > 47 ) {
				if ( 48 <= _widec && _widec <= 57 )
					goto tr3277;
			} else if ( _widec >= 45 )
				goto tr3211;
		} else
			goto tr67;
	} else if ( _widec > 63 ) {
		if ( _widec < 91 ) {
			if ( 64 <= _widec && _widec <= 90 )
				goto tr3211;
		} else if ( _widec > 96 ) {
			if ( _widec > 122 ) {
				if ( 123 <= _widec )
					goto tr67;
			} else if ( _widec >= 98 )
				goto tr3211;
		} else
			goto tr67;
	} else
		goto tr67;
	goto tr3276;
tr3279:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 202; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 202; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 202; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 202; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 202; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 202; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 202; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 202; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 202; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 202; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 202; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 202; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 202; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 202; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 202; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 202; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 202; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 202; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 202; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 202; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 202; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 202; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 202; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 202; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 202; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 202; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 202; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 202; goto st255;} }
	goto st202;
st202:
	if ( ++p == pe )
		goto _test_eof202;
case 202:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr715;
		case 32: goto tr715;
		case 40: goto tr716;
		case 41: goto tr717;
		case 68: goto st17;
		case 69: goto st25;
		case 78: goto st28;
		case 100: goto st17;
		case 101: goto st25;
		case 110: goto st28;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr718;
		case 1083: goto tr719;
	}
	goto tr731;
tr3280:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 203; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 203; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 203; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 203; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 203; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 203; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 203; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 203; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 203; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 203; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 203; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 203; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 203; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 203; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 203; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 203; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 203; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 203; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 203; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 203; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 203; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 203; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 203; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 203; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 203; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 203; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 203; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 203; goto st255;} }
	goto st203;
st203:
	if ( ++p == pe )
		goto _test_eof203;
case 203:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr715;
		case 32: goto tr715;
		case 40: goto tr716;
		case 41: goto tr717;
		case 72: goto st33;
		case 78: goto st37;
		case 83: goto st45;
		case 104: goto st33;
		case 110: goto st37;
		case 115: goto st45;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr718;
		case 1083: goto tr719;
	}
	goto tr731;
tr3281:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 204; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 204; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 204; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 204; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 204; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 204; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 204; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 204; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 204; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 204; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 204; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 204; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 204; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 204; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 204; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 204; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 204; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 204; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 204; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 204; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 204; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 204; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 204; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 204; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 204; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 204; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 204; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 204; goto st255;} }
	goto st204;
st204:
	if ( ++p == pe )
		goto _test_eof204;
case 204:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr715;
		case 32: goto tr715;
		case 40: goto tr716;
		case 41: goto tr717;
		case 85: goto st47;
		case 117: goto st47;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr718;
		case 1083: goto tr719;
	}
	goto tr731;
tr3282:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 205; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 205; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 205; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 205; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 205; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 205; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 205; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 205; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 205; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 205; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 205; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 205; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 205; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 205; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 205; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 205; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 205; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 205; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 205; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 205; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 205; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 205; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 205; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 205; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 205; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 205; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 205; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 205; goto st255;} }
	goto st205;
st205:
	if ( ++p == pe )
		goto _test_eof205;
case 205:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr715;
		case 32: goto tr715;
		case 40: goto tr716;
		case 41: goto tr717;
		case 73: goto st54;
		case 105: goto st54;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr718;
		case 1083: goto tr719;
	}
	goto tr731;
tr3283:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 206; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 206; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 206; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 206; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 206; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 206; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 206; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 206; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 206; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 206; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 206; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 206; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 206; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 206; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 206; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 206; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 206; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 206; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 206; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 206; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 206; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 206; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 206; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 206; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 206; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 206; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 206; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 206; goto st255;} }
	goto st206;
st206:
	if ( ++p == pe )
		goto _test_eof206;
case 206:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr715;
		case 32: goto tr715;
		case 40: goto tr716;
		case 41: goto tr717;
		case 78: goto st137;
		case 80: goto st62;
		case 110: goto st137;
		case 112: goto st62;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr718;
		case 1083: goto tr719;
	}
	goto tr731;
tr3284:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 207; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 207; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 207; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 207; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 207; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 207; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 207; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 207; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 207; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 207; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 207; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 207; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 207; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 207; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 207; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 207; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 207; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 207; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 207; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 207; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 207; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 207; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 207; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 207; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 207; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 207; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 207; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 207; goto st255;} }
	goto st207;
st207:
	if ( ++p == pe )
		goto _test_eof207;
case 207:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr715;
		case 32: goto tr715;
		case 40: goto tr716;
		case 41: goto tr717;
		case 69: goto st70;
		case 88: goto st72;
		case 101: goto st70;
		case 120: goto st72;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr718;
		case 1083: goto tr719;
	}
	goto tr731;
tr3285:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 208; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 208; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 208; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 208; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 208; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 208; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 208; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 208; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 208; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 208; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 208; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 208; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 208; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 208; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 208; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 208; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 208; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 208; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 208; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 208; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 208; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 208; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 208; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 208; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 208; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 208; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 208; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 208; goto st255;} }
	goto st208;
st208:
	if ( ++p == pe )
		goto _test_eof208;
case 208:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr715;
		case 32: goto tr715;
		case 40: goto tr716;
		case 41: goto tr717;
		case 51: goto st74;
		case 54: goto st76;
		case 79: goto st78;
		case 80: goto st80;
		case 111: goto st78;
		case 112: goto st80;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr718;
		case 1083: goto tr719;
	}
	goto tr731;
tr3286:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 209; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 209; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 209; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 209; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 209; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 209; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 209; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 209; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 209; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 209; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 209; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 209; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 209; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 209; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 209; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 209; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 209; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 209; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 209; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 209; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 209; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 209; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 209; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 209; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 209; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 209; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 209; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 209; goto st255;} }
	goto st209;
st209:
	if ( ++p == pe )
		goto _test_eof209;
case 209:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr715;
		case 32: goto tr715;
		case 40: goto tr716;
		case 41: goto tr717;
		case 73: goto st82;
		case 88: goto st86;
		case 105: goto st82;
		case 120: goto st86;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr718;
		case 1083: goto tr719;
	}
	goto tr731;
tr3287:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 210; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 210; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 210; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 210; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 210; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 210; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 210; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 210; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 210; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 210; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 210; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 210; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 210; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 210; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 210; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 210; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 210; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 210; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 210; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 210; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 210; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 210; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 210; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 210; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 210; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 210; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 210; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 210; goto st255;} }
	goto st210;
st210:
	if ( ++p == pe )
		goto _test_eof210;
case 210:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr715;
		case 32: goto tr715;
		case 40: goto tr716;
		case 41: goto tr717;
		case 65: goto st88;
		case 73: goto st92;
		case 83: goto st94;
		case 97: goto st88;
		case 105: goto st92;
		case 115: goto st94;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr718;
		case 1083: goto tr719;
	}
	goto tr731;
tr3288:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 211; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 211; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 211; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 211; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 211; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 211; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 211; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 211; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 211; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 211; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 211; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 211; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 211; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 211; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 211; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 211; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 211; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 211; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 211; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 211; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 211; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 211; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 211; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 211; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 211; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 211; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 211; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 211; goto st255;} }
	goto st211;
st211:
	if ( ++p == pe )
		goto _test_eof211;
case 211:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr715;
		case 32: goto tr715;
		case 40: goto tr716;
		case 41: goto tr717;
		case 84: goto st104;
		case 116: goto st104;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr718;
		case 1083: goto tr719;
	}
	goto tr731;
tr3289:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 212; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 212; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 212; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 212; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 212; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 212; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 212; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 212; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 212; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 212; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 212; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 212; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 212; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 212; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 212; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 212; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 212; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 212; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 212; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 212; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 212; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 212; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 212; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 212; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 212; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 212; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 212; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 212; goto st255;} }
	goto st212;
st212:
	if ( ++p == pe )
		goto _test_eof212;
case 212:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr715;
		case 32: goto tr715;
		case 40: goto tr716;
		case 41: goto tr717;
		case 80: goto st107;
		case 82: goto st108;
		case 84: goto st112;
		case 112: goto st107;
		case 114: goto st108;
		case 116: goto st112;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr718;
		case 1083: goto tr719;
	}
	goto tr731;
tr3290:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 213; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 213; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 213; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 213; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 213; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 213; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 213; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 213; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 213; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 213; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 213; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 213; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 213; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 213; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 213; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 213; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 213; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 213; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 213; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 213; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 213; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 213; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 213; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 213; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 213; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 213; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 213; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 213; goto st255;} }
	goto st213;
st213:
	if ( ++p == pe )
		goto _test_eof213;
case 213:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr715;
		case 32: goto tr715;
		case 40: goto tr716;
		case 41: goto tr717;
		case 79: goto st114;
		case 80: goto st116;
		case 82: goto st118;
		case 83: goto st120;
		case 111: goto st114;
		case 112: goto st116;
		case 114: goto st118;
		case 115: goto st120;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr718;
		case 1083: goto tr719;
	}
	goto tr731;
tr3291:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 214; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 214; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 214; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 214; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 214; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 214; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 214; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 214; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 214; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 214; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 214; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 214; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 214; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 214; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 214; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 214; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 214; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 214; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 214; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 214; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 214; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 214; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 214; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 214; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 214; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 214; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 214; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 214; goto st255;} }
	goto st214;
st214:
	if ( ++p == pe )
		goto _test_eof214;
case 214:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr715;
		case 32: goto tr715;
		case 40: goto tr716;
		case 41: goto tr717;
		case 76: goto st125;
		case 88: goto st128;
		case 89: goto st130;
		case 108: goto st125;
		case 120: goto st128;
		case 121: goto st130;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr718;
		case 1083: goto tr719;
	}
	goto tr731;
tr3212:
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 215; goto st255;} }
	goto st215;
st215:
	if ( ++p == pe )
		goto _test_eof215;
case 215:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr743;
		case 32: goto tr743;
		case 35: goto tr85;
		case 40: goto tr745;
		case 41: goto tr746;
		case 778: goto tr84;
		case 827: goto tr84;
		case 1034: goto tr747;
		case 1083: goto tr748;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr84;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr84;
	} else
		goto tr84;
	goto tr744;
tr743:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{ p--; }
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 216; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 216; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 216; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 216; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 216; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 216; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 216; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 216; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 216; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 216; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 216; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 216; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 216; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 216; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 216; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 216; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 216; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 216; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 216; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 216; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 216; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 216; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 216; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 216; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 216; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 216; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 216; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st216;
tr745:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	{ p--; }
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 216; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 216; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 216; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 216; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 216; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 216; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 216; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 216; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 216; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 216; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 216; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 216; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 216; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 216; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 216; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 216; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 216; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 216; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 216; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 216; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 216; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 216; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 216; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 216; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 216; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 216; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 216; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st216;
tr746:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	{ p--; }
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 216; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 216; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 216; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 216; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 216; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 216; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 216; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 216; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 216; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 216; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 216; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 216; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 216; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 216; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 216; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 216; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 216; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 216; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 216; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 216; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 216; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 216; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 216; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 216; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 216; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 216; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 216; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st216;
tr747:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{ p--; }
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 216; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 216; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 216; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 216; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 216; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 216; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 216; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 216; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 216; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 216; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 216; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 216; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 216; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 216; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 216; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 216; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 216; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 216; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 216; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 216; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 216; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 216; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 216; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 216; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 216; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 216; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 216; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st216;
st216:
	if ( ++p == pe )
		goto _test_eof216;
case 216:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st167;
		case 32: goto st167;
		case 40: goto tr691;
		case 41: goto tr692;
		case 65: goto tr5;
		case 67: goto tr6;
		case 68: goto tr7;
		case 69: goto tr8;
		case 72: goto tr9;
		case 73: goto tr10;
		case 75: goto tr11;
		case 76: goto tr12;
		case 77: goto tr13;
		case 78: goto tr14;
		case 80: goto tr15;
		case 82: goto tr16;
		case 83: goto tr17;
		case 84: goto tr18;
		case 97: goto tr5;
		case 99: goto tr6;
		case 100: goto tr7;
		case 101: goto tr8;
		case 104: goto tr9;
		case 105: goto tr10;
		case 107: goto tr11;
		case 108: goto tr12;
		case 109: goto tr13;
		case 110: goto tr14;
		case 112: goto tr15;
		case 114: goto tr16;
		case 115: goto tr17;
		case 116: goto tr18;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr693;
		case 1083: goto st168;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr4;
	goto tr95;
tr748:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{ p--; }
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 217; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 217; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 217; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 217; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 217; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 217; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 217; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 217; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 217; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 217; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 217; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 217; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 217; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 217; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 217; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 217; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 217; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 217; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 217; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 217; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 217; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 217; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 217; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 217; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 217; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 217; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 217; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st217;
st217:
	if ( ++p == pe )
		goto _test_eof217;
case 217:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto st7;
		case 778: goto tr81;
		case 800: goto st7;
		case 808: goto tr79;
		case 809: goto tr80;
		case 827: goto st8;
		case 1033: goto st218;
		case 1034: goto tr693;
		case 1056: goto st218;
		case 1064: goto tr750;
		case 1065: goto tr751;
		case 1083: goto st168;
	}
	if ( 896 <= _widec && _widec <= 1151 )
		goto st155;
	goto tr77;
tr750:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st218;
tr751:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st218;
st218:
	if ( ++p == pe )
		goto _test_eof218;
case 218:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto st7;
		case 778: goto tr81;
		case 800: goto st7;
		case 808: goto tr79;
		case 809: goto tr80;
		case 827: goto st8;
		case 1033: goto st218;
		case 1034: goto tr693;
		case 1056: goto st218;
		case 1064: goto tr750;
		case 1065: goto tr751;
		case 1083: goto st168;
	}
	if ( 896 <= _widec && _widec <= 1151 )
		goto st155;
	goto tr83;
tr753:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st219;
tr754:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st219;
tr702:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 219; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 219; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 219; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 219; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 219; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 219; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 219; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 219; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 219; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 219; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 219; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 219; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 219; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 219; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 219; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 219; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 219; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 219; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 219; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 219; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 219; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 219; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 219; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 219; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 219; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 219; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 219; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st219;
tr738:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 219; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 219; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 219; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 219; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 219; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 219; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 219; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 219; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 219; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 219; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 219; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 219; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 219; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 219; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 219; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 219; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 219; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 219; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 219; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 219; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 219; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 219; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 219; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 219; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 219; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 219; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 219; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st219;
tr3275:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 219; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 219; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 219; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 219; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 219; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 219; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 219; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 219; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 219; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 219; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 219; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 219; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 219; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 219; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 219; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 219; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 219; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 219; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 219; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 219; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 219; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 219; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 219; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 219; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 219; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 219; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 219; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st219;
st219:
	if ( ++p == pe )
		goto _test_eof219;
case 219:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto st170;
		case 778: goto tr81;
		case 800: goto st170;
		case 808: goto tr696;
		case 809: goto tr697;
		case 1033: goto st219;
		case 1034: goto tr693;
		case 1056: goto st219;
		case 1064: goto tr753;
		case 1065: goto tr754;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto st185;
	} else if ( _widec >= 640 )
		goto st8;
	goto tr77;
st220:
	if ( ++p == pe )
		goto _test_eof220;
case 220:
	switch( (*p) ) {
		case 65: goto st221;
		case 97: goto st221;
	}
	goto tr35;
st221:
	if ( ++p == pe )
		goto _test_eof221;
case 221:
	switch( (*p) ) {
		case 65: goto st222;
		case 97: goto st222;
	}
	goto tr35;
st222:
	if ( ++p == pe )
		goto _test_eof222;
case 222:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr757;
		case 32: goto tr757;
		case 40: goto tr758;
		case 41: goto tr759;
		case 2058: goto tr760;
		case 2107: goto tr761;
		case 2314: goto tr762;
		case 2363: goto tr762;
		case 2570: goto tr763;
		case 2619: goto tr764;
	}
	goto tr55;
st223:
	if ( ++p == pe )
		goto _test_eof223;
case 223:
	switch( (*p) ) {
		case 83: goto st224;
		case 115: goto st224;
	}
	goto tr35;
st224:
	if ( ++p == pe )
		goto _test_eof224;
case 224:
	switch( (*p) ) {
		case 68: goto st225;
		case 100: goto st225;
	}
	goto tr35;
st225:
	if ( ++p == pe )
		goto _test_eof225;
case 225:
	switch( (*p) ) {
		case 66: goto st226;
		case 98: goto st226;
	}
	goto tr35;
st226:
	if ( ++p == pe )
		goto _test_eof226;
case 226:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr768;
		case 32: goto tr768;
		case 40: goto tr769;
		case 41: goto tr770;
		case 2058: goto tr771;
		case 2107: goto tr772;
		case 2314: goto tr773;
		case 2363: goto tr773;
		case 2570: goto tr774;
		case 2619: goto tr775;
	}
	goto tr55;
st227:
	if ( ++p == pe )
		goto _test_eof227;
case 227:
	switch( (*p) ) {
		case 76: goto st228;
		case 108: goto st228;
	}
	goto tr35;
st228:
	if ( ++p == pe )
		goto _test_eof228;
case 228:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr777;
		case 32: goto tr777;
		case 40: goto tr778;
		case 41: goto tr779;
		case 2058: goto tr780;
		case 2107: goto tr781;
		case 2314: goto tr782;
		case 2363: goto tr782;
		case 2570: goto tr783;
		case 2619: goto tr784;
	}
	goto tr55;
tr3214:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st229;
st229:
	if ( ++p == pe )
		goto _test_eof229;
case 229:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr19;
		case 1034: goto tr113;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto st229;
	} else if ( _widec >= 640 )
		goto st144;
	goto tr77;
tr116:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 230; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 230; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 230; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 230; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 230; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 230; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 230; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 230; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 230; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 230; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 230; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 230; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 230; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 230; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 230; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 230; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 230; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 230; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 230; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 230; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 230; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 230; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 230; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 230; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 230; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 230; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 230; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st230;
st230:
	if ( ++p == pe )
		goto _test_eof230;
case 230:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st231;
		case 32: goto st231;
		case 40: goto tr786;
		case 41: goto tr787;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr81;
		case 1083: goto st8;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st144;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto st144;
	} else
		goto st144;
	goto tr77;
tr786:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st231;
tr787:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st231;
st231:
	if ( ++p == pe )
		goto _test_eof231;
case 231:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st231;
		case 32: goto st231;
		case 40: goto tr786;
		case 41: goto tr787;
		case 778: goto tr81;
		case 827: goto st8;
		case 1034: goto tr81;
		case 1083: goto st8;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto st144;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto st144;
	} else
		goto st144;
	goto tr83;
tr789:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st232;
tr790:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st232;
tr118:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 232; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 232; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 232; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 232; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 232; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 232; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 232; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 232; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 232; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 232; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 232; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 232; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 232; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 232; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 232; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 232; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 232; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 232; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 232; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 232; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 232; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 232; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 232; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 232; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 232; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 232; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 232; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st232;
tr3216:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 232; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 232; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 232; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 232; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 232; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 232; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 232; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 232; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 232; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 232; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 232; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 232; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 232; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 232; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 232; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 232; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 232; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 232; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 232; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 232; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 232; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 232; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 232; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 232; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 232; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 232; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 232; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st232;
st232:
	if ( ++p == pe )
		goto _test_eof232;
case 232:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto st231;
		case 778: goto tr81;
		case 800: goto st231;
		case 808: goto tr786;
		case 809: goto tr787;
		case 827: goto st8;
		case 1033: goto st232;
		case 1034: goto tr693;
		case 1056: goto st232;
		case 1064: goto tr789;
		case 1065: goto tr790;
		case 1083: goto st185;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto st229;
	} else if ( _widec >= 640 )
		goto st144;
	goto tr77;
tr792:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st233;
tr793:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st233;
tr685:
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 233; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 233; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 233; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 233; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 233; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 233; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 233; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 233; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 233; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 233; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 233; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 233; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 233; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 233; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 233; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 233; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 233; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 233; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 233; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 233; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 233; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 233; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 233; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 233; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 233; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 233; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 233; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st233;
tr92:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 233; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 233; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 233; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 233; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 233; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 233; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 233; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 233; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 233; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 233; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 233; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 233; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 233; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 233; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 233; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 233; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 233; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 233; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 233; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 233; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 233; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 233; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 233; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 233; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 233; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 233; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 233; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st233;
st233:
	if ( ++p == pe )
		goto _test_eof233;
case 233:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto st170;
		case 778: goto tr81;
		case 800: goto st170;
		case 808: goto tr696;
		case 809: goto tr697;
		case 1033: goto st233;
		case 1034: goto tr81;
		case 1056: goto st233;
		case 1064: goto tr792;
		case 1065: goto tr793;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto st12;
	} else if ( _widec >= 640 )
		goto st8;
	goto tr77;
tr666:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	goto st234;
st234:
	if ( ++p == pe )
		goto _test_eof234;
case 234:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr672;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st234;
	goto tr69;
tr795:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st235;
tr796:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st235;
tr675:
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 235; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 235; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 235; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 235; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 235; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 235; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 235; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 235; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 235; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 235; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 235; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 235; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 235; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 235; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 235; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 235; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 235; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 235; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 235; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 235; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 235; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 235; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 235; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 235; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 235; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 235; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 235; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st235;
tr668:
	{
		s->r_owner_length = s->dname_tmp_length;
	}
	{ s->r_type = KNOT_RRTYPE_A; }
	{
		rdata_tail = s->r_data;
	}
	{
		p--;
		switch (s->r_type) {
		case KNOT_RRTYPE_A:
			{stack[top++] = 235; goto st603;}
		case KNOT_RRTYPE_NS:
		case KNOT_RRTYPE_CNAME:
		case KNOT_RRTYPE_PTR:
		case KNOT_RRTYPE_DNAME:
			{stack[top++] = 235; goto st605;}
		case KNOT_RRTYPE_SOA:
			{stack[top++] = 235; goto st607;}
		case KNOT_RRTYPE_HINFO:
			{stack[top++] = 235; goto st639;}
		case KNOT_RRTYPE_MINFO:
		case KNOT_RRTYPE_RP:
			{stack[top++] = 235; goto st644;}
		case KNOT_RRTYPE_MX:
		case KNOT_RRTYPE_AFSDB:
		case KNOT_RRTYPE_RT:
		case KNOT_RRTYPE_KX:
		case KNOT_RRTYPE_LP:
			{stack[top++] = 235; goto st649;}
		case KNOT_RRTYPE_TXT:
		case KNOT_RRTYPE_SPF:
			{stack[top++] = 235; goto st654;}
		case KNOT_RRTYPE_AAAA:
			{stack[top++] = 235; goto st658;}
		case KNOT_RRTYPE_LOC:
			{stack[top++] = 235; goto st660;}
		case KNOT_RRTYPE_SRV:
			{stack[top++] = 235; goto st715;}
		case KNOT_RRTYPE_NAPTR:
			{stack[top++] = 235; goto st726;}
		case KNOT_RRTYPE_CERT:
			{stack[top++] = 235; goto st743;}
		case KNOT_RRTYPE_APL:
			{stack[top++] = 235; goto st754;}
		case KNOT_RRTYPE_DS:
		case KNOT_RRTYPE_CDS:
			{stack[top++] = 235; goto st765;}
		case KNOT_RRTYPE_SSHFP:
			{stack[top++] = 235; goto st778;}
		case KNOT_RRTYPE_IPSECKEY:
			{stack[top++] = 235; goto st788;}
		case KNOT_RRTYPE_RRSIG:
			{stack[top++] = 235; goto st827;}
		case KNOT_RRTYPE_NSEC:
			{stack[top++] = 235; goto st976;}
		case KNOT_RRTYPE_KEY:
		case KNOT_RRTYPE_DNSKEY:
		case KNOT_RRTYPE_CDNSKEY:
			{stack[top++] = 235; goto st979;}
		case KNOT_RRTYPE_DHCID:
			{stack[top++] = 235; goto st990;}
		case KNOT_RRTYPE_NSEC3:
			{stack[top++] = 235; goto st992;}
		case KNOT_RRTYPE_NSEC3PARAM:
			{stack[top++] = 235; goto st1021;}
		case KNOT_RRTYPE_TLSA:
			{stack[top++] = 235; goto st1034;}
		case KNOT_RRTYPE_NID:
		case KNOT_RRTYPE_L64:
			{stack[top++] = 235; goto st1052;}
		case KNOT_RRTYPE_L32:
			{stack[top++] = 235; goto st1047;}
		case KNOT_RRTYPE_EUI48:
			{stack[top++] = 235; goto st1065;}
		case KNOT_RRTYPE_EUI64:
			{stack[top++] = 235; goto st1071;}
		default:
			WARN(ZS_CANNOT_TEXT_DATA);
			{goto st253;}
		}
	}
	goto st235;
st235:
	if ( ++p == pe )
		goto _test_eof235;
case 235:
	_widec = (*p);
	if ( (*p) < 11 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(640 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
			}
		} else {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 58 ) {
		if ( (*p) > 59 ) {
			if ( 60 <= (*p) )
 {				_widec = (short)(640 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
			}
		} else if ( (*p) >= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 777: goto st7;
		case 778: goto tr81;
		case 800: goto st7;
		case 808: goto tr79;
		case 809: goto tr80;
		case 827: goto st8;
		case 1033: goto st235;
		case 1034: goto tr693;
		case 1056: goto st235;
		case 1064: goto tr795;
		case 1065: goto tr796;
		case 1083: goto st185;
	}
	if ( 896 <= _widec && _widec <= 1151 )
		goto st234;
	goto tr77;
tr3222:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 236; goto st255;} }
	goto st236;
tr3240:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 236; goto st255;} }
	goto st236;
st236:
	if ( ++p == pe )
		goto _test_eof236;
case 236:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr626;
		case 32: goto tr626;
		case 40: goto tr627;
		case 41: goto tr628;
		case 68: goto st17;
		case 69: goto st25;
		case 78: goto st28;
		case 100: goto st17;
		case 101: goto st25;
		case 110: goto st28;
		case 1034: goto tr629;
		case 1083: goto tr630;
	}
	goto tr797;
tr3223:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 237; goto st255;} }
	goto st237;
tr3241:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 237; goto st255;} }
	goto st237;
st237:
	if ( ++p == pe )
		goto _test_eof237;
case 237:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr626;
		case 32: goto tr626;
		case 40: goto tr627;
		case 41: goto tr628;
		case 72: goto st33;
		case 78: goto st37;
		case 83: goto st45;
		case 104: goto st33;
		case 110: goto st37;
		case 115: goto st45;
		case 1034: goto tr629;
		case 1083: goto tr630;
	}
	goto tr797;
tr3224:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 238; goto st255;} }
	goto st238;
tr3242:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 238; goto st255;} }
	goto st238;
st238:
	if ( ++p == pe )
		goto _test_eof238;
case 238:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr626;
		case 32: goto tr626;
		case 40: goto tr627;
		case 41: goto tr628;
		case 85: goto st47;
		case 117: goto st47;
		case 1034: goto tr629;
		case 1083: goto tr630;
	}
	goto tr797;
tr3225:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 239; goto st255;} }
	goto st239;
tr3243:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 239; goto st255;} }
	goto st239;
st239:
	if ( ++p == pe )
		goto _test_eof239;
case 239:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr626;
		case 32: goto tr626;
		case 40: goto tr627;
		case 41: goto tr628;
		case 73: goto st54;
		case 105: goto st54;
		case 1034: goto tr629;
		case 1083: goto tr630;
	}
	goto tr797;
tr3244:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 240; goto st255;} }
	goto st240;
st240:
	if ( ++p == pe )
		goto _test_eof240;
case 240:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr626;
		case 32: goto tr626;
		case 40: goto tr627;
		case 41: goto tr628;
		case 78: goto st137;
		case 80: goto st62;
		case 110: goto st137;
		case 112: goto st62;
		case 1034: goto tr629;
		case 1083: goto tr630;
	}
	goto tr797;
tr3227:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 241; goto st255;} }
	goto st241;
tr3245:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 241; goto st255;} }
	goto st241;
st241:
	if ( ++p == pe )
		goto _test_eof241;
case 241:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr626;
		case 32: goto tr626;
		case 40: goto tr627;
		case 41: goto tr628;
		case 69: goto st70;
		case 88: goto st72;
		case 101: goto st70;
		case 120: goto st72;
		case 1034: goto tr629;
		case 1083: goto tr630;
	}
	goto tr797;
tr3228:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 242; goto st255;} }
	goto st242;
tr3246:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 242; goto st255;} }
	goto st242;
st242:
	if ( ++p == pe )
		goto _test_eof242;
case 242:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr626;
		case 32: goto tr626;
		case 40: goto tr627;
		case 41: goto tr628;
		case 51: goto st74;
		case 54: goto st76;
		case 79: goto st78;
		case 80: goto st80;
		case 111: goto st78;
		case 112: goto st80;
		case 1034: goto tr629;
		case 1083: goto tr630;
	}
	goto tr797;
tr3229:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 243; goto st255;} }
	goto st243;
tr3247:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 243; goto st255;} }
	goto st243;
st243:
	if ( ++p == pe )
		goto _test_eof243;
case 243:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr626;
		case 32: goto tr626;
		case 40: goto tr627;
		case 41: goto tr628;
		case 73: goto st82;
		case 88: goto st86;
		case 105: goto st82;
		case 120: goto st86;
		case 1034: goto tr629;
		case 1083: goto tr630;
	}
	goto tr797;
tr3230:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 244; goto st255;} }
	goto st244;
tr3248:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 244; goto st255;} }
	goto st244;
st244:
	if ( ++p == pe )
		goto _test_eof244;
case 244:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr626;
		case 32: goto tr626;
		case 40: goto tr627;
		case 41: goto tr628;
		case 65: goto st88;
		case 73: goto st92;
		case 83: goto st94;
		case 97: goto st88;
		case 105: goto st92;
		case 115: goto st94;
		case 1034: goto tr629;
		case 1083: goto tr630;
	}
	goto tr797;
tr3231:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 245; goto st255;} }
	goto st245;
tr3249:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 245; goto st255;} }
	goto st245;
st245:
	if ( ++p == pe )
		goto _test_eof245;
case 245:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr626;
		case 32: goto tr626;
		case 40: goto tr627;
		case 41: goto tr628;
		case 84: goto st104;
		case 116: goto st104;
		case 1034: goto tr629;
		case 1083: goto tr630;
	}
	goto tr797;
tr3232:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 246; goto st255;} }
	goto st246;
tr3250:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 246; goto st255;} }
	goto st246;
st246:
	if ( ++p == pe )
		goto _test_eof246;
case 246:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr626;
		case 32: goto tr626;
		case 40: goto tr627;
		case 41: goto tr628;
		case 80: goto st107;
		case 82: goto st108;
		case 84: goto st112;
		case 112: goto st107;
		case 114: goto st108;
		case 116: goto st112;
		case 1034: goto tr629;
		case 1083: goto tr630;
	}
	goto tr797;
tr3233:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 247; goto st255;} }
	goto st247;
tr3251:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 247; goto st255;} }
	goto st247;
st247:
	if ( ++p == pe )
		goto _test_eof247;
case 247:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr626;
		case 32: goto tr626;
		case 40: goto tr627;
		case 41: goto tr628;
		case 79: goto st114;
		case 80: goto st116;
		case 82: goto st118;
		case 83: goto st120;
		case 111: goto st114;
		case 112: goto st116;
		case 114: goto st118;
		case 115: goto st120;
		case 1034: goto tr629;
		case 1083: goto tr630;
	}
	goto tr797;
tr3234:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 248; goto st255;} }
	goto st248;
tr3252:
	{
		s->r_class = s->default_class;
	}
	{
		s->r_ttl = s->default_ttl;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 248; goto st255;} }
	goto st248;
st248:
	if ( ++p == pe )
		goto _test_eof248;
case 248:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr626;
		case 32: goto tr626;
		case 40: goto tr627;
		case 41: goto tr628;
		case 76: goto st125;
		case 88: goto st128;
		case 89: goto st130;
		case 108: goto st125;
		case 120: goto st128;
		case 121: goto st130;
		case 1034: goto tr629;
		case 1083: goto tr630;
	}
	goto tr797;
tr3206:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st249;
tr3262:
	{
		s->stop = false;
	}
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st249;
st249:
	if ( ++p == pe )
		goto _test_eof249;
case 249:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr19;
		case 1034: goto tr21;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto st249;
	} else if ( _widec >= 640 )
		goto st144;
	goto tr83;
tr3236:
	{
		if (s->r_owner_length == 0) {
			WARN(ZS_BAD_PREVIOUS_OWNER);
			p--; {goto st253;}
		}
	}
	goto st250;
st250:
	if ( ++p == pe )
		goto _test_eof250;
case 250:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr19;
		case 1034: goto tr562;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto st250;
	} else if ( _widec >= 640 )
		goto st144;
	goto tr125;
tr3226:
	{
		s->r_class = s->default_class;
	}
	{
		s->dname = s->r_owner;
		s->r_owner_length = 0;
	}
	{ p--; {stack[top++] = 251; goto st255;} }
	goto st251;
st251:
	if ( ++p == pe )
		goto _test_eof251;
case 251:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr626;
		case 32: goto tr626;
		case 40: goto tr627;
		case 41: goto tr628;
		case 78: goto st59;
		case 80: goto st62;
		case 110: goto st59;
		case 112: goto st62;
		case 1034: goto tr629;
		case 1083: goto tr630;
	}
	goto tr797;
tr124:
	{
		if (s->number64 <= UINT32_MAX) {
			s->r_ttl = (uint32_t)(s->number64);
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st252;
st252:
	if ( ++p == pe )
		goto _test_eof252;
case 252:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr81;
		case 1034: goto tr129;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto st252;
	} else if ( _widec >= 640 )
		goto st8;
	goto tr125;
st253:
	if ( ++p == pe )
		goto _test_eof253;
case 253:
	if ( (*p) == 10 )
		goto tr799;
	goto tr798;
tr798:
	{
		s->buffer_length = 0;
	}
	{
		if ((*p) == '\r') {
			ERR(ZS_DOS_NEWLINE);
		}

		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st254;
tr800:
	{
		if ((*p) == '\r') {
			ERR(ZS_DOS_NEWLINE);
		}

		if (s->buffer_length < sizeof(s->buffer) - 1) {
			s->buffer[s->buffer_length++] = (*p);
		}
	}
	goto st254;
st254:
	if ( ++p == pe )
		goto _test_eof254;
case 254:
	if ( (*p) == 10 )
		goto tr801;
	goto tr800;
tr799:
	{
		s->buffer_length = 0;
	}
	{
		// Ending string in buffer.
		s->buffer[s->buffer_length++] = 0;

		// Error counter incrementation.
		s->error_counter++;

		// Initialization of fcall stack.
		top = 0;

		// Process error message.
		s->process_error(s);

		// Reset.
		s->error_code = ZS_OK;
		s->multiline = false;

		// In case of serious error, stop scanner.
		if (s->stop == true) {
			return;
		}
	}
	{
		s->line_counter++;
	}
	{ {goto st1077;} }
	goto st1087;
tr801:
	{
		// Ending string in buffer.
		s->buffer[s->buffer_length++] = 0;

		// Error counter incrementation.
		s->error_counter++;

		// Initialization of fcall stack.
		top = 0;

		// Process error message.
		s->process_error(s);

		// Reset.
		s->error_code = ZS_OK;
		s->multiline = false;

		// In case of serious error, stop scanner.
		if (s->stop == true) {
			return;
		}
	}
	{
		s->line_counter++;
	}
	{ {goto st1077;} }
	goto st1087;
st1087:
	if ( ++p == pe )
		goto _test_eof1087;
case 1087:
	goto st0;
st255:
	if ( ++p == pe )
		goto _test_eof255;
case 255:
	switch( (*p) ) {
		case 42: goto tr803;
		case 46: goto tr804;
		case 64: goto st263;
		case 92: goto tr806;
		case 95: goto tr803;
	}
	if ( (*p) < 65 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr803;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr803;
	} else
		goto tr803;
	goto tr802;
tr803:
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st256;
tr808:
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st256;
tr812:
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st256;
tr819:
	{
		s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st256;
st256:
	if ( ++p == pe )
		goto _test_eof256;
case 256:
	switch( (*p) ) {
		case 32: goto tr807;
		case 42: goto tr808;
		case 46: goto tr809;
		case 59: goto tr807;
		case 92: goto st258;
		case 95: goto tr808;
	}
	if ( (*p) < 45 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr807;
		} else if ( (*p) >= 9 )
			goto tr807;
	} else if ( (*p) > 57 ) {
		if ( (*p) > 90 ) {
			if ( 97 <= (*p) && (*p) <= 122 )
				goto tr808;
		} else if ( (*p) >= 65 )
			goto tr808;
	} else
		goto tr808;
	goto tr802;
tr807:
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		// Check for (relative + origin) dname length overflow.
		if (s->dname_tmp_length + s->zone_origin_length <= MAX_DNAME_LENGTH) {
			memcpy(s->dname + s->dname_tmp_length,
			       s->zone_origin,
			       s->zone_origin_length);

			s->dname_tmp_length += s->zone_origin_length;
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1088;
tr811:
	{
		// Enough room for the terminal label is garanteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1088;
tr818:
	{
		s->dname_tmp_length++;
	}
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		// Check for (relative + origin) dname length overflow.
		if (s->dname_tmp_length + s->zone_origin_length <= MAX_DNAME_LENGTH) {
			memcpy(s->dname + s->dname_tmp_length,
			       s->zone_origin,
			       s->zone_origin_length);

			s->dname_tmp_length += s->zone_origin_length;
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1088;
tr822:
	{
		// Copy already verified zone origin.
		memcpy(s->dname,
		       s->zone_origin,
		       s->zone_origin_length);

		s->dname_tmp_length = s->zone_origin_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1088;
st1088:
	if ( ++p == pe )
		goto _test_eof1088;
case 1088:
	goto st0;
tr809:
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st257;
tr820:
	{
		s->dname_tmp_length++;
	}
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st257;
st257:
	if ( ++p == pe )
		goto _test_eof257;
case 257:
	switch( (*p) ) {
		case 32: goto tr811;
		case 42: goto tr812;
		case 45: goto tr812;
		case 59: goto tr811;
		case 92: goto tr813;
		case 95: goto tr812;
	}
	if ( (*p) < 47 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr811;
		} else if ( (*p) >= 9 )
			goto tr811;
	} else if ( (*p) > 57 ) {
		if ( (*p) > 90 ) {
			if ( 97 <= (*p) && (*p) <= 122 )
				goto tr812;
		} else if ( (*p) >= 65 )
			goto tr812;
	} else
		goto tr812;
	goto tr802;
tr806:
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st258;
tr813:
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st258;
tr821:
	{
		s->dname_tmp_length++;
	}
	goto st258;
st258:
	if ( ++p == pe )
		goto _test_eof258;
case 258:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr814;
	goto tr808;
tr814:
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length] = 0;
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st259;
st259:
	if ( ++p == pe )
		goto _test_eof259;
case 259:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr816;
	goto tr815;
tr816:
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st260;
st260:
	if ( ++p == pe )
		goto _test_eof260;
case 260:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr817;
	goto tr815;
tr817:
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st261;
st261:
	if ( ++p == pe )
		goto _test_eof261;
case 261:
	switch( (*p) ) {
		case 32: goto tr818;
		case 42: goto tr819;
		case 46: goto tr820;
		case 59: goto tr818;
		case 92: goto tr821;
		case 95: goto tr819;
	}
	if ( (*p) < 45 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr818;
		} else if ( (*p) >= 9 )
			goto tr818;
	} else if ( (*p) > 57 ) {
		if ( (*p) > 90 ) {
			if ( 97 <= (*p) && (*p) <= 122 )
				goto tr819;
		} else if ( (*p) >= 65 )
			goto tr819;
	} else
		goto tr819;
	goto tr815;
tr804:
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	goto st262;
st262:
	if ( ++p == pe )
		goto _test_eof262;
case 262:
	switch( (*p) ) {
		case 32: goto tr811;
		case 59: goto tr811;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr811;
	} else if ( (*p) >= 9 )
		goto tr811;
	goto tr802;
st263:
	if ( ++p == pe )
		goto _test_eof263;
case 263:
	switch( (*p) ) {
		case 32: goto tr822;
		case 59: goto tr822;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr822;
	} else if ( (*p) >= 9 )
		goto tr822;
	goto tr802;
st264:
	if ( ++p == pe )
		goto _test_eof264;
case 264:
	switch( (*p) ) {
		case 34: goto st270;
		case 92: goto st266;
	}
	if ( (*p) > 58 ) {
		if ( 60 <= (*p) && (*p) <= 126 )
			goto tr824;
	} else if ( (*p) >= 33 )
		goto tr824;
	goto tr823;
tr824:
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;
			}

			*(rdata_tail++) = (*p);
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st265;
tr834:
	{
		rdata_tail++;
	}
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;
			}

			*(rdata_tail++) = (*p);
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st265;
st265:
	if ( ++p == pe )
		goto _test_eof265;
case 265:
	switch( (*p) ) {
		case 32: goto tr827;
		case 33: goto tr824;
		case 59: goto tr827;
		case 92: goto st266;
	}
	if ( (*p) < 35 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr827;
	} else if ( (*p) > 39 ) {
		if ( (*p) > 41 ) {
			if ( 42 <= (*p) && (*p) <= 126 )
				goto tr824;
		} else if ( (*p) >= 40 )
			goto tr828;
	} else
		goto tr824;
	goto tr823;
tr827:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1089;
tr833:
	{
		rdata_tail++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1089;
st1089:
	if ( ++p == pe )
		goto _test_eof1089;
case 1089:
	goto st0;
tr828:
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;
			}

			*(rdata_tail++) = (*p);
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1090;
tr835:
	{
		rdata_tail++;
	}
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;
			}

			*(rdata_tail++) = (*p);
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1090;
st1090:
	if ( ++p == pe )
		goto _test_eof1090;
case 1090:
	switch( (*p) ) {
		case 32: goto tr827;
		case 33: goto tr824;
		case 59: goto tr827;
		case 92: goto st266;
	}
	if ( (*p) < 35 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr827;
	} else if ( (*p) > 39 ) {
		if ( (*p) > 41 ) {
			if ( 42 <= (*p) && (*p) <= 126 )
				goto tr824;
		} else if ( (*p) >= 40 )
			goto tr828;
	} else
		goto tr824;
	goto tr823;
tr836:
	{
		rdata_tail++;
	}
	goto st266;
st266:
	if ( ++p == pe )
		goto _test_eof266;
case 266:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr829;
	goto tr824;
tr829:
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;
			}

			*rdata_tail = 0;
			s->item_length++;
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + '0')
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st267;
st267:
	if ( ++p == pe )
		goto _test_eof267;
case 267:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr831;
	goto tr830;
tr831:
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + '0')
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st268;
st268:
	if ( ++p == pe )
		goto _test_eof268;
case 268:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr832;
	goto tr830;
tr832:
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + '0')
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st269;
st269:
	if ( ++p == pe )
		goto _test_eof269;
case 269:
	switch( (*p) ) {
		case 32: goto tr833;
		case 33: goto tr834;
		case 59: goto tr833;
		case 92: goto tr836;
	}
	if ( (*p) < 35 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr833;
	} else if ( (*p) > 39 ) {
		if ( (*p) > 41 ) {
			if ( 42 <= (*p) && (*p) <= 126 )
				goto tr834;
		} else if ( (*p) >= 40 )
			goto tr835;
	} else
		goto tr834;
	goto tr830;
tr837:
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;
			}

			*(rdata_tail++) = (*p);
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st270;
tr844:
	{
		rdata_tail++;
	}
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;
			}

			*(rdata_tail++) = (*p);
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st270;
st270:
	if ( ++p == pe )
		goto _test_eof270;
case 270:
	_widec = (*p);
	if ( 10 <= (*p) && (*p) <= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr837;
		case 34: goto st271;
		case 92: goto st272;
		case 522: goto tr837;
	}
	if ( 32 <= _widec && _widec <= 126 )
		goto tr837;
	goto tr823;
tr845:
	{
		rdata_tail++;
	}
	goto st271;
st271:
	if ( ++p == pe )
		goto _test_eof271;
case 271:
	switch( (*p) ) {
		case 32: goto tr827;
		case 59: goto tr827;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr827;
	} else if ( (*p) >= 9 )
		goto tr827;
	goto tr840;
tr846:
	{
		rdata_tail++;
	}
	goto st272;
st272:
	if ( ++p == pe )
		goto _test_eof272;
case 272:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr841;
	goto tr837;
tr841:
	{
		if (rdata_tail <= rdata_stop) {
			// Split long string.
			if (s->long_string &&
			    rdata_tail - s->item_length_location == 1 + MAX_ITEM_LENGTH) {
				// _item_length_exit equivalent.
				*(s->item_length_location) = MAX_ITEM_LENGTH;
				// _item_length_init equivalent.
				s->item_length_location = rdata_tail++;
			}

			*rdata_tail = 0;
			s->item_length++;
		} else {
			WARN(ZS_TEXT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + '0')
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st273;
st273:
	if ( ++p == pe )
		goto _test_eof273;
case 273:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr842;
	goto tr830;
tr842:
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + '0')
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st274;
st274:
	if ( ++p == pe )
		goto _test_eof274;
case 274:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr843;
	goto tr830;
tr843:
	{
		if ((*rdata_tail < (UINT8_MAX / 10)) ||   // Dominant fast check.
			((*rdata_tail == (UINT8_MAX / 10)) && // Marginal case.
			 ((*p) <= (UINT8_MAX % 10) + '0')
			)
		   ) {
			*rdata_tail *= 10;
			*rdata_tail += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st275;
st275:
	if ( ++p == pe )
		goto _test_eof275;
case 275:
	_widec = (*p);
	if ( 10 <= (*p) && (*p) <= 10 ) {
		_widec = (short)(128 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr844;
		case 34: goto tr845;
		case 92: goto tr846;
		case 522: goto tr844;
	}
	if ( 32 <= _widec && _widec <= 126 )
		goto tr844;
	goto tr830;
st276:
	if ( ++p == pe )
		goto _test_eof276;
case 276:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st277;
		case 32: goto st277;
		case 40: goto tr849;
		case 41: goto tr850;
		case 1034: goto tr851;
		case 1083: goto st284;
	}
	goto tr847;
tr849:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st277;
tr850:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st277;
tr851:
	{
		s->line_counter++;
	}
	goto st277;
st277:
	if ( ++p == pe )
		goto _test_eof277;
case 277:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st277;
		case 32: goto st277;
		case 40: goto tr849;
		case 41: goto tr850;
		case 1034: goto tr851;
		case 1083: goto st284;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr854;
	goto tr853;
tr854:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st278;
tr859:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st278;
st278:
	if ( ++p == pe )
		goto _test_eof278;
case 278:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st279;
		case 32: goto st279;
		case 40: goto tr857;
		case 41: goto tr858;
		case 68: goto tr860;
		case 72: goto tr861;
		case 77: goto tr862;
		case 83: goto st281;
		case 87: goto tr864;
		case 100: goto tr860;
		case 104: goto tr861;
		case 109: goto tr862;
		case 115: goto st281;
		case 119: goto tr864;
		case 778: goto tr865;
		case 827: goto st280;
		case 1034: goto tr865;
		case 1083: goto st280;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr859;
	goto tr855;
tr857:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st279;
tr858:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st279;
tr876:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st279;
tr877:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st279;
tr878:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st279;
st279:
	if ( ++p == pe )
		goto _test_eof279;
case 279:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st279;
		case 32: goto st279;
		case 40: goto tr857;
		case 41: goto tr858;
		case 778: goto tr865;
		case 827: goto st280;
		case 1034: goto tr865;
		case 1083: goto st280;
	}
	goto tr867;
tr865:
	{
		if (s->number64 <= UINT32_MAX) {
			s->default_ttl = (uint32_t)(s->number64);
		} else {
			ERR(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto st1091;
tr880:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			s->default_ttl = (uint32_t)(s->number64);
		} else {
			ERR(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto st1091;
st1091:
	if ( ++p == pe )
		goto _test_eof1091;
case 1091:
	goto st0;
tr881:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st280;
st280:
	if ( ++p == pe )
		goto _test_eof280;
case 280:
	if ( (*p) == 10 )
		goto tr865;
	goto st280;
tr860:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st281;
tr861:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st281;
tr862:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st281;
tr864:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st281;
st281:
	if ( ++p == pe )
		goto _test_eof281;
case 281:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st279;
		case 32: goto st279;
		case 40: goto tr857;
		case 41: goto tr858;
		case 778: goto tr865;
		case 827: goto st280;
		case 1034: goto tr865;
		case 1083: goto st280;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr868;
	goto tr855;
tr870:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st282;
tr868:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st282;
tr879:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st282;
st282:
	if ( ++p == pe )
		goto _test_eof282;
case 282:
	switch( (*p) ) {
		case 68: goto tr871;
		case 72: goto tr872;
		case 77: goto tr873;
		case 83: goto st283;
		case 87: goto tr875;
		case 100: goto tr871;
		case 104: goto tr872;
		case 109: goto tr873;
		case 115: goto st283;
		case 119: goto tr875;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr870;
	goto tr869;
tr871:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st283;
tr872:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st283;
tr873:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st283;
tr875:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st283;
st283:
	if ( ++p == pe )
		goto _test_eof283;
case 283:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr876;
		case 32: goto tr876;
		case 40: goto tr877;
		case 41: goto tr878;
		case 778: goto tr880;
		case 827: goto tr881;
		case 1034: goto tr880;
		case 1083: goto tr881;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr879;
	goto tr855;
st284:
	if ( ++p == pe )
		goto _test_eof284;
case 284:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr851;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st284;
	goto tr847;
st285:
	if ( ++p == pe )
		goto _test_eof285;
case 285:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st286;
		case 32: goto st286;
		case 40: goto tr884;
		case 41: goto tr885;
		case 1034: goto tr886;
		case 1083: goto st296;
	}
	goto tr882;
tr884:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st286;
tr885:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st286;
tr886:
	{
		s->line_counter++;
	}
	goto st286;
st286:
	if ( ++p == pe )
		goto _test_eof286;
case 286:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st286;
		case 32: goto st286;
		case 40: goto tr884;
		case 41: goto tr885;
		case 42: goto tr888;
		case 46: goto tr889;
		case 92: goto tr890;
		case 95: goto tr888;
		case 1034: goto tr886;
		case 1083: goto st296;
	}
	if ( _widec < 65 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr888;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr888;
	} else
		goto tr888;
	goto tr882;
tr891:
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st287;
tr898:
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st287;
tr911:
	{
		s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st287;
tr888:
	{
		s->dname = s->zone_origin;
	}
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st287;
st287:
	if ( ++p == pe )
		goto _test_eof287;
case 287:
	switch( (*p) ) {
		case 42: goto tr891;
		case 46: goto tr892;
		case 92: goto st291;
		case 95: goto tr891;
	}
	if ( (*p) < 65 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr891;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr891;
	} else
		goto tr891;
	goto tr882;
tr892:
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st288;
tr912:
	{
		s->dname_tmp_length++;
	}
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st288;
st288:
	if ( ++p == pe )
		goto _test_eof288;
case 288:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr895;
		case 32: goto tr895;
		case 40: goto tr896;
		case 41: goto tr897;
		case 42: goto tr898;
		case 45: goto tr898;
		case 92: goto tr899;
		case 95: goto tr898;
		case 778: goto tr900;
		case 827: goto tr901;
		case 1034: goto tr900;
		case 1083: goto tr901;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr898;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr898;
	} else
		goto tr898;
	goto tr894;
tr903:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st289;
tr904:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st289;
tr895:
	{
		// Enough room for the terminal label is garanteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	goto st289;
tr896:
	{
		// Enough room for the terminal label is garanteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st289;
tr897:
	{
		// Enough room for the terminal label is garanteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st289;
st289:
	if ( ++p == pe )
		goto _test_eof289;
case 289:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st289;
		case 32: goto st289;
		case 40: goto tr903;
		case 41: goto tr904;
		case 778: goto tr905;
		case 827: goto st290;
		case 1034: goto tr905;
		case 1083: goto st290;
	}
	goto tr894;
tr900:
	{
		// Enough room for the terminal label is garanteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->zone_origin_length = s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto st1092;
tr905:
	{
		s->zone_origin_length = s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto st1092;
st1092:
	if ( ++p == pe )
		goto _test_eof1092;
case 1092:
	goto st0;
tr901:
	{
		// Enough room for the terminal label is garanteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	goto st290;
st290:
	if ( ++p == pe )
		goto _test_eof290;
case 290:
	if ( (*p) == 10 )
		goto tr905;
	goto st290;
tr899:
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st291;
tr913:
	{
		s->dname_tmp_length++;
	}
	goto st291;
tr890:
	{
		s->dname = s->zone_origin;
	}
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st291;
st291:
	if ( ++p == pe )
		goto _test_eof291;
case 291:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr907;
	goto tr891;
tr907:
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length] = 0;
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st292;
st292:
	if ( ++p == pe )
		goto _test_eof292;
case 292:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr909;
	goto tr908;
tr909:
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st293;
st293:
	if ( ++p == pe )
		goto _test_eof293;
case 293:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr910;
	goto tr908;
tr910:
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st294;
st294:
	if ( ++p == pe )
		goto _test_eof294;
case 294:
	switch( (*p) ) {
		case 42: goto tr911;
		case 46: goto tr912;
		case 92: goto tr913;
		case 95: goto tr911;
	}
	if ( (*p) < 65 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr911;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr911;
	} else
		goto tr911;
	goto tr908;
tr889:
	{
		s->dname = s->zone_origin;
	}
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	goto st295;
st295:
	if ( ++p == pe )
		goto _test_eof295;
case 295:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr895;
		case 32: goto tr895;
		case 40: goto tr896;
		case 41: goto tr897;
		case 778: goto tr900;
		case 827: goto tr901;
		case 1034: goto tr900;
		case 1083: goto tr901;
	}
	goto tr894;
st296:
	if ( ++p == pe )
		goto _test_eof296;
case 296:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr886;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st296;
	goto tr882;
st297:
	if ( ++p == pe )
		goto _test_eof297;
case 297:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st298;
		case 32: goto st298;
		case 40: goto tr915;
		case 41: goto tr916;
		case 1034: goto tr917;
		case 1083: goto st313;
	}
	goto st0;
tr915:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st298;
tr916:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st298;
tr917:
	{
		s->line_counter++;
	}
	goto st298;
st298:
	if ( ++p == pe )
		goto _test_eof298;
case 298:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st298;
		case 32: goto st298;
		case 40: goto tr915;
		case 41: goto tr916;
		case 1034: goto tr917;
		case 1083: goto st313;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr919;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr919;
	} else
		goto tr919;
	goto tr920;
tr919:
	{
		rdata_tail = s->r_data;
	}
	{ p--; {stack[top++] = 299; goto st264;} }
	goto st299;
st299:
	if ( ++p == pe )
		goto _test_eof299;
case 299:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr922;
		case 32: goto tr922;
		case 40: goto tr923;
		case 41: goto tr924;
		case 778: goto tr925;
		case 827: goto tr926;
		case 1034: goto tr925;
		case 1083: goto tr927;
	}
	goto tr921;
tr930:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st300;
tr931:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st300;
tr922:
	{
		*rdata_tail = 0; // Ending filename string.
		strncpy((char*)(s->include_filename), (char*)(s->r_data),
		        sizeof(s->include_filename));

		// Check for correct string copy.
		if (strlen(s->include_filename) !=
		    (size_t)(rdata_tail - s->r_data)) {
			ERR(ZS_BAD_INCLUDE_FILENAME);
			p--; {goto st253;}
		}

		// For detection whether origin is not present.
		s->dname = NULL;
	}
	goto st300;
tr923:
	{
		*rdata_tail = 0; // Ending filename string.
		strncpy((char*)(s->include_filename), (char*)(s->r_data),
		        sizeof(s->include_filename));

		// Check for correct string copy.
		if (strlen(s->include_filename) !=
		    (size_t)(rdata_tail - s->r_data)) {
			ERR(ZS_BAD_INCLUDE_FILENAME);
			p--; {goto st253;}
		}

		// For detection whether origin is not present.
		s->dname = NULL;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st300;
tr924:
	{
		*rdata_tail = 0; // Ending filename string.
		strncpy((char*)(s->include_filename), (char*)(s->r_data),
		        sizeof(s->include_filename));

		// Check for correct string copy.
		if (strlen(s->include_filename) !=
		    (size_t)(rdata_tail - s->r_data)) {
			ERR(ZS_BAD_INCLUDE_FILENAME);
			p--; {goto st253;}
		}

		// For detection whether origin is not present.
		s->dname = NULL;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st300;
st300:
	if ( ++p == pe )
		goto _test_eof300;
case 300:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st300;
		case 32: goto st300;
		case 40: goto tr930;
		case 41: goto tr931;
		case 42: goto tr932;
		case 46: goto tr933;
		case 92: goto tr934;
		case 95: goto tr932;
		case 778: goto tr935;
		case 827: goto st304;
		case 1034: goto tr935;
		case 1083: goto st310;
	}
	if ( _widec < 65 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr932;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr932;
	} else
		goto tr932;
	goto tr928;
tr939:
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st301;
tr945:
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st301;
tr956:
	{
		s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st301;
tr932:
	{
		s->dname = s->r_data;
	}
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	{
		// Check for maximum dname label length.
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length++] = (*p);
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st301;
st301:
	if ( ++p == pe )
		goto _test_eof301;
case 301:
	switch( (*p) ) {
		case 42: goto tr939;
		case 46: goto tr940;
		case 92: goto st305;
		case 95: goto tr939;
	}
	if ( (*p) < 65 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr939;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr939;
	} else
		goto tr939;
	goto tr938;
tr940:
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st302;
tr957:
	{
		s->dname_tmp_length++;
	}
	{
		// Check for maximum dname length overflow after each label.
		// (at least the next label length must follow).
		if (s->dname_tmp_length < MAX_DNAME_LENGTH) {
			(s->dname)[s->item_length_position] =
				(uint8_t)(s->item_length);
		} else {
			WARN(ZS_DNAME_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st302;
st302:
	if ( ++p == pe )
		goto _test_eof302;
case 302:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr942;
		case 32: goto tr942;
		case 40: goto tr943;
		case 41: goto tr944;
		case 42: goto tr945;
		case 45: goto tr945;
		case 92: goto tr946;
		case 95: goto tr945;
		case 778: goto tr947;
		case 827: goto tr948;
		case 1034: goto tr947;
		case 1083: goto tr948;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr945;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr945;
	} else
		goto tr945;
	goto tr928;
tr950:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st303;
tr951:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st303;
tr942:
	{
		// Enough room for the terminal label is garanteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->r_data_length = s->dname_tmp_length;
	}
	goto st303;
tr943:
	{
		// Enough room for the terminal label is garanteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->r_data_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st303;
tr944:
	{
		// Enough room for the terminal label is garanteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->r_data_length = s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st303;
st303:
	if ( ++p == pe )
		goto _test_eof303;
case 303:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st303;
		case 32: goto st303;
		case 40: goto tr950;
		case 41: goto tr951;
		case 778: goto tr935;
		case 827: goto st304;
		case 1034: goto tr935;
		case 1083: goto st304;
	}
	goto tr83;
tr925:
	{
		*rdata_tail = 0; // Ending filename string.
		strncpy((char*)(s->include_filename), (char*)(s->r_data),
		        sizeof(s->include_filename));

		// Check for correct string copy.
		if (strlen(s->include_filename) !=
		    (size_t)(rdata_tail - s->r_data)) {
			ERR(ZS_BAD_INCLUDE_FILENAME);
			p--; {goto st253;}
		}

		// For detection whether origin is not present.
		s->dname = NULL;
	}
	{
		char text_origin[4 * MAX_DNAME_LENGTH]; // Each char as \DDD.

		// Origin conversion from wire to text form.
		if (s->dname == NULL) { // Use current origin.
			wire_dname_to_str(s->zone_origin,
			                  s->zone_origin_length,
			                  text_origin);
		} else { // Use specified origin.
			wire_dname_to_str(s->r_data,
			                  s->r_data_length,
			                  text_origin);
		}

		// Relative file path.
		if (s->include_filename[0] != '/') {
			snprintf((char*)(s->buffer), sizeof(s->buffer),
			         "%s/%s", s->path, s->include_filename);
		} else {
			strncpy((char*)(s->buffer), (char*)(s->include_filename),
			        sizeof(s->buffer));
		}

		// Create new scanner for included zone file.
		zs_scanner_t *ss = zs_scanner_create(text_origin,
		                                     s->default_class,
		                                     s->default_ttl,
		                                     s->process_record,
		                                     s->process_error,
		                                     s->data);
		if (ss != NULL) {
			// Parse included zone file.
			ret = zs_scanner_parse_file(ss, (char*)(s->buffer));
			if (ret != 0) {
				// File internal errors are handled by error callback.
				if (ss->error_counter > 0) {
					ERR(ZS_UNPROCESSED_INCLUDE);
				// General include file error.
				} else {
					ERR(ss->error_code);
				}
				zs_scanner_free(ss);
				p--; {goto st253;}
			}
			zs_scanner_free(ss);
		} else {
			ERR(ZS_UNPROCESSED_INCLUDE);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto st1093;
tr935:
	{
		char text_origin[4 * MAX_DNAME_LENGTH]; // Each char as \DDD.

		// Origin conversion from wire to text form.
		if (s->dname == NULL) { // Use current origin.
			wire_dname_to_str(s->zone_origin,
			                  s->zone_origin_length,
			                  text_origin);
		} else { // Use specified origin.
			wire_dname_to_str(s->r_data,
			                  s->r_data_length,
			                  text_origin);
		}

		// Relative file path.
		if (s->include_filename[0] != '/') {
			snprintf((char*)(s->buffer), sizeof(s->buffer),
			         "%s/%s", s->path, s->include_filename);
		} else {
			strncpy((char*)(s->buffer), (char*)(s->include_filename),
			        sizeof(s->buffer));
		}

		// Create new scanner for included zone file.
		zs_scanner_t *ss = zs_scanner_create(text_origin,
		                                     s->default_class,
		                                     s->default_ttl,
		                                     s->process_record,
		                                     s->process_error,
		                                     s->data);
		if (ss != NULL) {
			// Parse included zone file.
			ret = zs_scanner_parse_file(ss, (char*)(s->buffer));
			if (ret != 0) {
				// File internal errors are handled by error callback.
				if (ss->error_counter > 0) {
					ERR(ZS_UNPROCESSED_INCLUDE);
				// General include file error.
				} else {
					ERR(ss->error_code);
				}
				zs_scanner_free(ss);
				p--; {goto st253;}
			}
			zs_scanner_free(ss);
		} else {
			ERR(ZS_UNPROCESSED_INCLUDE);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto st1093;
tr947:
	{
		// Enough room for the terminal label is garanteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->r_data_length = s->dname_tmp_length;
	}
	{
		char text_origin[4 * MAX_DNAME_LENGTH]; // Each char as \DDD.

		// Origin conversion from wire to text form.
		if (s->dname == NULL) { // Use current origin.
			wire_dname_to_str(s->zone_origin,
			                  s->zone_origin_length,
			                  text_origin);
		} else { // Use specified origin.
			wire_dname_to_str(s->r_data,
			                  s->r_data_length,
			                  text_origin);
		}

		// Relative file path.
		if (s->include_filename[0] != '/') {
			snprintf((char*)(s->buffer), sizeof(s->buffer),
			         "%s/%s", s->path, s->include_filename);
		} else {
			strncpy((char*)(s->buffer), (char*)(s->include_filename),
			        sizeof(s->buffer));
		}

		// Create new scanner for included zone file.
		zs_scanner_t *ss = zs_scanner_create(text_origin,
		                                     s->default_class,
		                                     s->default_ttl,
		                                     s->process_record,
		                                     s->process_error,
		                                     s->data);
		if (ss != NULL) {
			// Parse included zone file.
			ret = zs_scanner_parse_file(ss, (char*)(s->buffer));
			if (ret != 0) {
				// File internal errors are handled by error callback.
				if (ss->error_counter > 0) {
					ERR(ZS_UNPROCESSED_INCLUDE);
				// General include file error.
				} else {
					ERR(ss->error_code);
				}
				zs_scanner_free(ss);
				p--; {goto st253;}
			}
			zs_scanner_free(ss);
		} else {
			ERR(ZS_UNPROCESSED_INCLUDE);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		s->line_counter++;
	}
	goto st1093;
st1093:
	if ( ++p == pe )
		goto _test_eof1093;
case 1093:
	goto st0;
tr926:
	{
		*rdata_tail = 0; // Ending filename string.
		strncpy((char*)(s->include_filename), (char*)(s->r_data),
		        sizeof(s->include_filename));

		// Check for correct string copy.
		if (strlen(s->include_filename) !=
		    (size_t)(rdata_tail - s->r_data)) {
			ERR(ZS_BAD_INCLUDE_FILENAME);
			p--; {goto st253;}
		}

		// For detection whether origin is not present.
		s->dname = NULL;
	}
	goto st304;
tr948:
	{
		// Enough room for the terminal label is garanteed (_label_exit).
		(s->dname)[s->dname_tmp_length++] = 0;
	}
	{
		s->r_data_length = s->dname_tmp_length;
	}
	goto st304;
st304:
	if ( ++p == pe )
		goto _test_eof304;
case 304:
	if ( (*p) == 10 )
		goto tr935;
	goto st304;
tr946:
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st305;
tr958:
	{
		s->dname_tmp_length++;
	}
	goto st305;
tr934:
	{
		s->dname = s->r_data;
	}
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	{
		s->item_length = 0;
		s->item_length_position = s->dname_tmp_length++;
	}
	goto st305;
st305:
	if ( ++p == pe )
		goto _test_eof305;
case 305:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr952;
	goto tr939;
tr952:
	{
		if (s->item_length < MAX_LABEL_LENGTH) {
			(s->dname)[s->dname_tmp_length] = 0;
			s->item_length++;
		} else {
			WARN(ZS_LABEL_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st306;
st306:
	if ( ++p == pe )
		goto _test_eof306;
case 306:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr954;
	goto tr953;
tr954:
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st307;
st307:
	if ( ++p == pe )
		goto _test_eof307;
case 307:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr955;
	goto tr953;
tr955:
	{
		(s->dname)[s->dname_tmp_length] *= 10;
		(s->dname)[s->dname_tmp_length] += digit_to_num[(uint8_t)(*p)];
	}
	goto st308;
st308:
	if ( ++p == pe )
		goto _test_eof308;
case 308:
	switch( (*p) ) {
		case 42: goto tr956;
		case 46: goto tr957;
		case 92: goto tr958;
		case 95: goto tr956;
	}
	if ( (*p) < 65 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr956;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr956;
	} else
		goto tr956;
	goto tr953;
tr933:
	{
		s->dname = s->r_data;
	}
	{
		s->item_length_position = 0;
		s->dname_tmp_length = 0;
	}
	goto st309;
st309:
	if ( ++p == pe )
		goto _test_eof309;
case 309:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr942;
		case 32: goto tr942;
		case 40: goto tr943;
		case 41: goto tr944;
		case 778: goto tr947;
		case 827: goto tr948;
		case 1034: goto tr947;
		case 1083: goto tr948;
	}
	goto tr928;
tr927:
	{
		*rdata_tail = 0; // Ending filename string.
		strncpy((char*)(s->include_filename), (char*)(s->r_data),
		        sizeof(s->include_filename));

		// Check for correct string copy.
		if (strlen(s->include_filename) !=
		    (size_t)(rdata_tail - s->r_data)) {
			ERR(ZS_BAD_INCLUDE_FILENAME);
			p--; {goto st253;}
		}

		// For detection whether origin is not present.
		s->dname = NULL;
	}
	goto st310;
st310:
	if ( ++p == pe )
		goto _test_eof310;
case 310:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 778: goto tr935;
		case 1034: goto tr959;
	}
	if ( _widec > 895 ) {
		if ( 896 <= _widec && _widec <= 1151 )
			goto st310;
	} else if ( _widec >= 640 )
		goto st304;
	goto tr83;
tr959:
	{
		s->line_counter++;
	}
	{
		char text_origin[4 * MAX_DNAME_LENGTH]; // Each char as \DDD.

		// Origin conversion from wire to text form.
		if (s->dname == NULL) { // Use current origin.
			wire_dname_to_str(s->zone_origin,
			                  s->zone_origin_length,
			                  text_origin);
		} else { // Use specified origin.
			wire_dname_to_str(s->r_data,
			                  s->r_data_length,
			                  text_origin);
		}

		// Relative file path.
		if (s->include_filename[0] != '/') {
			snprintf((char*)(s->buffer), sizeof(s->buffer),
			         "%s/%s", s->path, s->include_filename);
		} else {
			strncpy((char*)(s->buffer), (char*)(s->include_filename),
			        sizeof(s->buffer));
		}

		// Create new scanner for included zone file.
		zs_scanner_t *ss = zs_scanner_create(text_origin,
		                                     s->default_class,
		                                     s->default_ttl,
		                                     s->process_record,
		                                     s->process_error,
		                                     s->data);
		if (ss != NULL) {
			// Parse included zone file.
			ret = zs_scanner_parse_file(ss, (char*)(s->buffer));
			if (ret != 0) {
				// File internal errors are handled by error callback.
				if (ss->error_counter > 0) {
					ERR(ZS_UNPROCESSED_INCLUDE);
				// General include file error.
				} else {
					ERR(ss->error_code);
				}
				zs_scanner_free(ss);
				p--; {goto st253;}
			}
			zs_scanner_free(ss);
		} else {
			ERR(ZS_UNPROCESSED_INCLUDE);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1094;
st1094:
	if ( ++p == pe )
		goto _test_eof1094;
case 1094:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st311;
		case 32: goto st311;
		case 40: goto tr961;
		case 41: goto tr962;
		case 42: goto tr932;
		case 46: goto tr933;
		case 92: goto tr934;
		case 95: goto tr932;
		case 1034: goto tr963;
		case 1083: goto st312;
	}
	if ( _widec < 65 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr932;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr932;
	} else
		goto tr932;
	goto tr938;
tr961:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st311;
tr962:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st311;
tr963:
	{
		s->line_counter++;
	}
	goto st311;
st311:
	if ( ++p == pe )
		goto _test_eof311;
case 311:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st311;
		case 32: goto st311;
		case 40: goto tr961;
		case 41: goto tr962;
		case 42: goto tr932;
		case 46: goto tr933;
		case 92: goto tr934;
		case 95: goto tr932;
		case 1034: goto tr963;
		case 1083: goto st312;
	}
	if ( _widec < 65 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr932;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr932;
	} else
		goto tr932;
	goto tr938;
st312:
	if ( ++p == pe )
		goto _test_eof312;
case 312:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr963;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st312;
	goto st0;
st313:
	if ( ++p == pe )
		goto _test_eof313;
case 313:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr917;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st313;
	goto st0;
st314:
	if ( ++p == pe )
		goto _test_eof314;
case 314:
	if ( (*p) == 43 )
		goto tr966;
	if ( (*p) < 65 ) {
		if ( 47 <= (*p) && (*p) <= 57 )
			goto tr966;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr966;
	} else
		goto tr966;
	goto tr965;
tr966:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_base64_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st315;
st315:
	if ( ++p == pe )
		goto _test_eof315;
case 315:
	if ( (*p) == 43 )
		goto tr967;
	if ( (*p) < 65 ) {
		if ( 47 <= (*p) && (*p) <= 57 )
			goto tr967;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr967;
	} else
		goto tr967;
	goto tr965;
tr967:
	{
		*(rdata_tail++) += second_left_base64_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = second_right_base64_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st316;
st316:
	if ( ++p == pe )
		goto _test_eof316;
case 316:
	switch( (*p) ) {
		case 43: goto tr968;
		case 61: goto st320;
	}
	if ( (*p) < 65 ) {
		if ( 47 <= (*p) && (*p) <= 57 )
			goto tr968;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr968;
	} else
		goto tr968;
	goto tr965;
tr968:
	{
		*(rdata_tail++) += third_left_base64_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = third_right_base64_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st317;
st317:
	if ( ++p == pe )
		goto _test_eof317;
case 317:
	switch( (*p) ) {
		case 43: goto tr970;
		case 61: goto st318;
	}
	if ( (*p) < 65 ) {
		if ( 47 <= (*p) && (*p) <= 57 )
			goto tr970;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr970;
	} else
		goto tr970;
	goto tr965;
tr972:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st318;
tr973:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st318;
tr974:
	{
		s->line_counter++;
	}
	goto st318;
tr970:
	{
		*(rdata_tail++) += fourth_base64_to_num[(uint8_t)(*p)];
	}
	goto st318;
st318:
	if ( ++p == pe )
		goto _test_eof318;
case 318:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st318;
		case 32: goto st318;
		case 40: goto tr972;
		case 41: goto tr973;
		case 43: goto tr966;
		case 2058: goto tr974;
		case 2107: goto st319;
		case 2314: goto tr976;
		case 2363: goto tr976;
		case 2570: goto tr977;
		case 2619: goto tr978;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr966;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr966;
	} else
		goto tr966;
	goto tr965;
st319:
	if ( ++p == pe )
		goto _test_eof319;
case 319:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr974;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st319;
	goto tr965;
tr976:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1095;
st1095:
	if ( ++p == pe )
		goto _test_eof1095;
case 1095:
	goto st0;
tr977:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1096;
st1096:
	if ( ++p == pe )
		goto _test_eof1096;
case 1096:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st318;
		case 32: goto st318;
		case 40: goto tr972;
		case 41: goto tr973;
		case 43: goto tr966;
		case 2058: goto tr974;
		case 2107: goto st319;
		case 2314: goto tr976;
		case 2363: goto tr976;
		case 2570: goto tr977;
		case 2619: goto tr978;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr966;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr966;
	} else
		goto tr966;
	goto tr965;
tr978:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1097;
st1097:
	if ( ++p == pe )
		goto _test_eof1097;
case 1097:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr974;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st319;
	goto tr965;
st320:
	if ( ++p == pe )
		goto _test_eof320;
case 320:
	if ( (*p) == 61 )
		goto st318;
	goto tr965;
st321:
	if ( ++p == pe )
		goto _test_eof321;
case 321:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr980;
		case 32: goto tr980;
		case 40: goto tr981;
		case 41: goto tr982;
		case 2058: goto tr983;
		case 2107: goto tr984;
		case 2314: goto tr985;
		case 2363: goto tr985;
		case 2570: goto tr986;
		case 2619: goto tr987;
	}
	goto tr979;
tr989:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr990:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1005:
	{
		s->line_counter++;
	}
	goto st322;
tr980:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	goto st322;
tr981:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr982:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr983:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	{
		s->line_counter++;
	}
	goto st322;
tr1010:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	goto st322;
tr1011:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1012:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1016:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1023:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	goto st322;
tr1024:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1025:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1026:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1040:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	goto st322;
tr1041:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1042:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1043:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1048:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	goto st322;
tr1049:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1050:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1051:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1058:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	goto st322;
tr1059:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1060:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1061:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1069:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	goto st322;
tr1070:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1071:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1072:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1083:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	goto st322;
tr1084:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1085:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1086:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1095:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	goto st322;
tr1096:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1097:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1098:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1106:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	goto st322;
tr1107:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1108:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1109:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1114:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	goto st322;
tr1115:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1116:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1117:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1127:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	goto st322;
tr1128:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1129:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1130:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1136:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	goto st322;
tr1137:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1138:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1139:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1148:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	goto st322;
tr1149:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1150:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1151:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1163:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	goto st322;
tr1164:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1165:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1166:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1174:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	goto st322;
tr1175:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1176:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1177:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1182:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	goto st322;
tr1183:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1184:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1185:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1195:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	goto st322;
tr1196:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1197:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1198:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1204:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	goto st322;
tr1205:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1206:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1207:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1213:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	goto st322;
tr1214:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1215:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1216:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1221:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	goto st322;
tr1222:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1223:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1224:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1234:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	goto st322;
tr1235:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1236:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1237:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1242:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	goto st322;
tr1243:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1244:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1245:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1256:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	goto st322;
tr1257:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1258:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1259:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1265:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	goto st322;
tr1266:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1267:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1268:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1273:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	goto st322;
tr1274:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1275:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1277:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1283:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	goto st322;
tr1284:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1285:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1287:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1292:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	goto st322;
tr1293:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1294:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1296:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1305:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	goto st322;
tr1306:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1307:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1308:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1315:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	goto st322;
tr1316:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1317:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1318:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1326:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	goto st322;
tr1327:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1328:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1329:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1337:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	goto st322;
tr1338:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1339:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1340:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1345:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	goto st322;
tr1346:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1347:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1348:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1358:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	goto st322;
tr1359:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1360:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1361:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1367:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	goto st322;
tr1368:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1369:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1370:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1376:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	goto st322;
tr1377:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1378:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1379:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1387:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	goto st322;
tr1388:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1389:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1390:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1400:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	goto st322;
tr1401:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1402:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1403:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1409:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	goto st322;
tr1410:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1411:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1412:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1420:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st322;
tr1421:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1422:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1424:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st322;
tr1432:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	goto st322;
tr1433:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1434:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1435:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	{
		s->line_counter++;
	}
	goto st322;
tr1441:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	goto st322;
tr1442:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st322;
tr1443:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st322;
tr1444:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	{
		s->line_counter++;
	}
	goto st322;
st322:
	if ( ++p == pe )
		goto _test_eof322;
case 322:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st322;
		case 32: goto st322;
		case 40: goto tr989;
		case 41: goto tr990;
		case 65: goto st323;
		case 67: goto st328;
		case 68: goto st343;
		case 69: goto st357;
		case 72: goto st364;
		case 73: goto st369;
		case 75: goto st377;
		case 76: goto st381;
		case 77: goto st389;
		case 78: goto st395;
		case 80: goto st411;
		case 82: goto st414;
		case 83: goto st421;
		case 84: goto st432;
		case 97: goto st323;
		case 99: goto st328;
		case 100: goto st343;
		case 101: goto st357;
		case 104: goto st364;
		case 105: goto st369;
		case 107: goto st377;
		case 108: goto st381;
		case 109: goto st389;
		case 110: goto st395;
		case 112: goto st411;
		case 114: goto st414;
		case 115: goto st421;
		case 116: goto st432;
		case 2058: goto tr1005;
		case 2107: goto st327;
		case 2314: goto tr1007;
		case 2363: goto tr1007;
		case 2570: goto tr1008;
		case 2619: goto tr1009;
	}
	goto tr979;
st323:
	if ( ++p == pe )
		goto _test_eof323;
case 323:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1010;
		case 32: goto tr1010;
		case 40: goto tr1011;
		case 41: goto tr1012;
		case 65: goto st324;
		case 70: goto st442;
		case 80: goto st446;
		case 97: goto st324;
		case 102: goto st442;
		case 112: goto st446;
		case 2058: goto tr1016;
		case 2107: goto tr1017;
		case 2314: goto tr1018;
		case 2363: goto tr1018;
		case 2570: goto tr1019;
		case 2619: goto tr1020;
	}
	goto tr979;
st324:
	if ( ++p == pe )
		goto _test_eof324;
case 324:
	switch( (*p) ) {
		case 65: goto st325;
		case 97: goto st325;
	}
	goto tr979;
st325:
	if ( ++p == pe )
		goto _test_eof325;
case 325:
	switch( (*p) ) {
		case 65: goto st326;
		case 97: goto st326;
	}
	goto tr979;
st326:
	if ( ++p == pe )
		goto _test_eof326;
case 326:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1023;
		case 32: goto tr1023;
		case 40: goto tr1024;
		case 41: goto tr1025;
		case 2058: goto tr1026;
		case 2107: goto tr1027;
		case 2314: goto tr1028;
		case 2363: goto tr1028;
		case 2570: goto tr1029;
		case 2619: goto tr1030;
	}
	goto tr979;
tr984:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	goto st327;
tr1017:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	goto st327;
tr1027:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	goto st327;
tr1044:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	goto st327;
tr1052:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	goto st327;
tr1062:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	goto st327;
tr1073:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	goto st327;
tr1087:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	goto st327;
tr1099:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	goto st327;
tr1110:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	goto st327;
tr1118:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	goto st327;
tr1131:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	goto st327;
tr1140:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	goto st327;
tr1152:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	goto st327;
tr1167:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	goto st327;
tr1178:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	goto st327;
tr1186:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	goto st327;
tr1199:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	goto st327;
tr1208:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	goto st327;
tr1217:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	goto st327;
tr1225:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	goto st327;
tr1238:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	goto st327;
tr1246:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	goto st327;
tr1260:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	goto st327;
tr1269:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	goto st327;
tr1278:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	goto st327;
tr1288:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	goto st327;
tr1297:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	goto st327;
tr1309:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	goto st327;
tr1319:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	goto st327;
tr1330:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	goto st327;
tr1341:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	goto st327;
tr1349:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	goto st327;
tr1362:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	goto st327;
tr1371:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	goto st327;
tr1380:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	goto st327;
tr1391:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	goto st327;
tr1404:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	goto st327;
tr1413:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	goto st327;
tr1425:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st327;
tr1436:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	goto st327;
tr1445:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	goto st327;
st327:
	if ( ++p == pe )
		goto _test_eof327;
case 327:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1005;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st327;
	goto tr979;
tr985:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1007:
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1018:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1028:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1045:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1053:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1063:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1074:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1088:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1100:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1111:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1119:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1132:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1141:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1153:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1168:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1179:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1187:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1200:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1209:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1218:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1226:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1239:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1247:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1261:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1270:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1279:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1289:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1298:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1310:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1320:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1331:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1342:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1350:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1363:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1372:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1381:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1392:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1405:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1414:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1426:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1437:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
tr1446:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1098;
st1098:
	if ( ++p == pe )
		goto _test_eof1098;
case 1098:
	goto st0;
tr986:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1008:
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1019:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1029:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1046:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1054:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1064:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1075:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1089:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1101:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1112:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1120:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1133:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1142:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1154:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1169:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1180:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1188:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1201:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1210:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1219:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1227:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1240:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1248:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1262:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1271:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1280:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1290:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1299:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1311:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1321:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1332:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1343:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1351:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1364:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1373:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1382:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1393:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1406:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1415:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1427:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1438:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
tr1447:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	{
		s->line_counter++;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1099;
st1099:
	if ( ++p == pe )
		goto _test_eof1099;
case 1099:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st322;
		case 32: goto st322;
		case 40: goto tr989;
		case 41: goto tr990;
		case 65: goto st323;
		case 67: goto st328;
		case 68: goto st343;
		case 69: goto st357;
		case 72: goto st364;
		case 73: goto st369;
		case 75: goto st377;
		case 76: goto st381;
		case 77: goto st389;
		case 78: goto st395;
		case 80: goto st411;
		case 82: goto st414;
		case 83: goto st421;
		case 84: goto st432;
		case 97: goto st323;
		case 99: goto st328;
		case 100: goto st343;
		case 101: goto st357;
		case 104: goto st364;
		case 105: goto st369;
		case 107: goto st377;
		case 108: goto st381;
		case 109: goto st389;
		case 110: goto st395;
		case 112: goto st411;
		case 114: goto st414;
		case 115: goto st421;
		case 116: goto st432;
		case 2058: goto tr1005;
		case 2107: goto st327;
		case 2314: goto tr1007;
		case 2363: goto tr1007;
		case 2570: goto tr1008;
		case 2619: goto tr1009;
	}
	goto tr979;
st328:
	if ( ++p == pe )
		goto _test_eof328;
case 328:
	switch( (*p) ) {
		case 68: goto st329;
		case 69: goto st336;
		case 78: goto st339;
		case 100: goto st329;
		case 101: goto st336;
		case 110: goto st339;
	}
	goto tr979;
st329:
	if ( ++p == pe )
		goto _test_eof329;
case 329:
	switch( (*p) ) {
		case 78: goto st330;
		case 83: goto st335;
		case 110: goto st330;
		case 115: goto st335;
	}
	goto tr979;
st330:
	if ( ++p == pe )
		goto _test_eof330;
case 330:
	switch( (*p) ) {
		case 83: goto st331;
		case 115: goto st331;
	}
	goto tr979;
st331:
	if ( ++p == pe )
		goto _test_eof331;
case 331:
	switch( (*p) ) {
		case 75: goto st332;
		case 107: goto st332;
	}
	goto tr979;
st332:
	if ( ++p == pe )
		goto _test_eof332;
case 332:
	switch( (*p) ) {
		case 69: goto st333;
		case 101: goto st333;
	}
	goto tr979;
st333:
	if ( ++p == pe )
		goto _test_eof333;
case 333:
	switch( (*p) ) {
		case 89: goto st334;
		case 121: goto st334;
	}
	goto tr979;
st334:
	if ( ++p == pe )
		goto _test_eof334;
case 334:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1040;
		case 32: goto tr1040;
		case 40: goto tr1041;
		case 41: goto tr1042;
		case 2058: goto tr1043;
		case 2107: goto tr1044;
		case 2314: goto tr1045;
		case 2363: goto tr1045;
		case 2570: goto tr1046;
		case 2619: goto tr1047;
	}
	goto tr979;
tr987:
	{
		memset(s->windows, 0, sizeof(s->windows));
		s->last_window = -1;
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1009:
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1020:
	{ window_add_bit(KNOT_RRTYPE_A, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1030:
	{ window_add_bit(KNOT_RRTYPE_AAAA, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1047:
	{ window_add_bit(KNOT_RRTYPE_CDNSKEY, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1055:
	{ window_add_bit(KNOT_RRTYPE_CDS, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1065:
	{ window_add_bit(KNOT_RRTYPE_CERT, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1076:
	{ window_add_bit(KNOT_RRTYPE_CNAME, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1090:
	{ window_add_bit(KNOT_RRTYPE_DHCID, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1102:
	{ window_add_bit(KNOT_RRTYPE_DNAME, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1113:
	{ window_add_bit(KNOT_RRTYPE_DNSKEY, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1121:
	{ window_add_bit(KNOT_RRTYPE_DS, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1134:
	{ window_add_bit(KNOT_RRTYPE_EUI48, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1143:
	{ window_add_bit(KNOT_RRTYPE_EUI64, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1155:
	{ window_add_bit(KNOT_RRTYPE_HINFO, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1170:
	{ window_add_bit(KNOT_RRTYPE_IPSECKEY, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1181:
	{ window_add_bit(KNOT_RRTYPE_KEY, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1189:
	{ window_add_bit(KNOT_RRTYPE_KX, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1202:
	{ window_add_bit(KNOT_RRTYPE_L32, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1211:
	{ window_add_bit(KNOT_RRTYPE_L64, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1220:
	{ window_add_bit(KNOT_RRTYPE_LOC, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1228:
	{ window_add_bit(KNOT_RRTYPE_LP, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1241:
	{ window_add_bit(KNOT_RRTYPE_MINFO, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1249:
	{ window_add_bit(KNOT_RRTYPE_MX, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1263:
	{ window_add_bit(KNOT_RRTYPE_NAPTR, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1272:
	{ window_add_bit(KNOT_RRTYPE_NID, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1281:
	{ window_add_bit(KNOT_RRTYPE_NS, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1291:
	{ window_add_bit(KNOT_RRTYPE_NSEC, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1300:
	{ window_add_bit(KNOT_RRTYPE_NSEC3, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1312:
	{ window_add_bit(KNOT_RRTYPE_NSEC3PARAM, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1322:
	{ window_add_bit(KNOT_RRTYPE_PTR, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1333:
	{ window_add_bit(KNOT_RRTYPE_RP, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1344:
	{ window_add_bit(KNOT_RRTYPE_RRSIG, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1352:
	{ window_add_bit(KNOT_RRTYPE_RT, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1365:
	{ window_add_bit(KNOT_RRTYPE_SOA, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1374:
	{ window_add_bit(KNOT_RRTYPE_SPF, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1383:
	{ window_add_bit(KNOT_RRTYPE_SRV, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1394:
	{ window_add_bit(KNOT_RRTYPE_SSHFP, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1407:
	{ window_add_bit(KNOT_RRTYPE_TLSA, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1416:
	{ window_add_bit(KNOT_RRTYPE_TXT, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1428:
	{
		if (s->number64 <= UINT16_MAX) {
			window_add_bit(s->number64, s);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1439:
	{ window_add_bit(KNOT_RRTYPE_AFSDB, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
tr1448:
	{ window_add_bit(KNOT_RRTYPE_APL, s); }
	{
		for (window = 0; window <= s->last_window; window++) {
			if ((s->windows[window]).length > 0) {
				if (rdata_tail + 2 + (s->windows[window]).length <= rdata_stop)
				{
					// Window number.
					*rdata_tail = (uint8_t)window;
					rdata_tail += 1;
					// Bitmap length.
					*rdata_tail = (s->windows[window]).length;
					rdata_tail += 1;
					// Copying bitmap.
					memcpy(rdata_tail,
					       (s->windows[window]).bitmap,
					       (s->windows[window]).length);
					rdata_tail += (s->windows[window]).length;
				} else {
					WARN(ZS_RDATA_OVERFLOW);
					p--; {goto st253;}
				}
			}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1100;
st1100:
	if ( ++p == pe )
		goto _test_eof1100;
case 1100:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1005;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st327;
	goto tr979;
st335:
	if ( ++p == pe )
		goto _test_eof335;
case 335:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1048;
		case 32: goto tr1048;
		case 40: goto tr1049;
		case 41: goto tr1050;
		case 2058: goto tr1051;
		case 2107: goto tr1052;
		case 2314: goto tr1053;
		case 2363: goto tr1053;
		case 2570: goto tr1054;
		case 2619: goto tr1055;
	}
	goto tr979;
st336:
	if ( ++p == pe )
		goto _test_eof336;
case 336:
	switch( (*p) ) {
		case 82: goto st337;
		case 114: goto st337;
	}
	goto tr979;
st337:
	if ( ++p == pe )
		goto _test_eof337;
case 337:
	switch( (*p) ) {
		case 84: goto st338;
		case 116: goto st338;
	}
	goto tr979;
st338:
	if ( ++p == pe )
		goto _test_eof338;
case 338:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1058;
		case 32: goto tr1058;
		case 40: goto tr1059;
		case 41: goto tr1060;
		case 2058: goto tr1061;
		case 2107: goto tr1062;
		case 2314: goto tr1063;
		case 2363: goto tr1063;
		case 2570: goto tr1064;
		case 2619: goto tr1065;
	}
	goto tr979;
st339:
	if ( ++p == pe )
		goto _test_eof339;
case 339:
	switch( (*p) ) {
		case 65: goto st340;
		case 97: goto st340;
	}
	goto tr979;
st340:
	if ( ++p == pe )
		goto _test_eof340;
case 340:
	switch( (*p) ) {
		case 77: goto st341;
		case 109: goto st341;
	}
	goto tr979;
st341:
	if ( ++p == pe )
		goto _test_eof341;
case 341:
	switch( (*p) ) {
		case 69: goto st342;
		case 101: goto st342;
	}
	goto tr979;
st342:
	if ( ++p == pe )
		goto _test_eof342;
case 342:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1069;
		case 32: goto tr1069;
		case 40: goto tr1070;
		case 41: goto tr1071;
		case 2058: goto tr1072;
		case 2107: goto tr1073;
		case 2314: goto tr1074;
		case 2363: goto tr1074;
		case 2570: goto tr1075;
		case 2619: goto tr1076;
	}
	goto tr979;
st343:
	if ( ++p == pe )
		goto _test_eof343;
case 343:
	switch( (*p) ) {
		case 72: goto st344;
		case 78: goto st348;
		case 83: goto st356;
		case 104: goto st344;
		case 110: goto st348;
		case 115: goto st356;
	}
	goto tr979;
st344:
	if ( ++p == pe )
		goto _test_eof344;
case 344:
	switch( (*p) ) {
		case 67: goto st345;
		case 99: goto st345;
	}
	goto tr979;
st345:
	if ( ++p == pe )
		goto _test_eof345;
case 345:
	switch( (*p) ) {
		case 73: goto st346;
		case 105: goto st346;
	}
	goto tr979;
st346:
	if ( ++p == pe )
		goto _test_eof346;
case 346:
	switch( (*p) ) {
		case 68: goto st347;
		case 100: goto st347;
	}
	goto tr979;
st347:
	if ( ++p == pe )
		goto _test_eof347;
case 347:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1083;
		case 32: goto tr1083;
		case 40: goto tr1084;
		case 41: goto tr1085;
		case 2058: goto tr1086;
		case 2107: goto tr1087;
		case 2314: goto tr1088;
		case 2363: goto tr1088;
		case 2570: goto tr1089;
		case 2619: goto tr1090;
	}
	goto tr979;
st348:
	if ( ++p == pe )
		goto _test_eof348;
case 348:
	switch( (*p) ) {
		case 65: goto st349;
		case 83: goto st352;
		case 97: goto st349;
		case 115: goto st352;
	}
	goto tr979;
st349:
	if ( ++p == pe )
		goto _test_eof349;
case 349:
	switch( (*p) ) {
		case 77: goto st350;
		case 109: goto st350;
	}
	goto tr979;
st350:
	if ( ++p == pe )
		goto _test_eof350;
case 350:
	switch( (*p) ) {
		case 69: goto st351;
		case 101: goto st351;
	}
	goto tr979;
st351:
	if ( ++p == pe )
		goto _test_eof351;
case 351:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1095;
		case 32: goto tr1095;
		case 40: goto tr1096;
		case 41: goto tr1097;
		case 2058: goto tr1098;
		case 2107: goto tr1099;
		case 2314: goto tr1100;
		case 2363: goto tr1100;
		case 2570: goto tr1101;
		case 2619: goto tr1102;
	}
	goto tr979;
st352:
	if ( ++p == pe )
		goto _test_eof352;
case 352:
	switch( (*p) ) {
		case 75: goto st353;
		case 107: goto st353;
	}
	goto tr979;
st353:
	if ( ++p == pe )
		goto _test_eof353;
case 353:
	switch( (*p) ) {
		case 69: goto st354;
		case 101: goto st354;
	}
	goto tr979;
st354:
	if ( ++p == pe )
		goto _test_eof354;
case 354:
	switch( (*p) ) {
		case 89: goto st355;
		case 121: goto st355;
	}
	goto tr979;
st355:
	if ( ++p == pe )
		goto _test_eof355;
case 355:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1106;
		case 32: goto tr1106;
		case 40: goto tr1107;
		case 41: goto tr1108;
		case 2058: goto tr1109;
		case 2107: goto tr1110;
		case 2314: goto tr1111;
		case 2363: goto tr1111;
		case 2570: goto tr1112;
		case 2619: goto tr1113;
	}
	goto tr979;
st356:
	if ( ++p == pe )
		goto _test_eof356;
case 356:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1114;
		case 32: goto tr1114;
		case 40: goto tr1115;
		case 41: goto tr1116;
		case 2058: goto tr1117;
		case 2107: goto tr1118;
		case 2314: goto tr1119;
		case 2363: goto tr1119;
		case 2570: goto tr1120;
		case 2619: goto tr1121;
	}
	goto tr979;
st357:
	if ( ++p == pe )
		goto _test_eof357;
case 357:
	switch( (*p) ) {
		case 85: goto st358;
		case 117: goto st358;
	}
	goto tr979;
st358:
	if ( ++p == pe )
		goto _test_eof358;
case 358:
	switch( (*p) ) {
		case 73: goto st359;
		case 105: goto st359;
	}
	goto tr979;
st359:
	if ( ++p == pe )
		goto _test_eof359;
case 359:
	switch( (*p) ) {
		case 52: goto st360;
		case 54: goto st362;
	}
	goto tr979;
st360:
	if ( ++p == pe )
		goto _test_eof360;
case 360:
	if ( (*p) == 56 )
		goto st361;
	goto tr979;
st361:
	if ( ++p == pe )
		goto _test_eof361;
case 361:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1127;
		case 32: goto tr1127;
		case 40: goto tr1128;
		case 41: goto tr1129;
		case 2058: goto tr1130;
		case 2107: goto tr1131;
		case 2314: goto tr1132;
		case 2363: goto tr1132;
		case 2570: goto tr1133;
		case 2619: goto tr1134;
	}
	goto tr979;
st362:
	if ( ++p == pe )
		goto _test_eof362;
case 362:
	if ( (*p) == 52 )
		goto st363;
	goto tr979;
st363:
	if ( ++p == pe )
		goto _test_eof363;
case 363:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1136;
		case 32: goto tr1136;
		case 40: goto tr1137;
		case 41: goto tr1138;
		case 2058: goto tr1139;
		case 2107: goto tr1140;
		case 2314: goto tr1141;
		case 2363: goto tr1141;
		case 2570: goto tr1142;
		case 2619: goto tr1143;
	}
	goto tr979;
st364:
	if ( ++p == pe )
		goto _test_eof364;
case 364:
	switch( (*p) ) {
		case 73: goto st365;
		case 105: goto st365;
	}
	goto tr979;
st365:
	if ( ++p == pe )
		goto _test_eof365;
case 365:
	switch( (*p) ) {
		case 78: goto st366;
		case 110: goto st366;
	}
	goto tr979;
st366:
	if ( ++p == pe )
		goto _test_eof366;
case 366:
	switch( (*p) ) {
		case 70: goto st367;
		case 102: goto st367;
	}
	goto tr979;
st367:
	if ( ++p == pe )
		goto _test_eof367;
case 367:
	switch( (*p) ) {
		case 79: goto st368;
		case 111: goto st368;
	}
	goto tr979;
st368:
	if ( ++p == pe )
		goto _test_eof368;
case 368:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1148;
		case 32: goto tr1148;
		case 40: goto tr1149;
		case 41: goto tr1150;
		case 2058: goto tr1151;
		case 2107: goto tr1152;
		case 2314: goto tr1153;
		case 2363: goto tr1153;
		case 2570: goto tr1154;
		case 2619: goto tr1155;
	}
	goto tr979;
st369:
	if ( ++p == pe )
		goto _test_eof369;
case 369:
	switch( (*p) ) {
		case 80: goto st370;
		case 112: goto st370;
	}
	goto tr979;
st370:
	if ( ++p == pe )
		goto _test_eof370;
case 370:
	switch( (*p) ) {
		case 83: goto st371;
		case 115: goto st371;
	}
	goto tr979;
st371:
	if ( ++p == pe )
		goto _test_eof371;
case 371:
	switch( (*p) ) {
		case 69: goto st372;
		case 101: goto st372;
	}
	goto tr979;
st372:
	if ( ++p == pe )
		goto _test_eof372;
case 372:
	switch( (*p) ) {
		case 67: goto st373;
		case 99: goto st373;
	}
	goto tr979;
st373:
	if ( ++p == pe )
		goto _test_eof373;
case 373:
	switch( (*p) ) {
		case 75: goto st374;
		case 107: goto st374;
	}
	goto tr979;
st374:
	if ( ++p == pe )
		goto _test_eof374;
case 374:
	switch( (*p) ) {
		case 69: goto st375;
		case 101: goto st375;
	}
	goto tr979;
st375:
	if ( ++p == pe )
		goto _test_eof375;
case 375:
	switch( (*p) ) {
		case 89: goto st376;
		case 121: goto st376;
	}
	goto tr979;
st376:
	if ( ++p == pe )
		goto _test_eof376;
case 376:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1163;
		case 32: goto tr1163;
		case 40: goto tr1164;
		case 41: goto tr1165;
		case 2058: goto tr1166;
		case 2107: goto tr1167;
		case 2314: goto tr1168;
		case 2363: goto tr1168;
		case 2570: goto tr1169;
		case 2619: goto tr1170;
	}
	goto tr979;
st377:
	if ( ++p == pe )
		goto _test_eof377;
case 377:
	switch( (*p) ) {
		case 69: goto st378;
		case 88: goto st380;
		case 101: goto st378;
		case 120: goto st380;
	}
	goto tr979;
st378:
	if ( ++p == pe )
		goto _test_eof378;
case 378:
	switch( (*p) ) {
		case 89: goto st379;
		case 121: goto st379;
	}
	goto tr979;
st379:
	if ( ++p == pe )
		goto _test_eof379;
case 379:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1174;
		case 32: goto tr1174;
		case 40: goto tr1175;
		case 41: goto tr1176;
		case 2058: goto tr1177;
		case 2107: goto tr1178;
		case 2314: goto tr1179;
		case 2363: goto tr1179;
		case 2570: goto tr1180;
		case 2619: goto tr1181;
	}
	goto tr979;
st380:
	if ( ++p == pe )
		goto _test_eof380;
case 380:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1182;
		case 32: goto tr1182;
		case 40: goto tr1183;
		case 41: goto tr1184;
		case 2058: goto tr1185;
		case 2107: goto tr1186;
		case 2314: goto tr1187;
		case 2363: goto tr1187;
		case 2570: goto tr1188;
		case 2619: goto tr1189;
	}
	goto tr979;
st381:
	if ( ++p == pe )
		goto _test_eof381;
case 381:
	switch( (*p) ) {
		case 51: goto st382;
		case 54: goto st384;
		case 79: goto st386;
		case 80: goto st388;
		case 111: goto st386;
		case 112: goto st388;
	}
	goto tr979;
st382:
	if ( ++p == pe )
		goto _test_eof382;
case 382:
	if ( (*p) == 50 )
		goto st383;
	goto tr979;
st383:
	if ( ++p == pe )
		goto _test_eof383;
case 383:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1195;
		case 32: goto tr1195;
		case 40: goto tr1196;
		case 41: goto tr1197;
		case 2058: goto tr1198;
		case 2107: goto tr1199;
		case 2314: goto tr1200;
		case 2363: goto tr1200;
		case 2570: goto tr1201;
		case 2619: goto tr1202;
	}
	goto tr979;
st384:
	if ( ++p == pe )
		goto _test_eof384;
case 384:
	if ( (*p) == 52 )
		goto st385;
	goto tr979;
st385:
	if ( ++p == pe )
		goto _test_eof385;
case 385:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1204;
		case 32: goto tr1204;
		case 40: goto tr1205;
		case 41: goto tr1206;
		case 2058: goto tr1207;
		case 2107: goto tr1208;
		case 2314: goto tr1209;
		case 2363: goto tr1209;
		case 2570: goto tr1210;
		case 2619: goto tr1211;
	}
	goto tr979;
st386:
	if ( ++p == pe )
		goto _test_eof386;
case 386:
	switch( (*p) ) {
		case 67: goto st387;
		case 99: goto st387;
	}
	goto tr979;
st387:
	if ( ++p == pe )
		goto _test_eof387;
case 387:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1213;
		case 32: goto tr1213;
		case 40: goto tr1214;
		case 41: goto tr1215;
		case 2058: goto tr1216;
		case 2107: goto tr1217;
		case 2314: goto tr1218;
		case 2363: goto tr1218;
		case 2570: goto tr1219;
		case 2619: goto tr1220;
	}
	goto tr979;
st388:
	if ( ++p == pe )
		goto _test_eof388;
case 388:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1221;
		case 32: goto tr1221;
		case 40: goto tr1222;
		case 41: goto tr1223;
		case 2058: goto tr1224;
		case 2107: goto tr1225;
		case 2314: goto tr1226;
		case 2363: goto tr1226;
		case 2570: goto tr1227;
		case 2619: goto tr1228;
	}
	goto tr979;
st389:
	if ( ++p == pe )
		goto _test_eof389;
case 389:
	switch( (*p) ) {
		case 73: goto st390;
		case 88: goto st394;
		case 105: goto st390;
		case 120: goto st394;
	}
	goto tr979;
st390:
	if ( ++p == pe )
		goto _test_eof390;
case 390:
	switch( (*p) ) {
		case 78: goto st391;
		case 110: goto st391;
	}
	goto tr979;
st391:
	if ( ++p == pe )
		goto _test_eof391;
case 391:
	switch( (*p) ) {
		case 70: goto st392;
		case 102: goto st392;
	}
	goto tr979;
st392:
	if ( ++p == pe )
		goto _test_eof392;
case 392:
	switch( (*p) ) {
		case 79: goto st393;
		case 111: goto st393;
	}
	goto tr979;
st393:
	if ( ++p == pe )
		goto _test_eof393;
case 393:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1234;
		case 32: goto tr1234;
		case 40: goto tr1235;
		case 41: goto tr1236;
		case 2058: goto tr1237;
		case 2107: goto tr1238;
		case 2314: goto tr1239;
		case 2363: goto tr1239;
		case 2570: goto tr1240;
		case 2619: goto tr1241;
	}
	goto tr979;
st394:
	if ( ++p == pe )
		goto _test_eof394;
case 394:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1242;
		case 32: goto tr1242;
		case 40: goto tr1243;
		case 41: goto tr1244;
		case 2058: goto tr1245;
		case 2107: goto tr1246;
		case 2314: goto tr1247;
		case 2363: goto tr1247;
		case 2570: goto tr1248;
		case 2619: goto tr1249;
	}
	goto tr979;
st395:
	if ( ++p == pe )
		goto _test_eof395;
case 395:
	switch( (*p) ) {
		case 65: goto st396;
		case 73: goto st400;
		case 83: goto st402;
		case 97: goto st396;
		case 105: goto st400;
		case 115: goto st402;
	}
	goto tr979;
st396:
	if ( ++p == pe )
		goto _test_eof396;
case 396:
	switch( (*p) ) {
		case 80: goto st397;
		case 112: goto st397;
	}
	goto tr979;
st397:
	if ( ++p == pe )
		goto _test_eof397;
case 397:
	switch( (*p) ) {
		case 84: goto st398;
		case 116: goto st398;
	}
	goto tr979;
st398:
	if ( ++p == pe )
		goto _test_eof398;
case 398:
	switch( (*p) ) {
		case 82: goto st399;
		case 114: goto st399;
	}
	goto tr979;
st399:
	if ( ++p == pe )
		goto _test_eof399;
case 399:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1256;
		case 32: goto tr1256;
		case 40: goto tr1257;
		case 41: goto tr1258;
		case 2058: goto tr1259;
		case 2107: goto tr1260;
		case 2314: goto tr1261;
		case 2363: goto tr1261;
		case 2570: goto tr1262;
		case 2619: goto tr1263;
	}
	goto tr979;
st400:
	if ( ++p == pe )
		goto _test_eof400;
case 400:
	switch( (*p) ) {
		case 68: goto st401;
		case 100: goto st401;
	}
	goto tr979;
st401:
	if ( ++p == pe )
		goto _test_eof401;
case 401:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1265;
		case 32: goto tr1265;
		case 40: goto tr1266;
		case 41: goto tr1267;
		case 2058: goto tr1268;
		case 2107: goto tr1269;
		case 2314: goto tr1270;
		case 2363: goto tr1270;
		case 2570: goto tr1271;
		case 2619: goto tr1272;
	}
	goto tr979;
st402:
	if ( ++p == pe )
		goto _test_eof402;
case 402:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1273;
		case 32: goto tr1273;
		case 40: goto tr1274;
		case 41: goto tr1275;
		case 69: goto st403;
		case 101: goto st403;
		case 2058: goto tr1277;
		case 2107: goto tr1278;
		case 2314: goto tr1279;
		case 2363: goto tr1279;
		case 2570: goto tr1280;
		case 2619: goto tr1281;
	}
	goto tr979;
st403:
	if ( ++p == pe )
		goto _test_eof403;
case 403:
	switch( (*p) ) {
		case 67: goto st404;
		case 99: goto st404;
	}
	goto tr979;
st404:
	if ( ++p == pe )
		goto _test_eof404;
case 404:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1283;
		case 32: goto tr1283;
		case 40: goto tr1284;
		case 41: goto tr1285;
		case 51: goto st405;
		case 2058: goto tr1287;
		case 2107: goto tr1288;
		case 2314: goto tr1289;
		case 2363: goto tr1289;
		case 2570: goto tr1290;
		case 2619: goto tr1291;
	}
	goto tr979;
st405:
	if ( ++p == pe )
		goto _test_eof405;
case 405:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1292;
		case 32: goto tr1292;
		case 40: goto tr1293;
		case 41: goto tr1294;
		case 80: goto st406;
		case 112: goto st406;
		case 2058: goto tr1296;
		case 2107: goto tr1297;
		case 2314: goto tr1298;
		case 2363: goto tr1298;
		case 2570: goto tr1299;
		case 2619: goto tr1300;
	}
	goto tr979;
st406:
	if ( ++p == pe )
		goto _test_eof406;
case 406:
	switch( (*p) ) {
		case 65: goto st407;
		case 97: goto st407;
	}
	goto tr979;
st407:
	if ( ++p == pe )
		goto _test_eof407;
case 407:
	switch( (*p) ) {
		case 82: goto st408;
		case 114: goto st408;
	}
	goto tr979;
st408:
	if ( ++p == pe )
		goto _test_eof408;
case 408:
	switch( (*p) ) {
		case 65: goto st409;
		case 97: goto st409;
	}
	goto tr979;
st409:
	if ( ++p == pe )
		goto _test_eof409;
case 409:
	switch( (*p) ) {
		case 77: goto st410;
		case 109: goto st410;
	}
	goto tr979;
st410:
	if ( ++p == pe )
		goto _test_eof410;
case 410:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1305;
		case 32: goto tr1305;
		case 40: goto tr1306;
		case 41: goto tr1307;
		case 2058: goto tr1308;
		case 2107: goto tr1309;
		case 2314: goto tr1310;
		case 2363: goto tr1310;
		case 2570: goto tr1311;
		case 2619: goto tr1312;
	}
	goto tr979;
st411:
	if ( ++p == pe )
		goto _test_eof411;
case 411:
	switch( (*p) ) {
		case 84: goto st412;
		case 116: goto st412;
	}
	goto tr979;
st412:
	if ( ++p == pe )
		goto _test_eof412;
case 412:
	switch( (*p) ) {
		case 82: goto st413;
		case 114: goto st413;
	}
	goto tr979;
st413:
	if ( ++p == pe )
		goto _test_eof413;
case 413:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1315;
		case 32: goto tr1315;
		case 40: goto tr1316;
		case 41: goto tr1317;
		case 2058: goto tr1318;
		case 2107: goto tr1319;
		case 2314: goto tr1320;
		case 2363: goto tr1320;
		case 2570: goto tr1321;
		case 2619: goto tr1322;
	}
	goto tr979;
st414:
	if ( ++p == pe )
		goto _test_eof414;
case 414:
	switch( (*p) ) {
		case 80: goto st415;
		case 82: goto st416;
		case 84: goto st420;
		case 112: goto st415;
		case 114: goto st416;
		case 116: goto st420;
	}
	goto tr979;
st415:
	if ( ++p == pe )
		goto _test_eof415;
case 415:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1326;
		case 32: goto tr1326;
		case 40: goto tr1327;
		case 41: goto tr1328;
		case 2058: goto tr1329;
		case 2107: goto tr1330;
		case 2314: goto tr1331;
		case 2363: goto tr1331;
		case 2570: goto tr1332;
		case 2619: goto tr1333;
	}
	goto tr979;
st416:
	if ( ++p == pe )
		goto _test_eof416;
case 416:
	switch( (*p) ) {
		case 83: goto st417;
		case 115: goto st417;
	}
	goto tr979;
st417:
	if ( ++p == pe )
		goto _test_eof417;
case 417:
	switch( (*p) ) {
		case 73: goto st418;
		case 105: goto st418;
	}
	goto tr979;
st418:
	if ( ++p == pe )
		goto _test_eof418;
case 418:
	switch( (*p) ) {
		case 71: goto st419;
		case 103: goto st419;
	}
	goto tr979;
st419:
	if ( ++p == pe )
		goto _test_eof419;
case 419:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1337;
		case 32: goto tr1337;
		case 40: goto tr1338;
		case 41: goto tr1339;
		case 2058: goto tr1340;
		case 2107: goto tr1341;
		case 2314: goto tr1342;
		case 2363: goto tr1342;
		case 2570: goto tr1343;
		case 2619: goto tr1344;
	}
	goto tr979;
st420:
	if ( ++p == pe )
		goto _test_eof420;
case 420:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1345;
		case 32: goto tr1345;
		case 40: goto tr1346;
		case 41: goto tr1347;
		case 2058: goto tr1348;
		case 2107: goto tr1349;
		case 2314: goto tr1350;
		case 2363: goto tr1350;
		case 2570: goto tr1351;
		case 2619: goto tr1352;
	}
	goto tr979;
st421:
	if ( ++p == pe )
		goto _test_eof421;
case 421:
	switch( (*p) ) {
		case 79: goto st422;
		case 80: goto st424;
		case 82: goto st426;
		case 83: goto st428;
		case 111: goto st422;
		case 112: goto st424;
		case 114: goto st426;
		case 115: goto st428;
	}
	goto tr979;
st422:
	if ( ++p == pe )
		goto _test_eof422;
case 422:
	switch( (*p) ) {
		case 65: goto st423;
		case 97: goto st423;
	}
	goto tr979;
st423:
	if ( ++p == pe )
		goto _test_eof423;
case 423:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1358;
		case 32: goto tr1358;
		case 40: goto tr1359;
		case 41: goto tr1360;
		case 2058: goto tr1361;
		case 2107: goto tr1362;
		case 2314: goto tr1363;
		case 2363: goto tr1363;
		case 2570: goto tr1364;
		case 2619: goto tr1365;
	}
	goto tr979;
st424:
	if ( ++p == pe )
		goto _test_eof424;
case 424:
	switch( (*p) ) {
		case 70: goto st425;
		case 102: goto st425;
	}
	goto tr979;
st425:
	if ( ++p == pe )
		goto _test_eof425;
case 425:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1367;
		case 32: goto tr1367;
		case 40: goto tr1368;
		case 41: goto tr1369;
		case 2058: goto tr1370;
		case 2107: goto tr1371;
		case 2314: goto tr1372;
		case 2363: goto tr1372;
		case 2570: goto tr1373;
		case 2619: goto tr1374;
	}
	goto tr979;
st426:
	if ( ++p == pe )
		goto _test_eof426;
case 426:
	switch( (*p) ) {
		case 86: goto st427;
		case 118: goto st427;
	}
	goto tr979;
st427:
	if ( ++p == pe )
		goto _test_eof427;
case 427:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1376;
		case 32: goto tr1376;
		case 40: goto tr1377;
		case 41: goto tr1378;
		case 2058: goto tr1379;
		case 2107: goto tr1380;
		case 2314: goto tr1381;
		case 2363: goto tr1381;
		case 2570: goto tr1382;
		case 2619: goto tr1383;
	}
	goto tr979;
st428:
	if ( ++p == pe )
		goto _test_eof428;
case 428:
	switch( (*p) ) {
		case 72: goto st429;
		case 104: goto st429;
	}
	goto tr979;
st429:
	if ( ++p == pe )
		goto _test_eof429;
case 429:
	switch( (*p) ) {
		case 70: goto st430;
		case 102: goto st430;
	}
	goto tr979;
st430:
	if ( ++p == pe )
		goto _test_eof430;
case 430:
	switch( (*p) ) {
		case 80: goto st431;
		case 112: goto st431;
	}
	goto tr979;
st431:
	if ( ++p == pe )
		goto _test_eof431;
case 431:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1387;
		case 32: goto tr1387;
		case 40: goto tr1388;
		case 41: goto tr1389;
		case 2058: goto tr1390;
		case 2107: goto tr1391;
		case 2314: goto tr1392;
		case 2363: goto tr1392;
		case 2570: goto tr1393;
		case 2619: goto tr1394;
	}
	goto tr979;
st432:
	if ( ++p == pe )
		goto _test_eof432;
case 432:
	switch( (*p) ) {
		case 76: goto st433;
		case 88: goto st436;
		case 89: goto st438;
		case 108: goto st433;
		case 120: goto st436;
		case 121: goto st438;
	}
	goto tr979;
st433:
	if ( ++p == pe )
		goto _test_eof433;
case 433:
	switch( (*p) ) {
		case 83: goto st434;
		case 115: goto st434;
	}
	goto tr979;
st434:
	if ( ++p == pe )
		goto _test_eof434;
case 434:
	switch( (*p) ) {
		case 65: goto st435;
		case 97: goto st435;
	}
	goto tr979;
st435:
	if ( ++p == pe )
		goto _test_eof435;
case 435:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1400;
		case 32: goto tr1400;
		case 40: goto tr1401;
		case 41: goto tr1402;
		case 2058: goto tr1403;
		case 2107: goto tr1404;
		case 2314: goto tr1405;
		case 2363: goto tr1405;
		case 2570: goto tr1406;
		case 2619: goto tr1407;
	}
	goto tr979;
st436:
	if ( ++p == pe )
		goto _test_eof436;
case 436:
	switch( (*p) ) {
		case 84: goto st437;
		case 116: goto st437;
	}
	goto tr979;
st437:
	if ( ++p == pe )
		goto _test_eof437;
case 437:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1409;
		case 32: goto tr1409;
		case 40: goto tr1410;
		case 41: goto tr1411;
		case 2058: goto tr1412;
		case 2107: goto tr1413;
		case 2314: goto tr1414;
		case 2363: goto tr1414;
		case 2570: goto tr1415;
		case 2619: goto tr1416;
	}
	goto tr979;
st438:
	if ( ++p == pe )
		goto _test_eof438;
case 438:
	switch( (*p) ) {
		case 80: goto st439;
		case 112: goto st439;
	}
	goto tr979;
st439:
	if ( ++p == pe )
		goto _test_eof439;
case 439:
	switch( (*p) ) {
		case 69: goto st440;
		case 101: goto st440;
	}
	goto tr979;
st440:
	if ( ++p == pe )
		goto _test_eof440;
case 440:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1419;
	goto tr979;
tr1419:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st441;
tr1423:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st441;
st441:
	if ( ++p == pe )
		goto _test_eof441;
case 441:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1420;
		case 32: goto tr1420;
		case 40: goto tr1421;
		case 41: goto tr1422;
		case 2058: goto tr1424;
		case 2107: goto tr1425;
		case 2314: goto tr1426;
		case 2363: goto tr1426;
		case 2570: goto tr1427;
		case 2619: goto tr1428;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1423;
	goto tr979;
st442:
	if ( ++p == pe )
		goto _test_eof442;
case 442:
	switch( (*p) ) {
		case 83: goto st443;
		case 115: goto st443;
	}
	goto tr979;
st443:
	if ( ++p == pe )
		goto _test_eof443;
case 443:
	switch( (*p) ) {
		case 68: goto st444;
		case 100: goto st444;
	}
	goto tr979;
st444:
	if ( ++p == pe )
		goto _test_eof444;
case 444:
	switch( (*p) ) {
		case 66: goto st445;
		case 98: goto st445;
	}
	goto tr979;
st445:
	if ( ++p == pe )
		goto _test_eof445;
case 445:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1432;
		case 32: goto tr1432;
		case 40: goto tr1433;
		case 41: goto tr1434;
		case 2058: goto tr1435;
		case 2107: goto tr1436;
		case 2314: goto tr1437;
		case 2363: goto tr1437;
		case 2570: goto tr1438;
		case 2619: goto tr1439;
	}
	goto tr979;
st446:
	if ( ++p == pe )
		goto _test_eof446;
case 446:
	switch( (*p) ) {
		case 76: goto st447;
		case 108: goto st447;
	}
	goto tr979;
st447:
	if ( ++p == pe )
		goto _test_eof447;
case 447:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1441;
		case 32: goto tr1441;
		case 40: goto tr1442;
		case 41: goto tr1443;
		case 2058: goto tr1444;
		case 2107: goto tr1445;
		case 2314: goto tr1446;
		case 2363: goto tr1446;
		case 2570: goto tr1447;
		case 2619: goto tr1448;
	}
	goto tr979;
st448:
	if ( ++p == pe )
		goto _test_eof448;
case 448:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st449;
		case 32: goto st449;
		case 40: goto tr1451;
		case 41: goto tr1452;
		case 1034: goto tr1453;
		case 1083: goto st456;
	}
	goto tr1449;
tr1451:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st449;
tr1452:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st449;
tr1453:
	{
		s->line_counter++;
	}
	goto st449;
st449:
	if ( ++p == pe )
		goto _test_eof449;
case 449:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st449;
		case 32: goto st449;
		case 40: goto tr1451;
		case 41: goto tr1452;
		case 1034: goto tr1453;
		case 1083: goto st456;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1456;
	goto tr1455;
tr1456:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st450;
tr1460:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st450;
st450:
	if ( ++p == pe )
		goto _test_eof450;
case 450:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1457;
		case 32: goto tr1457;
		case 40: goto tr1458;
		case 41: goto tr1459;
		case 1034: goto tr1461;
		case 1083: goto tr1462;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1460;
	goto tr1455;
tr1465:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st451;
tr1466:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st451;
tr1468:
	{
		s->line_counter++;
	}
	goto st451;
tr1457:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st451;
tr1458:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st451;
tr1459:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st451;
tr1461:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st451;
st451:
	if ( ++p == pe )
		goto _test_eof451;
case 451:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st451;
		case 32: goto st451;
		case 40: goto tr1465;
		case 41: goto tr1466;
		case 1034: goto tr1468;
		case 1083: goto st455;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1467;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1467;
	} else
		goto tr1467;
	goto tr1463;
tr1467:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st452;
st452:
	if ( ++p == pe )
		goto _test_eof452;
case 452:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1470;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr1470;
	} else
		goto tr1470;
	goto tr1463;
tr1472:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st453;
tr1473:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st453;
tr1474:
	{
		s->line_counter++;
	}
	goto st453;
tr1470:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st453;
st453:
	if ( ++p == pe )
		goto _test_eof453;
case 453:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st453;
		case 32: goto st453;
		case 40: goto tr1472;
		case 41: goto tr1473;
		case 2058: goto tr1474;
		case 2107: goto st454;
		case 2314: goto tr1476;
		case 2363: goto tr1476;
		case 2570: goto tr1477;
		case 2619: goto tr1478;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1467;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1467;
	} else
		goto tr1467;
	goto tr1463;
st454:
	if ( ++p == pe )
		goto _test_eof454;
case 454:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1474;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st454;
	goto tr1463;
tr1476:
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			WARN(ZS_BAD_RDATA_LENGTH);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1101;
st1101:
	if ( ++p == pe )
		goto _test_eof1101;
case 1101:
	goto st0;
tr1477:
	{
		s->line_counter++;
	}
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			WARN(ZS_BAD_RDATA_LENGTH);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1102;
st1102:
	if ( ++p == pe )
		goto _test_eof1102;
case 1102:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st453;
		case 32: goto st453;
		case 40: goto tr1472;
		case 41: goto tr1473;
		case 2058: goto tr1474;
		case 2107: goto st454;
		case 2314: goto tr1476;
		case 2363: goto tr1476;
		case 2570: goto tr1477;
		case 2619: goto tr1478;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1467;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1467;
	} else
		goto tr1467;
	goto tr1463;
tr1478:
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			WARN(ZS_BAD_RDATA_LENGTH);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1103;
st1103:
	if ( ++p == pe )
		goto _test_eof1103;
case 1103:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1474;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st454;
	goto tr1463;
tr1462:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st455;
st455:
	if ( ++p == pe )
		goto _test_eof455;
case 455:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1468;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st455;
	goto tr1449;
st456:
	if ( ++p == pe )
		goto _test_eof456;
case 456:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1453;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st456;
	goto tr1449;
st457:
	if ( ++p == pe )
		goto _test_eof457;
case 457:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st458;
		case 32: goto st458;
		case 40: goto tr1480;
		case 41: goto tr1481;
		case 1034: goto tr1482;
		case 1083: goto st466;
	}
	goto tr1449;
tr1480:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st458;
tr1481:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st458;
tr1482:
	{
		s->line_counter++;
	}
	goto st458;
st458:
	if ( ++p == pe )
		goto _test_eof458;
case 458:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st458;
		case 32: goto st458;
		case 40: goto tr1480;
		case 41: goto tr1481;
		case 48: goto tr1484;
		case 1034: goto tr1482;
		case 1083: goto st466;
	}
	if ( 49 <= _widec && _widec <= 57 )
		goto tr1485;
	goto tr1455;
tr1484:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st459;
st459:
	if ( ++p == pe )
		goto _test_eof459;
case 459:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1486;
		case 32: goto tr1486;
		case 40: goto tr1487;
		case 41: goto tr1488;
		case 778: goto tr1490;
		case 827: goto tr1490;
		case 1034: goto tr1491;
		case 1083: goto tr1492;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1489;
	goto tr1455;
tr1486:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1104;
tr1487:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1104;
tr1488:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1104;
tr1491:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1104;
st1104:
	if ( ++p == pe )
		goto _test_eof1104;
case 1104:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st460;
		case 32: goto st460;
		case 40: goto tr1494;
		case 41: goto tr1495;
		case 1034: goto tr1497;
		case 1083: goto st464;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1496;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1496;
	} else
		goto tr1496;
	goto tr1463;
tr1494:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st460;
tr1495:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st460;
tr1497:
	{
		s->line_counter++;
	}
	goto st460;
tr1508:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st460;
tr1509:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st460;
tr1510:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st460;
tr1511:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st460;
st460:
	if ( ++p == pe )
		goto _test_eof460;
case 460:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st460;
		case 32: goto st460;
		case 40: goto tr1494;
		case 41: goto tr1495;
		case 1034: goto tr1497;
		case 1083: goto st464;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1496;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1496;
	} else
		goto tr1496;
	goto tr1463;
tr1496:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st461;
st461:
	if ( ++p == pe )
		goto _test_eof461;
case 461:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1499;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr1499;
	} else
		goto tr1499;
	goto tr1463;
tr1501:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st462;
tr1502:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st462;
tr1503:
	{
		s->line_counter++;
	}
	goto st462;
tr1499:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st462;
st462:
	if ( ++p == pe )
		goto _test_eof462;
case 462:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st462;
		case 32: goto st462;
		case 40: goto tr1501;
		case 41: goto tr1502;
		case 2058: goto tr1503;
		case 2107: goto st463;
		case 2314: goto tr1505;
		case 2363: goto tr1505;
		case 2570: goto tr1506;
		case 2619: goto tr1507;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1496;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1496;
	} else
		goto tr1496;
	goto tr1463;
st463:
	if ( ++p == pe )
		goto _test_eof463;
case 463:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1503;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st463;
	goto tr1463;
tr1490:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1105;
tr1505:
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			WARN(ZS_BAD_RDATA_LENGTH);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1105;
st1105:
	if ( ++p == pe )
		goto _test_eof1105;
case 1105:
	goto tr1449;
tr1506:
	{
		s->line_counter++;
	}
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			WARN(ZS_BAD_RDATA_LENGTH);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1106;
st1106:
	if ( ++p == pe )
		goto _test_eof1106;
case 1106:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st462;
		case 32: goto st462;
		case 40: goto tr1501;
		case 41: goto tr1502;
		case 2058: goto tr1503;
		case 2107: goto st463;
		case 2314: goto tr1505;
		case 2363: goto tr1505;
		case 2570: goto tr1506;
		case 2619: goto tr1507;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr1496;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr1496;
	} else
		goto tr1496;
	goto tr1463;
tr1507:
	{
		if ((rdata_tail - s->r_data) != s->r_data_length) {
			WARN(ZS_BAD_RDATA_LENGTH);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1107;
st1107:
	if ( ++p == pe )
		goto _test_eof1107;
case 1107:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1503;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st463;
	goto tr1463;
tr1512:
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st464;
st464:
	if ( ++p == pe )
		goto _test_eof464;
case 464:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1497;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st464;
	goto tr1449;
tr1485:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st465;
tr1489:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st465;
st465:
	if ( ++p == pe )
		goto _test_eof465;
case 465:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1508;
		case 32: goto tr1508;
		case 40: goto tr1509;
		case 41: goto tr1510;
		case 1034: goto tr1511;
		case 1083: goto tr1512;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1489;
	goto tr1455;
tr1492:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	{
		if (s->number64 <= UINT16_MAX) {
			s->r_data_length = (uint16_t)(s->number64);
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1108;
st1108:
	if ( ++p == pe )
		goto _test_eof1108;
case 1108:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1497;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st464;
	goto tr1449;
st466:
	if ( ++p == pe )
		goto _test_eof466;
case 466:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1482;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st466;
	goto tr1449;
st467:
	if ( ++p == pe )
		goto _test_eof467;
case 467:
	switch( (*p) ) {
		case 68: goto st469;
		case 69: goto st484;
		case 73: goto st514;
		case 80: goto st522;
		case 82: goto st535;
		case 100: goto st469;
		case 101: goto st484;
		case 105: goto st514;
		case 112: goto st522;
		case 114: goto st535;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1514;
	goto tr1513;
tr1514:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st468;
tr1521:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st468;
st468:
	if ( ++p == pe )
		goto _test_eof468;
case 468:
	switch( (*p) ) {
		case 32: goto tr1520;
		case 59: goto tr1520;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1520;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1521;
	} else
		goto tr1520;
	goto tr1513;
tr1520:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1109;
tr1524:
	{
		*(rdata_tail++) = 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1109;
tr1526:
	{
		*(rdata_tail++) = 3;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1109;
tr1538:
	{
		*(rdata_tail++) = 6;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1109;
tr1547:
	{
		*(rdata_tail++) = 12;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1109;
tr1561:
	{
		*(rdata_tail++) = 13;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1109;
tr1570:
	{
		*(rdata_tail++) = 14;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1109;
tr1578:
	{
		*(rdata_tail++) = 252;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1109;
tr1589:
	{
		*(rdata_tail++) = 253;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1109;
tr1592:
	{
		*(rdata_tail++) = 254;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1109;
tr1599:
	{
		*(rdata_tail++) = 1;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1109;
tr1605:
	{
		*(rdata_tail++) = 5;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1109;
tr1617:
	{
		*(rdata_tail++) = 7;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1109;
tr1620:
	{
		*(rdata_tail++) = 8;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1109;
tr1623:
	{
		*(rdata_tail++) = 10;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1109;
st1109:
	if ( ++p == pe )
		goto _test_eof1109;
case 1109:
	goto st0;
st469:
	if ( ++p == pe )
		goto _test_eof469;
case 469:
	switch( (*p) ) {
		case 72: goto st470;
		case 83: goto st471;
		case 104: goto st470;
		case 115: goto st471;
	}
	goto tr1513;
st470:
	if ( ++p == pe )
		goto _test_eof470;
case 470:
	switch( (*p) ) {
		case 32: goto tr1524;
		case 59: goto tr1524;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1524;
	} else if ( (*p) >= 9 )
		goto tr1524;
	goto tr1513;
st471:
	if ( ++p == pe )
		goto _test_eof471;
case 471:
	switch( (*p) ) {
		case 65: goto st472;
		case 97: goto st472;
	}
	goto tr1513;
st472:
	if ( ++p == pe )
		goto _test_eof472;
case 472:
	switch( (*p) ) {
		case 32: goto tr1526;
		case 45: goto st473;
		case 59: goto tr1526;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1526;
	} else if ( (*p) >= 9 )
		goto tr1526;
	goto tr1513;
st473:
	if ( ++p == pe )
		goto _test_eof473;
case 473:
	switch( (*p) ) {
		case 78: goto st474;
		case 110: goto st474;
	}
	goto tr1513;
st474:
	if ( ++p == pe )
		goto _test_eof474;
case 474:
	switch( (*p) ) {
		case 83: goto st475;
		case 115: goto st475;
	}
	goto tr1513;
st475:
	if ( ++p == pe )
		goto _test_eof475;
case 475:
	switch( (*p) ) {
		case 69: goto st476;
		case 101: goto st476;
	}
	goto tr1513;
st476:
	if ( ++p == pe )
		goto _test_eof476;
case 476:
	switch( (*p) ) {
		case 67: goto st477;
		case 99: goto st477;
	}
	goto tr1513;
st477:
	if ( ++p == pe )
		goto _test_eof477;
case 477:
	if ( (*p) == 51 )
		goto st478;
	goto tr1513;
st478:
	if ( ++p == pe )
		goto _test_eof478;
case 478:
	if ( (*p) == 45 )
		goto st479;
	goto tr1513;
st479:
	if ( ++p == pe )
		goto _test_eof479;
case 479:
	switch( (*p) ) {
		case 83: goto st480;
		case 115: goto st480;
	}
	goto tr1513;
st480:
	if ( ++p == pe )
		goto _test_eof480;
case 480:
	switch( (*p) ) {
		case 72: goto st481;
		case 104: goto st481;
	}
	goto tr1513;
st481:
	if ( ++p == pe )
		goto _test_eof481;
case 481:
	switch( (*p) ) {
		case 65: goto st482;
		case 97: goto st482;
	}
	goto tr1513;
st482:
	if ( ++p == pe )
		goto _test_eof482;
case 482:
	if ( (*p) == 49 )
		goto st483;
	goto tr1513;
st483:
	if ( ++p == pe )
		goto _test_eof483;
case 483:
	switch( (*p) ) {
		case 32: goto tr1538;
		case 59: goto tr1538;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1538;
	} else if ( (*p) >= 9 )
		goto tr1538;
	goto tr1513;
st484:
	if ( ++p == pe )
		goto _test_eof484;
case 484:
	switch( (*p) ) {
		case 67: goto st485;
		case 99: goto st485;
	}
	goto tr1513;
st485:
	if ( ++p == pe )
		goto _test_eof485;
case 485:
	switch( (*p) ) {
		case 67: goto st486;
		case 68: goto st492;
		case 99: goto st486;
		case 100: goto st492;
	}
	goto tr1513;
st486:
	if ( ++p == pe )
		goto _test_eof486;
case 486:
	if ( (*p) == 45 )
		goto st487;
	goto tr1513;
st487:
	if ( ++p == pe )
		goto _test_eof487;
case 487:
	switch( (*p) ) {
		case 71: goto st488;
		case 103: goto st488;
	}
	goto tr1513;
st488:
	if ( ++p == pe )
		goto _test_eof488;
case 488:
	switch( (*p) ) {
		case 79: goto st489;
		case 111: goto st489;
	}
	goto tr1513;
st489:
	if ( ++p == pe )
		goto _test_eof489;
case 489:
	switch( (*p) ) {
		case 83: goto st490;
		case 115: goto st490;
	}
	goto tr1513;
st490:
	if ( ++p == pe )
		goto _test_eof490;
case 490:
	switch( (*p) ) {
		case 84: goto st491;
		case 116: goto st491;
	}
	goto tr1513;
st491:
	if ( ++p == pe )
		goto _test_eof491;
case 491:
	switch( (*p) ) {
		case 32: goto tr1547;
		case 59: goto tr1547;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1547;
	} else if ( (*p) >= 9 )
		goto tr1547;
	goto tr1513;
st492:
	if ( ++p == pe )
		goto _test_eof492;
case 492:
	switch( (*p) ) {
		case 83: goto st493;
		case 115: goto st493;
	}
	goto tr1513;
st493:
	if ( ++p == pe )
		goto _test_eof493;
case 493:
	switch( (*p) ) {
		case 65: goto st494;
		case 97: goto st494;
	}
	goto tr1513;
st494:
	if ( ++p == pe )
		goto _test_eof494;
case 494:
	switch( (*p) ) {
		case 80: goto st495;
		case 112: goto st495;
	}
	goto tr1513;
st495:
	if ( ++p == pe )
		goto _test_eof495;
case 495:
	switch( (*p) ) {
		case 50: goto st496;
		case 51: goto st505;
	}
	goto tr1513;
st496:
	if ( ++p == pe )
		goto _test_eof496;
case 496:
	if ( (*p) == 53 )
		goto st497;
	goto tr1513;
st497:
	if ( ++p == pe )
		goto _test_eof497;
case 497:
	if ( (*p) == 54 )
		goto st498;
	goto tr1513;
st498:
	if ( ++p == pe )
		goto _test_eof498;
case 498:
	switch( (*p) ) {
		case 83: goto st499;
		case 115: goto st499;
	}
	goto tr1513;
st499:
	if ( ++p == pe )
		goto _test_eof499;
case 499:
	switch( (*p) ) {
		case 72: goto st500;
		case 104: goto st500;
	}
	goto tr1513;
st500:
	if ( ++p == pe )
		goto _test_eof500;
case 500:
	switch( (*p) ) {
		case 65: goto st501;
		case 97: goto st501;
	}
	goto tr1513;
st501:
	if ( ++p == pe )
		goto _test_eof501;
case 501:
	if ( (*p) == 50 )
		goto st502;
	goto tr1513;
st502:
	if ( ++p == pe )
		goto _test_eof502;
case 502:
	if ( (*p) == 53 )
		goto st503;
	goto tr1513;
st503:
	if ( ++p == pe )
		goto _test_eof503;
case 503:
	if ( (*p) == 54 )
		goto st504;
	goto tr1513;
st504:
	if ( ++p == pe )
		goto _test_eof504;
case 504:
	switch( (*p) ) {
		case 32: goto tr1561;
		case 59: goto tr1561;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1561;
	} else if ( (*p) >= 9 )
		goto tr1561;
	goto tr1513;
st505:
	if ( ++p == pe )
		goto _test_eof505;
case 505:
	if ( (*p) == 56 )
		goto st506;
	goto tr1513;
st506:
	if ( ++p == pe )
		goto _test_eof506;
case 506:
	if ( (*p) == 52 )
		goto st507;
	goto tr1513;
st507:
	if ( ++p == pe )
		goto _test_eof507;
case 507:
	switch( (*p) ) {
		case 83: goto st508;
		case 115: goto st508;
	}
	goto tr1513;
st508:
	if ( ++p == pe )
		goto _test_eof508;
case 508:
	switch( (*p) ) {
		case 72: goto st509;
		case 104: goto st509;
	}
	goto tr1513;
st509:
	if ( ++p == pe )
		goto _test_eof509;
case 509:
	switch( (*p) ) {
		case 65: goto st510;
		case 97: goto st510;
	}
	goto tr1513;
st510:
	if ( ++p == pe )
		goto _test_eof510;
case 510:
	if ( (*p) == 51 )
		goto st511;
	goto tr1513;
st511:
	if ( ++p == pe )
		goto _test_eof511;
case 511:
	if ( (*p) == 56 )
		goto st512;
	goto tr1513;
st512:
	if ( ++p == pe )
		goto _test_eof512;
case 512:
	if ( (*p) == 52 )
		goto st513;
	goto tr1513;
st513:
	if ( ++p == pe )
		goto _test_eof513;
case 513:
	switch( (*p) ) {
		case 32: goto tr1570;
		case 59: goto tr1570;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1570;
	} else if ( (*p) >= 9 )
		goto tr1570;
	goto tr1513;
st514:
	if ( ++p == pe )
		goto _test_eof514;
case 514:
	switch( (*p) ) {
		case 78: goto st515;
		case 110: goto st515;
	}
	goto tr1513;
st515:
	if ( ++p == pe )
		goto _test_eof515;
case 515:
	switch( (*p) ) {
		case 68: goto st516;
		case 100: goto st516;
	}
	goto tr1513;
st516:
	if ( ++p == pe )
		goto _test_eof516;
case 516:
	switch( (*p) ) {
		case 73: goto st517;
		case 105: goto st517;
	}
	goto tr1513;
st517:
	if ( ++p == pe )
		goto _test_eof517;
case 517:
	switch( (*p) ) {
		case 82: goto st518;
		case 114: goto st518;
	}
	goto tr1513;
st518:
	if ( ++p == pe )
		goto _test_eof518;
case 518:
	switch( (*p) ) {
		case 69: goto st519;
		case 101: goto st519;
	}
	goto tr1513;
st519:
	if ( ++p == pe )
		goto _test_eof519;
case 519:
	switch( (*p) ) {
		case 67: goto st520;
		case 99: goto st520;
	}
	goto tr1513;
st520:
	if ( ++p == pe )
		goto _test_eof520;
case 520:
	switch( (*p) ) {
		case 84: goto st521;
		case 116: goto st521;
	}
	goto tr1513;
st521:
	if ( ++p == pe )
		goto _test_eof521;
case 521:
	switch( (*p) ) {
		case 32: goto tr1578;
		case 59: goto tr1578;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1578;
	} else if ( (*p) >= 9 )
		goto tr1578;
	goto tr1513;
st522:
	if ( ++p == pe )
		goto _test_eof522;
case 522:
	switch( (*p) ) {
		case 82: goto st523;
		case 114: goto st523;
	}
	goto tr1513;
st523:
	if ( ++p == pe )
		goto _test_eof523;
case 523:
	switch( (*p) ) {
		case 73: goto st524;
		case 105: goto st524;
	}
	goto tr1513;
st524:
	if ( ++p == pe )
		goto _test_eof524;
case 524:
	switch( (*p) ) {
		case 86: goto st525;
		case 118: goto st525;
	}
	goto tr1513;
st525:
	if ( ++p == pe )
		goto _test_eof525;
case 525:
	switch( (*p) ) {
		case 65: goto st526;
		case 97: goto st526;
	}
	goto tr1513;
st526:
	if ( ++p == pe )
		goto _test_eof526;
case 526:
	switch( (*p) ) {
		case 84: goto st527;
		case 116: goto st527;
	}
	goto tr1513;
st527:
	if ( ++p == pe )
		goto _test_eof527;
case 527:
	switch( (*p) ) {
		case 69: goto st528;
		case 101: goto st528;
	}
	goto tr1513;
st528:
	if ( ++p == pe )
		goto _test_eof528;
case 528:
	switch( (*p) ) {
		case 68: goto st529;
		case 79: goto st532;
		case 100: goto st529;
		case 111: goto st532;
	}
	goto tr1513;
st529:
	if ( ++p == pe )
		goto _test_eof529;
case 529:
	switch( (*p) ) {
		case 78: goto st530;
		case 110: goto st530;
	}
	goto tr1513;
st530:
	if ( ++p == pe )
		goto _test_eof530;
case 530:
	switch( (*p) ) {
		case 83: goto st531;
		case 115: goto st531;
	}
	goto tr1513;
st531:
	if ( ++p == pe )
		goto _test_eof531;
case 531:
	switch( (*p) ) {
		case 32: goto tr1589;
		case 59: goto tr1589;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1589;
	} else if ( (*p) >= 9 )
		goto tr1589;
	goto tr1513;
st532:
	if ( ++p == pe )
		goto _test_eof532;
case 532:
	switch( (*p) ) {
		case 73: goto st533;
		case 105: goto st533;
	}
	goto tr1513;
st533:
	if ( ++p == pe )
		goto _test_eof533;
case 533:
	switch( (*p) ) {
		case 68: goto st534;
		case 100: goto st534;
	}
	goto tr1513;
st534:
	if ( ++p == pe )
		goto _test_eof534;
case 534:
	switch( (*p) ) {
		case 32: goto tr1592;
		case 59: goto tr1592;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1592;
	} else if ( (*p) >= 9 )
		goto tr1592;
	goto tr1513;
st535:
	if ( ++p == pe )
		goto _test_eof535;
case 535:
	switch( (*p) ) {
		case 83: goto st536;
		case 115: goto st536;
	}
	goto tr1513;
st536:
	if ( ++p == pe )
		goto _test_eof536;
case 536:
	switch( (*p) ) {
		case 65: goto st537;
		case 97: goto st537;
	}
	goto tr1513;
st537:
	if ( ++p == pe )
		goto _test_eof537;
case 537:
	switch( (*p) ) {
		case 77: goto st538;
		case 83: goto st541;
		case 109: goto st538;
		case 115: goto st541;
	}
	goto tr1513;
st538:
	if ( ++p == pe )
		goto _test_eof538;
case 538:
	switch( (*p) ) {
		case 68: goto st539;
		case 100: goto st539;
	}
	goto tr1513;
st539:
	if ( ++p == pe )
		goto _test_eof539;
case 539:
	if ( (*p) == 53 )
		goto st540;
	goto tr1513;
st540:
	if ( ++p == pe )
		goto _test_eof540;
case 540:
	switch( (*p) ) {
		case 32: goto tr1599;
		case 59: goto tr1599;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1599;
	} else if ( (*p) >= 9 )
		goto tr1599;
	goto tr1513;
st541:
	if ( ++p == pe )
		goto _test_eof541;
case 541:
	switch( (*p) ) {
		case 72: goto st542;
		case 104: goto st542;
	}
	goto tr1513;
st542:
	if ( ++p == pe )
		goto _test_eof542;
case 542:
	switch( (*p) ) {
		case 65: goto st543;
		case 97: goto st543;
	}
	goto tr1513;
st543:
	if ( ++p == pe )
		goto _test_eof543;
case 543:
	switch( (*p) ) {
		case 49: goto st544;
		case 50: goto st556;
		case 53: goto st559;
	}
	goto tr1513;
st544:
	if ( ++p == pe )
		goto _test_eof544;
case 544:
	switch( (*p) ) {
		case 32: goto tr1605;
		case 45: goto st545;
		case 59: goto tr1605;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1605;
	} else if ( (*p) >= 9 )
		goto tr1605;
	goto tr1513;
st545:
	if ( ++p == pe )
		goto _test_eof545;
case 545:
	switch( (*p) ) {
		case 78: goto st546;
		case 110: goto st546;
	}
	goto tr1513;
st546:
	if ( ++p == pe )
		goto _test_eof546;
case 546:
	switch( (*p) ) {
		case 83: goto st547;
		case 115: goto st547;
	}
	goto tr1513;
st547:
	if ( ++p == pe )
		goto _test_eof547;
case 547:
	switch( (*p) ) {
		case 69: goto st548;
		case 101: goto st548;
	}
	goto tr1513;
st548:
	if ( ++p == pe )
		goto _test_eof548;
case 548:
	switch( (*p) ) {
		case 67: goto st549;
		case 99: goto st549;
	}
	goto tr1513;
st549:
	if ( ++p == pe )
		goto _test_eof549;
case 549:
	if ( (*p) == 51 )
		goto st550;
	goto tr1513;
st550:
	if ( ++p == pe )
		goto _test_eof550;
case 550:
	if ( (*p) == 45 )
		goto st551;
	goto tr1513;
st551:
	if ( ++p == pe )
		goto _test_eof551;
case 551:
	switch( (*p) ) {
		case 83: goto st552;
		case 115: goto st552;
	}
	goto tr1513;
st552:
	if ( ++p == pe )
		goto _test_eof552;
case 552:
	switch( (*p) ) {
		case 72: goto st553;
		case 104: goto st553;
	}
	goto tr1513;
st553:
	if ( ++p == pe )
		goto _test_eof553;
case 553:
	switch( (*p) ) {
		case 65: goto st554;
		case 97: goto st554;
	}
	goto tr1513;
st554:
	if ( ++p == pe )
		goto _test_eof554;
case 554:
	if ( (*p) == 49 )
		goto st555;
	goto tr1513;
st555:
	if ( ++p == pe )
		goto _test_eof555;
case 555:
	switch( (*p) ) {
		case 32: goto tr1617;
		case 59: goto tr1617;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1617;
	} else if ( (*p) >= 9 )
		goto tr1617;
	goto tr1513;
st556:
	if ( ++p == pe )
		goto _test_eof556;
case 556:
	if ( (*p) == 53 )
		goto st557;
	goto tr1513;
st557:
	if ( ++p == pe )
		goto _test_eof557;
case 557:
	if ( (*p) == 54 )
		goto st558;
	goto tr1513;
st558:
	if ( ++p == pe )
		goto _test_eof558;
case 558:
	switch( (*p) ) {
		case 32: goto tr1620;
		case 59: goto tr1620;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1620;
	} else if ( (*p) >= 9 )
		goto tr1620;
	goto tr1513;
st559:
	if ( ++p == pe )
		goto _test_eof559;
case 559:
	if ( (*p) == 49 )
		goto st560;
	goto tr1513;
st560:
	if ( ++p == pe )
		goto _test_eof560;
case 560:
	if ( (*p) == 50 )
		goto st561;
	goto tr1513;
st561:
	if ( ++p == pe )
		goto _test_eof561;
case 561:
	switch( (*p) ) {
		case 32: goto tr1623;
		case 59: goto tr1623;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1623;
	} else if ( (*p) >= 9 )
		goto tr1623;
	goto tr1513;
st562:
	if ( ++p == pe )
		goto _test_eof562;
case 562:
	switch( (*p) ) {
		case 65: goto st564;
		case 73: goto st570;
		case 79: goto st587;
		case 80: goto st590;
		case 83: goto st596;
		case 85: goto st600;
		case 97: goto st564;
		case 105: goto st570;
		case 111: goto st587;
		case 112: goto st590;
		case 115: goto st596;
		case 117: goto st600;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1625;
	goto tr1624;
tr1625:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st563;
tr1633:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st563;
st563:
	if ( ++p == pe )
		goto _test_eof563;
case 563:
	switch( (*p) ) {
		case 32: goto tr1632;
		case 59: goto tr1632;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1632;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1633;
	} else
		goto tr1632;
	goto tr1624;
tr1632:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1110;
tr1639:
	{
		*((uint16_t *)rdata_tail) = htons(7);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1110;
tr1648:
	{
		*((uint16_t *)rdata_tail) = htons(8);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1110;
tr1652:
	{
		*((uint16_t *)rdata_tail) = htons(6);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1110;
tr1655:
	{
		*((uint16_t *)rdata_tail) = htons(4);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1110;
tr1659:
	{
		*((uint16_t *)rdata_tail) = htons(5);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1110;
tr1662:
	{
		*((uint16_t *)rdata_tail) = htons(254);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1110;
tr1666:
	{
		*((uint16_t *)rdata_tail) = htons(3);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1110;
tr1669:
	{
		*((uint16_t *)rdata_tail) = htons(1);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1110;
tr1673:
	{
		*((uint16_t *)rdata_tail) = htons(2);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1110;
tr1676:
	{
		*((uint16_t *)rdata_tail) = htons(253);
		rdata_tail += 2;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1110;
st1110:
	if ( ++p == pe )
		goto _test_eof1110;
case 1110:
	goto st0;
st564:
	if ( ++p == pe )
		goto _test_eof564;
case 564:
	switch( (*p) ) {
		case 67: goto st565;
		case 99: goto st565;
	}
	goto tr1624;
st565:
	if ( ++p == pe )
		goto _test_eof565;
case 565:
	switch( (*p) ) {
		case 80: goto st566;
		case 112: goto st566;
	}
	goto tr1624;
st566:
	if ( ++p == pe )
		goto _test_eof566;
case 566:
	switch( (*p) ) {
		case 75: goto st567;
		case 107: goto st567;
	}
	goto tr1624;
st567:
	if ( ++p == pe )
		goto _test_eof567;
case 567:
	switch( (*p) ) {
		case 73: goto st568;
		case 105: goto st568;
	}
	goto tr1624;
st568:
	if ( ++p == pe )
		goto _test_eof568;
case 568:
	switch( (*p) ) {
		case 88: goto st569;
		case 120: goto st569;
	}
	goto tr1624;
st569:
	if ( ++p == pe )
		goto _test_eof569;
case 569:
	switch( (*p) ) {
		case 32: goto tr1639;
		case 59: goto tr1639;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1639;
	} else if ( (*p) >= 9 )
		goto tr1639;
	goto tr1624;
st570:
	if ( ++p == pe )
		goto _test_eof570;
case 570:
	switch( (*p) ) {
		case 65: goto st571;
		case 80: goto st577;
		case 83: goto st583;
		case 97: goto st571;
		case 112: goto st577;
		case 115: goto st583;
	}
	goto tr1624;
st571:
	if ( ++p == pe )
		goto _test_eof571;
case 571:
	switch( (*p) ) {
		case 67: goto st572;
		case 99: goto st572;
	}
	goto tr1624;
st572:
	if ( ++p == pe )
		goto _test_eof572;
case 572:
	switch( (*p) ) {
		case 80: goto st573;
		case 112: goto st573;
	}
	goto tr1624;
st573:
	if ( ++p == pe )
		goto _test_eof573;
case 573:
	switch( (*p) ) {
		case 75: goto st574;
		case 107: goto st574;
	}
	goto tr1624;
st574:
	if ( ++p == pe )
		goto _test_eof574;
case 574:
	switch( (*p) ) {
		case 73: goto st575;
		case 105: goto st575;
	}
	goto tr1624;
st575:
	if ( ++p == pe )
		goto _test_eof575;
case 575:
	switch( (*p) ) {
		case 88: goto st576;
		case 120: goto st576;
	}
	goto tr1624;
st576:
	if ( ++p == pe )
		goto _test_eof576;
case 576:
	switch( (*p) ) {
		case 32: goto tr1648;
		case 59: goto tr1648;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1648;
	} else if ( (*p) >= 9 )
		goto tr1648;
	goto tr1624;
st577:
	if ( ++p == pe )
		goto _test_eof577;
case 577:
	switch( (*p) ) {
		case 71: goto st578;
		case 75: goto st580;
		case 103: goto st578;
		case 107: goto st580;
	}
	goto tr1624;
st578:
	if ( ++p == pe )
		goto _test_eof578;
case 578:
	switch( (*p) ) {
		case 80: goto st579;
		case 112: goto st579;
	}
	goto tr1624;
st579:
	if ( ++p == pe )
		goto _test_eof579;
case 579:
	switch( (*p) ) {
		case 32: goto tr1652;
		case 59: goto tr1652;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1652;
	} else if ( (*p) >= 9 )
		goto tr1652;
	goto tr1624;
st580:
	if ( ++p == pe )
		goto _test_eof580;
case 580:
	switch( (*p) ) {
		case 73: goto st581;
		case 105: goto st581;
	}
	goto tr1624;
st581:
	if ( ++p == pe )
		goto _test_eof581;
case 581:
	switch( (*p) ) {
		case 88: goto st582;
		case 120: goto st582;
	}
	goto tr1624;
st582:
	if ( ++p == pe )
		goto _test_eof582;
case 582:
	switch( (*p) ) {
		case 32: goto tr1655;
		case 59: goto tr1655;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1655;
	} else if ( (*p) >= 9 )
		goto tr1655;
	goto tr1624;
st583:
	if ( ++p == pe )
		goto _test_eof583;
case 583:
	switch( (*p) ) {
		case 80: goto st584;
		case 112: goto st584;
	}
	goto tr1624;
st584:
	if ( ++p == pe )
		goto _test_eof584;
case 584:
	switch( (*p) ) {
		case 75: goto st585;
		case 107: goto st585;
	}
	goto tr1624;
st585:
	if ( ++p == pe )
		goto _test_eof585;
case 585:
	switch( (*p) ) {
		case 73: goto st586;
		case 105: goto st586;
	}
	goto tr1624;
st586:
	if ( ++p == pe )
		goto _test_eof586;
case 586:
	switch( (*p) ) {
		case 32: goto tr1659;
		case 59: goto tr1659;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1659;
	} else if ( (*p) >= 9 )
		goto tr1659;
	goto tr1624;
st587:
	if ( ++p == pe )
		goto _test_eof587;
case 587:
	switch( (*p) ) {
		case 73: goto st588;
		case 105: goto st588;
	}
	goto tr1624;
st588:
	if ( ++p == pe )
		goto _test_eof588;
case 588:
	switch( (*p) ) {
		case 68: goto st589;
		case 100: goto st589;
	}
	goto tr1624;
st589:
	if ( ++p == pe )
		goto _test_eof589;
case 589:
	switch( (*p) ) {
		case 32: goto tr1662;
		case 59: goto tr1662;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1662;
	} else if ( (*p) >= 9 )
		goto tr1662;
	goto tr1624;
st590:
	if ( ++p == pe )
		goto _test_eof590;
case 590:
	switch( (*p) ) {
		case 71: goto st591;
		case 75: goto st593;
		case 103: goto st591;
		case 107: goto st593;
	}
	goto tr1624;
st591:
	if ( ++p == pe )
		goto _test_eof591;
case 591:
	switch( (*p) ) {
		case 80: goto st592;
		case 112: goto st592;
	}
	goto tr1624;
st592:
	if ( ++p == pe )
		goto _test_eof592;
case 592:
	switch( (*p) ) {
		case 32: goto tr1666;
		case 59: goto tr1666;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1666;
	} else if ( (*p) >= 9 )
		goto tr1666;
	goto tr1624;
st593:
	if ( ++p == pe )
		goto _test_eof593;
case 593:
	switch( (*p) ) {
		case 73: goto st594;
		case 105: goto st594;
	}
	goto tr1624;
st594:
	if ( ++p == pe )
		goto _test_eof594;
case 594:
	switch( (*p) ) {
		case 88: goto st595;
		case 120: goto st595;
	}
	goto tr1624;
st595:
	if ( ++p == pe )
		goto _test_eof595;
case 595:
	switch( (*p) ) {
		case 32: goto tr1669;
		case 59: goto tr1669;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1669;
	} else if ( (*p) >= 9 )
		goto tr1669;
	goto tr1624;
st596:
	if ( ++p == pe )
		goto _test_eof596;
case 596:
	switch( (*p) ) {
		case 80: goto st597;
		case 112: goto st597;
	}
	goto tr1624;
st597:
	if ( ++p == pe )
		goto _test_eof597;
case 597:
	switch( (*p) ) {
		case 75: goto st598;
		case 107: goto st598;
	}
	goto tr1624;
st598:
	if ( ++p == pe )
		goto _test_eof598;
case 598:
	switch( (*p) ) {
		case 73: goto st599;
		case 105: goto st599;
	}
	goto tr1624;
st599:
	if ( ++p == pe )
		goto _test_eof599;
case 599:
	switch( (*p) ) {
		case 32: goto tr1673;
		case 59: goto tr1673;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1673;
	} else if ( (*p) >= 9 )
		goto tr1673;
	goto tr1624;
st600:
	if ( ++p == pe )
		goto _test_eof600;
case 600:
	switch( (*p) ) {
		case 82: goto st601;
		case 114: goto st601;
	}
	goto tr1624;
st601:
	if ( ++p == pe )
		goto _test_eof601;
case 601:
	switch( (*p) ) {
		case 73: goto st602;
		case 105: goto st602;
	}
	goto tr1624;
st602:
	if ( ++p == pe )
		goto _test_eof602;
case 602:
	switch( (*p) ) {
		case 32: goto tr1676;
		case 59: goto tr1676;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1676;
	} else if ( (*p) >= 9 )
		goto tr1676;
	goto tr1624;
st603:
	if ( ++p == pe )
		goto _test_eof603;
case 603:
	if ( (*p) == 46 )
		goto tr1678;
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1678;
	goto tr1677;
tr1678:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st604;
tr1680:
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st604;
st604:
	if ( ++p == pe )
		goto _test_eof604;
case 604:
	switch( (*p) ) {
		case 32: goto tr1679;
		case 46: goto tr1680;
		case 59: goto tr1679;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1679;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1680;
	} else
		goto tr1679;
	goto tr1677;
tr1679:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1111;
st1111:
	if ( ++p == pe )
		goto _test_eof1111;
case 1111:
	goto st0;
st605:
	if ( ++p == pe )
		goto _test_eof605;
case 605:
	switch( (*p) ) {
		case 42: goto tr1681;
		case 92: goto tr1681;
		case 95: goto tr1681;
	}
	if ( (*p) < 64 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr1681;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1681;
	} else
		goto tr1681;
	goto tr69;
tr1681:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 606; goto st255;} }
	goto st606;
st606:
	if ( ++p == pe )
		goto _test_eof606;
case 606:
	switch( (*p) ) {
		case 32: goto tr1682;
		case 59: goto tr1682;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1682;
	} else if ( (*p) >= 9 )
		goto tr1682;
	goto tr69;
tr1682:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1112;
st1112:
	if ( ++p == pe )
		goto _test_eof1112;
case 1112:
	goto st0;
st607:
	if ( ++p == pe )
		goto _test_eof607;
case 607:
	switch( (*p) ) {
		case 42: goto tr1683;
		case 92: goto tr1683;
		case 95: goto tr1683;
	}
	if ( (*p) < 64 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr1683;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1683;
	} else
		goto tr1683;
	goto tr69;
tr1683:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 608; goto st255;} }
	goto st608;
st608:
	if ( ++p == pe )
		goto _test_eof608;
case 608:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1684;
		case 32: goto tr1684;
		case 40: goto tr1685;
		case 41: goto tr1686;
		case 1034: goto tr1687;
		case 1083: goto tr1688;
	}
	goto tr69;
tr1690:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st609;
tr1691:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st609;
tr1693:
	{
		s->line_counter++;
	}
	goto st609;
tr1684:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st609;
tr1685:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st609;
tr1686:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st609;
tr1687:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st609;
st609:
	if ( ++p == pe )
		goto _test_eof609;
case 609:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st609;
		case 32: goto st609;
		case 40: goto tr1690;
		case 41: goto tr1691;
		case 42: goto tr1692;
		case 92: goto tr1692;
		case 95: goto tr1692;
		case 1034: goto tr1693;
		case 1083: goto st638;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr1692;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1692;
	} else
		goto tr1692;
	goto tr69;
tr1692:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 610; goto st255;} }
	goto st610;
st610:
	if ( ++p == pe )
		goto _test_eof610;
case 610:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1695;
		case 32: goto tr1695;
		case 40: goto tr1696;
		case 41: goto tr1697;
		case 1034: goto tr1698;
		case 1083: goto tr1699;
	}
	goto tr69;
tr1702:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st611;
tr1703:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st611;
tr1705:
	{
		s->line_counter++;
	}
	goto st611;
tr1695:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st611;
tr1696:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st611;
tr1697:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st611;
tr1698:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st611;
st611:
	if ( ++p == pe )
		goto _test_eof611;
case 611:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st611;
		case 32: goto st611;
		case 40: goto tr1702;
		case 41: goto tr1703;
		case 1034: goto tr1705;
		case 1083: goto st637;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1704;
	goto tr1700;
tr1704:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st612;
tr1710:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st612;
st612:
	if ( ++p == pe )
		goto _test_eof612;
case 612:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1707;
		case 32: goto tr1707;
		case 40: goto tr1708;
		case 41: goto tr1709;
		case 1034: goto tr1711;
		case 1083: goto tr1712;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1710;
	goto tr1700;
tr1714:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st613;
tr1715:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st613;
tr1717:
	{
		s->line_counter++;
	}
	goto st613;
tr1707:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st613;
tr1708:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st613;
tr1709:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st613;
tr1711:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st613;
st613:
	if ( ++p == pe )
		goto _test_eof613;
case 613:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st613;
		case 32: goto st613;
		case 40: goto tr1714;
		case 41: goto tr1715;
		case 1034: goto tr1717;
		case 1083: goto st636;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1716;
	goto tr1700;
tr1716:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st614;
tr1723:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st614;
st614:
	if ( ++p == pe )
		goto _test_eof614;
case 614:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1720;
		case 32: goto tr1720;
		case 40: goto tr1721;
		case 41: goto tr1722;
		case 68: goto tr1724;
		case 72: goto tr1725;
		case 77: goto tr1726;
		case 83: goto st633;
		case 87: goto tr1728;
		case 100: goto tr1724;
		case 104: goto tr1725;
		case 109: goto tr1726;
		case 115: goto st633;
		case 119: goto tr1728;
		case 1034: goto tr1729;
		case 1083: goto tr1730;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1723;
	goto tr1719;
tr1732:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st615;
tr1733:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st615;
tr1735:
	{
		s->line_counter++;
	}
	goto st615;
tr1720:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st615;
tr1721:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st615;
tr1722:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st615;
tr1729:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st615;
tr1820:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st615;
tr1821:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st615;
tr1822:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st615;
tr1824:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st615;
st615:
	if ( ++p == pe )
		goto _test_eof615;
case 615:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st615;
		case 32: goto st615;
		case 40: goto tr1732;
		case 41: goto tr1733;
		case 1034: goto tr1735;
		case 1083: goto st632;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1734;
	goto tr1700;
tr1734:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st616;
tr1740:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st616;
st616:
	if ( ++p == pe )
		goto _test_eof616;
case 616:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1737;
		case 32: goto tr1737;
		case 40: goto tr1738;
		case 41: goto tr1739;
		case 68: goto tr1741;
		case 72: goto tr1742;
		case 77: goto tr1743;
		case 83: goto st629;
		case 87: goto tr1745;
		case 100: goto tr1741;
		case 104: goto tr1742;
		case 109: goto tr1743;
		case 115: goto st629;
		case 119: goto tr1745;
		case 1034: goto tr1746;
		case 1083: goto tr1747;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1740;
	goto tr1719;
tr1749:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st617;
tr1750:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st617;
tr1752:
	{
		s->line_counter++;
	}
	goto st617;
tr1737:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st617;
tr1738:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st617;
tr1739:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st617;
tr1746:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st617;
tr1807:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st617;
tr1808:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st617;
tr1809:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st617;
tr1811:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st617;
st617:
	if ( ++p == pe )
		goto _test_eof617;
case 617:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st617;
		case 32: goto st617;
		case 40: goto tr1749;
		case 41: goto tr1750;
		case 1034: goto tr1752;
		case 1083: goto st628;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1751;
	goto tr1700;
tr1751:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st618;
tr1757:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st618;
st618:
	if ( ++p == pe )
		goto _test_eof618;
case 618:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1754;
		case 32: goto tr1754;
		case 40: goto tr1755;
		case 41: goto tr1756;
		case 68: goto tr1758;
		case 72: goto tr1759;
		case 77: goto tr1760;
		case 83: goto st625;
		case 87: goto tr1762;
		case 100: goto tr1758;
		case 104: goto tr1759;
		case 109: goto tr1760;
		case 115: goto st625;
		case 119: goto tr1762;
		case 1034: goto tr1763;
		case 1083: goto tr1764;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1757;
	goto tr1719;
tr1766:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st619;
tr1767:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st619;
tr1769:
	{
		s->line_counter++;
	}
	goto st619;
tr1754:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st619;
tr1755:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st619;
tr1756:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st619;
tr1763:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st619;
tr1794:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st619;
tr1795:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st619;
tr1796:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st619;
tr1798:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st619;
st619:
	if ( ++p == pe )
		goto _test_eof619;
case 619:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st619;
		case 32: goto st619;
		case 40: goto tr1766;
		case 41: goto tr1767;
		case 1034: goto tr1769;
		case 1083: goto st624;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1768;
	goto tr1700;
tr1768:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st620;
tr1772:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st620;
st620:
	if ( ++p == pe )
		goto _test_eof620;
case 620:
	switch( (*p) ) {
		case 32: goto tr1771;
		case 59: goto tr1771;
		case 68: goto tr1773;
		case 72: goto tr1774;
		case 77: goto tr1775;
		case 83: goto st621;
		case 87: goto tr1777;
		case 100: goto tr1773;
		case 104: goto tr1774;
		case 109: goto tr1775;
		case 115: goto st621;
		case 119: goto tr1777;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1771;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1772;
	} else
		goto tr1771;
	goto tr1719;
tr1771:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1113;
tr1785:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1113;
st1113:
	if ( ++p == pe )
		goto _test_eof1113;
case 1113:
	goto st0;
tr1773:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st621;
tr1774:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st621;
tr1775:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st621;
tr1777:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st621;
st621:
	if ( ++p == pe )
		goto _test_eof621;
case 621:
	switch( (*p) ) {
		case 32: goto tr1771;
		case 59: goto tr1771;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1771;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1778;
	} else
		goto tr1771;
	goto tr1719;
tr1779:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st622;
tr1778:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st622;
tr1786:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st622;
st622:
	if ( ++p == pe )
		goto _test_eof622;
case 622:
	switch( (*p) ) {
		case 68: goto tr1780;
		case 72: goto tr1781;
		case 77: goto tr1782;
		case 83: goto st623;
		case 87: goto tr1784;
		case 100: goto tr1780;
		case 104: goto tr1781;
		case 109: goto tr1782;
		case 115: goto st623;
		case 119: goto tr1784;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1779;
	goto tr1719;
tr1780:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st623;
tr1781:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st623;
tr1782:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st623;
tr1784:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st623;
st623:
	if ( ++p == pe )
		goto _test_eof623;
case 623:
	switch( (*p) ) {
		case 32: goto tr1785;
		case 59: goto tr1785;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr1785;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr1786;
	} else
		goto tr1785;
	goto tr1719;
tr1764:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st624;
tr1799:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st624;
st624:
	if ( ++p == pe )
		goto _test_eof624;
case 624:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1769;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st624;
	goto tr69;
tr1758:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st625;
tr1759:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st625;
tr1760:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st625;
tr1762:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st625;
st625:
	if ( ++p == pe )
		goto _test_eof625;
case 625:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1754;
		case 32: goto tr1754;
		case 40: goto tr1755;
		case 41: goto tr1756;
		case 1034: goto tr1763;
		case 1083: goto tr1764;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1787;
	goto tr1719;
tr1788:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st626;
tr1787:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st626;
tr1797:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st626;
st626:
	if ( ++p == pe )
		goto _test_eof626;
case 626:
	switch( (*p) ) {
		case 68: goto tr1789;
		case 72: goto tr1790;
		case 77: goto tr1791;
		case 83: goto st627;
		case 87: goto tr1793;
		case 100: goto tr1789;
		case 104: goto tr1790;
		case 109: goto tr1791;
		case 115: goto st627;
		case 119: goto tr1793;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1788;
	goto tr1719;
tr1789:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st627;
tr1790:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st627;
tr1791:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st627;
tr1793:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st627;
st627:
	if ( ++p == pe )
		goto _test_eof627;
case 627:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1794;
		case 32: goto tr1794;
		case 40: goto tr1795;
		case 41: goto tr1796;
		case 1034: goto tr1798;
		case 1083: goto tr1799;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1797;
	goto tr1719;
tr1747:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st628;
tr1812:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st628;
st628:
	if ( ++p == pe )
		goto _test_eof628;
case 628:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1752;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st628;
	goto tr69;
tr1741:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st629;
tr1742:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st629;
tr1743:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st629;
tr1745:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st629;
st629:
	if ( ++p == pe )
		goto _test_eof629;
case 629:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1737;
		case 32: goto tr1737;
		case 40: goto tr1738;
		case 41: goto tr1739;
		case 1034: goto tr1746;
		case 1083: goto tr1747;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1800;
	goto tr1719;
tr1801:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st630;
tr1800:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st630;
tr1810:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st630;
st630:
	if ( ++p == pe )
		goto _test_eof630;
case 630:
	switch( (*p) ) {
		case 68: goto tr1802;
		case 72: goto tr1803;
		case 77: goto tr1804;
		case 83: goto st631;
		case 87: goto tr1806;
		case 100: goto tr1802;
		case 104: goto tr1803;
		case 109: goto tr1804;
		case 115: goto st631;
		case 119: goto tr1806;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1801;
	goto tr1719;
tr1802:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st631;
tr1803:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st631;
tr1804:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st631;
tr1806:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st631;
st631:
	if ( ++p == pe )
		goto _test_eof631;
case 631:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1807;
		case 32: goto tr1807;
		case 40: goto tr1808;
		case 41: goto tr1809;
		case 1034: goto tr1811;
		case 1083: goto tr1812;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1810;
	goto tr1719;
tr1730:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st632;
tr1825:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st632;
st632:
	if ( ++p == pe )
		goto _test_eof632;
case 632:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1735;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st632;
	goto tr69;
tr1724:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st633;
tr1725:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st633;
tr1726:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st633;
tr1728:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st633;
st633:
	if ( ++p == pe )
		goto _test_eof633;
case 633:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1720;
		case 32: goto tr1720;
		case 40: goto tr1721;
		case 41: goto tr1722;
		case 1034: goto tr1729;
		case 1083: goto tr1730;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1813;
	goto tr1719;
tr1814:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st634;
tr1813:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st634;
tr1823:
	{
		if (s->number64 + s->number64_tmp < UINT32_MAX) {
			s->number64 += s->number64_tmp;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st634;
st634:
	if ( ++p == pe )
		goto _test_eof634;
case 634:
	switch( (*p) ) {
		case 68: goto tr1815;
		case 72: goto tr1816;
		case 77: goto tr1817;
		case 83: goto st635;
		case 87: goto tr1819;
		case 100: goto tr1815;
		case 104: goto tr1816;
		case 109: goto tr1817;
		case 115: goto st635;
		case 119: goto tr1819;
	}
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1814;
	goto tr1719;
tr1815:
	{ if (s->number64 <= (UINT32_MAX / 86400)) {
	                  s->number64 *= 86400;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st635;
tr1816:
	{ if (s->number64 <= (UINT32_MAX / 3600)) {
	                  s->number64 *= 3600;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st635;
tr1817:
	{ if (s->number64 <= (UINT32_MAX / 60)) {
	                  s->number64 *= 60;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st635;
tr1819:
	{ if (s->number64 <= (UINT32_MAX / 604800)) {
	                  s->number64 *= 604800;
	              } else {
	                  WARN(ZS_NUMBER32_OVERFLOW);
	                  p--; {goto st253;}
	              }
	            }
	goto st635;
st635:
	if ( ++p == pe )
		goto _test_eof635;
case 635:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1820;
		case 32: goto tr1820;
		case 40: goto tr1821;
		case 41: goto tr1822;
		case 1034: goto tr1824;
		case 1083: goto tr1825;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1823;
	goto tr1719;
tr1712:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st636;
st636:
	if ( ++p == pe )
		goto _test_eof636;
case 636:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1717;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st636;
	goto tr69;
tr1699:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st637;
st637:
	if ( ++p == pe )
		goto _test_eof637;
case 637:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1705;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st637;
	goto tr69;
tr1688:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st638;
st638:
	if ( ++p == pe )
		goto _test_eof638;
case 638:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1693;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st638;
	goto tr69;
st639:
	if ( ++p == pe )
		goto _test_eof639;
case 639:
	switch( (*p) ) {
		case 32: goto tr69;
		case 59: goto tr69;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr69;
	} else if ( (*p) >= 9 )
		goto tr69;
	goto tr1826;
tr1826:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{ p--; {stack[top++] = 640; goto st264;} }
	goto st640;
st640:
	if ( ++p == pe )
		goto _test_eof640;
case 640:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1827;
		case 32: goto tr1827;
		case 40: goto tr1828;
		case 41: goto tr1829;
		case 1034: goto tr1830;
		case 1083: goto tr1831;
	}
	goto tr69;
tr1834:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st641;
tr1835:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st641;
tr1836:
	{
		s->line_counter++;
	}
	goto st641;
tr1827:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st641;
tr1828:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st641;
tr1829:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st641;
tr1830:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st641;
st641:
	if ( ++p == pe )
		goto _test_eof641;
case 641:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st641;
		case 32: goto st641;
		case 40: goto tr1834;
		case 41: goto tr1835;
		case 1034: goto tr1836;
		case 1083: goto st643;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1832;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr1832;
	} else
		goto tr1832;
	goto tr69;
tr1832:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{ p--; {stack[top++] = 642; goto st264;} }
	goto st642;
st642:
	if ( ++p == pe )
		goto _test_eof642;
case 642:
	switch( (*p) ) {
		case 32: goto tr1838;
		case 59: goto tr1838;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1838;
	} else if ( (*p) >= 9 )
		goto tr1838;
	goto tr69;
tr1838:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1114;
st1114:
	if ( ++p == pe )
		goto _test_eof1114;
case 1114:
	goto st0;
tr1831:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st643;
st643:
	if ( ++p == pe )
		goto _test_eof643;
case 643:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1836;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st643;
	goto tr69;
st644:
	if ( ++p == pe )
		goto _test_eof644;
case 644:
	switch( (*p) ) {
		case 42: goto tr1839;
		case 92: goto tr1839;
		case 95: goto tr1839;
	}
	if ( (*p) < 64 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr1839;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr1839;
	} else
		goto tr1839;
	goto tr69;
tr1839:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 645; goto st255;} }
	goto st645;
st645:
	if ( ++p == pe )
		goto _test_eof645;
case 645:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1840;
		case 32: goto tr1840;
		case 40: goto tr1841;
		case 41: goto tr1842;
		case 1034: goto tr1843;
		case 1083: goto tr1844;
	}
	goto tr69;
tr1846:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st646;
tr1847:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st646;
tr1849:
	{
		s->line_counter++;
	}
	goto st646;
tr1840:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st646;
tr1841:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st646;
tr1842:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st646;
tr1843:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st646;
st646:
	if ( ++p == pe )
		goto _test_eof646;
case 646:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st646;
		case 32: goto st646;
		case 40: goto tr1846;
		case 41: goto tr1847;
		case 42: goto tr1848;
		case 92: goto tr1848;
		case 95: goto tr1848;
		case 1034: goto tr1849;
		case 1083: goto st648;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr1848;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1848;
	} else
		goto tr1848;
	goto tr69;
tr1848:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 647; goto st255;} }
	goto st647;
st647:
	if ( ++p == pe )
		goto _test_eof647;
case 647:
	switch( (*p) ) {
		case 32: goto tr1851;
		case 59: goto tr1851;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1851;
	} else if ( (*p) >= 9 )
		goto tr1851;
	goto tr69;
tr1851:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1115;
st1115:
	if ( ++p == pe )
		goto _test_eof1115;
case 1115:
	goto st0;
tr1844:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st648;
st648:
	if ( ++p == pe )
		goto _test_eof648;
case 648:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1849;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st648;
	goto tr69;
st649:
	if ( ++p == pe )
		goto _test_eof649;
case 649:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1852;
	goto tr1700;
tr1852:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st650;
tr1856:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st650;
st650:
	if ( ++p == pe )
		goto _test_eof650;
case 650:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1853;
		case 32: goto tr1853;
		case 40: goto tr1854;
		case 41: goto tr1855;
		case 1034: goto tr1857;
		case 1083: goto tr1858;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1856;
	goto tr1700;
tr1860:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st651;
tr1861:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st651;
tr1863:
	{
		s->line_counter++;
	}
	goto st651;
tr1853:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st651;
tr1854:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st651;
tr1855:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st651;
tr1857:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st651;
st651:
	if ( ++p == pe )
		goto _test_eof651;
case 651:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st651;
		case 32: goto st651;
		case 40: goto tr1860;
		case 41: goto tr1861;
		case 42: goto tr1862;
		case 92: goto tr1862;
		case 95: goto tr1862;
		case 1034: goto tr1863;
		case 1083: goto st653;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr1862;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr1862;
	} else
		goto tr1862;
	goto tr69;
tr1862:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 652; goto st255;} }
	goto st652;
st652:
	if ( ++p == pe )
		goto _test_eof652;
case 652:
	switch( (*p) ) {
		case 32: goto tr1865;
		case 59: goto tr1865;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1865;
	} else if ( (*p) >= 9 )
		goto tr1865;
	goto tr69;
tr1865:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1116;
st1116:
	if ( ++p == pe )
		goto _test_eof1116;
case 1116:
	goto st0;
tr1858:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st653;
st653:
	if ( ++p == pe )
		goto _test_eof653;
case 653:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1863;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st653;
	goto tr69;
st654:
	if ( ++p == pe )
		goto _test_eof654;
case 654:
	switch( (*p) ) {
		case 32: goto tr1867;
		case 59: goto tr1867;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr1867;
	} else if ( (*p) >= 9 )
		goto tr1867;
	goto tr1866;
tr1876:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{ p--; {stack[top++] = 655; goto st264;} }
	goto st655;
tr1866:
	{
		s->long_string = true;
	}
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{ p--; {stack[top++] = 655; goto st264;} }
	goto st655;
st655:
	if ( ++p == pe )
		goto _test_eof655;
case 655:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1868;
		case 32: goto tr1868;
		case 40: goto tr1869;
		case 41: goto tr1870;
		case 2058: goto tr1871;
		case 2107: goto tr1872;
		case 2314: goto tr1873;
		case 2363: goto tr1873;
		case 2570: goto tr1874;
		case 2619: goto tr1875;
	}
	goto tr1867;
tr1878:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st656;
tr1879:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st656;
tr1880:
	{
		s->line_counter++;
	}
	goto st656;
tr1868:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st656;
tr1869:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st656;
tr1870:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st656;
tr1871:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st656;
st656:
	if ( ++p == pe )
		goto _test_eof656;
case 656:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st656;
		case 32: goto st656;
		case 40: goto tr1878;
		case 41: goto tr1879;
		case 2058: goto tr1880;
		case 2107: goto st657;
		case 2314: goto tr1882;
		case 2363: goto tr1882;
		case 2570: goto tr1883;
		case 2619: goto tr1884;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1876;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr1876;
	} else
		goto tr1876;
	goto tr1867;
tr1872:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st657;
st657:
	if ( ++p == pe )
		goto _test_eof657;
case 657:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1880;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st657;
	goto tr1867;
tr1873:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->long_string = false;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1117;
tr1882:
	{
		s->long_string = false;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1117;
st1117:
	if ( ++p == pe )
		goto _test_eof1117;
case 1117:
	goto st0;
tr1874:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	{
		s->long_string = false;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1118;
tr1883:
	{
		s->line_counter++;
	}
	{
		s->long_string = false;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1118;
st1118:
	if ( ++p == pe )
		goto _test_eof1118;
case 1118:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st656;
		case 32: goto st656;
		case 40: goto tr1878;
		case 41: goto tr1879;
		case 2058: goto tr1880;
		case 2107: goto st657;
		case 2314: goto tr1882;
		case 2363: goto tr1882;
		case 2570: goto tr1883;
		case 2619: goto tr1884;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr1876;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr1876;
	} else
		goto tr1876;
	goto tr1867;
tr1875:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->long_string = false;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1119;
tr1884:
	{
		s->long_string = false;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1119;
st1119:
	if ( ++p == pe )
		goto _test_eof1119;
case 1119:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1880;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st657;
	goto tr1867;
st658:
	if ( ++p == pe )
		goto _test_eof658;
case 658:
	if ( (*p) == 46 )
		goto tr1885;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 58 )
			goto tr1885;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr1885;
	} else
		goto tr1885;
	goto tr1677;
tr1885:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st659;
tr1887:
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st659;
st659:
	if ( ++p == pe )
		goto _test_eof659;
case 659:
	switch( (*p) ) {
		case 32: goto tr1886;
		case 46: goto tr1887;
		case 59: goto tr1886;
	}
	if ( (*p) < 48 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr1886;
		} else if ( (*p) >= 9 )
			goto tr1886;
	} else if ( (*p) > 58 ) {
		if ( (*p) > 70 ) {
			if ( 97 <= (*p) && (*p) <= 102 )
				goto tr1887;
		} else if ( (*p) >= 65 )
			goto tr1887;
	} else
		goto tr1887;
	goto tr1677;
tr1886:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1120;
st1120:
	if ( ++p == pe )
		goto _test_eof1120;
case 1120:
	goto st0;
st660:
	if ( ++p == pe )
		goto _test_eof660;
case 660:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1889;
	goto tr1888;
tr1893:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st661;
tr1889:
	{
		memset(&(s->loc), 0, sizeof(s->loc));
		// Defaults.
		s->loc.siz = 100;
		s->loc.vp  = 1000;
		s->loc.hp  = 1000000;
		s->loc.lat_sign  = 1;
		s->loc.long_sign = 1;
		s->loc.alt_sign  = 1;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st661;
st661:
	if ( ++p == pe )
		goto _test_eof661;
case 661:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1890;
		case 32: goto tr1890;
		case 40: goto tr1891;
		case 41: goto tr1892;
		case 1034: goto tr1894;
		case 1083: goto tr1895;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1893;
	goto tr1888;
tr1897:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st662;
tr1898:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st662;
tr1902:
	{
		s->line_counter++;
	}
	goto st662;
tr1890:
	{
		if (s->number64 <= 90) {
			s->loc.d1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st662;
tr1891:
	{
		if (s->number64 <= 90) {
			s->loc.d1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st662;
tr1892:
	{
		if (s->number64 <= 90) {
			s->loc.d1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st662;
tr1894:
	{
		if (s->number64 <= 90) {
			s->loc.d1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st662;
st662:
	if ( ++p == pe )
		goto _test_eof662;
case 662:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st662;
		case 32: goto st662;
		case 40: goto tr1897;
		case 41: goto tr1898;
		case 78: goto st667;
		case 83: goto st709;
		case 1034: goto tr1902;
		case 1083: goto st714;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1899;
	goto tr1888;
tr1899:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st663;
tr1907:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st663;
st663:
	if ( ++p == pe )
		goto _test_eof663;
case 663:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1904;
		case 32: goto tr1904;
		case 40: goto tr1905;
		case 41: goto tr1906;
		case 1034: goto tr1908;
		case 1083: goto tr1909;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1907;
	goto tr1888;
tr1911:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st664;
tr1912:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st664;
tr1914:
	{
		s->line_counter++;
	}
	goto st664;
tr1904:
	{
		if (s->number64 <= 59) {
			s->loc.m1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st664;
tr1905:
	{
		if (s->number64 <= 59) {
			s->loc.m1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st664;
tr1906:
	{
		if (s->number64 <= 59) {
			s->loc.m1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st664;
tr1908:
	{
		if (s->number64 <= 59) {
			s->loc.m1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st664;
st664:
	if ( ++p == pe )
		goto _test_eof664;
case 664:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st664;
		case 32: goto st664;
		case 40: goto tr1911;
		case 41: goto tr1912;
		case 78: goto st667;
		case 83: goto st709;
		case 1034: goto tr1914;
		case 1083: goto st713;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1913;
	goto tr1888;
tr1920:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st665;
tr1913:
	{
		s->decimals = 3;
	}
	{
		s->decimal_counter = 0;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st665;
st665:
	if ( ++p == pe )
		goto _test_eof665;
case 665:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1916;
		case 32: goto tr1916;
		case 40: goto tr1917;
		case 41: goto tr1918;
		case 46: goto st711;
		case 1034: goto tr1921;
		case 1083: goto tr1922;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1920;
	goto tr1888;
tr1924:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st666;
tr1925:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st666;
tr1926:
	{
		s->line_counter++;
	}
	goto st666;
tr1916:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st666;
tr1917:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st666;
tr1918:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st666;
tr1921:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st666;
tr2115:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st666;
tr2116:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st666;
tr2117:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st666;
tr2119:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st666;
st666:
	if ( ++p == pe )
		goto _test_eof666;
case 666:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st666;
		case 32: goto st666;
		case 40: goto tr1924;
		case 41: goto tr1925;
		case 78: goto st667;
		case 83: goto st709;
		case 1034: goto tr1926;
		case 1083: goto st710;
	}
	goto tr1888;
st667:
	if ( ++p == pe )
		goto _test_eof667;
case 667:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st668;
		case 32: goto st668;
		case 40: goto tr1929;
		case 41: goto tr1930;
		case 1034: goto tr1931;
		case 1083: goto st708;
	}
	goto tr1888;
tr1929:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st668;
tr1930:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st668;
tr1931:
	{
		s->line_counter++;
	}
	goto st668;
tr2110:
	{
		s->loc.lat_sign = -1;
	}
	goto st668;
tr2111:
	{
		s->loc.lat_sign = -1;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st668;
tr2112:
	{
		s->loc.lat_sign = -1;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st668;
tr2113:
	{
		s->loc.lat_sign = -1;
	}
	{
		s->line_counter++;
	}
	goto st668;
st668:
	if ( ++p == pe )
		goto _test_eof668;
case 668:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st668;
		case 32: goto st668;
		case 40: goto tr1929;
		case 41: goto tr1930;
		case 1034: goto tr1931;
		case 1083: goto st708;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1933;
	goto tr1888;
tr1933:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st669;
tr1937:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st669;
st669:
	if ( ++p == pe )
		goto _test_eof669;
case 669:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1934;
		case 32: goto tr1934;
		case 40: goto tr1935;
		case 41: goto tr1936;
		case 1034: goto tr1938;
		case 1083: goto tr1939;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1937;
	goto tr1888;
tr1941:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st670;
tr1942:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st670;
tr1946:
	{
		s->line_counter++;
	}
	goto st670;
tr1934:
	{
		if (s->number64 <= 180) {
			s->loc.d2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st670;
tr1935:
	{
		if (s->number64 <= 180) {
			s->loc.d2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st670;
tr1936:
	{
		if (s->number64 <= 180) {
			s->loc.d2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st670;
tr1938:
	{
		if (s->number64 <= 180) {
			s->loc.d2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st670;
st670:
	if ( ++p == pe )
		goto _test_eof670;
case 670:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st670;
		case 32: goto st670;
		case 40: goto tr1941;
		case 41: goto tr1942;
		case 69: goto st675;
		case 87: goto st702;
		case 1034: goto tr1946;
		case 1083: goto st707;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1943;
	goto tr1888;
tr1943:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st671;
tr1951:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st671;
st671:
	if ( ++p == pe )
		goto _test_eof671;
case 671:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1948;
		case 32: goto tr1948;
		case 40: goto tr1949;
		case 41: goto tr1950;
		case 1034: goto tr1952;
		case 1083: goto tr1953;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1951;
	goto tr1888;
tr1955:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st672;
tr1956:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st672;
tr1958:
	{
		s->line_counter++;
	}
	goto st672;
tr1948:
	{
		if (s->number64 <= 59) {
			s->loc.m2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st672;
tr1949:
	{
		if (s->number64 <= 59) {
			s->loc.m2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st672;
tr1950:
	{
		if (s->number64 <= 59) {
			s->loc.m2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st672;
tr1952:
	{
		if (s->number64 <= 59) {
			s->loc.m2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st672;
st672:
	if ( ++p == pe )
		goto _test_eof672;
case 672:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st672;
		case 32: goto st672;
		case 40: goto tr1955;
		case 41: goto tr1956;
		case 69: goto st675;
		case 87: goto st702;
		case 1034: goto tr1958;
		case 1083: goto st706;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1957;
	goto tr1888;
tr1964:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st673;
tr1957:
	{
		s->decimals = 3;
	}
	{
		s->decimal_counter = 0;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st673;
st673:
	if ( ++p == pe )
		goto _test_eof673;
case 673:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1960;
		case 32: goto tr1960;
		case 40: goto tr1961;
		case 41: goto tr1962;
		case 46: goto st704;
		case 1034: goto tr1965;
		case 1083: goto tr1966;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1964;
	goto tr1888;
tr1968:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st674;
tr1969:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st674;
tr1970:
	{
		s->line_counter++;
	}
	goto st674;
tr1960:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st674;
tr1961:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st674;
tr1962:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st674;
tr1965:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st674;
tr2103:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st674;
tr2104:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st674;
tr2105:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st674;
tr2107:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st674;
st674:
	if ( ++p == pe )
		goto _test_eof674;
case 674:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st674;
		case 32: goto st674;
		case 40: goto tr1968;
		case 41: goto tr1969;
		case 69: goto st675;
		case 87: goto st702;
		case 1034: goto tr1970;
		case 1083: goto st703;
	}
	goto tr1888;
st675:
	if ( ++p == pe )
		goto _test_eof675;
case 675:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st676;
		case 32: goto st676;
		case 40: goto tr1973;
		case 41: goto tr1974;
		case 1034: goto tr1975;
		case 1083: goto st701;
	}
	goto tr1888;
tr1973:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st676;
tr1974:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st676;
tr1975:
	{
		s->line_counter++;
	}
	goto st676;
tr2098:
	{
		s->loc.long_sign = -1;
	}
	goto st676;
tr2099:
	{
		s->loc.long_sign = -1;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st676;
tr2100:
	{
		s->loc.long_sign = -1;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st676;
tr2101:
	{
		s->loc.long_sign = -1;
	}
	{
		s->line_counter++;
	}
	goto st676;
st676:
	if ( ++p == pe )
		goto _test_eof676;
case 676:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st676;
		case 32: goto st676;
		case 40: goto tr1973;
		case 41: goto tr1974;
		case 45: goto st677;
		case 1034: goto tr1975;
		case 1083: goto st701;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1978;
	goto tr1888;
st677:
	if ( ++p == pe )
		goto _test_eof677;
case 677:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr1979;
	goto tr1888;
tr1984:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st678;
tr1978:
	{
		s->decimals = 2;
	}
	{
		s->decimal_counter = 0;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st678;
tr1979:
	{
		s->loc.alt_sign = -1;
	}
	{
		s->decimals = 2;
	}
	{
		s->decimal_counter = 0;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st678;
st678:
	if ( ++p == pe )
		goto _test_eof678;
case 678:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1980;
		case 32: goto tr1980;
		case 40: goto tr1981;
		case 41: goto tr1982;
		case 46: goto st698;
		case 109: goto tr1985;
		case 2058: goto tr1986;
		case 2107: goto tr1987;
		case 2314: goto tr1988;
		case 2363: goto tr1988;
		case 2570: goto tr1989;
		case 2619: goto tr1990;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1984;
	goto tr1888;
tr1992:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st679;
tr1993:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st679;
tr1995:
	{
		s->line_counter++;
	}
	goto st679;
tr1980:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st679;
tr1981:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st679;
tr1982:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st679;
tr1986:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st679;
tr2087:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st679;
tr2088:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st679;
tr2089:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st679;
tr2092:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st679;
st679:
	if ( ++p == pe )
		goto _test_eof679;
case 679:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st679;
		case 32: goto st679;
		case 40: goto tr1992;
		case 41: goto tr1993;
		case 2058: goto tr1995;
		case 2107: goto st697;
		case 2314: goto tr1997;
		case 2363: goto tr1997;
		case 2570: goto tr1998;
		case 2619: goto tr1999;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1994;
	goto tr1888;
tr2004:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st680;
tr1994:
	{
		s->decimals = 2;
	}
	{
		s->decimal_counter = 0;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st680;
st680:
	if ( ++p == pe )
		goto _test_eof680;
case 680:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2000;
		case 32: goto tr2000;
		case 40: goto tr2001;
		case 41: goto tr2002;
		case 46: goto st694;
		case 109: goto tr2005;
		case 2058: goto tr2006;
		case 2107: goto tr2007;
		case 2314: goto tr2008;
		case 2363: goto tr2008;
		case 2570: goto tr2009;
		case 2619: goto tr2010;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2004;
	goto tr1888;
tr2012:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st681;
tr2013:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st681;
tr2015:
	{
		s->line_counter++;
	}
	goto st681;
tr2000:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st681;
tr2001:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st681;
tr2002:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st681;
tr2006:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st681;
tr2076:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st681;
tr2077:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st681;
tr2078:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st681;
tr2081:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st681;
st681:
	if ( ++p == pe )
		goto _test_eof681;
case 681:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st681;
		case 32: goto st681;
		case 40: goto tr2012;
		case 41: goto tr2013;
		case 2058: goto tr2015;
		case 2107: goto st693;
		case 2314: goto tr1997;
		case 2363: goto tr1997;
		case 2570: goto tr2017;
		case 2619: goto tr2018;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2014;
	goto tr1888;
tr2023:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st682;
tr2014:
	{
		s->decimals = 2;
	}
	{
		s->decimal_counter = 0;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st682;
st682:
	if ( ++p == pe )
		goto _test_eof682;
case 682:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2019;
		case 32: goto tr2019;
		case 40: goto tr2020;
		case 41: goto tr2021;
		case 46: goto st690;
		case 109: goto tr2024;
		case 2058: goto tr2025;
		case 2107: goto tr2026;
		case 2314: goto tr2027;
		case 2363: goto tr2027;
		case 2570: goto tr2028;
		case 2619: goto tr2029;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2023;
	goto tr1888;
tr2031:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st683;
tr2032:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st683;
tr2034:
	{
		s->line_counter++;
	}
	goto st683;
tr2019:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st683;
tr2020:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st683;
tr2021:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st683;
tr2025:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st683;
tr2065:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st683;
tr2066:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st683;
tr2067:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st683;
tr2070:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st683;
st683:
	if ( ++p == pe )
		goto _test_eof683;
case 683:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st683;
		case 32: goto st683;
		case 40: goto tr2031;
		case 41: goto tr2032;
		case 2058: goto tr2034;
		case 2107: goto st689;
		case 2314: goto tr1997;
		case 2363: goto tr1997;
		case 2570: goto tr2036;
		case 2619: goto tr2037;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2033;
	goto tr1888;
tr2042:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st684;
tr2033:
	{
		s->decimals = 2;
	}
	{
		s->decimal_counter = 0;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st684;
st684:
	if ( ++p == pe )
		goto _test_eof684;
case 684:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2038;
		case 32: goto tr2038;
		case 40: goto tr2039;
		case 41: goto tr2040;
		case 46: goto st687;
		case 109: goto tr2038;
		case 2058: goto tr2043;
		case 2107: goto tr2044;
		case 2314: goto tr2045;
		case 2363: goto tr2045;
		case 2570: goto tr2046;
		case 2619: goto tr2047;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2042;
	goto tr1888;
tr2049:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st685;
tr2050:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st685;
tr2051:
	{
		s->line_counter++;
	}
	goto st685;
tr2038:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st685;
tr2039:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st685;
tr2040:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st685;
tr2043:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st685;
tr2055:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st685;
tr2056:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st685;
tr2057:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st685;
tr2059:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st685;
st685:
	if ( ++p == pe )
		goto _test_eof685;
case 685:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st685;
		case 32: goto st685;
		case 40: goto tr2049;
		case 41: goto tr2050;
		case 2058: goto tr2051;
		case 2107: goto st686;
		case 2314: goto tr1997;
		case 2363: goto tr1997;
		case 2570: goto tr2053;
		case 2619: goto tr2054;
	}
	goto tr1888;
tr2044:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st686;
tr2060:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st686;
st686:
	if ( ++p == pe )
		goto _test_eof686;
case 686:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2051;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st686;
	goto tr1888;
tr1988:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1121;
tr1997:
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1121;
tr2008:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1121;
tr2027:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1121;
tr2045:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1121;
tr2061:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1121;
tr2072:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1121;
tr2083:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1121;
tr2094:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1121;
st1121:
	if ( ++p == pe )
		goto _test_eof1121;
case 1121:
	goto st0;
tr2053:
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1122;
tr2046:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1122;
tr2062:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1122;
st1122:
	if ( ++p == pe )
		goto _test_eof1122;
case 1122:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st685;
		case 32: goto st685;
		case 40: goto tr2049;
		case 41: goto tr2050;
		case 2058: goto tr2051;
		case 2107: goto st686;
		case 2314: goto tr1997;
		case 2363: goto tr1997;
		case 2570: goto tr2053;
		case 2619: goto tr2054;
	}
	goto tr1888;
tr2054:
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1123;
tr2047:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1123;
tr2063:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.vp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1123;
st1123:
	if ( ++p == pe )
		goto _test_eof1123;
case 1123:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2051;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st686;
	goto tr1888;
st687:
	if ( ++p == pe )
		goto _test_eof687;
case 687:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2055;
		case 32: goto tr2055;
		case 40: goto tr2056;
		case 41: goto tr2057;
		case 109: goto tr2055;
		case 2058: goto tr2059;
		case 2107: goto tr2060;
		case 2314: goto tr2061;
		case 2363: goto tr2061;
		case 2570: goto tr2062;
		case 2619: goto tr2063;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2058;
	goto tr1888;
tr2058:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st688;
tr2064:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st688;
st688:
	if ( ++p == pe )
		goto _test_eof688;
case 688:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2038;
		case 32: goto tr2038;
		case 40: goto tr2039;
		case 41: goto tr2040;
		case 109: goto tr2038;
		case 2058: goto tr2043;
		case 2107: goto tr2044;
		case 2314: goto tr2045;
		case 2363: goto tr2045;
		case 2570: goto tr2046;
		case 2619: goto tr2047;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2064;
	goto tr1888;
tr2026:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st689;
tr2071:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st689;
st689:
	if ( ++p == pe )
		goto _test_eof689;
case 689:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2034;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st689;
	goto tr1888;
tr2036:
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1124;
tr2028:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1124;
tr2073:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1124;
st1124:
	if ( ++p == pe )
		goto _test_eof1124;
case 1124:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st683;
		case 32: goto st683;
		case 40: goto tr2031;
		case 41: goto tr2032;
		case 2058: goto tr2034;
		case 2107: goto st689;
		case 2314: goto tr1997;
		case 2363: goto tr1997;
		case 2570: goto tr2036;
		case 2619: goto tr2037;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2033;
	goto tr1888;
tr2037:
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1125;
tr2029:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1125;
tr2074:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1125;
st1125:
	if ( ++p == pe )
		goto _test_eof1125;
case 1125:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2034;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st689;
	goto tr1888;
st690:
	if ( ++p == pe )
		goto _test_eof690;
case 690:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2065;
		case 32: goto tr2065;
		case 40: goto tr2066;
		case 41: goto tr2067;
		case 109: goto tr2069;
		case 2058: goto tr2070;
		case 2107: goto tr2071;
		case 2314: goto tr2072;
		case 2363: goto tr2072;
		case 2570: goto tr2073;
		case 2619: goto tr2074;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2068;
	goto tr1888;
tr2068:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st691;
tr2075:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st691;
st691:
	if ( ++p == pe )
		goto _test_eof691;
case 691:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2019;
		case 32: goto tr2019;
		case 40: goto tr2020;
		case 41: goto tr2021;
		case 109: goto tr2024;
		case 2058: goto tr2025;
		case 2107: goto tr2026;
		case 2314: goto tr2027;
		case 2363: goto tr2027;
		case 2570: goto tr2028;
		case 2619: goto tr2029;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2075;
	goto tr1888;
tr2024:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st692;
tr2069:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.hp = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st692;
st692:
	if ( ++p == pe )
		goto _test_eof692;
case 692:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st683;
		case 32: goto st683;
		case 40: goto tr2031;
		case 41: goto tr2032;
		case 2058: goto tr2034;
		case 2107: goto st689;
		case 2314: goto tr1997;
		case 2363: goto tr1997;
		case 2570: goto tr2036;
		case 2619: goto tr2037;
	}
	goto tr1888;
tr2007:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st693;
tr2082:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st693;
st693:
	if ( ++p == pe )
		goto _test_eof693;
case 693:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2015;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st693;
	goto tr1888;
tr2017:
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1126;
tr2009:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1126;
tr2084:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1126;
st1126:
	if ( ++p == pe )
		goto _test_eof1126;
case 1126:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st681;
		case 32: goto st681;
		case 40: goto tr2012;
		case 41: goto tr2013;
		case 2058: goto tr2015;
		case 2107: goto st693;
		case 2314: goto tr1997;
		case 2363: goto tr1997;
		case 2570: goto tr2017;
		case 2619: goto tr2018;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2014;
	goto tr1888;
tr2018:
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1127;
tr2010:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1127;
tr2085:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1127;
st1127:
	if ( ++p == pe )
		goto _test_eof1127;
case 1127:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2015;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st693;
	goto tr1888;
st694:
	if ( ++p == pe )
		goto _test_eof694;
case 694:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2076;
		case 32: goto tr2076;
		case 40: goto tr2077;
		case 41: goto tr2078;
		case 109: goto tr2080;
		case 2058: goto tr2081;
		case 2107: goto tr2082;
		case 2314: goto tr2083;
		case 2363: goto tr2083;
		case 2570: goto tr2084;
		case 2619: goto tr2085;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2079;
	goto tr1888;
tr2079:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st695;
tr2086:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st695;
st695:
	if ( ++p == pe )
		goto _test_eof695;
case 695:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2000;
		case 32: goto tr2000;
		case 40: goto tr2001;
		case 41: goto tr2002;
		case 109: goto tr2005;
		case 2058: goto tr2006;
		case 2107: goto tr2007;
		case 2314: goto tr2008;
		case 2363: goto tr2008;
		case 2570: goto tr2009;
		case 2619: goto tr2010;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2086;
	goto tr1888;
tr2005:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st696;
tr2080:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 9000000000ULL) {
			s->loc.siz = s->number64;
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st696;
st696:
	if ( ++p == pe )
		goto _test_eof696;
case 696:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st681;
		case 32: goto st681;
		case 40: goto tr2012;
		case 41: goto tr2013;
		case 2058: goto tr2015;
		case 2107: goto st693;
		case 2314: goto tr1997;
		case 2363: goto tr1997;
		case 2570: goto tr2017;
		case 2619: goto tr2018;
	}
	goto tr1888;
tr1987:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st697;
tr2093:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st697;
st697:
	if ( ++p == pe )
		goto _test_eof697;
case 697:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1995;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st697;
	goto tr1888;
tr1989:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1128;
tr1998:
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1128;
tr2095:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1128;
st1128:
	if ( ++p == pe )
		goto _test_eof1128;
case 1128:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st679;
		case 32: goto st679;
		case 40: goto tr1992;
		case 41: goto tr1993;
		case 2058: goto tr1995;
		case 2107: goto st697;
		case 2314: goto tr1997;
		case 2363: goto tr1997;
		case 2570: goto tr1998;
		case 2619: goto tr1999;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr1994;
	goto tr1888;
tr1990:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1129;
tr1999:
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1129;
tr2096:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	{
		// Write version.
		*(rdata_tail) = 0;
		rdata_tail += 1;
		// Write size.
		*(rdata_tail) = loc64to8(s->loc.siz);
		rdata_tail += 1;
		// Write horizontal precision.
		*(rdata_tail) = loc64to8(s->loc.hp);
		rdata_tail += 1;
		// Write vertical precision.
		*(rdata_tail) = loc64to8(s->loc.vp);
		rdata_tail += 1;
		// Write latitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LAT_ZERO + s->loc.lat_sign *
			(3600000 * s->loc.d1 + 60000 * s->loc.m1 + s->loc.s1));
		rdata_tail += 4;
		// Write longitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_LONG_ZERO + s->loc.long_sign *
			(3600000 * s->loc.d2 + 60000 * s->loc.m2 + s->loc.s2));
		rdata_tail += 4;
		// Write altitude.
		*((uint32_t *)rdata_tail) = htonl(LOC_ALT_ZERO + s->loc.alt_sign *
			(s->loc.alt));
		rdata_tail += 4;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1129;
st1129:
	if ( ++p == pe )
		goto _test_eof1129;
case 1129:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1995;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st697;
	goto tr1888;
st698:
	if ( ++p == pe )
		goto _test_eof698;
case 698:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2087;
		case 32: goto tr2087;
		case 40: goto tr2088;
		case 41: goto tr2089;
		case 109: goto tr2091;
		case 2058: goto tr2092;
		case 2107: goto tr2093;
		case 2314: goto tr2094;
		case 2363: goto tr2094;
		case 2570: goto tr2095;
		case 2619: goto tr2096;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2090;
	goto tr1888;
tr2090:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st699;
tr2097:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st699;
st699:
	if ( ++p == pe )
		goto _test_eof699;
case 699:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr1980;
		case 32: goto tr1980;
		case 40: goto tr1981;
		case 41: goto tr1982;
		case 109: goto tr1985;
		case 2058: goto tr1986;
		case 2107: goto tr1987;
		case 2314: goto tr1988;
		case 2363: goto tr1988;
		case 2570: goto tr1989;
		case 2619: goto tr1990;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2097;
	goto tr1888;
tr1985:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st700;
tr2091:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if ((s->loc.alt_sign ==  1 && s->number64 <= 4284967295) ||
		    (s->loc.alt_sign == -1 && s->number64 <=   10000000))
		{
			s->loc.alt = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st700;
st700:
	if ( ++p == pe )
		goto _test_eof700;
case 700:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st679;
		case 32: goto st679;
		case 40: goto tr1992;
		case 41: goto tr1993;
		case 2058: goto tr1995;
		case 2107: goto st697;
		case 2314: goto tr1997;
		case 2363: goto tr1997;
		case 2570: goto tr1998;
		case 2619: goto tr1999;
	}
	goto tr1888;
tr2102:
	{
		s->loc.long_sign = -1;
	}
	goto st701;
st701:
	if ( ++p == pe )
		goto _test_eof701;
case 701:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1975;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st701;
	goto tr1888;
st702:
	if ( ++p == pe )
		goto _test_eof702;
case 702:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2098;
		case 32: goto tr2098;
		case 40: goto tr2099;
		case 41: goto tr2100;
		case 1034: goto tr2101;
		case 1083: goto tr2102;
	}
	goto tr1888;
tr1966:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st703;
tr2108:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st703;
st703:
	if ( ++p == pe )
		goto _test_eof703;
case 703:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1970;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st703;
	goto tr1888;
st704:
	if ( ++p == pe )
		goto _test_eof704;
case 704:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2103;
		case 32: goto tr2103;
		case 40: goto tr2104;
		case 41: goto tr2105;
		case 1034: goto tr2107;
		case 1083: goto tr2108;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2106;
	goto tr1888;
tr2106:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st705;
tr2109:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st705;
st705:
	if ( ++p == pe )
		goto _test_eof705;
case 705:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1960;
		case 32: goto tr1960;
		case 40: goto tr1961;
		case 41: goto tr1962;
		case 1034: goto tr1965;
		case 1083: goto tr1966;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2109;
	goto tr1888;
tr1953:
	{
		if (s->number64 <= 59) {
			s->loc.m2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st706;
st706:
	if ( ++p == pe )
		goto _test_eof706;
case 706:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1958;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st706;
	goto tr1888;
tr1939:
	{
		if (s->number64 <= 180) {
			s->loc.d2 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st707;
st707:
	if ( ++p == pe )
		goto _test_eof707;
case 707:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1946;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st707;
	goto tr1888;
tr2114:
	{
		s->loc.lat_sign = -1;
	}
	goto st708;
st708:
	if ( ++p == pe )
		goto _test_eof708;
case 708:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1931;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st708;
	goto tr1888;
st709:
	if ( ++p == pe )
		goto _test_eof709;
case 709:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2110;
		case 32: goto tr2110;
		case 40: goto tr2111;
		case 41: goto tr2112;
		case 1034: goto tr2113;
		case 1083: goto tr2114;
	}
	goto tr1888;
tr1922:
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st710;
tr2120:
	{
		s->number64_tmp = s->number64;
	}
	{
		if (s->decimal_counter == 0 && s->number64 < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals);
		} else if (s->decimal_counter <= s->decimals &&
				 s->number64_tmp < UINT32_MAX) {
			s->number64 *= pow(10, s->decimals - s->decimal_counter);
			s->number64 += s->number64_tmp * pow(10, s->decimals);
		} else {
			WARN(ZS_FLOAT_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->number64 <= 59999) {
			s->loc.s1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st710;
st710:
	if ( ++p == pe )
		goto _test_eof710;
case 710:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1926;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st710;
	goto tr1888;
st711:
	if ( ++p == pe )
		goto _test_eof711;
case 711:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2115;
		case 32: goto tr2115;
		case 40: goto tr2116;
		case 41: goto tr2117;
		case 1034: goto tr2119;
		case 1083: goto tr2120;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2118;
	goto tr1888;
tr2118:
	{
		s->number64_tmp = s->number64;
	}
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st712;
tr2121:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->decimal_counter++;
	}
	goto st712;
st712:
	if ( ++p == pe )
		goto _test_eof712;
case 712:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr1916;
		case 32: goto tr1916;
		case 40: goto tr1917;
		case 41: goto tr1918;
		case 1034: goto tr1921;
		case 1083: goto tr1922;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2121;
	goto tr1888;
tr1909:
	{
		if (s->number64 <= 59) {
			s->loc.m1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st713;
st713:
	if ( ++p == pe )
		goto _test_eof713;
case 713:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1914;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st713;
	goto tr1888;
tr1895:
	{
		if (s->number64 <= 90) {
			s->loc.d1 = (uint32_t)(s->number64);
		} else {
			WARN(ZS_BAD_NUMBER);
			p--; {goto st253;}
		}
	}
	goto st714;
st714:
	if ( ++p == pe )
		goto _test_eof714;
case 714:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr1902;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st714;
	goto tr1888;
st715:
	if ( ++p == pe )
		goto _test_eof715;
case 715:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2122;
	goto tr1700;
tr2122:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st716;
tr2126:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st716;
st716:
	if ( ++p == pe )
		goto _test_eof716;
case 716:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2123;
		case 32: goto tr2123;
		case 40: goto tr2124;
		case 41: goto tr2125;
		case 1034: goto tr2127;
		case 1083: goto tr2128;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2126;
	goto tr1700;
tr2130:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st717;
tr2131:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st717;
tr2133:
	{
		s->line_counter++;
	}
	goto st717;
tr2123:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st717;
tr2124:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st717;
tr2125:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st717;
tr2127:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st717;
st717:
	if ( ++p == pe )
		goto _test_eof717;
case 717:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st717;
		case 32: goto st717;
		case 40: goto tr2130;
		case 41: goto tr2131;
		case 1034: goto tr2133;
		case 1083: goto st725;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2132;
	goto tr1700;
tr2132:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st718;
tr2138:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st718;
st718:
	if ( ++p == pe )
		goto _test_eof718;
case 718:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2135;
		case 32: goto tr2135;
		case 40: goto tr2136;
		case 41: goto tr2137;
		case 1034: goto tr2139;
		case 1083: goto tr2140;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2138;
	goto tr1700;
tr2142:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st719;
tr2143:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st719;
tr2145:
	{
		s->line_counter++;
	}
	goto st719;
tr2135:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st719;
tr2136:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st719;
tr2137:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st719;
tr2139:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st719;
st719:
	if ( ++p == pe )
		goto _test_eof719;
case 719:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st719;
		case 32: goto st719;
		case 40: goto tr2142;
		case 41: goto tr2143;
		case 1034: goto tr2145;
		case 1083: goto st724;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2144;
	goto tr1700;
tr2144:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st720;
tr2150:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st720;
st720:
	if ( ++p == pe )
		goto _test_eof720;
case 720:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2147;
		case 32: goto tr2147;
		case 40: goto tr2148;
		case 41: goto tr2149;
		case 1034: goto tr2151;
		case 1083: goto tr2152;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2150;
	goto tr1700;
tr2154:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st721;
tr2155:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st721;
tr2157:
	{
		s->line_counter++;
	}
	goto st721;
tr2147:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st721;
tr2148:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st721;
tr2149:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st721;
tr2151:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st721;
st721:
	if ( ++p == pe )
		goto _test_eof721;
case 721:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st721;
		case 32: goto st721;
		case 40: goto tr2154;
		case 41: goto tr2155;
		case 42: goto tr2156;
		case 92: goto tr2156;
		case 95: goto tr2156;
		case 1034: goto tr2157;
		case 1083: goto st723;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr2156;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2156;
	} else
		goto tr2156;
	goto tr69;
tr2156:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 722; goto st255;} }
	goto st722;
st722:
	if ( ++p == pe )
		goto _test_eof722;
case 722:
	switch( (*p) ) {
		case 32: goto tr2159;
		case 59: goto tr2159;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr2159;
	} else if ( (*p) >= 9 )
		goto tr2159;
	goto tr69;
tr2159:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1130;
st1130:
	if ( ++p == pe )
		goto _test_eof1130;
case 1130:
	goto st0;
tr2152:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st723;
st723:
	if ( ++p == pe )
		goto _test_eof723;
case 723:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2157;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st723;
	goto tr69;
tr2140:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st724;
st724:
	if ( ++p == pe )
		goto _test_eof724;
case 724:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2145;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st724;
	goto tr69;
tr2128:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st725;
st725:
	if ( ++p == pe )
		goto _test_eof725;
case 725:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2133;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st725;
	goto tr69;
st726:
	if ( ++p == pe )
		goto _test_eof726;
case 726:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2160;
	goto tr1700;
tr2160:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st727;
tr2164:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st727;
st727:
	if ( ++p == pe )
		goto _test_eof727;
case 727:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2161;
		case 32: goto tr2161;
		case 40: goto tr2162;
		case 41: goto tr2163;
		case 1034: goto tr2165;
		case 1083: goto tr2166;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2164;
	goto tr1700;
tr2168:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st728;
tr2169:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st728;
tr2171:
	{
		s->line_counter++;
	}
	goto st728;
tr2161:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st728;
tr2162:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st728;
tr2163:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st728;
tr2165:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st728;
st728:
	if ( ++p == pe )
		goto _test_eof728;
case 728:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st728;
		case 32: goto st728;
		case 40: goto tr2168;
		case 41: goto tr2169;
		case 1034: goto tr2171;
		case 1083: goto st742;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2170;
	goto tr1700;
tr2170:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st729;
tr2176:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st729;
st729:
	if ( ++p == pe )
		goto _test_eof729;
case 729:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2173;
		case 32: goto tr2173;
		case 40: goto tr2174;
		case 41: goto tr2175;
		case 1034: goto tr2177;
		case 1083: goto tr2178;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2176;
	goto tr1700;
tr2181:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st730;
tr2182:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st730;
tr2183:
	{
		s->line_counter++;
	}
	goto st730;
tr2173:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st730;
tr2174:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st730;
tr2175:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st730;
tr2177:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st730;
st730:
	if ( ++p == pe )
		goto _test_eof730;
case 730:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st730;
		case 32: goto st730;
		case 40: goto tr2181;
		case 41: goto tr2182;
		case 1034: goto tr2183;
		case 1083: goto st741;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr2179;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr2179;
	} else
		goto tr2179;
	goto tr69;
tr2179:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{ p--; {stack[top++] = 731; goto st264;} }
	goto st731;
st731:
	if ( ++p == pe )
		goto _test_eof731;
case 731:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2185;
		case 32: goto tr2185;
		case 40: goto tr2186;
		case 41: goto tr2187;
		case 1034: goto tr2188;
		case 1083: goto tr2189;
	}
	goto tr69;
tr2192:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st732;
tr2193:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st732;
tr2194:
	{
		s->line_counter++;
	}
	goto st732;
tr2185:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st732;
tr2186:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st732;
tr2187:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st732;
tr2188:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st732;
st732:
	if ( ++p == pe )
		goto _test_eof732;
case 732:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st732;
		case 32: goto st732;
		case 40: goto tr2192;
		case 41: goto tr2193;
		case 1034: goto tr2194;
		case 1083: goto st740;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr2190;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr2190;
	} else
		goto tr2190;
	goto tr69;
tr2190:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{ p--; {stack[top++] = 733; goto st264;} }
	goto st733;
st733:
	if ( ++p == pe )
		goto _test_eof733;
case 733:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2196;
		case 32: goto tr2196;
		case 40: goto tr2197;
		case 41: goto tr2198;
		case 1034: goto tr2199;
		case 1083: goto tr2200;
	}
	goto tr69;
tr2203:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st734;
tr2204:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st734;
tr2205:
	{
		s->line_counter++;
	}
	goto st734;
tr2196:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st734;
tr2197:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st734;
tr2198:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st734;
tr2199:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st734;
st734:
	if ( ++p == pe )
		goto _test_eof734;
case 734:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st734;
		case 32: goto st734;
		case 40: goto tr2203;
		case 41: goto tr2204;
		case 1034: goto tr2205;
		case 1083: goto st739;
	}
	if ( _widec < 11 ) {
		if ( _widec <= 8 )
			goto tr2201;
	} else if ( _widec > 58 ) {
		if ( 60 <= _widec )
			goto tr2201;
	} else
		goto tr2201;
	goto tr69;
tr2201:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{ p--; {stack[top++] = 735; goto st264;} }
	goto st735;
st735:
	if ( ++p == pe )
		goto _test_eof735;
case 735:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2207;
		case 32: goto tr2207;
		case 40: goto tr2208;
		case 41: goto tr2209;
		case 1034: goto tr2210;
		case 1083: goto tr2211;
	}
	goto tr69;
tr2213:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st736;
tr2214:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st736;
tr2216:
	{
		s->line_counter++;
	}
	goto st736;
tr2207:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st736;
tr2208:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st736;
tr2209:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st736;
tr2210:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st736;
st736:
	if ( ++p == pe )
		goto _test_eof736;
case 736:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st736;
		case 32: goto st736;
		case 40: goto tr2213;
		case 41: goto tr2214;
		case 42: goto tr2215;
		case 92: goto tr2215;
		case 95: goto tr2215;
		case 1034: goto tr2216;
		case 1083: goto st738;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr2215;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2215;
	} else
		goto tr2215;
	goto tr69;
tr2215:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 737; goto st255;} }
	goto st737;
st737:
	if ( ++p == pe )
		goto _test_eof737;
case 737:
	switch( (*p) ) {
		case 32: goto tr2218;
		case 59: goto tr2218;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr2218;
	} else if ( (*p) >= 9 )
		goto tr2218;
	goto tr69;
tr2218:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1131;
st1131:
	if ( ++p == pe )
		goto _test_eof1131;
case 1131:
	goto st0;
tr2211:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st738;
st738:
	if ( ++p == pe )
		goto _test_eof738;
case 738:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2216;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st738;
	goto tr69;
tr2200:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st739;
st739:
	if ( ++p == pe )
		goto _test_eof739;
case 739:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2205;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st739;
	goto tr69;
tr2189:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st740;
st740:
	if ( ++p == pe )
		goto _test_eof740;
case 740:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2194;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st740;
	goto tr69;
tr2178:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st741;
st741:
	if ( ++p == pe )
		goto _test_eof741;
case 741:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2183;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st741;
	goto tr69;
tr2166:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st742;
st742:
	if ( ++p == pe )
		goto _test_eof742;
case 742:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2171;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st742;
	goto tr69;
st743:
	if ( ++p == pe )
		goto _test_eof743;
case 743:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr2219;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr2219;
	} else
		goto tr2219;
	goto tr69;
tr2219:
	{ p--; {stack[top++] = 744; goto st562;} }
	goto st744;
st744:
	if ( ++p == pe )
		goto _test_eof744;
case 744:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st745;
		case 32: goto st745;
		case 40: goto tr2221;
		case 41: goto tr2222;
		case 1034: goto tr2223;
		case 1083: goto st753;
	}
	goto tr69;
tr2221:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st745;
tr2222:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st745;
tr2223:
	{
		s->line_counter++;
	}
	goto st745;
st745:
	if ( ++p == pe )
		goto _test_eof745;
case 745:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st745;
		case 32: goto st745;
		case 40: goto tr2221;
		case 41: goto tr2222;
		case 1034: goto tr2223;
		case 1083: goto st753;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2225;
	goto tr1700;
tr2225:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st746;
tr2229:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st746;
st746:
	if ( ++p == pe )
		goto _test_eof746;
case 746:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2226;
		case 32: goto tr2226;
		case 40: goto tr2227;
		case 41: goto tr2228;
		case 1034: goto tr2230;
		case 1083: goto tr2231;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2229;
	goto tr1700;
tr2233:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st747;
tr2234:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st747;
tr2236:
	{
		s->line_counter++;
	}
	goto st747;
tr2226:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st747;
tr2227:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st747;
tr2228:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st747;
tr2230:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st747;
st747:
	if ( ++p == pe )
		goto _test_eof747;
case 747:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st747;
		case 32: goto st747;
		case 40: goto tr2233;
		case 41: goto tr2234;
		case 1034: goto tr2236;
		case 1083: goto st752;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2235;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2235;
	} else
		goto tr2235;
	goto tr69;
tr2235:
	{ p--; {stack[top++] = 748; goto st467;} }
	goto st748;
st748:
	if ( ++p == pe )
		goto _test_eof748;
case 748:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st749;
		case 32: goto st749;
		case 40: goto tr2239;
		case 41: goto tr2240;
		case 1034: goto tr2241;
		case 1083: goto st751;
	}
	goto tr69;
tr2239:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st749;
tr2240:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st749;
tr2241:
	{
		s->line_counter++;
	}
	goto st749;
st749:
	if ( ++p == pe )
		goto _test_eof749;
case 749:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st749;
		case 32: goto st749;
		case 40: goto tr2239;
		case 41: goto tr2240;
		case 43: goto tr2243;
		case 1034: goto tr2241;
		case 1083: goto st751;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr2243;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2243;
	} else
		goto tr2243;
	goto tr69;
tr2243:
	{ p--; {stack[top++] = 750; goto st314;} }
	goto st750;
st750:
	if ( ++p == pe )
		goto _test_eof750;
case 750:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1152 + ((*p) - -128));
			if ( 
 !s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1152 + ((*p) - -128));
		if ( 
 !s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 1546: goto tr2244;
		case 1595: goto tr2244;
	}
	goto tr69;
tr2244:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1132;
st1132:
	if ( ++p == pe )
		goto _test_eof1132;
case 1132:
	goto st0;
st751:
	if ( ++p == pe )
		goto _test_eof751;
case 751:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2241;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st751;
	goto tr69;
tr2231:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st752;
st752:
	if ( ++p == pe )
		goto _test_eof752;
case 752:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2236;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st752;
	goto tr69;
st753:
	if ( ++p == pe )
		goto _test_eof753;
case 753:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2223;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st753;
	goto tr69;
tr2248:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st754;
tr2249:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st754;
tr2252:
	{
		s->line_counter++;
	}
	goto st754;
tr2265:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st253;}
		}
	}
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st253;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	goto st754;
tr2266:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st253;}
		}
	}
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st253;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st754;
tr2267:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st253;}
		}
	}
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st253;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st754;
tr2269:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st253;}
		}
	}
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st253;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	{
		s->line_counter++;
	}
	goto st754;
st754:
	if ( ++p == pe )
		goto _test_eof754;
case 754:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st754;
		case 32: goto st754;
		case 33: goto tr2247;
		case 40: goto tr2248;
		case 41: goto tr2249;
		case 49: goto tr2250;
		case 50: goto tr2251;
		case 2058: goto tr2252;
		case 2107: goto st761;
		case 2314: goto tr2254;
		case 2363: goto tr2254;
		case 2570: goto tr2255;
		case 2619: goto tr2256;
	}
	goto tr2245;
tr2247:
	{
		memset(&(s->apl), 0, sizeof(s->apl));
	}
	{
		s->apl.excl_flag = 128; // dec 128  = bin 10000000.
	}
	goto st755;
st755:
	if ( ++p == pe )
		goto _test_eof755;
case 755:
	switch( (*p) ) {
		case 49: goto tr2257;
		case 50: goto tr2258;
	}
	goto tr2245;
tr2250:
	{
		memset(&(s->apl), 0, sizeof(s->apl));
	}
	{
		s->apl.addr_family = 1;
	}
	goto st756;
tr2257:
	{
		s->apl.addr_family = 1;
	}
	goto st756;
st756:
	if ( ++p == pe )
		goto _test_eof756;
case 756:
	if ( (*p) == 58 )
		goto st757;
	goto tr2245;
st757:
	if ( ++p == pe )
		goto _test_eof757;
case 757:
	if ( (*p) == 46 )
		goto tr2261;
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2261;
	goto tr2260;
tr2261:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st758;
tr2262:
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st758;
st758:
	if ( ++p == pe )
		goto _test_eof758;
case 758:
	if ( (*p) == 47 )
		goto tr2263;
	if ( 46 <= (*p) && (*p) <= 57 )
		goto tr2262;
	goto tr2260;
tr2263:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	goto st759;
tr2277:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	goto st759;
st759:
	if ( ++p == pe )
		goto _test_eof759;
case 759:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2264;
	goto tr2245;
tr2264:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st760;
tr2268:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st760;
st760:
	if ( ++p == pe )
		goto _test_eof760;
case 760:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto tr2265;
		case 32: goto tr2265;
		case 40: goto tr2266;
		case 41: goto tr2267;
		case 2058: goto tr2269;
		case 2107: goto tr2270;
		case 2314: goto tr2271;
		case 2363: goto tr2271;
		case 2570: goto tr2272;
		case 2619: goto tr2273;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2268;
	goto tr2245;
tr2270:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st253;}
		}
	}
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st253;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	goto st761;
st761:
	if ( ++p == pe )
		goto _test_eof761;
case 761:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2252;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st761;
	goto tr69;
tr2254:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1133;
tr2271:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st253;}
		}
	}
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st253;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1133;
st1133:
	if ( ++p == pe )
		goto _test_eof1133;
case 1133:
	goto st0;
tr2255:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1134;
tr2272:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st253;}
		}
	}
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st253;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1134;
st1134:
	if ( ++p == pe )
		goto _test_eof1134;
case 1134:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st754;
		case 32: goto st754;
		case 33: goto tr2247;
		case 40: goto tr2248;
		case 41: goto tr2249;
		case 49: goto tr2250;
		case 50: goto tr2251;
		case 2058: goto tr2252;
		case 2107: goto st761;
		case 2314: goto tr2254;
		case 2363: goto tr2254;
		case 2570: goto tr2255;
		case 2619: goto tr2256;
	}
	goto tr2245;
tr2251:
	{
		memset(&(s->apl), 0, sizeof(s->apl));
	}
	{
		s->apl.addr_family = 2;
	}
	goto st762;
tr2258:
	{
		s->apl.addr_family = 2;
	}
	goto st762;
st762:
	if ( ++p == pe )
		goto _test_eof762;
case 762:
	if ( (*p) == 58 )
		goto st763;
	goto tr2245;
st763:
	if ( ++p == pe )
		goto _test_eof763;
case 763:
	if ( (*p) == 46 )
		goto tr2275;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 58 )
			goto tr2275;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr2275;
	} else
		goto tr2275;
	goto tr2260;
tr2275:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st764;
tr2276:
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st764;
st764:
	if ( ++p == pe )
		goto _test_eof764;
case 764:
	if ( (*p) == 47 )
		goto tr2277;
	if ( (*p) < 65 ) {
		if ( 46 <= (*p) && (*p) <= 58 )
			goto tr2276;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr2276;
	} else
		goto tr2276;
	goto tr2260;
tr2256:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1135;
tr2273:
	{
		if ((s->apl.addr_family == 1 && s->number64 <= 32) ||
		    (s->apl.addr_family == 2 && s->number64 <= 128)) {
			s->apl.prefix_length = (uint8_t)(s->number64);
		} else {
			WARN(ZS_BAD_APL);
			p--; {goto st253;}
		}
	}
	{
		// Write address family.
		*((uint16_t *)rdata_tail) = htons(s->apl.addr_family);
		rdata_tail += 2;
		// Write prefix length in bites.
		*(rdata_tail) = s->apl.prefix_length;
		rdata_tail += 1;
		// Copy address to buffer.
		uint8_t len;
		switch (s->apl.addr_family) {
		case 1:
			len = INET4_ADDR_LENGTH;
			memcpy(s->buffer, &(addr4.s_addr), len);
			break;
		case 2:
			len = INET6_ADDR_LENGTH;
			memcpy(s->buffer, &(addr6.s6_addr), len);
			break;
		default:
			WARN(ZS_BAD_APL);
			p--; {goto st253;}
		}
		// Find prefix without trailing zeroes.
		while (len > 0) {
			if ((s->buffer[len - 1] & 255) != 0) {
				break;
			}
			len--;
		}
		// Write negation flag + prefix length in bytes.
		*(rdata_tail) = len + s->apl.excl_flag;
		rdata_tail += 1;
		// Write address prefix non-null data.
		memcpy(rdata_tail, s->buffer, len);
		rdata_tail += len;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1135;
st1135:
	if ( ++p == pe )
		goto _test_eof1135;
case 1135:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2252;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st761;
	goto tr69;
st765:
	if ( ++p == pe )
		goto _test_eof765;
case 765:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2278;
	goto tr1700;
tr2278:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st766;
tr2282:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st766;
st766:
	if ( ++p == pe )
		goto _test_eof766;
case 766:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2279;
		case 32: goto tr2279;
		case 40: goto tr2280;
		case 41: goto tr2281;
		case 1034: goto tr2283;
		case 1083: goto tr2284;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2282;
	goto tr1700;
tr2286:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st767;
tr2287:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st767;
tr2289:
	{
		s->line_counter++;
	}
	goto st767;
tr2279:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st767;
tr2280:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st767;
tr2281:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st767;
tr2283:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st767;
st767:
	if ( ++p == pe )
		goto _test_eof767;
case 767:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st767;
		case 32: goto st767;
		case 40: goto tr2286;
		case 41: goto tr2287;
		case 1034: goto tr2289;
		case 1083: goto st777;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2288;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2288;
	} else
		goto tr2288;
	goto tr69;
tr2288:
	{ p--; {stack[top++] = 768; goto st467;} }
	goto st768;
st768:
	if ( ++p == pe )
		goto _test_eof768;
case 768:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st769;
		case 32: goto st769;
		case 40: goto tr2292;
		case 41: goto tr2293;
		case 1034: goto tr2294;
		case 1083: goto st776;
	}
	goto tr69;
tr2292:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st769;
tr2293:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st769;
tr2294:
	{
		s->line_counter++;
	}
	goto st769;
st769:
	if ( ++p == pe )
		goto _test_eof769;
case 769:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st769;
		case 32: goto st769;
		case 40: goto tr2292;
		case 41: goto tr2293;
		case 1034: goto tr2294;
		case 1083: goto st776;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2296;
	goto tr1700;
tr2296:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st770;
tr2300:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st770;
st770:
	if ( ++p == pe )
		goto _test_eof770;
case 770:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2297;
		case 32: goto tr2297;
		case 40: goto tr2298;
		case 41: goto tr2299;
		case 1034: goto tr2301;
		case 1083: goto tr2302;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2300;
	goto tr1700;
tr2305:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st771;
tr2306:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st771;
tr2308:
	{
		s->line_counter++;
	}
	goto st771;
tr2297:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st771;
tr2298:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st771;
tr2299:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st771;
tr2301:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st771;
st771:
	if ( ++p == pe )
		goto _test_eof771;
case 771:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st771;
		case 32: goto st771;
		case 40: goto tr2305;
		case 41: goto tr2306;
		case 1034: goto tr2308;
		case 1083: goto st775;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2307;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2307;
	} else
		goto tr2307;
	goto tr2303;
tr2307:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st772;
st772:
	if ( ++p == pe )
		goto _test_eof772;
case 772:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr2310;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr2310;
	} else
		goto tr2310;
	goto tr2303;
tr2312:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st773;
tr2313:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st773;
tr2314:
	{
		s->line_counter++;
	}
	goto st773;
tr2310:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st773;
st773:
	if ( ++p == pe )
		goto _test_eof773;
case 773:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st773;
		case 32: goto st773;
		case 40: goto tr2312;
		case 41: goto tr2313;
		case 2058: goto tr2314;
		case 2107: goto st774;
		case 2314: goto tr2316;
		case 2363: goto tr2316;
		case 2570: goto tr2317;
		case 2619: goto tr2318;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2307;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2307;
	} else
		goto tr2307;
	goto tr2303;
st774:
	if ( ++p == pe )
		goto _test_eof774;
case 774:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2314;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st774;
	goto tr2303;
tr2316:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1136;
st1136:
	if ( ++p == pe )
		goto _test_eof1136;
case 1136:
	goto st0;
tr2317:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1137;
st1137:
	if ( ++p == pe )
		goto _test_eof1137;
case 1137:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st773;
		case 32: goto st773;
		case 40: goto tr2312;
		case 41: goto tr2313;
		case 2058: goto tr2314;
		case 2107: goto st774;
		case 2314: goto tr2316;
		case 2363: goto tr2316;
		case 2570: goto tr2317;
		case 2619: goto tr2318;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2307;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2307;
	} else
		goto tr2307;
	goto tr2303;
tr2318:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1138;
st1138:
	if ( ++p == pe )
		goto _test_eof1138;
case 1138:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2314;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st774;
	goto tr2303;
tr2302:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st775;
st775:
	if ( ++p == pe )
		goto _test_eof775;
case 775:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2308;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st775;
	goto tr69;
st776:
	if ( ++p == pe )
		goto _test_eof776;
case 776:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2294;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st776;
	goto tr69;
tr2284:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st777;
st777:
	if ( ++p == pe )
		goto _test_eof777;
case 777:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2289;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st777;
	goto tr69;
st778:
	if ( ++p == pe )
		goto _test_eof778;
case 778:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2319;
	goto tr1700;
tr2319:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st779;
tr2323:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st779;
st779:
	if ( ++p == pe )
		goto _test_eof779;
case 779:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2320;
		case 32: goto tr2320;
		case 40: goto tr2321;
		case 41: goto tr2322;
		case 1034: goto tr2324;
		case 1083: goto tr2325;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2323;
	goto tr1700;
tr2327:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st780;
tr2328:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st780;
tr2330:
	{
		s->line_counter++;
	}
	goto st780;
tr2320:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st780;
tr2321:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st780;
tr2322:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st780;
tr2324:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st780;
st780:
	if ( ++p == pe )
		goto _test_eof780;
case 780:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st780;
		case 32: goto st780;
		case 40: goto tr2327;
		case 41: goto tr2328;
		case 1034: goto tr2330;
		case 1083: goto st787;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2329;
	goto tr1700;
tr2329:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st781;
tr2335:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st781;
st781:
	if ( ++p == pe )
		goto _test_eof781;
case 781:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2332;
		case 32: goto tr2332;
		case 40: goto tr2333;
		case 41: goto tr2334;
		case 1034: goto tr2336;
		case 1083: goto tr2337;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2335;
	goto tr1700;
tr2339:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st782;
tr2340:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st782;
tr2342:
	{
		s->line_counter++;
	}
	goto st782;
tr2332:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st782;
tr2333:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st782;
tr2334:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st782;
tr2336:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st782;
st782:
	if ( ++p == pe )
		goto _test_eof782;
case 782:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st782;
		case 32: goto st782;
		case 40: goto tr2339;
		case 41: goto tr2340;
		case 1034: goto tr2342;
		case 1083: goto st786;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2341;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2341;
	} else
		goto tr2341;
	goto tr2303;
tr2341:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st783;
st783:
	if ( ++p == pe )
		goto _test_eof783;
case 783:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr2344;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr2344;
	} else
		goto tr2344;
	goto tr2303;
tr2346:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st784;
tr2347:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st784;
tr2348:
	{
		s->line_counter++;
	}
	goto st784;
tr2344:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st784;
st784:
	if ( ++p == pe )
		goto _test_eof784;
case 784:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st784;
		case 32: goto st784;
		case 40: goto tr2346;
		case 41: goto tr2347;
		case 2058: goto tr2348;
		case 2107: goto st785;
		case 2314: goto tr2350;
		case 2363: goto tr2350;
		case 2570: goto tr2351;
		case 2619: goto tr2352;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2341;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2341;
	} else
		goto tr2341;
	goto tr2303;
st785:
	if ( ++p == pe )
		goto _test_eof785;
case 785:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2348;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st785;
	goto tr2303;
tr2350:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1139;
st1139:
	if ( ++p == pe )
		goto _test_eof1139;
case 1139:
	goto st0;
tr2351:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1140;
st1140:
	if ( ++p == pe )
		goto _test_eof1140;
case 1140:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st784;
		case 32: goto st784;
		case 40: goto tr2346;
		case 41: goto tr2347;
		case 2058: goto tr2348;
		case 2107: goto st785;
		case 2314: goto tr2350;
		case 2363: goto tr2350;
		case 2570: goto tr2351;
		case 2619: goto tr2352;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2341;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2341;
	} else
		goto tr2341;
	goto tr2303;
tr2352:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1141;
st1141:
	if ( ++p == pe )
		goto _test_eof1141;
case 1141:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2348;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st785;
	goto tr2303;
tr2337:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st786;
st786:
	if ( ++p == pe )
		goto _test_eof786;
case 786:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2342;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st786;
	goto tr69;
tr2325:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st787;
st787:
	if ( ++p == pe )
		goto _test_eof787;
case 787:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2330;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st787;
	goto tr69;
st788:
	if ( ++p == pe )
		goto _test_eof788;
case 788:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2353;
	goto tr1700;
tr2353:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st789;
tr2357:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st789;
st789:
	if ( ++p == pe )
		goto _test_eof789;
case 789:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2354;
		case 32: goto tr2354;
		case 40: goto tr2355;
		case 41: goto tr2356;
		case 1034: goto tr2358;
		case 1083: goto tr2359;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2357;
	goto tr1700;
tr2362:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st790;
tr2363:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st790;
tr2368:
	{
		s->line_counter++;
	}
	goto st790;
tr2354:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st790;
tr2355:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st790;
tr2356:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st790;
tr2358:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st790;
st790:
	if ( ++p == pe )
		goto _test_eof790;
case 790:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st790;
		case 32: goto st790;
		case 40: goto tr2362;
		case 41: goto tr2363;
		case 48: goto tr2364;
		case 49: goto tr2365;
		case 50: goto tr2366;
		case 51: goto tr2367;
		case 1034: goto tr2368;
		case 1083: goto st826;
	}
	goto tr2360;
tr2364:
	{
		*(rdata_tail++) = 0;
	}
	goto st791;
st791:
	if ( ++p == pe )
		goto _test_eof791;
case 791:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st792;
		case 32: goto st792;
		case 40: goto tr2371;
		case 41: goto tr2372;
		case 1034: goto tr2373;
		case 1083: goto st804;
	}
	goto tr2360;
tr2371:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st792;
tr2372:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st792;
tr2373:
	{
		s->line_counter++;
	}
	goto st792;
st792:
	if ( ++p == pe )
		goto _test_eof792;
case 792:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st792;
		case 32: goto st792;
		case 40: goto tr2371;
		case 41: goto tr2372;
		case 1034: goto tr2373;
		case 1083: goto st804;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2376;
	goto tr2375;
tr2376:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st793;
tr2380:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st793;
st793:
	if ( ++p == pe )
		goto _test_eof793;
case 793:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2377;
		case 32: goto tr2377;
		case 40: goto tr2378;
		case 41: goto tr2379;
		case 1034: goto tr2381;
		case 1083: goto tr2382;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2380;
	goto tr2375;
tr2384:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st794;
tr2385:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st794;
tr2387:
	{
		s->line_counter++;
	}
	goto st794;
tr2377:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st794;
tr2378:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st794;
tr2379:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st794;
tr2381:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st794;
st794:
	if ( ++p == pe )
		goto _test_eof794;
case 794:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st794;
		case 32: goto st794;
		case 40: goto tr2384;
		case 41: goto tr2385;
		case 46: goto st795;
		case 1034: goto tr2387;
		case 1083: goto st803;
	}
	goto tr2360;
st795:
	if ( ++p == pe )
		goto _test_eof795;
case 795:
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
				if ( 
 !s->multiline  ) _widec += 512;
				if ( 
 s->number64 != 0  ) _widec += 1024;
				if ( 
 s->number64 == 0  ) _widec += 2048;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(5760 + ((*p) - -128));
			if ( 
 s->number64 != 0  ) _widec += 256;
			if ( 
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if ( 
 s->number64 != 0  ) _widec += 256;
				if ( 
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
				if ( 
 !s->multiline  ) _widec += 512;
				if ( 
 s->number64 != 0  ) _widec += 1024;
				if ( 
 s->number64 == 0  ) _widec += 2048;
			}
		} else {
			_widec = (short)(5760 + ((*p) - -128));
			if ( 
 s->number64 != 0  ) _widec += 256;
			if ( 
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(5760 + ((*p) - -128));
		if ( 
 s->number64 != 0  ) _widec += 256;
		if ( 
 s->number64 == 0  ) _widec += 512;
	}
	switch( _widec ) {
		case 6153: goto st796;
		case 6176: goto st796;
		case 6184: goto tr2391;
		case 6185: goto tr2392;
		case 6409: goto st799;
		case 6432: goto st799;
		case 6440: goto tr2394;
		case 6441: goto tr2395;
		case 6665: goto st801;
		case 6688: goto st801;
		case 6696: goto tr2397;
		case 6697: goto tr2398;
		case 9482: goto tr2399;
		case 9531: goto tr2399;
		case 9738: goto tr2399;
		case 9787: goto tr2399;
		case 10250: goto tr2400;
		case 10299: goto st798;
		case 10506: goto tr2399;
		case 10555: goto tr2399;
		case 10762: goto tr2402;
		case 10811: goto tr2403;
		case 11274: goto tr2404;
		case 11323: goto st800;
		case 11530: goto tr2399;
		case 11579: goto tr2399;
		case 11786: goto tr2406;
		case 11835: goto tr2407;
		case 12298: goto tr2408;
		case 12347: goto st802;
		case 12554: goto tr2399;
		case 12603: goto tr2399;
		case 12810: goto tr2410;
		case 12859: goto tr2411;
	}
	goto tr2389;
tr2391:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st796;
tr2392:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st796;
tr2400:
	{
		s->line_counter++;
	}
	goto st796;
tr2516:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st796;
tr2517:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st796;
tr2518:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st796;
tr2526:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st796;
tr2435:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	goto st796;
tr2436:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st796;
tr2437:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st796;
tr2445:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	goto st796;
tr2476:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	goto st796;
tr2477:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st796;
tr2478:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st796;
tr2486:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	goto st796;
st796:
	if ( ++p == pe )
		goto _test_eof796;
case 796:
	_widec = (*p);
	if ( (*p) < 43 ) {
		if ( (*p) < 32 ) {
			if ( (*p) > 9 ) {
				if ( 10 <= (*p) && (*p) <= 10 ) {
					_widec = (short)(3200 + ((*p) - -128));
					if ( 
 s->multiline  ) _widec += 256;
					if ( 
 s->number64 != 0  ) _widec += 512;
				}
			} else if ( (*p) >= 9 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if ( 
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 32 ) {
			if ( (*p) > 40 ) {
				if ( 41 <= (*p) && (*p) <= 41 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if ( 
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 40 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if ( 
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(2688 + ((*p) - -128));
			if ( 
 s->number64 != 0  ) _widec += 256;
		}
	} else if ( (*p) > 43 ) {
		if ( (*p) < 59 ) {
			if ( (*p) > 47 ) {
				if ( 48 <= (*p) && (*p) <= 57 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if ( 
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 47 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if ( 
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 59 ) {
			if ( (*p) > 90 ) {
				if ( 97 <= (*p) && (*p) <= 122 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if ( 
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 65 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if ( 
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(3200 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 s->number64 != 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(2688 + ((*p) - -128));
		if ( 
 s->number64 != 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 3081: goto st796;
		case 3104: goto st796;
		case 3112: goto tr2391;
		case 3113: goto tr2392;
		case 3115: goto tr2413;
		case 4106: goto tr2400;
		case 4155: goto st798;
	}
	if ( _widec < 3137 ) {
		if ( 3119 <= _widec && _widec <= 3129 )
			goto tr2413;
	} else if ( _widec > 3162 ) {
		if ( 3169 <= _widec && _widec <= 3194 )
			goto tr2413;
	} else
		goto tr2413;
	goto tr2412;
tr2413:
	{ p--; {stack[top++] = 797; goto st314;} }
	goto st797;
st797:
	if ( ++p == pe )
		goto _test_eof797;
case 797:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1152 + ((*p) - -128));
			if ( 
 !s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1152 + ((*p) - -128));
		if ( 
 !s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 1546: goto tr2399;
		case 1595: goto tr2399;
	}
	goto tr2412;
tr2399:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1142;
tr2444:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1142;
tr2525:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1142;
tr2485:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1142;
st1142:
	if ( ++p == pe )
		goto _test_eof1142;
case 1142:
	goto st0;
tr2527:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st798;
tr2446:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	goto st798;
tr2487:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	goto st798;
st798:
	if ( ++p == pe )
		goto _test_eof798;
case 798:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(3200 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 s->number64 != 0  ) _widec += 512;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(3200 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 s->number64 != 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(3200 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 s->number64 != 0  ) _widec += 512;
	}
	if ( _widec == 4106 )
		goto tr2400;
	if ( 3968 <= _widec && _widec <= 4223 )
		goto st798;
	goto tr2412;
tr2394:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st799;
tr2395:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st799;
tr2404:
	{
		s->line_counter++;
	}
	goto st799;
tr2519:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st799;
tr2520:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st799;
tr2521:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st799;
tr2530:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st799;
tr2438:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	goto st799;
tr2439:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st799;
tr2440:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st799;
tr2449:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	goto st799;
tr2479:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	goto st799;
tr2480:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st799;
tr2481:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st799;
tr2490:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	goto st799;
st799:
	if ( ++p == pe )
		goto _test_eof799;
case 799:
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(12928 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
				if ( 
 !s->multiline  ) _widec += 512;
				if ( 
 s->number64 == 0  ) _widec += 1024;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(4224 + ((*p) - -128));
			if ( 
 s->number64 == 0  ) _widec += 256;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(4224 + ((*p) - -128));
				if ( 
 s->number64 == 0  ) _widec += 256;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(12928 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
				if ( 
 !s->multiline  ) _widec += 512;
				if ( 
 s->number64 == 0  ) _widec += 1024;
			}
		} else {
			_widec = (short)(4224 + ((*p) - -128));
			if ( 
 s->number64 == 0  ) _widec += 256;
		}
	} else {
		_widec = (short)(4224 + ((*p) - -128));
		if ( 
 s->number64 == 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 4617: goto st799;
		case 4640: goto st799;
		case 4648: goto tr2394;
		case 4649: goto tr2395;
		case 13578: goto tr2399;
		case 13627: goto tr2399;
		case 13834: goto tr2399;
		case 13883: goto tr2399;
		case 14346: goto tr2404;
		case 14395: goto st800;
		case 14602: goto tr2399;
		case 14651: goto tr2399;
		case 14858: goto tr2406;
		case 14907: goto tr2407;
	}
	goto tr2412;
tr2531:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st800;
tr2450:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	goto st800;
tr2491:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	goto st800;
st800:
	if ( ++p == pe )
		goto _test_eof800;
case 800:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(4736 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(4736 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(4736 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 s->number64 == 0  ) _widec += 512;
	}
	if ( _widec == 5642 )
		goto tr2404;
	if ( 5504 <= _widec && _widec <= 5759 )
		goto st800;
	goto tr2412;
tr2406:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1143;
tr2451:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1143;
tr2492:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1143;
tr2532:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1143;
st1143:
	if ( ++p == pe )
		goto _test_eof1143;
case 1143:
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(12928 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
				if ( 
 !s->multiline  ) _widec += 512;
				if ( 
 s->number64 == 0  ) _widec += 1024;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(4224 + ((*p) - -128));
			if ( 
 s->number64 == 0  ) _widec += 256;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(4224 + ((*p) - -128));
				if ( 
 s->number64 == 0  ) _widec += 256;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(12928 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
				if ( 
 !s->multiline  ) _widec += 512;
				if ( 
 s->number64 == 0  ) _widec += 1024;
			}
		} else {
			_widec = (short)(4224 + ((*p) - -128));
			if ( 
 s->number64 == 0  ) _widec += 256;
		}
	} else {
		_widec = (short)(4224 + ((*p) - -128));
		if ( 
 s->number64 == 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 4617: goto st799;
		case 4640: goto st799;
		case 4648: goto tr2394;
		case 4649: goto tr2395;
		case 13578: goto tr2399;
		case 13627: goto tr2399;
		case 13834: goto tr2399;
		case 13883: goto tr2399;
		case 14346: goto tr2404;
		case 14395: goto st800;
		case 14602: goto tr2399;
		case 14651: goto tr2399;
		case 14858: goto tr2406;
		case 14907: goto tr2407;
	}
	goto tr2412;
tr2407:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1144;
tr2452:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1144;
tr2533:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1144;
tr2493:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1144;
st1144:
	if ( ++p == pe )
		goto _test_eof1144;
case 1144:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(4736 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(4736 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(4736 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 s->number64 == 0  ) _widec += 512;
	}
	if ( _widec == 5642 )
		goto tr2404;
	if ( 5504 <= _widec && _widec <= 5759 )
		goto st800;
	goto tr2412;
tr2397:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st801;
tr2398:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st801;
tr2408:
	{
		s->line_counter++;
	}
	goto st801;
tr2522:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st801;
tr2523:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st801;
tr2524:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st801;
tr2534:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st801;
tr2441:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	goto st801;
tr2442:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st801;
tr2443:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st801;
tr2453:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	goto st801;
tr2482:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	goto st801;
tr2483:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st801;
tr2484:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st801;
tr2494:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	goto st801;
st801:
	if ( ++p == pe )
		goto _test_eof801;
case 801:
	_widec = (*p);
	if ( (*p) < 43 ) {
		if ( (*p) < 32 ) {
			if ( (*p) > 9 ) {
				if ( 10 <= (*p) && (*p) <= 10 ) {
					_widec = (short)(8832 + ((*p) - -128));
					if ( 
 s->multiline  ) _widec += 256;
					if ( 
 !s->multiline  ) _widec += 512;
					if ( 
 s->number64 != 0  ) _widec += 1024;
					if ( 
 s->number64 == 0  ) _widec += 2048;
				}
			} else if ( (*p) >= 9 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if ( 
 s->number64 != 0  ) _widec += 256;
				if ( 
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 32 ) {
			if ( (*p) > 40 ) {
				if ( 41 <= (*p) && (*p) <= 41 ) {
					_widec = (short)(5760 + ((*p) - -128));
					if ( 
 s->number64 != 0  ) _widec += 256;
					if ( 
 s->number64 == 0  ) _widec += 512;
				}
			} else if ( (*p) >= 40 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if ( 
 s->number64 != 0  ) _widec += 256;
				if ( 
 s->number64 == 0  ) _widec += 512;
			}
		} else {
			_widec = (short)(5760 + ((*p) - -128));
			if ( 
 s->number64 != 0  ) _widec += 256;
			if ( 
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 43 ) {
		if ( (*p) < 59 ) {
			if ( (*p) > 47 ) {
				if ( 48 <= (*p) && (*p) <= 57 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if ( 
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 47 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if ( 
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 59 ) {
			if ( (*p) > 90 ) {
				if ( 97 <= (*p) && (*p) <= 122 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if ( 
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 65 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if ( 
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(8832 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
			if ( 
 s->number64 != 0  ) _widec += 1024;
			if ( 
 s->number64 == 0  ) _widec += 2048;
		}
	} else {
		_widec = (short)(2688 + ((*p) - -128));
		if ( 
 s->number64 != 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 3115: goto tr2413;
		case 6153: goto st796;
		case 6176: goto st796;
		case 6184: goto tr2391;
		case 6185: goto tr2392;
		case 6409: goto st799;
		case 6432: goto st799;
		case 6440: goto tr2394;
		case 6441: goto tr2395;
		case 6665: goto st801;
		case 6688: goto st801;
		case 6696: goto tr2397;
		case 6697: goto tr2398;
		case 9482: goto tr2399;
		case 9531: goto tr2399;
		case 9738: goto tr2399;
		case 9787: goto tr2399;
		case 10250: goto tr2400;
		case 10299: goto st798;
		case 10506: goto tr2399;
		case 10555: goto tr2399;
		case 10762: goto tr2402;
		case 10811: goto tr2403;
		case 11274: goto tr2404;
		case 11323: goto st800;
		case 11530: goto tr2399;
		case 11579: goto tr2399;
		case 11786: goto tr2406;
		case 11835: goto tr2407;
		case 12298: goto tr2408;
		case 12347: goto st802;
		case 12554: goto tr2399;
		case 12603: goto tr2399;
		case 12810: goto tr2410;
		case 12859: goto tr2411;
	}
	if ( _widec < 3137 ) {
		if ( 3119 <= _widec && _widec <= 3129 )
			goto tr2413;
	} else if ( _widec > 3162 ) {
		if ( 3169 <= _widec && _widec <= 3194 )
			goto tr2413;
	} else
		goto tr2413;
	goto tr2412;
tr2402:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1145;
tr2447:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1145;
tr2488:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1145;
tr2528:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1145;
st1145:
	if ( ++p == pe )
		goto _test_eof1145;
case 1145:
	_widec = (*p);
	if ( (*p) < 43 ) {
		if ( (*p) < 32 ) {
			if ( (*p) > 9 ) {
				if ( 10 <= (*p) && (*p) <= 10 ) {
					_widec = (short)(3200 + ((*p) - -128));
					if ( 
 s->multiline  ) _widec += 256;
					if ( 
 s->number64 != 0  ) _widec += 512;
				}
			} else if ( (*p) >= 9 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if ( 
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 32 ) {
			if ( (*p) > 40 ) {
				if ( 41 <= (*p) && (*p) <= 41 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if ( 
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 40 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if ( 
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(2688 + ((*p) - -128));
			if ( 
 s->number64 != 0  ) _widec += 256;
		}
	} else if ( (*p) > 43 ) {
		if ( (*p) < 59 ) {
			if ( (*p) > 47 ) {
				if ( 48 <= (*p) && (*p) <= 57 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if ( 
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 47 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if ( 
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 59 ) {
			if ( (*p) > 90 ) {
				if ( 97 <= (*p) && (*p) <= 122 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if ( 
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 65 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if ( 
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(3200 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 s->number64 != 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(2688 + ((*p) - -128));
		if ( 
 s->number64 != 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 3081: goto st796;
		case 3104: goto st796;
		case 3112: goto tr2391;
		case 3113: goto tr2392;
		case 3115: goto tr2413;
		case 4106: goto tr2400;
		case 4155: goto st798;
	}
	if ( _widec < 3137 ) {
		if ( 3119 <= _widec && _widec <= 3129 )
			goto tr2413;
	} else if ( _widec > 3162 ) {
		if ( 3169 <= _widec && _widec <= 3194 )
			goto tr2413;
	} else
		goto tr2413;
	goto tr2412;
tr2403:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1146;
tr2448:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1146;
tr2529:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1146;
tr2489:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1146;
st1146:
	if ( ++p == pe )
		goto _test_eof1146;
case 1146:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(3200 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 s->number64 != 0  ) _widec += 512;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(3200 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 s->number64 != 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(3200 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 s->number64 != 0  ) _widec += 512;
	}
	if ( _widec == 4106 )
		goto tr2400;
	if ( 3968 <= _widec && _widec <= 4223 )
		goto st798;
	goto tr2412;
tr2535:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st802;
tr2454:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	goto st802;
tr2495:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	goto st802;
st802:
	if ( ++p == pe )
		goto _test_eof802;
case 802:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(6784 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 s->number64 != 0  ) _widec += 512;
			if ( 
 s->number64 == 0  ) _widec += 1024;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(6784 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 s->number64 != 0  ) _widec += 512;
			if ( 
 s->number64 == 0  ) _widec += 1024;
		}
	} else {
		_widec = (short)(6784 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 s->number64 != 0  ) _widec += 512;
		if ( 
 s->number64 == 0  ) _widec += 1024;
	}
	switch( _widec ) {
		case 7690: goto tr2400;
		case 8202: goto tr2404;
		case 8714: goto tr2408;
	}
	if ( _widec < 8064 ) {
		if ( 7552 <= _widec && _widec <= 7807 )
			goto st798;
	} else if ( _widec > 8319 ) {
		if ( 8576 <= _widec && _widec <= 8831 )
			goto st802;
	} else
		goto st800;
	goto tr2412;
tr2410:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1147;
tr2455:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1147;
tr2496:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1147;
tr2536:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1147;
st1147:
	if ( ++p == pe )
		goto _test_eof1147;
case 1147:
	_widec = (*p);
	if ( (*p) < 43 ) {
		if ( (*p) < 32 ) {
			if ( (*p) > 9 ) {
				if ( 10 <= (*p) && (*p) <= 10 ) {
					_widec = (short)(8832 + ((*p) - -128));
					if ( 
 s->multiline  ) _widec += 256;
					if ( 
 !s->multiline  ) _widec += 512;
					if ( 
 s->number64 != 0  ) _widec += 1024;
					if ( 
 s->number64 == 0  ) _widec += 2048;
				}
			} else if ( (*p) >= 9 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if ( 
 s->number64 != 0  ) _widec += 256;
				if ( 
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 32 ) {
			if ( (*p) > 40 ) {
				if ( 41 <= (*p) && (*p) <= 41 ) {
					_widec = (short)(5760 + ((*p) - -128));
					if ( 
 s->number64 != 0  ) _widec += 256;
					if ( 
 s->number64 == 0  ) _widec += 512;
				}
			} else if ( (*p) >= 40 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if ( 
 s->number64 != 0  ) _widec += 256;
				if ( 
 s->number64 == 0  ) _widec += 512;
			}
		} else {
			_widec = (short)(5760 + ((*p) - -128));
			if ( 
 s->number64 != 0  ) _widec += 256;
			if ( 
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 43 ) {
		if ( (*p) < 59 ) {
			if ( (*p) > 47 ) {
				if ( 48 <= (*p) && (*p) <= 57 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if ( 
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 47 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if ( 
 s->number64 != 0  ) _widec += 256;
			}
		} else if ( (*p) > 59 ) {
			if ( (*p) > 90 ) {
				if ( 97 <= (*p) && (*p) <= 122 ) {
					_widec = (short)(2688 + ((*p) - -128));
					if ( 
 s->number64 != 0  ) _widec += 256;
				}
			} else if ( (*p) >= 65 ) {
				_widec = (short)(2688 + ((*p) - -128));
				if ( 
 s->number64 != 0  ) _widec += 256;
			}
		} else {
			_widec = (short)(8832 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
			if ( 
 s->number64 != 0  ) _widec += 1024;
			if ( 
 s->number64 == 0  ) _widec += 2048;
		}
	} else {
		_widec = (short)(2688 + ((*p) - -128));
		if ( 
 s->number64 != 0  ) _widec += 256;
	}
	switch( _widec ) {
		case 3115: goto tr2413;
		case 6153: goto st796;
		case 6176: goto st796;
		case 6184: goto tr2391;
		case 6185: goto tr2392;
		case 6409: goto st799;
		case 6432: goto st799;
		case 6440: goto tr2394;
		case 6441: goto tr2395;
		case 6665: goto st801;
		case 6688: goto st801;
		case 6696: goto tr2397;
		case 6697: goto tr2398;
		case 9482: goto tr2399;
		case 9531: goto tr2399;
		case 9738: goto tr2399;
		case 9787: goto tr2399;
		case 10250: goto tr2400;
		case 10299: goto st798;
		case 10506: goto tr2399;
		case 10555: goto tr2399;
		case 10762: goto tr2402;
		case 10811: goto tr2403;
		case 11274: goto tr2404;
		case 11323: goto st800;
		case 11530: goto tr2399;
		case 11579: goto tr2399;
		case 11786: goto tr2406;
		case 11835: goto tr2407;
		case 12298: goto tr2408;
		case 12347: goto st802;
		case 12554: goto tr2399;
		case 12603: goto tr2399;
		case 12810: goto tr2410;
		case 12859: goto tr2411;
	}
	if ( _widec < 3137 ) {
		if ( 3119 <= _widec && _widec <= 3129 )
			goto tr2413;
	} else if ( _widec > 3162 ) {
		if ( 3169 <= _widec && _widec <= 3194 )
			goto tr2413;
	} else
		goto tr2413;
	goto tr2412;
tr2411:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr2456:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr2537:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
tr2497:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET6, (char *)s->buffer, &addr6) <= 0) {
			WARN(ZS_BAD_IPV6);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr6.s6_addr), INET6_ADDR_LENGTH);
		rdata_tail += INET6_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1148;
st1148:
	if ( ++p == pe )
		goto _test_eof1148;
case 1148:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(6784 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 s->number64 != 0  ) _widec += 512;
			if ( 
 s->number64 == 0  ) _widec += 1024;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(6784 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 s->number64 != 0  ) _widec += 512;
			if ( 
 s->number64 == 0  ) _widec += 1024;
		}
	} else {
		_widec = (short)(6784 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 s->number64 != 0  ) _widec += 512;
		if ( 
 s->number64 == 0  ) _widec += 1024;
	}
	switch( _widec ) {
		case 7690: goto tr2400;
		case 8202: goto tr2404;
		case 8714: goto tr2408;
	}
	if ( _widec < 8064 ) {
		if ( 7552 <= _widec && _widec <= 7807 )
			goto st798;
	} else if ( _widec > 8319 ) {
		if ( 8576 <= _widec && _widec <= 8831 )
			goto st802;
	} else
		goto st800;
	goto tr2412;
tr2382:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st803;
st803:
	if ( ++p == pe )
		goto _test_eof803;
case 803:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2387;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st803;
	goto tr2360;
st804:
	if ( ++p == pe )
		goto _test_eof804;
case 804:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2373;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st804;
	goto tr2360;
tr2365:
	{
		*(rdata_tail++) = 1;
	}
	goto st805;
st805:
	if ( ++p == pe )
		goto _test_eof805;
case 805:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st806;
		case 32: goto st806;
		case 40: goto tr2415;
		case 41: goto tr2416;
		case 1034: goto tr2417;
		case 1083: goto st811;
	}
	goto tr2360;
tr2415:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st806;
tr2416:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st806;
tr2417:
	{
		s->line_counter++;
	}
	goto st806;
st806:
	if ( ++p == pe )
		goto _test_eof806;
case 806:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st806;
		case 32: goto st806;
		case 40: goto tr2415;
		case 41: goto tr2416;
		case 1034: goto tr2417;
		case 1083: goto st811;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2419;
	goto tr2375;
tr2419:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st807;
tr2423:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st807;
st807:
	if ( ++p == pe )
		goto _test_eof807;
case 807:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2420;
		case 32: goto tr2420;
		case 40: goto tr2421;
		case 41: goto tr2422;
		case 1034: goto tr2424;
		case 1083: goto tr2425;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2423;
	goto tr2375;
tr2428:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st808;
tr2429:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st808;
tr2431:
	{
		s->line_counter++;
	}
	goto st808;
tr2420:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st808;
tr2421:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st808;
tr2422:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st808;
tr2424:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st808;
st808:
	if ( ++p == pe )
		goto _test_eof808;
case 808:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st808;
		case 32: goto st808;
		case 40: goto tr2428;
		case 41: goto tr2429;
		case 46: goto tr2430;
		case 1034: goto tr2431;
		case 1083: goto st810;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2430;
	goto tr2426;
tr2430:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st809;
tr2434:
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st809;
st809:
	if ( ++p == pe )
		goto _test_eof809;
case 809:
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
				if ( 
 !s->multiline  ) _widec += 512;
				if ( 
 s->number64 != 0  ) _widec += 1024;
				if ( 
 s->number64 == 0  ) _widec += 2048;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(5760 + ((*p) - -128));
			if ( 
 s->number64 != 0  ) _widec += 256;
			if ( 
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if ( 
 s->number64 != 0  ) _widec += 256;
				if ( 
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
				if ( 
 !s->multiline  ) _widec += 512;
				if ( 
 s->number64 != 0  ) _widec += 1024;
				if ( 
 s->number64 == 0  ) _widec += 2048;
			}
		} else {
			_widec = (short)(5760 + ((*p) - -128));
			if ( 
 s->number64 != 0  ) _widec += 256;
			if ( 
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(5760 + ((*p) - -128));
		if ( 
 s->number64 != 0  ) _widec += 256;
		if ( 
 s->number64 == 0  ) _widec += 512;
	}
	switch( _widec ) {
		case 46: goto tr2434;
		case 6153: goto tr2435;
		case 6176: goto tr2435;
		case 6184: goto tr2436;
		case 6185: goto tr2437;
		case 6409: goto tr2438;
		case 6432: goto tr2438;
		case 6440: goto tr2439;
		case 6441: goto tr2440;
		case 6665: goto tr2441;
		case 6688: goto tr2441;
		case 6696: goto tr2442;
		case 6697: goto tr2443;
		case 9482: goto tr2444;
		case 9531: goto tr2444;
		case 9738: goto tr2444;
		case 9787: goto tr2444;
		case 10250: goto tr2445;
		case 10299: goto tr2446;
		case 10506: goto tr2444;
		case 10555: goto tr2444;
		case 10762: goto tr2447;
		case 10811: goto tr2448;
		case 11274: goto tr2449;
		case 11323: goto tr2450;
		case 11530: goto tr2444;
		case 11579: goto tr2444;
		case 11786: goto tr2451;
		case 11835: goto tr2452;
		case 12298: goto tr2453;
		case 12347: goto tr2454;
		case 12554: goto tr2444;
		case 12603: goto tr2444;
		case 12810: goto tr2455;
		case 12859: goto tr2456;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2434;
	goto tr2433;
tr2425:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st810;
st810:
	if ( ++p == pe )
		goto _test_eof810;
case 810:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2431;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st810;
	goto tr2360;
st811:
	if ( ++p == pe )
		goto _test_eof811;
case 811:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2417;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st811;
	goto tr2360;
tr2366:
	{
		*(rdata_tail++) = 2;
	}
	goto st812;
st812:
	if ( ++p == pe )
		goto _test_eof812;
case 812:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st813;
		case 32: goto st813;
		case 40: goto tr2458;
		case 41: goto tr2459;
		case 1034: goto tr2460;
		case 1083: goto st818;
	}
	goto tr2360;
tr2458:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st813;
tr2459:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st813;
tr2460:
	{
		s->line_counter++;
	}
	goto st813;
st813:
	if ( ++p == pe )
		goto _test_eof813;
case 813:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st813;
		case 32: goto st813;
		case 40: goto tr2458;
		case 41: goto tr2459;
		case 1034: goto tr2460;
		case 1083: goto st818;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2462;
	goto tr2375;
tr2462:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st814;
tr2466:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st814;
st814:
	if ( ++p == pe )
		goto _test_eof814;
case 814:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2463;
		case 32: goto tr2463;
		case 40: goto tr2464;
		case 41: goto tr2465;
		case 1034: goto tr2467;
		case 1083: goto tr2468;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2466;
	goto tr2375;
tr2470:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st815;
tr2471:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st815;
tr2473:
	{
		s->line_counter++;
	}
	goto st815;
tr2463:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st815;
tr2464:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st815;
tr2465:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st815;
tr2467:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st815;
st815:
	if ( ++p == pe )
		goto _test_eof815;
case 815:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st815;
		case 32: goto st815;
		case 40: goto tr2470;
		case 41: goto tr2471;
		case 46: goto tr2472;
		case 1034: goto tr2473;
		case 1083: goto st817;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 58 )
			goto tr2472;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2472;
	} else
		goto tr2472;
	goto tr2426;
tr2472:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st816;
tr2475:
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st816;
st816:
	if ( ++p == pe )
		goto _test_eof816;
case 816:
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
				if ( 
 !s->multiline  ) _widec += 512;
				if ( 
 s->number64 != 0  ) _widec += 1024;
				if ( 
 s->number64 == 0  ) _widec += 2048;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(5760 + ((*p) - -128));
			if ( 
 s->number64 != 0  ) _widec += 256;
			if ( 
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if ( 
 s->number64 != 0  ) _widec += 256;
				if ( 
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
				if ( 
 !s->multiline  ) _widec += 512;
				if ( 
 s->number64 != 0  ) _widec += 1024;
				if ( 
 s->number64 == 0  ) _widec += 2048;
			}
		} else {
			_widec = (short)(5760 + ((*p) - -128));
			if ( 
 s->number64 != 0  ) _widec += 256;
			if ( 
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(5760 + ((*p) - -128));
		if ( 
 s->number64 != 0  ) _widec += 256;
		if ( 
 s->number64 == 0  ) _widec += 512;
	}
	switch( _widec ) {
		case 46: goto tr2475;
		case 6153: goto tr2476;
		case 6176: goto tr2476;
		case 6184: goto tr2477;
		case 6185: goto tr2478;
		case 6409: goto tr2479;
		case 6432: goto tr2479;
		case 6440: goto tr2480;
		case 6441: goto tr2481;
		case 6665: goto tr2482;
		case 6688: goto tr2482;
		case 6696: goto tr2483;
		case 6697: goto tr2484;
		case 9482: goto tr2485;
		case 9531: goto tr2485;
		case 9738: goto tr2485;
		case 9787: goto tr2485;
		case 10250: goto tr2486;
		case 10299: goto tr2487;
		case 10506: goto tr2485;
		case 10555: goto tr2485;
		case 10762: goto tr2488;
		case 10811: goto tr2489;
		case 11274: goto tr2490;
		case 11323: goto tr2491;
		case 11530: goto tr2485;
		case 11579: goto tr2485;
		case 11786: goto tr2492;
		case 11835: goto tr2493;
		case 12298: goto tr2494;
		case 12347: goto tr2495;
		case 12554: goto tr2485;
		case 12603: goto tr2485;
		case 12810: goto tr2496;
		case 12859: goto tr2497;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 58 )
			goto tr2475;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr2475;
	} else
		goto tr2475;
	goto tr2433;
tr2468:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st817;
st817:
	if ( ++p == pe )
		goto _test_eof817;
case 817:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2473;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st817;
	goto tr2360;
st818:
	if ( ++p == pe )
		goto _test_eof818;
case 818:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2460;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st818;
	goto tr2360;
tr2367:
	{
		*(rdata_tail++) = 3;
	}
	goto st819;
st819:
	if ( ++p == pe )
		goto _test_eof819;
case 819:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st820;
		case 32: goto st820;
		case 40: goto tr2499;
		case 41: goto tr2500;
		case 1034: goto tr2501;
		case 1083: goto st825;
	}
	goto tr2360;
tr2499:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st820;
tr2500:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st820;
tr2501:
	{
		s->line_counter++;
	}
	goto st820;
st820:
	if ( ++p == pe )
		goto _test_eof820;
case 820:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st820;
		case 32: goto st820;
		case 40: goto tr2499;
		case 41: goto tr2500;
		case 1034: goto tr2501;
		case 1083: goto st825;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2503;
	goto tr2375;
tr2503:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st821;
tr2507:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st821;
st821:
	if ( ++p == pe )
		goto _test_eof821;
case 821:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2504;
		case 32: goto tr2504;
		case 40: goto tr2505;
		case 41: goto tr2506;
		case 1034: goto tr2508;
		case 1083: goto tr2509;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2507;
	goto tr2375;
tr2511:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st822;
tr2512:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st822;
tr2514:
	{
		s->line_counter++;
	}
	goto st822;
tr2504:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st822;
tr2505:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st822;
tr2506:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st822;
tr2508:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st822;
st822:
	if ( ++p == pe )
		goto _test_eof822;
case 822:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st822;
		case 32: goto st822;
		case 40: goto tr2511;
		case 41: goto tr2512;
		case 42: goto tr2513;
		case 92: goto tr2513;
		case 95: goto tr2513;
		case 1034: goto tr2514;
		case 1083: goto st824;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr2513;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2513;
	} else
		goto tr2513;
	goto tr2360;
tr2513:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 823; goto st255;} }
	goto st823;
st823:
	if ( ++p == pe )
		goto _test_eof823;
case 823:
	_widec = (*p);
	if ( (*p) < 32 ) {
		if ( (*p) > 9 ) {
			if ( 10 <= (*p) && (*p) <= 10 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
				if ( 
 !s->multiline  ) _widec += 512;
				if ( 
 s->number64 != 0  ) _widec += 1024;
				if ( 
 s->number64 == 0  ) _widec += 2048;
			}
		} else if ( (*p) >= 9 ) {
			_widec = (short)(5760 + ((*p) - -128));
			if ( 
 s->number64 != 0  ) _widec += 256;
			if ( 
 s->number64 == 0  ) _widec += 512;
		}
	} else if ( (*p) > 32 ) {
		if ( (*p) < 41 ) {
			if ( 40 <= (*p) && (*p) <= 40 ) {
				_widec = (short)(5760 + ((*p) - -128));
				if ( 
 s->number64 != 0  ) _widec += 256;
				if ( 
 s->number64 == 0  ) _widec += 512;
			}
		} else if ( (*p) > 41 ) {
			if ( 59 <= (*p) && (*p) <= 59 ) {
				_widec = (short)(8832 + ((*p) - -128));
				if ( 
 s->multiline  ) _widec += 256;
				if ( 
 !s->multiline  ) _widec += 512;
				if ( 
 s->number64 != 0  ) _widec += 1024;
				if ( 
 s->number64 == 0  ) _widec += 2048;
			}
		} else {
			_widec = (short)(5760 + ((*p) - -128));
			if ( 
 s->number64 != 0  ) _widec += 256;
			if ( 
 s->number64 == 0  ) _widec += 512;
		}
	} else {
		_widec = (short)(5760 + ((*p) - -128));
		if ( 
 s->number64 != 0  ) _widec += 256;
		if ( 
 s->number64 == 0  ) _widec += 512;
	}
	switch( _widec ) {
		case 6153: goto tr2516;
		case 6176: goto tr2516;
		case 6184: goto tr2517;
		case 6185: goto tr2518;
		case 6409: goto tr2519;
		case 6432: goto tr2519;
		case 6440: goto tr2520;
		case 6441: goto tr2521;
		case 6665: goto tr2522;
		case 6688: goto tr2522;
		case 6696: goto tr2523;
		case 6697: goto tr2524;
		case 9482: goto tr2525;
		case 9531: goto tr2525;
		case 9738: goto tr2525;
		case 9787: goto tr2525;
		case 10250: goto tr2526;
		case 10299: goto tr2527;
		case 10506: goto tr2525;
		case 10555: goto tr2525;
		case 10762: goto tr2528;
		case 10811: goto tr2529;
		case 11274: goto tr2530;
		case 11323: goto tr2531;
		case 11530: goto tr2525;
		case 11579: goto tr2525;
		case 11786: goto tr2532;
		case 11835: goto tr2533;
		case 12298: goto tr2534;
		case 12347: goto tr2535;
		case 12554: goto tr2525;
		case 12603: goto tr2525;
		case 12810: goto tr2536;
		case 12859: goto tr2537;
	}
	goto tr2389;
tr2509:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st824;
st824:
	if ( ++p == pe )
		goto _test_eof824;
case 824:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2514;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st824;
	goto tr2360;
st825:
	if ( ++p == pe )
		goto _test_eof825;
case 825:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2501;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st825;
	goto tr2360;
tr2359:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st826;
st826:
	if ( ++p == pe )
		goto _test_eof826;
case 826:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2368;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st826;
	goto tr69;
st827:
	if ( ++p == pe )
		goto _test_eof827;
case 827:
	switch( (*p) ) {
		case 65: goto st828;
		case 67: goto st862;
		case 68: goto st877;
		case 69: goto st891;
		case 72: goto st898;
		case 73: goto st903;
		case 75: goto st911;
		case 76: goto st915;
		case 77: goto st923;
		case 78: goto st929;
		case 80: goto st945;
		case 82: goto st948;
		case 83: goto st955;
		case 84: goto st966;
		case 97: goto st828;
		case 99: goto st862;
		case 100: goto st877;
		case 101: goto st891;
		case 104: goto st898;
		case 105: goto st903;
		case 107: goto st911;
		case 108: goto st915;
		case 109: goto st923;
		case 110: goto st929;
		case 112: goto st945;
		case 114: goto st948;
		case 115: goto st955;
		case 116: goto st966;
	}
	goto tr2538;
st828:
	if ( ++p == pe )
		goto _test_eof828;
case 828:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2553;
		case 32: goto tr2553;
		case 40: goto tr2554;
		case 41: goto tr2555;
		case 65: goto st853;
		case 70: goto st856;
		case 80: goto st860;
		case 97: goto st853;
		case 102: goto st856;
		case 112: goto st860;
		case 1034: goto tr2559;
		case 1083: goto tr2560;
	}
	goto tr2538;
tr2562:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2563:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2565:
	{
		s->line_counter++;
	}
	goto st829;
tr2949:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st829;
tr2950:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2951:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2953:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st829;
tr2553:
	{ type_num(KNOT_RRTYPE_A, &rdata_tail); }
	goto st829;
tr2554:
	{ type_num(KNOT_RRTYPE_A, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2555:
	{ type_num(KNOT_RRTYPE_A, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2559:
	{ type_num(KNOT_RRTYPE_A, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2648:
	{ type_num(KNOT_RRTYPE_AAAA, &rdata_tail); }
	goto st829;
tr2649:
	{ type_num(KNOT_RRTYPE_AAAA, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2650:
	{ type_num(KNOT_RRTYPE_AAAA, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2651:
	{ type_num(KNOT_RRTYPE_AAAA, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2656:
	{ type_num(KNOT_RRTYPE_AFSDB, &rdata_tail); }
	goto st829;
tr2657:
	{ type_num(KNOT_RRTYPE_AFSDB, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2658:
	{ type_num(KNOT_RRTYPE_AFSDB, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2659:
	{ type_num(KNOT_RRTYPE_AFSDB, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2662:
	{ type_num(KNOT_RRTYPE_APL, &rdata_tail); }
	goto st829;
tr2663:
	{ type_num(KNOT_RRTYPE_APL, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2664:
	{ type_num(KNOT_RRTYPE_APL, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2665:
	{ type_num(KNOT_RRTYPE_APL, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2676:
	{ type_num(KNOT_RRTYPE_CDNSKEY, &rdata_tail); }
	goto st829;
tr2677:
	{ type_num(KNOT_RRTYPE_CDNSKEY, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2678:
	{ type_num(KNOT_RRTYPE_CDNSKEY, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2679:
	{ type_num(KNOT_RRTYPE_CDNSKEY, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2681:
	{ type_num(KNOT_RRTYPE_CDS, &rdata_tail); }
	goto st829;
tr2682:
	{ type_num(KNOT_RRTYPE_CDS, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2683:
	{ type_num(KNOT_RRTYPE_CDS, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2684:
	{ type_num(KNOT_RRTYPE_CDS, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2688:
	{ type_num(KNOT_RRTYPE_CERT, &rdata_tail); }
	goto st829;
tr2689:
	{ type_num(KNOT_RRTYPE_CERT, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2690:
	{ type_num(KNOT_RRTYPE_CERT, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2691:
	{ type_num(KNOT_RRTYPE_CERT, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2696:
	{ type_num(KNOT_RRTYPE_CNAME, &rdata_tail); }
	goto st829;
tr2697:
	{ type_num(KNOT_RRTYPE_CNAME, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2698:
	{ type_num(KNOT_RRTYPE_CNAME, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2699:
	{ type_num(KNOT_RRTYPE_CNAME, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2707:
	{ type_num(KNOT_RRTYPE_DHCID, &rdata_tail); }
	goto st829;
tr2708:
	{ type_num(KNOT_RRTYPE_DHCID, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2709:
	{ type_num(KNOT_RRTYPE_DHCID, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2710:
	{ type_num(KNOT_RRTYPE_DHCID, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2716:
	{ type_num(KNOT_RRTYPE_DNAME, &rdata_tail); }
	goto st829;
tr2717:
	{ type_num(KNOT_RRTYPE_DNAME, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2718:
	{ type_num(KNOT_RRTYPE_DNAME, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2719:
	{ type_num(KNOT_RRTYPE_DNAME, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2724:
	{ type_num(KNOT_RRTYPE_DNSKEY, &rdata_tail); }
	goto st829;
tr2725:
	{ type_num(KNOT_RRTYPE_DNSKEY, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2726:
	{ type_num(KNOT_RRTYPE_DNSKEY, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2727:
	{ type_num(KNOT_RRTYPE_DNSKEY, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2729:
	{ type_num(KNOT_RRTYPE_DS, &rdata_tail); }
	goto st829;
tr2730:
	{ type_num(KNOT_RRTYPE_DS, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2731:
	{ type_num(KNOT_RRTYPE_DS, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2732:
	{ type_num(KNOT_RRTYPE_DS, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2739:
	{ type_num(KNOT_RRTYPE_EUI48, &rdata_tail); }
	goto st829;
tr2740:
	{ type_num(KNOT_RRTYPE_EUI48, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2741:
	{ type_num(KNOT_RRTYPE_EUI48, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2742:
	{ type_num(KNOT_RRTYPE_EUI48, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2745:
	{ type_num(KNOT_RRTYPE_EUI64, &rdata_tail); }
	goto st829;
tr2746:
	{ type_num(KNOT_RRTYPE_EUI64, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2747:
	{ type_num(KNOT_RRTYPE_EUI64, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2748:
	{ type_num(KNOT_RRTYPE_EUI64, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2754:
	{ type_num(KNOT_RRTYPE_HINFO, &rdata_tail); }
	goto st829;
tr2755:
	{ type_num(KNOT_RRTYPE_HINFO, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2756:
	{ type_num(KNOT_RRTYPE_HINFO, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2757:
	{ type_num(KNOT_RRTYPE_HINFO, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2766:
	{ type_num(KNOT_RRTYPE_IPSECKEY, &rdata_tail); }
	goto st829;
tr2767:
	{ type_num(KNOT_RRTYPE_IPSECKEY, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2768:
	{ type_num(KNOT_RRTYPE_IPSECKEY, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2769:
	{ type_num(KNOT_RRTYPE_IPSECKEY, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2774:
	{ type_num(KNOT_RRTYPE_KEY, &rdata_tail); }
	goto st829;
tr2775:
	{ type_num(KNOT_RRTYPE_KEY, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2776:
	{ type_num(KNOT_RRTYPE_KEY, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2777:
	{ type_num(KNOT_RRTYPE_KEY, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2779:
	{ type_num(KNOT_RRTYPE_KX, &rdata_tail); }
	goto st829;
tr2780:
	{ type_num(KNOT_RRTYPE_KX, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2781:
	{ type_num(KNOT_RRTYPE_KX, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2782:
	{ type_num(KNOT_RRTYPE_KX, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2789:
	{ type_num(KNOT_RRTYPE_L32, &rdata_tail); }
	goto st829;
tr2790:
	{ type_num(KNOT_RRTYPE_L32, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2791:
	{ type_num(KNOT_RRTYPE_L32, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2792:
	{ type_num(KNOT_RRTYPE_L32, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2795:
	{ type_num(KNOT_RRTYPE_L64, &rdata_tail); }
	goto st829;
tr2796:
	{ type_num(KNOT_RRTYPE_L64, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2797:
	{ type_num(KNOT_RRTYPE_L64, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2798:
	{ type_num(KNOT_RRTYPE_L64, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2801:
	{ type_num(KNOT_RRTYPE_LOC, &rdata_tail); }
	goto st829;
tr2802:
	{ type_num(KNOT_RRTYPE_LOC, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2803:
	{ type_num(KNOT_RRTYPE_LOC, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2804:
	{ type_num(KNOT_RRTYPE_LOC, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2806:
	{ type_num(KNOT_RRTYPE_LP, &rdata_tail); }
	goto st829;
tr2807:
	{ type_num(KNOT_RRTYPE_LP, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2808:
	{ type_num(KNOT_RRTYPE_LP, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2809:
	{ type_num(KNOT_RRTYPE_LP, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2816:
	{ type_num(KNOT_RRTYPE_MINFO, &rdata_tail); }
	goto st829;
tr2817:
	{ type_num(KNOT_RRTYPE_MINFO, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2818:
	{ type_num(KNOT_RRTYPE_MINFO, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2819:
	{ type_num(KNOT_RRTYPE_MINFO, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2821:
	{ type_num(KNOT_RRTYPE_MX, &rdata_tail); }
	goto st829;
tr2822:
	{ type_num(KNOT_RRTYPE_MX, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2823:
	{ type_num(KNOT_RRTYPE_MX, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2824:
	{ type_num(KNOT_RRTYPE_MX, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2832:
	{ type_num(KNOT_RRTYPE_NAPTR, &rdata_tail); }
	goto st829;
tr2833:
	{ type_num(KNOT_RRTYPE_NAPTR, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2834:
	{ type_num(KNOT_RRTYPE_NAPTR, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2835:
	{ type_num(KNOT_RRTYPE_NAPTR, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2838:
	{ type_num(KNOT_RRTYPE_NID, &rdata_tail); }
	goto st829;
tr2839:
	{ type_num(KNOT_RRTYPE_NID, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2840:
	{ type_num(KNOT_RRTYPE_NID, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2841:
	{ type_num(KNOT_RRTYPE_NID, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2843:
	{ type_num(KNOT_RRTYPE_NS, &rdata_tail); }
	goto st829;
tr2844:
	{ type_num(KNOT_RRTYPE_NS, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2845:
	{ type_num(KNOT_RRTYPE_NS, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2847:
	{ type_num(KNOT_RRTYPE_NS, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2850:
	{ type_num(KNOT_RRTYPE_NSEC, &rdata_tail); }
	goto st829;
tr2851:
	{ type_num(KNOT_RRTYPE_NSEC, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2852:
	{ type_num(KNOT_RRTYPE_NSEC, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2854:
	{ type_num(KNOT_RRTYPE_NSEC, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2856:
	{ type_num(KNOT_RRTYPE_NSEC3, &rdata_tail); }
	goto st829;
tr2857:
	{ type_num(KNOT_RRTYPE_NSEC3, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2858:
	{ type_num(KNOT_RRTYPE_NSEC3, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2860:
	{ type_num(KNOT_RRTYPE_NSEC3, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2866:
	{ type_num(KNOT_RRTYPE_NSEC3PARAM, &rdata_tail); }
	goto st829;
tr2867:
	{ type_num(KNOT_RRTYPE_NSEC3PARAM, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2868:
	{ type_num(KNOT_RRTYPE_NSEC3PARAM, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2869:
	{ type_num(KNOT_RRTYPE_NSEC3PARAM, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2873:
	{ type_num(KNOT_RRTYPE_PTR, &rdata_tail); }
	goto st829;
tr2874:
	{ type_num(KNOT_RRTYPE_PTR, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2875:
	{ type_num(KNOT_RRTYPE_PTR, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2876:
	{ type_num(KNOT_RRTYPE_PTR, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2881:
	{ type_num(KNOT_RRTYPE_RP, &rdata_tail); }
	goto st829;
tr2882:
	{ type_num(KNOT_RRTYPE_RP, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2883:
	{ type_num(KNOT_RRTYPE_RP, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2884:
	{ type_num(KNOT_RRTYPE_RP, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2889:
	{ type_num(KNOT_RRTYPE_RRSIG, &rdata_tail); }
	goto st829;
tr2890:
	{ type_num(KNOT_RRTYPE_RRSIG, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2891:
	{ type_num(KNOT_RRTYPE_RRSIG, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2892:
	{ type_num(KNOT_RRTYPE_RRSIG, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2894:
	{ type_num(KNOT_RRTYPE_RT, &rdata_tail); }
	goto st829;
tr2895:
	{ type_num(KNOT_RRTYPE_RT, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2896:
	{ type_num(KNOT_RRTYPE_RT, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2897:
	{ type_num(KNOT_RRTYPE_RT, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2904:
	{ type_num(KNOT_RRTYPE_SOA, &rdata_tail); }
	goto st829;
tr2905:
	{ type_num(KNOT_RRTYPE_SOA, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2906:
	{ type_num(KNOT_RRTYPE_SOA, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2907:
	{ type_num(KNOT_RRTYPE_SOA, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2910:
	{ type_num(KNOT_RRTYPE_SPF, &rdata_tail); }
	goto st829;
tr2911:
	{ type_num(KNOT_RRTYPE_SPF, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2912:
	{ type_num(KNOT_RRTYPE_SPF, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2913:
	{ type_num(KNOT_RRTYPE_SPF, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2916:
	{ type_num(KNOT_RRTYPE_SRV, &rdata_tail); }
	goto st829;
tr2917:
	{ type_num(KNOT_RRTYPE_SRV, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2918:
	{ type_num(KNOT_RRTYPE_SRV, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2919:
	{ type_num(KNOT_RRTYPE_SRV, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2924:
	{ type_num(KNOT_RRTYPE_SSHFP, &rdata_tail); }
	goto st829;
tr2925:
	{ type_num(KNOT_RRTYPE_SSHFP, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2926:
	{ type_num(KNOT_RRTYPE_SSHFP, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2927:
	{ type_num(KNOT_RRTYPE_SSHFP, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2934:
	{ type_num(KNOT_RRTYPE_TLSA, &rdata_tail); }
	goto st829;
tr2935:
	{ type_num(KNOT_RRTYPE_TLSA, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2936:
	{ type_num(KNOT_RRTYPE_TLSA, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2937:
	{ type_num(KNOT_RRTYPE_TLSA, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
tr2940:
	{ type_num(KNOT_RRTYPE_TXT, &rdata_tail); }
	goto st829;
tr2941:
	{ type_num(KNOT_RRTYPE_TXT, &rdata_tail); }
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st829;
tr2942:
	{ type_num(KNOT_RRTYPE_TXT, &rdata_tail); }
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st829;
tr2943:
	{ type_num(KNOT_RRTYPE_TXT, &rdata_tail); }
	{
		s->line_counter++;
	}
	goto st829;
st829:
	if ( ++p == pe )
		goto _test_eof829;
case 829:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st829;
		case 32: goto st829;
		case 40: goto tr2562;
		case 41: goto tr2563;
		case 1034: goto tr2565;
		case 1083: goto st852;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2564;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2564;
	} else
		goto tr2564;
	goto tr69;
tr2564:
	{ p--; {stack[top++] = 830; goto st467;} }
	goto st830;
st830:
	if ( ++p == pe )
		goto _test_eof830;
case 830:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st831;
		case 32: goto st831;
		case 40: goto tr2568;
		case 41: goto tr2569;
		case 1034: goto tr2570;
		case 1083: goto st851;
	}
	goto tr69;
tr2568:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st831;
tr2569:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st831;
tr2570:
	{
		s->line_counter++;
	}
	goto st831;
st831:
	if ( ++p == pe )
		goto _test_eof831;
case 831:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st831;
		case 32: goto st831;
		case 40: goto tr2568;
		case 41: goto tr2569;
		case 1034: goto tr2570;
		case 1083: goto st851;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2572;
	goto tr1700;
tr2572:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st832;
tr2576:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st832;
st832:
	if ( ++p == pe )
		goto _test_eof832;
case 832:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2573;
		case 32: goto tr2573;
		case 40: goto tr2574;
		case 41: goto tr2575;
		case 1034: goto tr2577;
		case 1083: goto tr2578;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2576;
	goto tr1700;
tr2580:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st833;
tr2581:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st833;
tr2583:
	{
		s->line_counter++;
	}
	goto st833;
tr2573:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st833;
tr2574:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st833;
tr2575:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st833;
tr2577:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st833;
st833:
	if ( ++p == pe )
		goto _test_eof833;
case 833:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st833;
		case 32: goto st833;
		case 40: goto tr2580;
		case 41: goto tr2581;
		case 1034: goto tr2583;
		case 1083: goto st850;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2582;
	goto tr1700;
tr2582:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st834;
tr2588:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st834;
st834:
	if ( ++p == pe )
		goto _test_eof834;
case 834:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2585;
		case 32: goto tr2585;
		case 40: goto tr2586;
		case 41: goto tr2587;
		case 1034: goto tr2589;
		case 1083: goto tr2590;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2588;
	goto tr1700;
tr2593:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st835;
tr2594:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st835;
tr2596:
	{
		s->line_counter++;
	}
	goto st835;
tr2585:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st835;
tr2586:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st835;
tr2587:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st835;
tr2589:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st835;
st835:
	if ( ++p == pe )
		goto _test_eof835;
case 835:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st835;
		case 32: goto st835;
		case 40: goto tr2593;
		case 41: goto tr2594;
		case 1034: goto tr2596;
		case 1083: goto st849;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2595;
	goto tr2591;
tr2595:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st836;
tr2601:
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st836;
st836:
	if ( ++p == pe )
		goto _test_eof836;
case 836:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2598;
		case 32: goto tr2598;
		case 40: goto tr2599;
		case 41: goto tr2600;
		case 1034: goto tr2602;
		case 1083: goto tr2603;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2601;
	goto tr2591;
tr2605:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st837;
tr2606:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st837;
tr2608:
	{
		s->line_counter++;
	}
	goto st837;
tr2598:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st253;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st253;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st253;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st253;}
		}
	}
	goto st837;
tr2599:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st253;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st253;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st253;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st837;
tr2600:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st253;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st253;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st253;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st837;
tr2602:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st253;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st253;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st253;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st837;
st837:
	if ( ++p == pe )
		goto _test_eof837;
case 837:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st837;
		case 32: goto st837;
		case 40: goto tr2605;
		case 41: goto tr2606;
		case 1034: goto tr2608;
		case 1083: goto st848;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2607;
	goto tr2591;
tr2607:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st838;
tr2613:
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st838;
st838:
	if ( ++p == pe )
		goto _test_eof838;
case 838:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2610;
		case 32: goto tr2610;
		case 40: goto tr2611;
		case 41: goto tr2612;
		case 1034: goto tr2614;
		case 1083: goto tr2615;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2613;
	goto tr2591;
tr2617:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st839;
tr2618:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st839;
tr2620:
	{
		s->line_counter++;
	}
	goto st839;
tr2610:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st253;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st253;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st253;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st253;}
		}
	}
	goto st839;
tr2611:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st253;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st253;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st253;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st839;
tr2612:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st253;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st253;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st253;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st839;
tr2614:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st253;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st253;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st253;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st839;
st839:
	if ( ++p == pe )
		goto _test_eof839;
case 839:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st839;
		case 32: goto st839;
		case 40: goto tr2617;
		case 41: goto tr2618;
		case 1034: goto tr2620;
		case 1083: goto st847;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2619;
	goto tr1700;
tr2619:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st840;
tr2625:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st840;
st840:
	if ( ++p == pe )
		goto _test_eof840;
case 840:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2622;
		case 32: goto tr2622;
		case 40: goto tr2623;
		case 41: goto tr2624;
		case 1034: goto tr2626;
		case 1083: goto tr2627;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2625;
	goto tr1700;
tr2629:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st841;
tr2630:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st841;
tr2632:
	{
		s->line_counter++;
	}
	goto st841;
tr2622:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st841;
tr2623:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st841;
tr2624:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st841;
tr2626:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st841;
st841:
	if ( ++p == pe )
		goto _test_eof841;
case 841:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st841;
		case 32: goto st841;
		case 40: goto tr2629;
		case 41: goto tr2630;
		case 42: goto tr2631;
		case 92: goto tr2631;
		case 95: goto tr2631;
		case 1034: goto tr2632;
		case 1083: goto st846;
	}
	if ( _widec < 64 ) {
		if ( 45 <= _widec && _widec <= 57 )
			goto tr2631;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2631;
	} else
		goto tr2631;
	goto tr69;
tr2631:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 842; goto st255;} }
	goto st842;
st842:
	if ( ++p == pe )
		goto _test_eof842;
case 842:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2634;
		case 32: goto tr2634;
		case 40: goto tr2635;
		case 41: goto tr2636;
		case 1034: goto tr2637;
		case 1083: goto tr2638;
	}
	goto tr69;
tr2640:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st843;
tr2641:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st843;
tr2643:
	{
		s->line_counter++;
	}
	goto st843;
tr2634:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st843;
tr2635:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st843;
tr2636:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st843;
tr2637:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{
		s->line_counter++;
	}
	goto st843;
st843:
	if ( ++p == pe )
		goto _test_eof843;
case 843:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st843;
		case 32: goto st843;
		case 40: goto tr2640;
		case 41: goto tr2641;
		case 43: goto tr2642;
		case 1034: goto tr2643;
		case 1083: goto st845;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr2642;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2642;
	} else
		goto tr2642;
	goto tr69;
tr2642:
	{ p--; {stack[top++] = 844; goto st314;} }
	goto st844;
st844:
	if ( ++p == pe )
		goto _test_eof844;
case 844:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1152 + ((*p) - -128));
			if ( 
 !s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1152 + ((*p) - -128));
		if ( 
 !s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 1546: goto tr2645;
		case 1595: goto tr2645;
	}
	goto tr69;
tr2645:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1149;
st1149:
	if ( ++p == pe )
		goto _test_eof1149;
case 1149:
	goto st0;
tr2638:
	{
		rdata_tail += s->dname_tmp_length;
	}
	goto st845;
st845:
	if ( ++p == pe )
		goto _test_eof845;
case 845:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2643;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st845;
	goto tr69;
tr2627:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st846;
st846:
	if ( ++p == pe )
		goto _test_eof846;
case 846:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2632;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st846;
	goto tr69;
tr2615:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st253;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st253;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st253;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st253;}
		}
	}
	goto st847;
st847:
	if ( ++p == pe )
		goto _test_eof847;
case 847:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2620;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st847;
	goto tr69;
tr2603:
	{
		s->buffer[s->buffer_length] = 0;

		if (s->buffer_length == 14) { // Date; 14 = len("YYYYMMDDHHmmSS").
			ret = date_to_timestamp(s->buffer, &timestamp);

			if (ret == ZS_OK) {
				*((uint32_t *)rdata_tail) = htonl(timestamp);
				rdata_tail += 4;
			} else {
				WARN(ret);
				p--; {goto st253;}
			}
		} else if (s->buffer_length <= 10) { // Timestamp format.
			char *end;

			s->number64 = strtoull((char *)(s->buffer), &end,  10);

			if (end == (char *)(s->buffer) || *end != '\0') {
				WARN(ZS_BAD_TIMESTAMP);
				p--; {goto st253;}
			}

			if (s->number64 <= UINT32_MAX) {
				*((uint32_t *)rdata_tail) = htonl((uint32_t)s->number64);
				rdata_tail += 4;
			} else {
				WARN(ZS_NUMBER32_OVERFLOW);
				p--; {goto st253;}
			}
		} else {
			WARN(ZS_BAD_TIMESTAMP_LENGTH);
			p--; {goto st253;}
		}
	}
	goto st848;
st848:
	if ( ++p == pe )
		goto _test_eof848;
case 848:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2608;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st848;
	goto tr69;
tr2590:
	{
		if (s->number64 <= UINT32_MAX) {
			*((uint32_t *)rdata_tail) = htonl((uint32_t)(s->number64));
			rdata_tail += 4;
		} else {
			WARN(ZS_NUMBER32_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st849;
st849:
	if ( ++p == pe )
		goto _test_eof849;
case 849:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2596;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st849;
	goto tr69;
tr2578:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st850;
st850:
	if ( ++p == pe )
		goto _test_eof850;
case 850:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2583;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st850;
	goto tr69;
st851:
	if ( ++p == pe )
		goto _test_eof851;
case 851:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2570;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st851;
	goto tr69;
tr2954:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st852;
tr2560:
	{ type_num(KNOT_RRTYPE_A, &rdata_tail); }
	goto st852;
tr2652:
	{ type_num(KNOT_RRTYPE_AAAA, &rdata_tail); }
	goto st852;
tr2660:
	{ type_num(KNOT_RRTYPE_AFSDB, &rdata_tail); }
	goto st852;
tr2666:
	{ type_num(KNOT_RRTYPE_APL, &rdata_tail); }
	goto st852;
tr2680:
	{ type_num(KNOT_RRTYPE_CDNSKEY, &rdata_tail); }
	goto st852;
tr2685:
	{ type_num(KNOT_RRTYPE_CDS, &rdata_tail); }
	goto st852;
tr2692:
	{ type_num(KNOT_RRTYPE_CERT, &rdata_tail); }
	goto st852;
tr2700:
	{ type_num(KNOT_RRTYPE_CNAME, &rdata_tail); }
	goto st852;
tr2711:
	{ type_num(KNOT_RRTYPE_DHCID, &rdata_tail); }
	goto st852;
tr2720:
	{ type_num(KNOT_RRTYPE_DNAME, &rdata_tail); }
	goto st852;
tr2728:
	{ type_num(KNOT_RRTYPE_DNSKEY, &rdata_tail); }
	goto st852;
tr2733:
	{ type_num(KNOT_RRTYPE_DS, &rdata_tail); }
	goto st852;
tr2743:
	{ type_num(KNOT_RRTYPE_EUI48, &rdata_tail); }
	goto st852;
tr2749:
	{ type_num(KNOT_RRTYPE_EUI64, &rdata_tail); }
	goto st852;
tr2758:
	{ type_num(KNOT_RRTYPE_HINFO, &rdata_tail); }
	goto st852;
tr2770:
	{ type_num(KNOT_RRTYPE_IPSECKEY, &rdata_tail); }
	goto st852;
tr2778:
	{ type_num(KNOT_RRTYPE_KEY, &rdata_tail); }
	goto st852;
tr2783:
	{ type_num(KNOT_RRTYPE_KX, &rdata_tail); }
	goto st852;
tr2793:
	{ type_num(KNOT_RRTYPE_L32, &rdata_tail); }
	goto st852;
tr2799:
	{ type_num(KNOT_RRTYPE_L64, &rdata_tail); }
	goto st852;
tr2805:
	{ type_num(KNOT_RRTYPE_LOC, &rdata_tail); }
	goto st852;
tr2810:
	{ type_num(KNOT_RRTYPE_LP, &rdata_tail); }
	goto st852;
tr2820:
	{ type_num(KNOT_RRTYPE_MINFO, &rdata_tail); }
	goto st852;
tr2825:
	{ type_num(KNOT_RRTYPE_MX, &rdata_tail); }
	goto st852;
tr2836:
	{ type_num(KNOT_RRTYPE_NAPTR, &rdata_tail); }
	goto st852;
tr2842:
	{ type_num(KNOT_RRTYPE_NID, &rdata_tail); }
	goto st852;
tr2848:
	{ type_num(KNOT_RRTYPE_NS, &rdata_tail); }
	goto st852;
tr2855:
	{ type_num(KNOT_RRTYPE_NSEC, &rdata_tail); }
	goto st852;
tr2861:
	{ type_num(KNOT_RRTYPE_NSEC3, &rdata_tail); }
	goto st852;
tr2870:
	{ type_num(KNOT_RRTYPE_NSEC3PARAM, &rdata_tail); }
	goto st852;
tr2877:
	{ type_num(KNOT_RRTYPE_PTR, &rdata_tail); }
	goto st852;
tr2885:
	{ type_num(KNOT_RRTYPE_RP, &rdata_tail); }
	goto st852;
tr2893:
	{ type_num(KNOT_RRTYPE_RRSIG, &rdata_tail); }
	goto st852;
tr2898:
	{ type_num(KNOT_RRTYPE_RT, &rdata_tail); }
	goto st852;
tr2908:
	{ type_num(KNOT_RRTYPE_SOA, &rdata_tail); }
	goto st852;
tr2914:
	{ type_num(KNOT_RRTYPE_SPF, &rdata_tail); }
	goto st852;
tr2920:
	{ type_num(KNOT_RRTYPE_SRV, &rdata_tail); }
	goto st852;
tr2928:
	{ type_num(KNOT_RRTYPE_SSHFP, &rdata_tail); }
	goto st852;
tr2938:
	{ type_num(KNOT_RRTYPE_TLSA, &rdata_tail); }
	goto st852;
tr2944:
	{ type_num(KNOT_RRTYPE_TXT, &rdata_tail); }
	goto st852;
st852:
	if ( ++p == pe )
		goto _test_eof852;
case 852:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2565;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st852;
	goto tr69;
st853:
	if ( ++p == pe )
		goto _test_eof853;
case 853:
	switch( (*p) ) {
		case 65: goto st854;
		case 97: goto st854;
	}
	goto tr2538;
st854:
	if ( ++p == pe )
		goto _test_eof854;
case 854:
	switch( (*p) ) {
		case 65: goto st855;
		case 97: goto st855;
	}
	goto tr2538;
st855:
	if ( ++p == pe )
		goto _test_eof855;
case 855:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2648;
		case 32: goto tr2648;
		case 40: goto tr2649;
		case 41: goto tr2650;
		case 1034: goto tr2651;
		case 1083: goto tr2652;
	}
	goto tr2538;
st856:
	if ( ++p == pe )
		goto _test_eof856;
case 856:
	switch( (*p) ) {
		case 83: goto st857;
		case 115: goto st857;
	}
	goto tr2538;
st857:
	if ( ++p == pe )
		goto _test_eof857;
case 857:
	switch( (*p) ) {
		case 68: goto st858;
		case 100: goto st858;
	}
	goto tr2538;
st858:
	if ( ++p == pe )
		goto _test_eof858;
case 858:
	switch( (*p) ) {
		case 66: goto st859;
		case 98: goto st859;
	}
	goto tr2538;
st859:
	if ( ++p == pe )
		goto _test_eof859;
case 859:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2656;
		case 32: goto tr2656;
		case 40: goto tr2657;
		case 41: goto tr2658;
		case 1034: goto tr2659;
		case 1083: goto tr2660;
	}
	goto tr2538;
st860:
	if ( ++p == pe )
		goto _test_eof860;
case 860:
	switch( (*p) ) {
		case 76: goto st861;
		case 108: goto st861;
	}
	goto tr2538;
st861:
	if ( ++p == pe )
		goto _test_eof861;
case 861:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2662;
		case 32: goto tr2662;
		case 40: goto tr2663;
		case 41: goto tr2664;
		case 1034: goto tr2665;
		case 1083: goto tr2666;
	}
	goto tr2538;
st862:
	if ( ++p == pe )
		goto _test_eof862;
case 862:
	switch( (*p) ) {
		case 68: goto st863;
		case 69: goto st870;
		case 78: goto st873;
		case 100: goto st863;
		case 101: goto st870;
		case 110: goto st873;
	}
	goto tr2538;
st863:
	if ( ++p == pe )
		goto _test_eof863;
case 863:
	switch( (*p) ) {
		case 78: goto st864;
		case 83: goto st869;
		case 110: goto st864;
		case 115: goto st869;
	}
	goto tr2538;
st864:
	if ( ++p == pe )
		goto _test_eof864;
case 864:
	switch( (*p) ) {
		case 83: goto st865;
		case 115: goto st865;
	}
	goto tr2538;
st865:
	if ( ++p == pe )
		goto _test_eof865;
case 865:
	switch( (*p) ) {
		case 75: goto st866;
		case 107: goto st866;
	}
	goto tr2538;
st866:
	if ( ++p == pe )
		goto _test_eof866;
case 866:
	switch( (*p) ) {
		case 69: goto st867;
		case 101: goto st867;
	}
	goto tr2538;
st867:
	if ( ++p == pe )
		goto _test_eof867;
case 867:
	switch( (*p) ) {
		case 89: goto st868;
		case 121: goto st868;
	}
	goto tr2538;
st868:
	if ( ++p == pe )
		goto _test_eof868;
case 868:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2676;
		case 32: goto tr2676;
		case 40: goto tr2677;
		case 41: goto tr2678;
		case 1034: goto tr2679;
		case 1083: goto tr2680;
	}
	goto tr2538;
st869:
	if ( ++p == pe )
		goto _test_eof869;
case 869:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2681;
		case 32: goto tr2681;
		case 40: goto tr2682;
		case 41: goto tr2683;
		case 1034: goto tr2684;
		case 1083: goto tr2685;
	}
	goto tr2538;
st870:
	if ( ++p == pe )
		goto _test_eof870;
case 870:
	switch( (*p) ) {
		case 82: goto st871;
		case 114: goto st871;
	}
	goto tr2538;
st871:
	if ( ++p == pe )
		goto _test_eof871;
case 871:
	switch( (*p) ) {
		case 84: goto st872;
		case 116: goto st872;
	}
	goto tr2538;
st872:
	if ( ++p == pe )
		goto _test_eof872;
case 872:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2688;
		case 32: goto tr2688;
		case 40: goto tr2689;
		case 41: goto tr2690;
		case 1034: goto tr2691;
		case 1083: goto tr2692;
	}
	goto tr2538;
st873:
	if ( ++p == pe )
		goto _test_eof873;
case 873:
	switch( (*p) ) {
		case 65: goto st874;
		case 97: goto st874;
	}
	goto tr2538;
st874:
	if ( ++p == pe )
		goto _test_eof874;
case 874:
	switch( (*p) ) {
		case 77: goto st875;
		case 109: goto st875;
	}
	goto tr2538;
st875:
	if ( ++p == pe )
		goto _test_eof875;
case 875:
	switch( (*p) ) {
		case 69: goto st876;
		case 101: goto st876;
	}
	goto tr2538;
st876:
	if ( ++p == pe )
		goto _test_eof876;
case 876:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2696;
		case 32: goto tr2696;
		case 40: goto tr2697;
		case 41: goto tr2698;
		case 1034: goto tr2699;
		case 1083: goto tr2700;
	}
	goto tr2538;
st877:
	if ( ++p == pe )
		goto _test_eof877;
case 877:
	switch( (*p) ) {
		case 72: goto st878;
		case 78: goto st882;
		case 83: goto st890;
		case 104: goto st878;
		case 110: goto st882;
		case 115: goto st890;
	}
	goto tr2538;
st878:
	if ( ++p == pe )
		goto _test_eof878;
case 878:
	switch( (*p) ) {
		case 67: goto st879;
		case 99: goto st879;
	}
	goto tr2538;
st879:
	if ( ++p == pe )
		goto _test_eof879;
case 879:
	switch( (*p) ) {
		case 73: goto st880;
		case 105: goto st880;
	}
	goto tr2538;
st880:
	if ( ++p == pe )
		goto _test_eof880;
case 880:
	switch( (*p) ) {
		case 68: goto st881;
		case 100: goto st881;
	}
	goto tr2538;
st881:
	if ( ++p == pe )
		goto _test_eof881;
case 881:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2707;
		case 32: goto tr2707;
		case 40: goto tr2708;
		case 41: goto tr2709;
		case 1034: goto tr2710;
		case 1083: goto tr2711;
	}
	goto tr2538;
st882:
	if ( ++p == pe )
		goto _test_eof882;
case 882:
	switch( (*p) ) {
		case 65: goto st883;
		case 83: goto st886;
		case 97: goto st883;
		case 115: goto st886;
	}
	goto tr2538;
st883:
	if ( ++p == pe )
		goto _test_eof883;
case 883:
	switch( (*p) ) {
		case 77: goto st884;
		case 109: goto st884;
	}
	goto tr2538;
st884:
	if ( ++p == pe )
		goto _test_eof884;
case 884:
	switch( (*p) ) {
		case 69: goto st885;
		case 101: goto st885;
	}
	goto tr2538;
st885:
	if ( ++p == pe )
		goto _test_eof885;
case 885:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2716;
		case 32: goto tr2716;
		case 40: goto tr2717;
		case 41: goto tr2718;
		case 1034: goto tr2719;
		case 1083: goto tr2720;
	}
	goto tr2538;
st886:
	if ( ++p == pe )
		goto _test_eof886;
case 886:
	switch( (*p) ) {
		case 75: goto st887;
		case 107: goto st887;
	}
	goto tr2538;
st887:
	if ( ++p == pe )
		goto _test_eof887;
case 887:
	switch( (*p) ) {
		case 69: goto st888;
		case 101: goto st888;
	}
	goto tr2538;
st888:
	if ( ++p == pe )
		goto _test_eof888;
case 888:
	switch( (*p) ) {
		case 89: goto st889;
		case 121: goto st889;
	}
	goto tr2538;
st889:
	if ( ++p == pe )
		goto _test_eof889;
case 889:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2724;
		case 32: goto tr2724;
		case 40: goto tr2725;
		case 41: goto tr2726;
		case 1034: goto tr2727;
		case 1083: goto tr2728;
	}
	goto tr2538;
st890:
	if ( ++p == pe )
		goto _test_eof890;
case 890:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2729;
		case 32: goto tr2729;
		case 40: goto tr2730;
		case 41: goto tr2731;
		case 1034: goto tr2732;
		case 1083: goto tr2733;
	}
	goto tr2538;
st891:
	if ( ++p == pe )
		goto _test_eof891;
case 891:
	switch( (*p) ) {
		case 85: goto st892;
		case 117: goto st892;
	}
	goto tr2538;
st892:
	if ( ++p == pe )
		goto _test_eof892;
case 892:
	switch( (*p) ) {
		case 73: goto st893;
		case 105: goto st893;
	}
	goto tr2538;
st893:
	if ( ++p == pe )
		goto _test_eof893;
case 893:
	switch( (*p) ) {
		case 52: goto st894;
		case 54: goto st896;
	}
	goto tr2538;
st894:
	if ( ++p == pe )
		goto _test_eof894;
case 894:
	if ( (*p) == 56 )
		goto st895;
	goto tr2538;
st895:
	if ( ++p == pe )
		goto _test_eof895;
case 895:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2739;
		case 32: goto tr2739;
		case 40: goto tr2740;
		case 41: goto tr2741;
		case 1034: goto tr2742;
		case 1083: goto tr2743;
	}
	goto tr2538;
st896:
	if ( ++p == pe )
		goto _test_eof896;
case 896:
	if ( (*p) == 52 )
		goto st897;
	goto tr2538;
st897:
	if ( ++p == pe )
		goto _test_eof897;
case 897:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2745;
		case 32: goto tr2745;
		case 40: goto tr2746;
		case 41: goto tr2747;
		case 1034: goto tr2748;
		case 1083: goto tr2749;
	}
	goto tr2538;
st898:
	if ( ++p == pe )
		goto _test_eof898;
case 898:
	switch( (*p) ) {
		case 73: goto st899;
		case 105: goto st899;
	}
	goto tr2538;
st899:
	if ( ++p == pe )
		goto _test_eof899;
case 899:
	switch( (*p) ) {
		case 78: goto st900;
		case 110: goto st900;
	}
	goto tr2538;
st900:
	if ( ++p == pe )
		goto _test_eof900;
case 900:
	switch( (*p) ) {
		case 70: goto st901;
		case 102: goto st901;
	}
	goto tr2538;
st901:
	if ( ++p == pe )
		goto _test_eof901;
case 901:
	switch( (*p) ) {
		case 79: goto st902;
		case 111: goto st902;
	}
	goto tr2538;
st902:
	if ( ++p == pe )
		goto _test_eof902;
case 902:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2754;
		case 32: goto tr2754;
		case 40: goto tr2755;
		case 41: goto tr2756;
		case 1034: goto tr2757;
		case 1083: goto tr2758;
	}
	goto tr2538;
st903:
	if ( ++p == pe )
		goto _test_eof903;
case 903:
	switch( (*p) ) {
		case 80: goto st904;
		case 112: goto st904;
	}
	goto tr2538;
st904:
	if ( ++p == pe )
		goto _test_eof904;
case 904:
	switch( (*p) ) {
		case 83: goto st905;
		case 115: goto st905;
	}
	goto tr2538;
st905:
	if ( ++p == pe )
		goto _test_eof905;
case 905:
	switch( (*p) ) {
		case 69: goto st906;
		case 101: goto st906;
	}
	goto tr2538;
st906:
	if ( ++p == pe )
		goto _test_eof906;
case 906:
	switch( (*p) ) {
		case 67: goto st907;
		case 99: goto st907;
	}
	goto tr2538;
st907:
	if ( ++p == pe )
		goto _test_eof907;
case 907:
	switch( (*p) ) {
		case 75: goto st908;
		case 107: goto st908;
	}
	goto tr2538;
st908:
	if ( ++p == pe )
		goto _test_eof908;
case 908:
	switch( (*p) ) {
		case 69: goto st909;
		case 101: goto st909;
	}
	goto tr2538;
st909:
	if ( ++p == pe )
		goto _test_eof909;
case 909:
	switch( (*p) ) {
		case 89: goto st910;
		case 121: goto st910;
	}
	goto tr2538;
st910:
	if ( ++p == pe )
		goto _test_eof910;
case 910:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2766;
		case 32: goto tr2766;
		case 40: goto tr2767;
		case 41: goto tr2768;
		case 1034: goto tr2769;
		case 1083: goto tr2770;
	}
	goto tr2538;
st911:
	if ( ++p == pe )
		goto _test_eof911;
case 911:
	switch( (*p) ) {
		case 69: goto st912;
		case 88: goto st914;
		case 101: goto st912;
		case 120: goto st914;
	}
	goto tr2538;
st912:
	if ( ++p == pe )
		goto _test_eof912;
case 912:
	switch( (*p) ) {
		case 89: goto st913;
		case 121: goto st913;
	}
	goto tr2538;
st913:
	if ( ++p == pe )
		goto _test_eof913;
case 913:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2774;
		case 32: goto tr2774;
		case 40: goto tr2775;
		case 41: goto tr2776;
		case 1034: goto tr2777;
		case 1083: goto tr2778;
	}
	goto tr2538;
st914:
	if ( ++p == pe )
		goto _test_eof914;
case 914:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2779;
		case 32: goto tr2779;
		case 40: goto tr2780;
		case 41: goto tr2781;
		case 1034: goto tr2782;
		case 1083: goto tr2783;
	}
	goto tr2538;
st915:
	if ( ++p == pe )
		goto _test_eof915;
case 915:
	switch( (*p) ) {
		case 51: goto st916;
		case 54: goto st918;
		case 79: goto st920;
		case 80: goto st922;
		case 111: goto st920;
		case 112: goto st922;
	}
	goto tr2538;
st916:
	if ( ++p == pe )
		goto _test_eof916;
case 916:
	if ( (*p) == 50 )
		goto st917;
	goto tr2538;
st917:
	if ( ++p == pe )
		goto _test_eof917;
case 917:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2789;
		case 32: goto tr2789;
		case 40: goto tr2790;
		case 41: goto tr2791;
		case 1034: goto tr2792;
		case 1083: goto tr2793;
	}
	goto tr2538;
st918:
	if ( ++p == pe )
		goto _test_eof918;
case 918:
	if ( (*p) == 52 )
		goto st919;
	goto tr2538;
st919:
	if ( ++p == pe )
		goto _test_eof919;
case 919:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2795;
		case 32: goto tr2795;
		case 40: goto tr2796;
		case 41: goto tr2797;
		case 1034: goto tr2798;
		case 1083: goto tr2799;
	}
	goto tr2538;
st920:
	if ( ++p == pe )
		goto _test_eof920;
case 920:
	switch( (*p) ) {
		case 67: goto st921;
		case 99: goto st921;
	}
	goto tr2538;
st921:
	if ( ++p == pe )
		goto _test_eof921;
case 921:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2801;
		case 32: goto tr2801;
		case 40: goto tr2802;
		case 41: goto tr2803;
		case 1034: goto tr2804;
		case 1083: goto tr2805;
	}
	goto tr2538;
st922:
	if ( ++p == pe )
		goto _test_eof922;
case 922:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2806;
		case 32: goto tr2806;
		case 40: goto tr2807;
		case 41: goto tr2808;
		case 1034: goto tr2809;
		case 1083: goto tr2810;
	}
	goto tr2538;
st923:
	if ( ++p == pe )
		goto _test_eof923;
case 923:
	switch( (*p) ) {
		case 73: goto st924;
		case 88: goto st928;
		case 105: goto st924;
		case 120: goto st928;
	}
	goto tr2538;
st924:
	if ( ++p == pe )
		goto _test_eof924;
case 924:
	switch( (*p) ) {
		case 78: goto st925;
		case 110: goto st925;
	}
	goto tr2538;
st925:
	if ( ++p == pe )
		goto _test_eof925;
case 925:
	switch( (*p) ) {
		case 70: goto st926;
		case 102: goto st926;
	}
	goto tr2538;
st926:
	if ( ++p == pe )
		goto _test_eof926;
case 926:
	switch( (*p) ) {
		case 79: goto st927;
		case 111: goto st927;
	}
	goto tr2538;
st927:
	if ( ++p == pe )
		goto _test_eof927;
case 927:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2816;
		case 32: goto tr2816;
		case 40: goto tr2817;
		case 41: goto tr2818;
		case 1034: goto tr2819;
		case 1083: goto tr2820;
	}
	goto tr2538;
st928:
	if ( ++p == pe )
		goto _test_eof928;
case 928:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2821;
		case 32: goto tr2821;
		case 40: goto tr2822;
		case 41: goto tr2823;
		case 1034: goto tr2824;
		case 1083: goto tr2825;
	}
	goto tr2538;
st929:
	if ( ++p == pe )
		goto _test_eof929;
case 929:
	switch( (*p) ) {
		case 65: goto st930;
		case 73: goto st934;
		case 83: goto st936;
		case 97: goto st930;
		case 105: goto st934;
		case 115: goto st936;
	}
	goto tr2538;
st930:
	if ( ++p == pe )
		goto _test_eof930;
case 930:
	switch( (*p) ) {
		case 80: goto st931;
		case 112: goto st931;
	}
	goto tr2538;
st931:
	if ( ++p == pe )
		goto _test_eof931;
case 931:
	switch( (*p) ) {
		case 84: goto st932;
		case 116: goto st932;
	}
	goto tr2538;
st932:
	if ( ++p == pe )
		goto _test_eof932;
case 932:
	switch( (*p) ) {
		case 82: goto st933;
		case 114: goto st933;
	}
	goto tr2538;
st933:
	if ( ++p == pe )
		goto _test_eof933;
case 933:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2832;
		case 32: goto tr2832;
		case 40: goto tr2833;
		case 41: goto tr2834;
		case 1034: goto tr2835;
		case 1083: goto tr2836;
	}
	goto tr2538;
st934:
	if ( ++p == pe )
		goto _test_eof934;
case 934:
	switch( (*p) ) {
		case 68: goto st935;
		case 100: goto st935;
	}
	goto tr2538;
st935:
	if ( ++p == pe )
		goto _test_eof935;
case 935:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2838;
		case 32: goto tr2838;
		case 40: goto tr2839;
		case 41: goto tr2840;
		case 1034: goto tr2841;
		case 1083: goto tr2842;
	}
	goto tr2538;
st936:
	if ( ++p == pe )
		goto _test_eof936;
case 936:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2843;
		case 32: goto tr2843;
		case 40: goto tr2844;
		case 41: goto tr2845;
		case 69: goto st937;
		case 101: goto st937;
		case 1034: goto tr2847;
		case 1083: goto tr2848;
	}
	goto tr2538;
st937:
	if ( ++p == pe )
		goto _test_eof937;
case 937:
	switch( (*p) ) {
		case 67: goto st938;
		case 99: goto st938;
	}
	goto tr2538;
st938:
	if ( ++p == pe )
		goto _test_eof938;
case 938:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2850;
		case 32: goto tr2850;
		case 40: goto tr2851;
		case 41: goto tr2852;
		case 51: goto st939;
		case 1034: goto tr2854;
		case 1083: goto tr2855;
	}
	goto tr2538;
st939:
	if ( ++p == pe )
		goto _test_eof939;
case 939:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2856;
		case 32: goto tr2856;
		case 40: goto tr2857;
		case 41: goto tr2858;
		case 80: goto st940;
		case 112: goto st940;
		case 1034: goto tr2860;
		case 1083: goto tr2861;
	}
	goto tr2538;
st940:
	if ( ++p == pe )
		goto _test_eof940;
case 940:
	switch( (*p) ) {
		case 65: goto st941;
		case 97: goto st941;
	}
	goto tr2538;
st941:
	if ( ++p == pe )
		goto _test_eof941;
case 941:
	switch( (*p) ) {
		case 82: goto st942;
		case 114: goto st942;
	}
	goto tr2538;
st942:
	if ( ++p == pe )
		goto _test_eof942;
case 942:
	switch( (*p) ) {
		case 65: goto st943;
		case 97: goto st943;
	}
	goto tr2538;
st943:
	if ( ++p == pe )
		goto _test_eof943;
case 943:
	switch( (*p) ) {
		case 77: goto st944;
		case 109: goto st944;
	}
	goto tr2538;
st944:
	if ( ++p == pe )
		goto _test_eof944;
case 944:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2866;
		case 32: goto tr2866;
		case 40: goto tr2867;
		case 41: goto tr2868;
		case 1034: goto tr2869;
		case 1083: goto tr2870;
	}
	goto tr2538;
st945:
	if ( ++p == pe )
		goto _test_eof945;
case 945:
	switch( (*p) ) {
		case 84: goto st946;
		case 116: goto st946;
	}
	goto tr2538;
st946:
	if ( ++p == pe )
		goto _test_eof946;
case 946:
	switch( (*p) ) {
		case 82: goto st947;
		case 114: goto st947;
	}
	goto tr2538;
st947:
	if ( ++p == pe )
		goto _test_eof947;
case 947:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2873;
		case 32: goto tr2873;
		case 40: goto tr2874;
		case 41: goto tr2875;
		case 1034: goto tr2876;
		case 1083: goto tr2877;
	}
	goto tr2538;
st948:
	if ( ++p == pe )
		goto _test_eof948;
case 948:
	switch( (*p) ) {
		case 80: goto st949;
		case 82: goto st950;
		case 84: goto st954;
		case 112: goto st949;
		case 114: goto st950;
		case 116: goto st954;
	}
	goto tr2538;
st949:
	if ( ++p == pe )
		goto _test_eof949;
case 949:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2881;
		case 32: goto tr2881;
		case 40: goto tr2882;
		case 41: goto tr2883;
		case 1034: goto tr2884;
		case 1083: goto tr2885;
	}
	goto tr2538;
st950:
	if ( ++p == pe )
		goto _test_eof950;
case 950:
	switch( (*p) ) {
		case 83: goto st951;
		case 115: goto st951;
	}
	goto tr2538;
st951:
	if ( ++p == pe )
		goto _test_eof951;
case 951:
	switch( (*p) ) {
		case 73: goto st952;
		case 105: goto st952;
	}
	goto tr2538;
st952:
	if ( ++p == pe )
		goto _test_eof952;
case 952:
	switch( (*p) ) {
		case 71: goto st953;
		case 103: goto st953;
	}
	goto tr2538;
st953:
	if ( ++p == pe )
		goto _test_eof953;
case 953:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2889;
		case 32: goto tr2889;
		case 40: goto tr2890;
		case 41: goto tr2891;
		case 1034: goto tr2892;
		case 1083: goto tr2893;
	}
	goto tr2538;
st954:
	if ( ++p == pe )
		goto _test_eof954;
case 954:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2894;
		case 32: goto tr2894;
		case 40: goto tr2895;
		case 41: goto tr2896;
		case 1034: goto tr2897;
		case 1083: goto tr2898;
	}
	goto tr2538;
st955:
	if ( ++p == pe )
		goto _test_eof955;
case 955:
	switch( (*p) ) {
		case 79: goto st956;
		case 80: goto st958;
		case 82: goto st960;
		case 83: goto st962;
		case 111: goto st956;
		case 112: goto st958;
		case 114: goto st960;
		case 115: goto st962;
	}
	goto tr2538;
st956:
	if ( ++p == pe )
		goto _test_eof956;
case 956:
	switch( (*p) ) {
		case 65: goto st957;
		case 97: goto st957;
	}
	goto tr2538;
st957:
	if ( ++p == pe )
		goto _test_eof957;
case 957:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2904;
		case 32: goto tr2904;
		case 40: goto tr2905;
		case 41: goto tr2906;
		case 1034: goto tr2907;
		case 1083: goto tr2908;
	}
	goto tr2538;
st958:
	if ( ++p == pe )
		goto _test_eof958;
case 958:
	switch( (*p) ) {
		case 70: goto st959;
		case 102: goto st959;
	}
	goto tr2538;
st959:
	if ( ++p == pe )
		goto _test_eof959;
case 959:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2910;
		case 32: goto tr2910;
		case 40: goto tr2911;
		case 41: goto tr2912;
		case 1034: goto tr2913;
		case 1083: goto tr2914;
	}
	goto tr2538;
st960:
	if ( ++p == pe )
		goto _test_eof960;
case 960:
	switch( (*p) ) {
		case 86: goto st961;
		case 118: goto st961;
	}
	goto tr2538;
st961:
	if ( ++p == pe )
		goto _test_eof961;
case 961:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2916;
		case 32: goto tr2916;
		case 40: goto tr2917;
		case 41: goto tr2918;
		case 1034: goto tr2919;
		case 1083: goto tr2920;
	}
	goto tr2538;
st962:
	if ( ++p == pe )
		goto _test_eof962;
case 962:
	switch( (*p) ) {
		case 72: goto st963;
		case 104: goto st963;
	}
	goto tr2538;
st963:
	if ( ++p == pe )
		goto _test_eof963;
case 963:
	switch( (*p) ) {
		case 70: goto st964;
		case 102: goto st964;
	}
	goto tr2538;
st964:
	if ( ++p == pe )
		goto _test_eof964;
case 964:
	switch( (*p) ) {
		case 80: goto st965;
		case 112: goto st965;
	}
	goto tr2538;
st965:
	if ( ++p == pe )
		goto _test_eof965;
case 965:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2924;
		case 32: goto tr2924;
		case 40: goto tr2925;
		case 41: goto tr2926;
		case 1034: goto tr2927;
		case 1083: goto tr2928;
	}
	goto tr2538;
st966:
	if ( ++p == pe )
		goto _test_eof966;
case 966:
	switch( (*p) ) {
		case 76: goto st967;
		case 88: goto st970;
		case 89: goto st972;
		case 108: goto st967;
		case 120: goto st970;
		case 121: goto st972;
	}
	goto tr2538;
st967:
	if ( ++p == pe )
		goto _test_eof967;
case 967:
	switch( (*p) ) {
		case 83: goto st968;
		case 115: goto st968;
	}
	goto tr2538;
st968:
	if ( ++p == pe )
		goto _test_eof968;
case 968:
	switch( (*p) ) {
		case 65: goto st969;
		case 97: goto st969;
	}
	goto tr2538;
st969:
	if ( ++p == pe )
		goto _test_eof969;
case 969:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2934;
		case 32: goto tr2934;
		case 40: goto tr2935;
		case 41: goto tr2936;
		case 1034: goto tr2937;
		case 1083: goto tr2938;
	}
	goto tr2538;
st970:
	if ( ++p == pe )
		goto _test_eof970;
case 970:
	switch( (*p) ) {
		case 84: goto st971;
		case 116: goto st971;
	}
	goto tr2538;
st971:
	if ( ++p == pe )
		goto _test_eof971;
case 971:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2940;
		case 32: goto tr2940;
		case 40: goto tr2941;
		case 41: goto tr2942;
		case 1034: goto tr2943;
		case 1083: goto tr2944;
	}
	goto tr2538;
st972:
	if ( ++p == pe )
		goto _test_eof972;
case 972:
	switch( (*p) ) {
		case 80: goto st973;
		case 112: goto st973;
	}
	goto tr2538;
st973:
	if ( ++p == pe )
		goto _test_eof973;
case 973:
	switch( (*p) ) {
		case 69: goto st974;
		case 101: goto st974;
	}
	goto tr2538;
st974:
	if ( ++p == pe )
		goto _test_eof974;
case 974:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2948;
	goto tr2947;
tr2948:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st975;
tr2952:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st975;
st975:
	if ( ++p == pe )
		goto _test_eof975;
case 975:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2949;
		case 32: goto tr2949;
		case 40: goto tr2950;
		case 41: goto tr2951;
		case 1034: goto tr2953;
		case 1083: goto tr2954;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2952;
	goto tr2947;
st976:
	if ( ++p == pe )
		goto _test_eof976;
case 976:
	switch( (*p) ) {
		case 42: goto tr2955;
		case 92: goto tr2955;
		case 95: goto tr2955;
	}
	if ( (*p) < 64 ) {
		if ( 45 <= (*p) && (*p) <= 57 )
			goto tr2955;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr2955;
	} else
		goto tr2955;
	goto tr69;
tr2955:
	{
		s->dname = rdata_tail;
	}
	{ p--; {stack[top++] = 977; goto st255;} }
	goto st977;
st977:
	if ( ++p == pe )
		goto _test_eof977;
case 977:
	switch( (*p) ) {
		case 32: goto tr2956;
		case 59: goto tr2956;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr2956;
	} else if ( (*p) >= 9 )
		goto tr2956;
	goto tr69;
tr2956:
	{
		rdata_tail += s->dname_tmp_length;
	}
	{ p--; {stack[top++] = 978; goto st321;} }
	goto st978;
st978:
	if ( ++p == pe )
		goto _test_eof978;
case 978:
	switch( (*p) ) {
		case 32: goto tr2957;
		case 59: goto tr2957;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr2957;
	} else if ( (*p) >= 9 )
		goto tr2957;
	goto tr69;
tr2957:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1150;
st1150:
	if ( ++p == pe )
		goto _test_eof1150;
case 1150:
	goto st0;
st979:
	if ( ++p == pe )
		goto _test_eof979;
case 979:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2958;
	goto tr1700;
tr2958:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st980;
tr2962:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st980;
st980:
	if ( ++p == pe )
		goto _test_eof980;
case 980:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2959;
		case 32: goto tr2959;
		case 40: goto tr2960;
		case 41: goto tr2961;
		case 1034: goto tr2963;
		case 1083: goto tr2964;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2962;
	goto tr1700;
tr2966:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st981;
tr2967:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st981;
tr2969:
	{
		s->line_counter++;
	}
	goto st981;
tr2959:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st981;
tr2960:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st981;
tr2961:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st981;
tr2963:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st981;
st981:
	if ( ++p == pe )
		goto _test_eof981;
case 981:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st981;
		case 32: goto st981;
		case 40: goto tr2966;
		case 41: goto tr2967;
		case 1034: goto tr2969;
		case 1083: goto st989;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2968;
	goto tr1700;
tr2968:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st982;
tr2974:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st982;
st982:
	if ( ++p == pe )
		goto _test_eof982;
case 982:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2971;
		case 32: goto tr2971;
		case 40: goto tr2972;
		case 41: goto tr2973;
		case 1034: goto tr2975;
		case 1083: goto tr2976;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2974;
	goto tr1700;
tr2978:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st983;
tr2979:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st983;
tr2981:
	{
		s->line_counter++;
	}
	goto st983;
tr2971:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st983;
tr2972:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st983;
tr2973:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st983;
tr2975:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st983;
st983:
	if ( ++p == pe )
		goto _test_eof983;
case 983:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st983;
		case 32: goto st983;
		case 40: goto tr2978;
		case 41: goto tr2979;
		case 1034: goto tr2981;
		case 1083: goto st988;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr2980;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2980;
	} else
		goto tr2980;
	goto tr69;
tr2980:
	{ p--; {stack[top++] = 984; goto st467;} }
	goto st984;
st984:
	if ( ++p == pe )
		goto _test_eof984;
case 984:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st985;
		case 32: goto st985;
		case 40: goto tr2984;
		case 41: goto tr2985;
		case 1034: goto tr2986;
		case 1083: goto st987;
	}
	goto tr69;
tr2984:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st985;
tr2985:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st985;
tr2986:
	{
		s->line_counter++;
	}
	goto st985;
st985:
	if ( ++p == pe )
		goto _test_eof985;
case 985:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st985;
		case 32: goto st985;
		case 40: goto tr2984;
		case 41: goto tr2985;
		case 43: goto tr2988;
		case 1034: goto tr2986;
		case 1083: goto st987;
	}
	if ( _widec < 65 ) {
		if ( 47 <= _widec && _widec <= 57 )
			goto tr2988;
	} else if ( _widec > 90 ) {
		if ( 97 <= _widec && _widec <= 122 )
			goto tr2988;
	} else
		goto tr2988;
	goto tr69;
tr2988:
	{ p--; {stack[top++] = 986; goto st314;} }
	goto st986;
st986:
	if ( ++p == pe )
		goto _test_eof986;
case 986:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1152 + ((*p) - -128));
			if ( 
 !s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1152 + ((*p) - -128));
		if ( 
 !s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 1546: goto tr2989;
		case 1595: goto tr2989;
	}
	goto tr69;
tr2989:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1151;
st1151:
	if ( ++p == pe )
		goto _test_eof1151;
case 1151:
	goto st0;
st987:
	if ( ++p == pe )
		goto _test_eof987;
case 987:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2986;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st987;
	goto tr69;
tr2976:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st988;
st988:
	if ( ++p == pe )
		goto _test_eof988;
case 988:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2981;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st988;
	goto tr69;
tr2964:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st989;
st989:
	if ( ++p == pe )
		goto _test_eof989;
case 989:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr2969;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st989;
	goto tr69;
st990:
	if ( ++p == pe )
		goto _test_eof990;
case 990:
	if ( (*p) == 43 )
		goto tr2990;
	if ( (*p) < 65 ) {
		if ( 47 <= (*p) && (*p) <= 57 )
			goto tr2990;
	} else if ( (*p) > 90 ) {
		if ( 97 <= (*p) && (*p) <= 122 )
			goto tr2990;
	} else
		goto tr2990;
	goto tr69;
tr2990:
	{ p--; {stack[top++] = 991; goto st314;} }
	goto st991;
st991:
	if ( ++p == pe )
		goto _test_eof991;
case 991:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1152 + ((*p) - -128));
			if ( 
 !s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1152 + ((*p) - -128));
		if ( 
 !s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 1546: goto tr2991;
		case 1595: goto tr2991;
	}
	goto tr69;
tr2991:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1152;
st1152:
	if ( ++p == pe )
		goto _test_eof1152;
case 1152:
	goto st0;
st992:
	if ( ++p == pe )
		goto _test_eof992;
case 992:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr2992;
	goto tr1700;
tr2992:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st993;
tr2996:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st993;
st993:
	if ( ++p == pe )
		goto _test_eof993;
case 993:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr2993;
		case 32: goto tr2993;
		case 40: goto tr2994;
		case 41: goto tr2995;
		case 1034: goto tr2997;
		case 1083: goto tr2998;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr2996;
	goto tr1700;
tr3000:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st994;
tr3001:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st994;
tr3003:
	{
		s->line_counter++;
	}
	goto st994;
tr2993:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st994;
tr2994:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st994;
tr2995:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st994;
tr2997:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st994;
st994:
	if ( ++p == pe )
		goto _test_eof994;
case 994:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st994;
		case 32: goto st994;
		case 40: goto tr3000;
		case 41: goto tr3001;
		case 1034: goto tr3003;
		case 1083: goto st1020;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3002;
	goto tr1700;
tr3002:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st995;
tr3008:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st995;
st995:
	if ( ++p == pe )
		goto _test_eof995;
case 995:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3005;
		case 32: goto tr3005;
		case 40: goto tr3006;
		case 41: goto tr3007;
		case 1034: goto tr3009;
		case 1083: goto tr3010;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3008;
	goto tr1700;
tr3012:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st996;
tr3013:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st996;
tr3015:
	{
		s->line_counter++;
	}
	goto st996;
tr3005:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st996;
tr3006:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st996;
tr3007:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st996;
tr3009:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st996;
st996:
	if ( ++p == pe )
		goto _test_eof996;
case 996:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st996;
		case 32: goto st996;
		case 40: goto tr3012;
		case 41: goto tr3013;
		case 1034: goto tr3015;
		case 1083: goto st1019;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3014;
	goto tr1700;
tr3014:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st997;
tr3020:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st997;
st997:
	if ( ++p == pe )
		goto _test_eof997;
case 997:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3017;
		case 32: goto tr3017;
		case 40: goto tr3018;
		case 41: goto tr3019;
		case 1034: goto tr3021;
		case 1083: goto tr3022;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3020;
	goto tr1700;
tr3024:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st998;
tr3025:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st998;
tr3028:
	{
		s->line_counter++;
	}
	goto st998;
tr3017:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st998;
tr3018:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st998;
tr3019:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st998;
tr3021:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st998;
st998:
	if ( ++p == pe )
		goto _test_eof998;
case 998:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st998;
		case 32: goto st998;
		case 40: goto tr3024;
		case 41: goto tr3025;
		case 45: goto tr3026;
		case 1034: goto tr3028;
		case 1083: goto st1018;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3027;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3027;
	} else
		goto tr3027;
	goto tr2303;
tr3026:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st999;
st999:
	if ( ++p == pe )
		goto _test_eof999;
case 999:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3030;
		case 32: goto tr3030;
		case 40: goto tr3031;
		case 41: goto tr3032;
		case 1034: goto tr3033;
		case 1083: goto tr3034;
	}
	goto tr2303;
tr3037:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1000;
tr3038:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1000;
tr3040:
	{
		s->line_counter++;
	}
	goto st1000;
tr3030:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1000;
tr3031:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1000;
tr3032:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1000;
tr3033:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1000;
st1000:
	if ( ++p == pe )
		goto _test_eof1000;
case 1000:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1000;
		case 32: goto st1000;
		case 40: goto tr3037;
		case 41: goto tr3038;
		case 1034: goto tr3040;
		case 1083: goto st1015;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3039;
	} else if ( _widec > 86 ) {
		if ( 97 <= _widec && _widec <= 118 )
			goto tr3039;
	} else
		goto tr3039;
	goto tr3035;
tr3039:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_base32hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1001;
tr3054:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_base32hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1001;
st1001:
	if ( ++p == pe )
		goto _test_eof1001;
case 1001:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3042;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3042;
	} else
		goto tr3042;
	goto tr3035;
tr3042:
	{
		*(rdata_tail++) += second_left_base32hex_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = second_right_base32hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1002;
st1002:
	if ( ++p == pe )
		goto _test_eof1002;
case 1002:
	if ( (*p) == 61 )
		goto st1013;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3043;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3043;
	} else
		goto tr3043;
	goto tr3035;
tr3043:
	{
		*rdata_tail += third_base32hex_to_num[(uint8_t)(*p)];
	}
	goto st1003;
st1003:
	if ( ++p == pe )
		goto _test_eof1003;
case 1003:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3045;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3045;
	} else
		goto tr3045;
	goto tr3035;
tr3045:
	{
		*(rdata_tail++) += fourth_left_base32hex_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = fourth_right_base32hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1004;
st1004:
	if ( ++p == pe )
		goto _test_eof1004;
case 1004:
	if ( (*p) == 61 )
		goto st1012;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3046;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3046;
	} else
		goto tr3046;
	goto tr3035;
tr3046:
	{
		*(rdata_tail++) += fifth_left_base32hex_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = fifth_right_base32hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1005;
st1005:
	if ( ++p == pe )
		goto _test_eof1005;
case 1005:
	if ( (*p) == 61 )
		goto st1010;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3048;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3048;
	} else
		goto tr3048;
	goto tr3035;
tr3048:
	{
		*rdata_tail += sixth_base32hex_to_num[(uint8_t)(*p)];
	}
	goto st1006;
st1006:
	if ( ++p == pe )
		goto _test_eof1006;
case 1006:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3050;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3050;
	} else
		goto tr3050;
	goto tr3035;
tr3050:
	{
		*(rdata_tail++) += seventh_left_base32hex_to_num[(uint8_t)(*p)];

		if (rdata_tail <= rdata_stop) {
			*rdata_tail = seventh_right_base32hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1007;
st1007:
	if ( ++p == pe )
		goto _test_eof1007;
case 1007:
	if ( (*p) == 61 )
		goto st1008;
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3051;
	} else if ( (*p) > 86 ) {
		if ( 97 <= (*p) && (*p) <= 118 )
			goto tr3051;
	} else
		goto tr3051;
	goto tr3035;
tr3051:
	{
		*(rdata_tail++) += eighth_base32hex_to_num[(uint8_t)(*p)];
	}
	goto st1008;
st1008:
	if ( ++p == pe )
		goto _test_eof1008;
case 1008:
	switch( (*p) ) {
		case 32: goto tr3053;
		case 59: goto tr3053;
	}
	if ( (*p) < 48 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr3053;
		} else if ( (*p) >= 9 )
			goto tr3053;
	} else if ( (*p) > 57 ) {
		if ( (*p) > 86 ) {
			if ( 97 <= (*p) && (*p) <= 118 )
				goto tr3054;
		} else if ( (*p) >= 65 )
			goto tr3054;
	} else
		goto tr3054;
	goto tr3035;
tr3053:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{ p--; {stack[top++] = 1009; goto st321;} }
	goto st1009;
st1009:
	if ( ++p == pe )
		goto _test_eof1009;
case 1009:
	switch( (*p) ) {
		case 32: goto tr3055;
		case 59: goto tr3055;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3055;
	} else if ( (*p) >= 9 )
		goto tr3055;
	goto tr69;
tr3055:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1153;
st1153:
	if ( ++p == pe )
		goto _test_eof1153;
case 1153:
	goto st0;
st1010:
	if ( ++p == pe )
		goto _test_eof1010;
case 1010:
	if ( (*p) == 61 )
		goto st1011;
	goto tr3035;
st1011:
	if ( ++p == pe )
		goto _test_eof1011;
case 1011:
	if ( (*p) == 61 )
		goto st1008;
	goto tr3035;
st1012:
	if ( ++p == pe )
		goto _test_eof1012;
case 1012:
	if ( (*p) == 61 )
		goto st1010;
	goto tr3035;
st1013:
	if ( ++p == pe )
		goto _test_eof1013;
case 1013:
	if ( (*p) == 61 )
		goto st1014;
	goto tr3035;
st1014:
	if ( ++p == pe )
		goto _test_eof1014;
case 1014:
	if ( (*p) == 61 )
		goto st1012;
	goto tr3035;
tr3034:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1015;
st1015:
	if ( ++p == pe )
		goto _test_eof1015;
case 1015:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3040;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st1015;
	goto tr69;
tr3059:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1016;
tr3027:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1016;
st1016:
	if ( ++p == pe )
		goto _test_eof1016;
case 1016:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3058;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3058;
	} else
		goto tr3058;
	goto tr2303;
tr3058:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1017;
st1017:
	if ( ++p == pe )
		goto _test_eof1017;
case 1017:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3030;
		case 32: goto tr3030;
		case 40: goto tr3031;
		case 41: goto tr3032;
		case 1034: goto tr3033;
		case 1083: goto tr3034;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3059;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3059;
	} else
		goto tr3059;
	goto tr2303;
tr3022:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1018;
st1018:
	if ( ++p == pe )
		goto _test_eof1018;
case 1018:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3028;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st1018;
	goto tr69;
tr3010:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1019;
st1019:
	if ( ++p == pe )
		goto _test_eof1019;
case 1019:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3015;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st1019;
	goto tr69;
tr2998:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1020;
st1020:
	if ( ++p == pe )
		goto _test_eof1020;
case 1020:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3003;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st1020;
	goto tr69;
st1021:
	if ( ++p == pe )
		goto _test_eof1021;
case 1021:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3060;
	goto tr1700;
tr3060:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1022;
tr3064:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1022;
st1022:
	if ( ++p == pe )
		goto _test_eof1022;
case 1022:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3061;
		case 32: goto tr3061;
		case 40: goto tr3062;
		case 41: goto tr3063;
		case 1034: goto tr3065;
		case 1083: goto tr3066;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3064;
	goto tr1700;
tr3068:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1023;
tr3069:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1023;
tr3071:
	{
		s->line_counter++;
	}
	goto st1023;
tr3061:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1023;
tr3062:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1023;
tr3063:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1023;
tr3065:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1023;
st1023:
	if ( ++p == pe )
		goto _test_eof1023;
case 1023:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1023;
		case 32: goto st1023;
		case 40: goto tr3068;
		case 41: goto tr3069;
		case 1034: goto tr3071;
		case 1083: goto st1033;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3070;
	goto tr1700;
tr3070:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1024;
tr3076:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1024;
st1024:
	if ( ++p == pe )
		goto _test_eof1024;
case 1024:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3073;
		case 32: goto tr3073;
		case 40: goto tr3074;
		case 41: goto tr3075;
		case 1034: goto tr3077;
		case 1083: goto tr3078;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3076;
	goto tr1700;
tr3080:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1025;
tr3081:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1025;
tr3083:
	{
		s->line_counter++;
	}
	goto st1025;
tr3073:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1025;
tr3074:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1025;
tr3075:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1025;
tr3077:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1025;
st1025:
	if ( ++p == pe )
		goto _test_eof1025;
case 1025:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1025;
		case 32: goto st1025;
		case 40: goto tr3080;
		case 41: goto tr3081;
		case 1034: goto tr3083;
		case 1083: goto st1032;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3082;
	goto tr1700;
tr3082:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1026;
tr3088:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1026;
st1026:
	if ( ++p == pe )
		goto _test_eof1026;
case 1026:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3085;
		case 32: goto tr3085;
		case 40: goto tr3086;
		case 41: goto tr3087;
		case 1034: goto tr3089;
		case 1083: goto tr3090;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3088;
	goto tr1700;
tr3092:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1027;
tr3093:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1027;
tr3096:
	{
		s->line_counter++;
	}
	goto st1027;
tr3085:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1027;
tr3086:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1027;
tr3087:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1027;
tr3089:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1027;
st1027:
	if ( ++p == pe )
		goto _test_eof1027;
case 1027:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1027;
		case 32: goto st1027;
		case 40: goto tr3092;
		case 41: goto tr3093;
		case 45: goto tr3094;
		case 1034: goto tr3096;
		case 1083: goto st1031;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3095;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3095;
	} else
		goto tr3095;
	goto tr2303;
tr3094:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1028;
st1028:
	if ( ++p == pe )
		goto _test_eof1028;
case 1028:
	switch( (*p) ) {
		case 32: goto tr3098;
		case 59: goto tr3098;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3098;
	} else if ( (*p) >= 9 )
		goto tr3098;
	goto tr2303;
tr3098:
	{
		s->item_length = rdata_tail - s->item_length_location - 1;

		if (s->item_length <= MAX_ITEM_LENGTH) {
			*(s->item_length_location) = (uint8_t)(s->item_length);
		} else {
			WARN(ZS_ITEM_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1154;
st1154:
	if ( ++p == pe )
		goto _test_eof1154;
case 1154:
	goto st0;
tr3100:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1029;
tr3095:
	{
		if (rdata_tail <= rdata_stop) {
			s->item_length_location = rdata_tail++;
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1029;
st1029:
	if ( ++p == pe )
		goto _test_eof1029;
case 1029:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3099;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3099;
	} else
		goto tr3099;
	goto tr2303;
tr3099:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1030;
st1030:
	if ( ++p == pe )
		goto _test_eof1030;
case 1030:
	switch( (*p) ) {
		case 32: goto tr3098;
		case 59: goto tr3098;
	}
	if ( (*p) < 48 ) {
		if ( (*p) > 10 ) {
			if ( 40 <= (*p) && (*p) <= 41 )
				goto tr3098;
		} else if ( (*p) >= 9 )
			goto tr3098;
	} else if ( (*p) > 57 ) {
		if ( (*p) > 70 ) {
			if ( 97 <= (*p) && (*p) <= 102 )
				goto tr3100;
		} else if ( (*p) >= 65 )
			goto tr3100;
	} else
		goto tr3100;
	goto tr2303;
tr3090:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1031;
st1031:
	if ( ++p == pe )
		goto _test_eof1031;
case 1031:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3096;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st1031;
	goto tr69;
tr3078:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1032;
st1032:
	if ( ++p == pe )
		goto _test_eof1032;
case 1032:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3083;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st1032;
	goto tr69;
tr3066:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1033;
st1033:
	if ( ++p == pe )
		goto _test_eof1033;
case 1033:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3071;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st1033;
	goto tr69;
st1034:
	if ( ++p == pe )
		goto _test_eof1034;
case 1034:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3101;
	goto tr1700;
tr3101:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1035;
tr3105:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1035;
st1035:
	if ( ++p == pe )
		goto _test_eof1035;
case 1035:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3102;
		case 32: goto tr3102;
		case 40: goto tr3103;
		case 41: goto tr3104;
		case 1034: goto tr3106;
		case 1083: goto tr3107;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3105;
	goto tr1700;
tr3109:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1036;
tr3110:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1036;
tr3112:
	{
		s->line_counter++;
	}
	goto st1036;
tr3102:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1036;
tr3103:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1036;
tr3104:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1036;
tr3106:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1036;
st1036:
	if ( ++p == pe )
		goto _test_eof1036;
case 1036:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1036;
		case 32: goto st1036;
		case 40: goto tr3109;
		case 41: goto tr3110;
		case 1034: goto tr3112;
		case 1083: goto st1046;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3111;
	goto tr1700;
tr3111:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1037;
tr3117:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1037;
st1037:
	if ( ++p == pe )
		goto _test_eof1037;
case 1037:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3114;
		case 32: goto tr3114;
		case 40: goto tr3115;
		case 41: goto tr3116;
		case 1034: goto tr3118;
		case 1083: goto tr3119;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3117;
	goto tr1700;
tr3121:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1038;
tr3122:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1038;
tr3124:
	{
		s->line_counter++;
	}
	goto st1038;
tr3114:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1038;
tr3115:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1038;
tr3116:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1038;
tr3118:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1038;
st1038:
	if ( ++p == pe )
		goto _test_eof1038;
case 1038:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1038;
		case 32: goto st1038;
		case 40: goto tr3121;
		case 41: goto tr3122;
		case 1034: goto tr3124;
		case 1083: goto st1045;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3123;
	goto tr1700;
tr3123:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1039;
tr3129:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1039;
st1039:
	if ( ++p == pe )
		goto _test_eof1039;
case 1039:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3126;
		case 32: goto tr3126;
		case 40: goto tr3127;
		case 41: goto tr3128;
		case 1034: goto tr3130;
		case 1083: goto tr3131;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3129;
	goto tr1700;
tr3133:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1040;
tr3134:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1040;
tr3136:
	{
		s->line_counter++;
	}
	goto st1040;
tr3126:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1040;
tr3127:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1040;
tr3128:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1040;
tr3130:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1040;
st1040:
	if ( ++p == pe )
		goto _test_eof1040;
case 1040:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1040;
		case 32: goto st1040;
		case 40: goto tr3133;
		case 41: goto tr3134;
		case 1034: goto tr3136;
		case 1083: goto st1044;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3135;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3135;
	} else
		goto tr3135;
	goto tr2303;
tr3135:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1041;
st1041:
	if ( ++p == pe )
		goto _test_eof1041;
case 1041:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3138;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3138;
	} else
		goto tr3138;
	goto tr2303;
tr3140:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1042;
tr3141:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1042;
tr3142:
	{
		s->line_counter++;
	}
	goto st1042;
tr3138:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1042;
st1042:
	if ( ++p == pe )
		goto _test_eof1042;
case 1042:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st1042;
		case 32: goto st1042;
		case 40: goto tr3140;
		case 41: goto tr3141;
		case 2058: goto tr3142;
		case 2107: goto st1043;
		case 2314: goto tr3144;
		case 2363: goto tr3144;
		case 2570: goto tr3145;
		case 2619: goto tr3146;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3135;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3135;
	} else
		goto tr3135;
	goto tr2303;
st1043:
	if ( ++p == pe )
		goto _test_eof1043;
case 1043:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3142;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st1043;
	goto tr2303;
tr3144:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1155;
st1155:
	if ( ++p == pe )
		goto _test_eof1155;
case 1155:
	goto st0;
tr3145:
	{
		s->line_counter++;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1156;
st1156:
	if ( ++p == pe )
		goto _test_eof1156;
case 1156:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(1664 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
			if ( 
 !s->multiline  ) _widec += 512;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(1664 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
		if ( 
 !s->multiline  ) _widec += 512;
	}
	switch( _widec ) {
		case 9: goto st1042;
		case 32: goto st1042;
		case 40: goto tr3140;
		case 41: goto tr3141;
		case 2058: goto tr3142;
		case 2107: goto st1043;
		case 2314: goto tr3144;
		case 2363: goto tr3144;
		case 2570: goto tr3145;
		case 2619: goto tr3146;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3135;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3135;
	} else
		goto tr3135;
	goto tr2303;
tr3146:
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1157;
st1157:
	if ( ++p == pe )
		goto _test_eof1157;
case 1157:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3142;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st1043;
	goto tr2303;
tr3131:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1044;
st1044:
	if ( ++p == pe )
		goto _test_eof1044;
case 1044:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3136;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st1044;
	goto tr69;
tr3119:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1045;
st1045:
	if ( ++p == pe )
		goto _test_eof1045;
case 1045:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3124;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st1045;
	goto tr69;
tr3107:
	{
		if (s->number64 <= UINT8_MAX) {
			*rdata_tail = (uint8_t)(s->number64);
			rdata_tail += 1;
		} else {
			WARN(ZS_NUMBER8_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1046;
st1046:
	if ( ++p == pe )
		goto _test_eof1046;
case 1046:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3112;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st1046;
	goto tr69;
st1047:
	if ( ++p == pe )
		goto _test_eof1047;
case 1047:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3147;
	goto tr1700;
tr3147:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1048;
tr3151:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1048;
st1048:
	if ( ++p == pe )
		goto _test_eof1048;
case 1048:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3148;
		case 32: goto tr3148;
		case 40: goto tr3149;
		case 41: goto tr3150;
		case 1034: goto tr3152;
		case 1083: goto tr3153;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3151;
	goto tr1700;
tr3155:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1049;
tr3156:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1049;
tr3158:
	{
		s->line_counter++;
	}
	goto st1049;
tr3148:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1049;
tr3149:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1049;
tr3150:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1049;
tr3152:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1049;
st1049:
	if ( ++p == pe )
		goto _test_eof1049;
case 1049:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1049;
		case 32: goto st1049;
		case 40: goto tr3155;
		case 41: goto tr3156;
		case 46: goto tr3157;
		case 1034: goto tr3158;
		case 1083: goto st1051;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3157;
	goto tr1677;
tr3157:
	{
		s->buffer_length = 0;
	}
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1050;
tr3161:
	{
		if (s->buffer_length < MAX_RDATA_LENGTH) {
			s->buffer[s->buffer_length++] = (*p);
		}
		else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1050;
st1050:
	if ( ++p == pe )
		goto _test_eof1050;
case 1050:
	switch( (*p) ) {
		case 32: goto tr3160;
		case 46: goto tr3161;
		case 59: goto tr3160;
	}
	if ( (*p) < 40 ) {
		if ( 9 <= (*p) && (*p) <= 10 )
			goto tr3160;
	} else if ( (*p) > 41 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3161;
	} else
		goto tr3160;
	goto tr1677;
tr3160:
	{
		s->buffer[s->buffer_length] = 0;

		if (inet_pton(AF_INET, (char *)s->buffer, &addr4) <= 0) {
			WARN(ZS_BAD_IPV4);
			p--; {goto st253;}
		}
	}
	{
		memcpy(rdata_tail, &(addr4.s_addr), INET4_ADDR_LENGTH);
		rdata_tail += INET4_ADDR_LENGTH;
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1158;
st1158:
	if ( ++p == pe )
		goto _test_eof1158;
case 1158:
	goto st0;
tr3153:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1051;
st1051:
	if ( ++p == pe )
		goto _test_eof1051;
case 1051:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3158;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st1051;
	goto tr69;
st1052:
	if ( ++p == pe )
		goto _test_eof1052;
case 1052:
	if ( 48 <= (*p) && (*p) <= 57 )
		goto tr3162;
	goto tr1700;
tr3162:
	{
		s->number64 = 0;
	}
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1053;
tr3166:
	{
		// Overflow check: 10*(s->number64) + fc - '0' <= UINT64_MAX
		if ((s->number64 < (UINT64_MAX / 10)) ||   // Dominant fast check.
			((s->number64 == (UINT64_MAX / 10)) && // Marginal case.
			 ((uint8_t)(*p) <= (UINT64_MAX % 10) + '0')
			)
		   ) {
			s->number64 *= 10;
			s->number64 += digit_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_NUMBER64_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1053;
st1053:
	if ( ++p == pe )
		goto _test_eof1053;
case 1053:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto tr3163;
		case 32: goto tr3163;
		case 40: goto tr3164;
		case 41: goto tr3165;
		case 1034: goto tr3167;
		case 1083: goto tr3168;
	}
	if ( 48 <= _widec && _widec <= 57 )
		goto tr3166;
	goto tr1700;
tr3170:
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1054;
tr3171:
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1054;
tr3173:
	{
		s->line_counter++;
	}
	goto st1054;
tr3163:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1054;
tr3164:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == true) {
			ERR(ZS_LEFT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = true;
	}
	goto st1054;
tr3165:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		if (s->multiline == false) {
			ERR(ZS_RIGHT_PARENTHESIS);
			p--; {goto st253;}
		}
		s->multiline = false;
	}
	goto st1054;
tr3167:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	{
		s->line_counter++;
	}
	goto st1054;
st1054:
	if ( ++p == pe )
		goto _test_eof1054;
case 1054:
	_widec = (*p);
	if ( (*p) > 10 ) {
		if ( 59 <= (*p) && (*p) <= 59 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) >= 10 ) {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	switch( _widec ) {
		case 9: goto st1054;
		case 32: goto st1054;
		case 40: goto tr3170;
		case 41: goto tr3171;
		case 1034: goto tr3173;
		case 1083: goto st1064;
	}
	if ( _widec < 65 ) {
		if ( 48 <= _widec && _widec <= 57 )
			goto tr3172;
	} else if ( _widec > 70 ) {
		if ( 97 <= _widec && _widec <= 102 )
			goto tr3172;
	} else
		goto tr3172;
	goto tr2303;
tr3172:
	{
		s->item_length = 0;
	}
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1055;
st1055:
	if ( ++p == pe )
		goto _test_eof1055;
case 1055:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3175;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3175;
	} else
		goto tr3175;
	goto tr2303;
tr3175:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1056;
st1056:
	if ( ++p == pe )
		goto _test_eof1056;
case 1056:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3176;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3176;
	} else
		goto tr3176;
	goto tr2303;
tr3176:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1057;
st1057:
	if ( ++p == pe )
		goto _test_eof1057;
case 1057:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3177;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3177;
	} else
		goto tr3177;
	goto tr2303;
tr3177:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1058;
st1058:
	if ( ++p == pe )
		goto _test_eof1058;
case 1058:
	if ( (*p) == 58 )
		goto tr3179;
	goto tr3178;
tr3179:
	{
		s->item_length++;
	}
	goto st1059;
st1059:
	if ( ++p == pe )
		goto _test_eof1059;
case 1059:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3180;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3180;
	} else
		goto tr3180;
	goto tr2303;
tr3180:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1060;
st1060:
	if ( ++p == pe )
		goto _test_eof1060;
case 1060:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3181;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3181;
	} else
		goto tr3181;
	goto tr2303;
tr3181:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1061;
st1061:
	if ( ++p == pe )
		goto _test_eof1061;
case 1061:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3182;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3182;
	} else
		goto tr3182;
	goto tr2303;
tr3182:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1062;
st1062:
	if ( ++p == pe )
		goto _test_eof1062;
case 1062:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3183;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3183;
	} else
		goto tr3183;
	goto tr2303;
tr3183:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1063;
st1063:
	if ( ++p == pe )
		goto _test_eof1063;
case 1063:
	switch( (*p) ) {
		case 32: goto tr3185;
		case 58: goto tr3179;
		case 59: goto tr3185;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3185;
	} else if ( (*p) >= 9 )
		goto tr3185;
	goto tr3184;
tr3185:
	{
		s->item_length++;
	}
	{
		if (s->item_length != 4) {
			WARN(ZS_BAD_L64_LENGTH);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1159;
st1159:
	if ( ++p == pe )
		goto _test_eof1159;
case 1159:
	goto st0;
tr3168:
	{
		if (s->number64 <= UINT16_MAX) {
			*((uint16_t *)rdata_tail) = htons((uint16_t)(s->number64));
			rdata_tail += 2;
		} else {
			WARN(ZS_NUMBER16_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1064;
st1064:
	if ( ++p == pe )
		goto _test_eof1064;
case 1064:
	_widec = (*p);
	if ( (*p) < 10 ) {
		if ( (*p) <= 9 ) {
			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else if ( (*p) > 10 ) {
		if ( 11 <= (*p) )
 {			_widec = (short)(640 + ((*p) - -128));
			if ( 
 s->multiline  ) _widec += 256;
		}
	} else {
		_widec = (short)(640 + ((*p) - -128));
		if ( 
 s->multiline  ) _widec += 256;
	}
	if ( _widec == 1034 )
		goto tr3173;
	if ( 896 <= _widec && _widec <= 1151 )
		goto st1064;
	goto tr69;
st1065:
	if ( ++p == pe )
		goto _test_eof1065;
case 1065:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3186;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3186;
	} else
		goto tr3186;
	goto tr2303;
tr3186:
	{
		s->item_length = 0;
	}
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1066;
st1066:
	if ( ++p == pe )
		goto _test_eof1066;
case 1066:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3187;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3187;
	} else
		goto tr3187;
	goto tr2303;
tr3187:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1067;
st1067:
	if ( ++p == pe )
		goto _test_eof1067;
case 1067:
	if ( (*p) == 45 )
		goto tr3189;
	goto tr3188;
tr3189:
	{
		s->item_length++;
	}
	goto st1068;
st1068:
	if ( ++p == pe )
		goto _test_eof1068;
case 1068:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3190;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3190;
	} else
		goto tr3190;
	goto tr2303;
tr3190:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1069;
st1069:
	if ( ++p == pe )
		goto _test_eof1069;
case 1069:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3191;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3191;
	} else
		goto tr3191;
	goto tr2303;
tr3191:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1070;
st1070:
	if ( ++p == pe )
		goto _test_eof1070;
case 1070:
	switch( (*p) ) {
		case 32: goto tr3192;
		case 45: goto tr3189;
		case 59: goto tr3192;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3192;
	} else if ( (*p) >= 9 )
		goto tr3192;
	goto tr3188;
tr3192:
	{
		s->item_length++;
	}
	{
		if (s->item_length != 6) {
			WARN(ZS_BAD_EUI_LENGTH);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1160;
st1160:
	if ( ++p == pe )
		goto _test_eof1160;
case 1160:
	goto st0;
st1071:
	if ( ++p == pe )
		goto _test_eof1071;
case 1071:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3193;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3193;
	} else
		goto tr3193;
	goto tr2303;
tr3193:
	{
		s->item_length = 0;
	}
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1072;
st1072:
	if ( ++p == pe )
		goto _test_eof1072;
case 1072:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3194;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3194;
	} else
		goto tr3194;
	goto tr2303;
tr3194:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1073;
st1073:
	if ( ++p == pe )
		goto _test_eof1073;
case 1073:
	if ( (*p) == 45 )
		goto tr3195;
	goto tr3188;
tr3195:
	{
		s->item_length++;
	}
	goto st1074;
st1074:
	if ( ++p == pe )
		goto _test_eof1074;
case 1074:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3196;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3196;
	} else
		goto tr3196;
	goto tr2303;
tr3196:
	{
		if (rdata_tail <= rdata_stop) {
			*rdata_tail = first_hex_to_num[(uint8_t)(*p)];
		} else {
			WARN(ZS_RDATA_OVERFLOW);
			p--; {goto st253;}
		}
	}
	goto st1075;
st1075:
	if ( ++p == pe )
		goto _test_eof1075;
case 1075:
	if ( (*p) < 65 ) {
		if ( 48 <= (*p) && (*p) <= 57 )
			goto tr3197;
	} else if ( (*p) > 70 ) {
		if ( 97 <= (*p) && (*p) <= 102 )
			goto tr3197;
	} else
		goto tr3197;
	goto tr2303;
tr3197:
	{
		*rdata_tail += second_hex_to_num[(uint8_t)(*p)];
		rdata_tail++;
	}
	goto st1076;
st1076:
	if ( ++p == pe )
		goto _test_eof1076;
case 1076:
	switch( (*p) ) {
		case 32: goto tr3198;
		case 45: goto tr3195;
		case 59: goto tr3198;
	}
	if ( (*p) > 10 ) {
		if ( 40 <= (*p) && (*p) <= 41 )
			goto tr3198;
	} else if ( (*p) >= 9 )
		goto tr3198;
	goto tr3188;
tr3198:
	{
		s->item_length++;
	}
	{
		if (s->item_length != 8) {
			WARN(ZS_BAD_EUI_LENGTH);
			p--; {goto st253;}
		}
	}
	{
		p--; {cs = stack[--top];goto _again;}
	}
	goto st1161;
st1161:
	if ( ++p == pe )
		goto _test_eof1161;
case 1161:
	goto st0;
	}
	_test_eof1077: cs = 1077; goto _test_eof; 
	_test_eof1: cs = 1; goto _test_eof; 
	_test_eof2: cs = 2; goto _test_eof; 
	_test_eof3: cs = 3; goto _test_eof; 
	_test_eof4: cs = 4; goto _test_eof; 
	_test_eof5: cs = 5; goto _test_eof; 
	_test_eof6: cs = 6; goto _test_eof; 
	_test_eof7: cs = 7; goto _test_eof; 
	_test_eof8: cs = 8; goto _test_eof; 
	_test_eof9: cs = 9; goto _test_eof; 
	_test_eof10: cs = 10; goto _test_eof; 
	_test_eof11: cs = 11; goto _test_eof; 
	_test_eof12: cs = 12; goto _test_eof; 
	_test_eof1078: cs = 1078; goto _test_eof; 
	_test_eof13: cs = 13; goto _test_eof; 
	_test_eof14: cs = 14; goto _test_eof; 
	_test_eof15: cs = 15; goto _test_eof; 
	_test_eof16: cs = 16; goto _test_eof; 
	_test_eof17: cs = 17; goto _test_eof; 
	_test_eof18: cs = 18; goto _test_eof; 
	_test_eof19: cs = 19; goto _test_eof; 
	_test_eof20: cs = 20; goto _test_eof; 
	_test_eof21: cs = 21; goto _test_eof; 
	_test_eof22: cs = 22; goto _test_eof; 
	_test_eof23: cs = 23; goto _test_eof; 
	_test_eof24: cs = 24; goto _test_eof; 
	_test_eof25: cs = 25; goto _test_eof; 
	_test_eof26: cs = 26; goto _test_eof; 
	_test_eof27: cs = 27; goto _test_eof; 
	_test_eof28: cs = 28; goto _test_eof; 
	_test_eof29: cs = 29; goto _test_eof; 
	_test_eof30: cs = 30; goto _test_eof; 
	_test_eof31: cs = 31; goto _test_eof; 
	_test_eof32: cs = 32; goto _test_eof; 
	_test_eof33: cs = 33; goto _test_eof; 
	_test_eof34: cs = 34; goto _test_eof; 
	_test_eof35: cs = 35; goto _test_eof; 
	_test_eof36: cs = 36; goto _test_eof; 
	_test_eof37: cs = 37; goto _test_eof; 
	_test_eof38: cs = 38; goto _test_eof; 
	_test_eof39: cs = 39; goto _test_eof; 
	_test_eof40: cs = 40; goto _test_eof; 
	_test_eof41: cs = 41; goto _test_eof; 
	_test_eof42: cs = 42; goto _test_eof; 
	_test_eof43: cs = 43; goto _test_eof; 
	_test_eof44: cs = 44; goto _test_eof; 
	_test_eof45: cs = 45; goto _test_eof; 
	_test_eof46: cs = 46; goto _test_eof; 
	_test_eof47: cs = 47; goto _test_eof; 
	_test_eof48: cs = 48; goto _test_eof; 
	_test_eof49: cs = 49; goto _test_eof; 
	_test_eof50: cs = 50; goto _test_eof; 
	_test_eof51: cs = 51; goto _test_eof; 
	_test_eof52: cs = 52; goto _test_eof; 
	_test_eof53: cs = 53; goto _test_eof; 
	_test_eof54: cs = 54; goto _test_eof; 
	_test_eof55: cs = 55; goto _test_eof; 
	_test_eof56: cs = 56; goto _test_eof; 
	_test_eof57: cs = 57; goto _test_eof; 
	_test_eof58: cs = 58; goto _test_eof; 
	_test_eof59: cs = 59; goto _test_eof; 
	_test_eof60: cs = 60; goto _test_eof; 
	_test_eof61: cs = 61; goto _test_eof; 
	_test_eof62: cs = 62; goto _test_eof; 
	_test_eof63: cs = 63; goto _test_eof; 
	_test_eof64: cs = 64; goto _test_eof; 
	_test_eof65: cs = 65; goto _test_eof; 
	_test_eof66: cs = 66; goto _test_eof; 
	_test_eof67: cs = 67; goto _test_eof; 
	_test_eof68: cs = 68; goto _test_eof; 
	_test_eof69: cs = 69; goto _test_eof; 
	_test_eof70: cs = 70; goto _test_eof; 
	_test_eof71: cs = 71; goto _test_eof; 
	_test_eof72: cs = 72; goto _test_eof; 
	_test_eof73: cs = 73; goto _test_eof; 
	_test_eof74: cs = 74; goto _test_eof; 
	_test_eof75: cs = 75; goto _test_eof; 
	_test_eof76: cs = 76; goto _test_eof; 
	_test_eof77: cs = 77; goto _test_eof; 
	_test_eof78: cs = 78; goto _test_eof; 
	_test_eof79: cs = 79; goto _test_eof; 
	_test_eof80: cs = 80; goto _test_eof; 
	_test_eof81: cs = 81; goto _test_eof; 
	_test_eof82: cs = 82; goto _test_eof; 
	_test_eof83: cs = 83; goto _test_eof; 
	_test_eof84: cs = 84; goto _test_eof; 
	_test_eof85: cs = 85; goto _test_eof; 
	_test_eof86: cs = 86; goto _test_eof; 
	_test_eof87: cs = 87; goto _test_eof; 
	_test_eof88: cs = 88; goto _test_eof; 
	_test_eof89: cs = 89; goto _test_eof; 
	_test_eof90: cs = 90; goto _test_eof; 
	_test_eof91: cs = 91; goto _test_eof; 
	_test_eof92: cs = 92; goto _test_eof; 
	_test_eof93: cs = 93; goto _test_eof; 
	_test_eof94: cs = 94; goto _test_eof; 
	_test_eof95: cs = 95; goto _test_eof; 
	_test_eof96: cs = 96; goto _test_eof; 
	_test_eof97: cs = 97; goto _test_eof; 
	_test_eof98: cs = 98; goto _test_eof; 
	_test_eof99: cs = 99; goto _test_eof; 
	_test_eof100: cs = 100; goto _test_eof; 
	_test_eof101: cs = 101; goto _test_eof; 
	_test_eof102: cs = 102; goto _test_eof; 
	_test_eof103: cs = 103; goto _test_eof; 
	_test_eof104: cs = 104; goto _test_eof; 
	_test_eof105: cs = 105; goto _test_eof; 
	_test_eof106: cs = 106; goto _test_eof; 
	_test_eof107: cs = 107; goto _test_eof; 
	_test_eof108: cs = 108; goto _test_eof; 
	_test_eof109: cs = 109; goto _test_eof; 
	_test_eof110: cs = 110; goto _test_eof; 
	_test_eof111: cs = 111; goto _test_eof; 
	_test_eof112: cs = 112; goto _test_eof; 
	_test_eof113: cs = 113; goto _test_eof; 
	_test_eof114: cs = 114; goto _test_eof; 
	_test_eof115: cs = 115; goto _test_eof; 
	_test_eof116: cs = 116; goto _test_eof; 
	_test_eof117: cs = 117; goto _test_eof; 
	_test_eof118: cs = 118; goto _test_eof; 
	_test_eof119: cs = 119; goto _test_eof; 
	_test_eof120: cs = 120; goto _test_eof; 
	_test_eof121: cs = 121; goto _test_eof; 
	_test_eof122: cs = 122; goto _test_eof; 
	_test_eof123: cs = 123; goto _test_eof; 
	_test_eof124: cs = 124; goto _test_eof; 
	_test_eof125: cs = 125; goto _test_eof; 
	_test_eof126: cs = 126; goto _test_eof; 
	_test_eof127: cs = 127; goto _test_eof; 
	_test_eof128: cs = 128; goto _test_eof; 
	_test_eof129: cs = 129; goto _test_eof; 
	_test_eof130: cs = 130; goto _test_eof; 
	_test_eof131: cs = 131; goto _test_eof; 
	_test_eof132: cs = 132; goto _test_eof; 
	_test_eof133: cs = 133; goto _test_eof; 
	_test_eof134: cs = 134; goto _test_eof; 
	_test_eof1079: cs = 1079; goto _test_eof; 
	_test_eof135: cs = 135; goto _test_eof; 
	_test_eof136: cs = 136; goto _test_eof; 
	_test_eof137: cs = 137; goto _test_eof; 
	_test_eof138: cs = 138; goto _test_eof; 
	_test_eof139: cs = 139; goto _test_eof; 
	_test_eof140: cs = 140; goto _test_eof; 
	_test_eof141: cs = 141; goto _test_eof; 
	_test_eof142: cs = 142; goto _test_eof; 
	_test_eof143: cs = 143; goto _test_eof; 
	_test_eof144: cs = 144; goto _test_eof; 
	_test_eof1080: cs = 1080; goto _test_eof; 
	_test_eof145: cs = 145; goto _test_eof; 
	_test_eof146: cs = 146; goto _test_eof; 
	_test_eof147: cs = 147; goto _test_eof; 
	_test_eof148: cs = 148; goto _test_eof; 
	_test_eof149: cs = 149; goto _test_eof; 
	_test_eof150: cs = 150; goto _test_eof; 
	_test_eof151: cs = 151; goto _test_eof; 
	_test_eof152: cs = 152; goto _test_eof; 
	_test_eof1081: cs = 1081; goto _test_eof; 
	_test_eof153: cs = 153; goto _test_eof; 
	_test_eof154: cs = 154; goto _test_eof; 
	_test_eof155: cs = 155; goto _test_eof; 
	_test_eof1082: cs = 1082; goto _test_eof; 
	_test_eof156: cs = 156; goto _test_eof; 
	_test_eof157: cs = 157; goto _test_eof; 
	_test_eof158: cs = 158; goto _test_eof; 
	_test_eof159: cs = 159; goto _test_eof; 
	_test_eof160: cs = 160; goto _test_eof; 
	_test_eof161: cs = 161; goto _test_eof; 
	_test_eof162: cs = 162; goto _test_eof; 
	_test_eof163: cs = 163; goto _test_eof; 
	_test_eof164: cs = 164; goto _test_eof; 
	_test_eof165: cs = 165; goto _test_eof; 
	_test_eof166: cs = 166; goto _test_eof; 
	_test_eof1083: cs = 1083; goto _test_eof; 
	_test_eof167: cs = 167; goto _test_eof; 
	_test_eof168: cs = 168; goto _test_eof; 
	_test_eof169: cs = 169; goto _test_eof; 
	_test_eof170: cs = 170; goto _test_eof; 
	_test_eof1084: cs = 1084; goto _test_eof; 
	_test_eof171: cs = 171; goto _test_eof; 
	_test_eof172: cs = 172; goto _test_eof; 
	_test_eof173: cs = 173; goto _test_eof; 
	_test_eof174: cs = 174; goto _test_eof; 
	_test_eof175: cs = 175; goto _test_eof; 
	_test_eof176: cs = 176; goto _test_eof; 
	_test_eof177: cs = 177; goto _test_eof; 
	_test_eof178: cs = 178; goto _test_eof; 
	_test_eof179: cs = 179; goto _test_eof; 
	_test_eof180: cs = 180; goto _test_eof; 
	_test_eof181: cs = 181; goto _test_eof; 
	_test_eof182: cs = 182; goto _test_eof; 
	_test_eof183: cs = 183; goto _test_eof; 
	_test_eof184: cs = 184; goto _test_eof; 
	_test_eof185: cs = 185; goto _test_eof; 
	_test_eof1085: cs = 1085; goto _test_eof; 
	_test_eof186: cs = 186; goto _test_eof; 
	_test_eof187: cs = 187; goto _test_eof; 
	_test_eof188: cs = 188; goto _test_eof; 
	_test_eof189: cs = 189; goto _test_eof; 
	_test_eof190: cs = 190; goto _test_eof; 
	_test_eof191: cs = 191; goto _test_eof; 
	_test_eof192: cs = 192; goto _test_eof; 
	_test_eof193: cs = 193; goto _test_eof; 
	_test_eof194: cs = 194; goto _test_eof; 
	_test_eof195: cs = 195; goto _test_eof; 
	_test_eof196: cs = 196; goto _test_eof; 
	_test_eof197: cs = 197; goto _test_eof; 
	_test_eof198: cs = 198; goto _test_eof; 
	_test_eof199: cs = 199; goto _test_eof; 
	_test_eof200: cs = 200; goto _test_eof; 
	_test_eof201: cs = 201; goto _test_eof; 
	_test_eof1086: cs = 1086; goto _test_eof; 
	_test_eof202: cs = 202; goto _test_eof; 
	_test_eof203: cs = 203; goto _test_eof; 
	_test_eof204: cs = 204; goto _test_eof; 
	_test_eof205: cs = 205; goto _test_eof; 
	_test_eof206: cs = 206; goto _test_eof; 
	_test_eof207: cs = 207; goto _test_eof; 
	_test_eof208: cs = 208; goto _test_eof; 
	_test_eof209: cs = 209; goto _test_eof; 
	_test_eof210: cs = 210; goto _test_eof; 
	_test_eof211: cs = 211; goto _test_eof; 
	_test_eof212: cs = 212; goto _test_eof; 
	_test_eof213: cs = 213; goto _test_eof; 
	_test_eof214: cs = 214; goto _test_eof; 
	_test_eof215: cs = 215; goto _test_eof; 
	_test_eof216: cs = 216; goto _test_eof; 
	_test_eof217: cs = 217; goto _test_eof; 
	_test_eof218: cs = 218; goto _test_eof; 
	_test_eof219: cs = 219; goto _test_eof; 
	_test_eof220: cs = 220; goto _test_eof; 
	_test_eof221: cs = 221; goto _test_eof; 
	_test_eof222: cs = 222; goto _test_eof; 
	_test_eof223: cs = 223; goto _test_eof; 
	_test_eof224: cs = 224; goto _test_eof; 
	_test_eof225: cs = 225; goto _test_eof; 
	_test_eof226: cs = 226; goto _test_eof; 
	_test_eof227: cs = 227; goto _test_eof; 
	_test_eof228: cs = 228; goto _test_eof; 
	_test_eof229: cs = 229; goto _test_eof; 
	_test_eof230: cs = 230; goto _test_eof; 
	_test_eof231: cs = 231; goto _test_eof; 
	_test_eof232: cs = 232; goto _test_eof; 
	_test_eof233: cs = 233; goto _test_eof; 
	_test_eof234: cs = 234; goto _test_eof; 
	_test_eof235: cs = 235; goto _test_eof; 
	_test_eof236: cs = 236; goto _test_eof; 
	_test_eof237: cs = 237; goto _test_eof; 
	_test_eof238: cs = 238; goto _test_eof; 
	_test_eof239: cs = 239; goto _test_eof; 
	_test_eof240: cs = 240; goto _test_eof; 
	_test_eof241: cs = 241; goto _test_eof; 
	_test_eof242: cs = 242; goto _test_eof; 
	_test_eof243: cs = 243; goto _test_eof; 
	_test_eof244: cs = 244; goto _test_eof; 
	_test_eof245: cs = 245; goto _test_eof; 
	_test_eof246: cs = 246; goto _test_eof; 
	_test_eof247: cs = 247; goto _test_eof; 
	_test_eof248: cs = 248; goto _test_eof; 
	_test_eof249: cs = 249; goto _test_eof; 
	_test_eof250: cs = 250; goto _test_eof; 
	_test_eof251: cs = 251; goto _test_eof; 
	_test_eof252: cs = 252; goto _test_eof; 
	_test_eof253: cs = 253; goto _test_eof; 
	_test_eof254: cs = 254; goto _test_eof; 
	_test_eof1087: cs = 1087; goto _test_eof; 
	_test_eof255: cs = 255; goto _test_eof; 
	_test_eof256: cs = 256; goto _test_eof; 
	_test_eof1088: cs = 1088; goto _test_eof; 
	_test_eof257: cs = 257; goto _test_eof; 
	_test_eof258: cs = 258; goto _test_eof; 
	_test_eof259: cs = 259; goto _test_eof; 
	_test_eof260: cs = 260; goto _test_eof; 
	_test_eof261: cs = 261; goto _test_eof; 
	_test_eof262: cs = 262; goto _test_eof; 
	_test_eof263: cs = 263; goto _test_eof; 
	_test_eof264: cs = 264; goto _test_eof; 
	_test_eof265: cs = 265; goto _test_eof; 
	_test_eof1089: cs = 1089; goto _test_eof; 
	_test_eof1090: cs = 1090; goto _test_eof; 
	_test_eof266: cs = 266; goto _test_eof; 
	_test_eof267: cs = 267; goto _test_eof; 
	_test_eof268: cs = 268; goto _test_eof; 
	_test_eof269: cs = 269; goto _test_eof; 
	_test_eof270: cs = 270; goto _test_eof; 
	_test_eof271: cs = 271; goto _test_eof; 
	_test_eof272: cs = 272; goto _test_eof; 
	_test_eof273: cs = 273; goto _test_eof; 
	_test_eof274: cs = 274; goto _test_eof; 
	_test_eof275: cs = 275; goto _test_eof; 
	_test_eof276: cs = 276; goto _test_eof; 
	_test_eof277: cs = 277; goto _test_eof; 
	_test_eof278: cs = 278; goto _test_eof; 
	_test_eof279: cs = 279; goto _test_eof; 
	_test_eof1091: cs = 1091; goto _test_eof; 
	_test_eof280: cs = 280; goto _test_eof; 
	_test_eof281: cs = 281; goto _test_eof; 
	_test_eof282: cs = 282; goto _test_eof; 
	_test_eof283: cs = 283; goto _test_eof; 
	_test_eof284: cs = 284; goto _test_eof; 
	_test_eof285: cs = 285; goto _test_eof; 
	_test_eof286: cs = 286; goto _test_eof; 
	_test_eof287: cs = 287; goto _test_eof; 
	_test_eof288: cs = 288; goto _test_eof; 
	_test_eof289: cs = 289; goto _test_eof; 
	_test_eof1092: cs = 1092; goto _test_eof; 
	_test_eof290: cs = 290; goto _test_eof; 
	_test_eof291: cs = 291; goto _test_eof; 
	_test_eof292: cs = 292; goto _test_eof; 
	_test_eof293: cs = 293; goto _test_eof; 
	_test_eof294: cs = 294; goto _test_eof; 
	_test_eof295: cs = 295; goto _test_eof; 
	_test_eof296: cs = 296; goto _test_eof; 
	_test_eof297: cs = 297; goto _test_eof; 
	_test_eof298: cs = 298; goto _test_eof; 
	_test_eof299: cs = 299; goto _test_eof; 
	_test_eof300: cs = 300; goto _test_eof; 
	_test_eof301: cs = 301; goto _test_eof; 
	_test_eof302: cs = 302; goto _test_eof; 
	_test_eof303: cs = 303; goto _test_eof; 
	_test_eof1093: cs = 1093; goto _test_eof; 
	_test_eof304: cs = 304; goto _test_eof; 
	_test_eof305: cs = 305; goto _test_eof; 
	_test_eof306: cs = 306; goto _test_eof; 
	_test_eof307: cs = 307; goto _test_eof; 
	_test_eof308: cs = 308; goto _test_eof; 
	_test_eof309: cs = 309; goto _test_eof; 
	_test_eof310: cs = 310; goto _test_eof; 
	_test_eof1094: cs = 1094; goto _test_eof; 
	_test_eof311: cs = 311; goto _test_eof; 
	_test_eof312: cs = 312; goto _test_eof; 
	_test_eof313: cs = 313; goto _test_eof; 
	_test_eof314: cs = 314; goto _test_eof; 
	_test_eof315: cs = 315; goto _test_eof; 
	_test_eof316: cs = 316; goto _test_eof; 
	_test_eof317: cs = 317; goto _test_eof; 
	_test_eof318: cs = 318; goto _test_eof; 
	_test_eof319: cs = 319; goto _test_eof; 
	_test_eof1095: cs = 1095; goto _test_eof; 
	_test_eof1096: cs = 1096; goto _test_eof; 
	_test_eof1097: cs = 1097; goto _test_eof; 
	_test_eof320: cs = 320; goto _test_eof; 
	_test_eof321: cs = 321; goto _test_eof; 
	_test_eof322: cs = 322; goto _test_eof; 
	_test_eof323: cs = 323; goto _test_eof; 
	_test_eof324: cs = 324; goto _test_eof; 
	_test_eof325: cs = 325; goto _test_eof; 
	_test_eof326: cs = 326; goto _test_eof; 
	_test_eof327: cs = 327; goto _test_eof; 
	_test_eof1098: cs = 1098; goto _test_eof; 
	_test_eof1099: cs = 1099; goto _test_eof; 
	_test_eof328: cs = 328; goto _test_eof; 
	_test_eof329: cs = 329; goto _test_eof; 
	_test_eof330: cs = 330; goto _test_eof; 
	_test_eof331: cs = 331; goto _test_eof; 
	_test_eof332: cs = 332; goto _test_eof; 
	_test_eof333: cs = 333; goto _test_eof; 
	_test_eof334: cs = 334; goto _test_eof; 
	_test_eof1100: cs = 1100; goto _test_eof; 
	_test_eof335: cs = 335; goto _test_eof; 
	_test_eof336: cs = 336; goto _test_eof; 
	_test_eof337: cs = 337; goto _test_eof; 
	_test_eof338: cs = 338; goto _test_eof; 
	_test_eof339: cs = 339; goto _test_eof; 
	_test_eof340: cs = 340; goto _test_eof; 
	_test_eof341: cs = 341; goto _test_eof; 
	_test_eof342: cs = 342; goto _test_eof; 
	_test_eof343: cs = 343; goto _test_eof; 
	_test_eof344: cs = 344; goto _test_eof; 
	_test_eof345: cs = 345; goto _test_eof; 
	_test_eof346: cs = 346; goto _test_eof; 
	_test_eof347: cs = 347; goto _test_eof; 
	_test_eof348: cs = 348; goto _test_eof; 
	_test_eof349: cs = 349; goto _test_eof; 
	_test_eof350: cs = 350; goto _test_eof; 
	_test_eof351: cs = 351; goto _test_eof; 
	_test_eof352: cs = 352; goto _test_eof; 
	_test_eof353: cs = 353; goto _test_eof; 
	_test_eof354: cs = 354; goto _test_eof; 
	_test_eof355: cs = 355; goto _test_eof; 
	_test_eof356: cs = 356; goto _test_eof; 
	_test_eof357: cs = 357; goto _test_eof; 
	_test_eof358: cs = 358; goto _test_eof; 
	_test_eof359: cs = 359; goto _test_eof; 
	_test_eof360: cs = 360; goto _test_eof; 
	_test_eof361: cs = 361; goto _test_eof; 
	_test_eof362: cs = 362; goto _test_eof; 
	_test_eof363: cs = 363; goto _test_eof; 
	_test_eof364: cs = 364; goto _test_eof; 
	_test_eof365: cs = 365; goto _test_eof; 
	_test_eof366: cs = 366; goto _test_eof; 
	_test_eof367: cs = 367; goto _test_eof; 
	_test_eof368: cs = 368; goto _test_eof; 
	_test_eof369: cs = 369; goto _test_eof; 
	_test_eof370: cs = 370; goto _test_eof; 
	_test_eof371: cs = 371; goto _test_eof; 
	_test_eof372: cs = 372; goto _test_eof; 
	_test_eof373: cs = 373; goto _test_eof; 
	_test_eof374: cs = 374; goto _test_eof; 
	_test_eof375: cs = 375; goto _test_eof; 
	_test_eof376: cs = 376; goto _test_eof; 
	_test_eof377: cs = 377; goto _test_eof; 
	_test_eof378: cs = 378; goto _test_eof; 
	_test_eof379: cs = 379; goto _test_eof; 
	_test_eof380: cs = 380; goto _test_eof; 
	_test_eof381: cs = 381; goto _test_eof; 
	_test_eof382: cs = 382; goto _test_eof; 
	_test_eof383: cs = 383; goto _test_eof; 
	_test_eof384: cs = 384; goto _test_eof; 
	_test_eof385: cs = 385; goto _test_eof; 
	_test_eof386: cs = 386; goto _test_eof; 
	_test_eof387: cs = 387; goto _test_eof; 
	_test_eof388: cs = 388; goto _test_eof; 
	_test_eof389: cs = 389; goto _test_eof; 
	_test_eof390: cs = 390; goto _test_eof; 
	_test_eof391: cs = 391; goto _test_eof; 
	_test_eof392: cs = 392; goto _test_eof; 
	_test_eof393: cs = 393; goto _test_eof; 
	_test_eof394: cs = 394; goto _test_eof; 
	_test_eof395: cs = 395; goto _test_eof; 
	_test_eof396: cs = 396; goto _test_eof; 
	_test_eof397: cs = 397; goto _test_eof; 
	_test_eof398: cs = 398; goto _test_eof; 
	_test_eof399: cs = 399; goto _test_eof; 
	_test_eof400: cs = 400; goto _test_eof; 
	_test_eof401: cs = 401; goto _test_eof; 
	_test_eof402: cs = 402; goto _test_eof; 
	_test_eof403: cs = 403; goto _test_eof; 
	_test_eof404: cs = 404; goto _test_eof; 
	_test_eof405: cs = 405; goto _test_eof; 
	_test_eof406: cs = 406; goto _test_eof; 
	_test_eof407: cs = 407; goto _test_eof; 
	_test_eof408: cs = 408; goto _test_eof; 
	_test_eof409: cs = 409; goto _test_eof; 
	_test_eof410: cs = 410; goto _test_eof; 
	_test_eof411: cs = 411; goto _test_eof; 
	_test_eof412: cs = 412; goto _test_eof; 
	_test_eof413: cs = 413; goto _test_eof; 
	_test_eof414: cs = 414; goto _test_eof; 
	_test_eof415: cs = 415; goto _test_eof; 
	_test_eof416: cs = 416; goto _test_eof; 
	_test_eof417: cs = 417; goto _test_eof; 
	_test_eof418: cs = 418; goto _test_eof; 
	_test_eof419: cs = 419; goto _test_eof; 
	_test_eof420: cs = 420; goto _test_eof; 
	_test_eof421: cs = 421; goto _test_eof; 
	_test_eof422: cs = 422; goto _test_eof; 
	_test_eof423: cs = 423; goto _test_eof; 
	_test_eof424: cs = 424; goto _test_eof; 
	_test_eof425: cs = 425; goto _test_eof; 
	_test_eof426: cs = 426; goto _test_eof; 
	_test_eof427: cs = 427; goto _test_eof; 
	_test_eof428: cs = 428; goto _test_eof; 
	_test_eof429: cs = 429; goto _test_eof; 
	_test_eof430: cs = 430; goto _test_eof; 
	_test_eof431: cs = 431; goto _test_eof; 
	_test_eof432: cs = 432; goto _test_eof; 
	_test_eof433: cs = 433; goto _test_eof; 
	_test_eof434: cs = 434; goto _test_eof; 
	_test_eof435: cs = 435; goto _test_eof; 
	_test_eof436: cs = 436; goto _test_eof; 
	_test_eof437: cs = 437; goto _test_eof; 
	_test_eof438: cs = 438; goto _test_eof; 
	_test_eof439: cs = 439; goto _test_eof; 
	_test_eof440: cs = 440; goto _test_eof; 
	_test_eof441: cs = 441; goto _test_eof; 
	_test_eof442: cs = 442; goto _test_eof; 
	_test_eof443: cs = 443; goto _test_eof; 
	_test_eof444: cs = 444; goto _test_eof; 
	_test_eof445: cs = 445; goto _test_eof; 
	_test_eof446: cs = 446; goto _test_eof; 
	_test_eof447: cs = 447; goto _test_eof; 
	_test_eof448: cs = 448; goto _test_eof; 
	_test_eof449: cs = 449; goto _test_eof; 
	_test_eof450: cs = 450; goto _test_eof; 
	_test_eof451: cs = 451; goto _test_eof; 
	_test_eof452: cs = 452; goto _test_eof; 
	_test_eof453: cs = 453; goto _test_eof; 
	_test_eof454: cs = 454; goto _test_eof; 
	_test_eof1101: cs = 1101; goto _test_eof; 
	_test_eof1102: cs = 1102; goto _test_eof; 
	_test_eof1103: cs = 1103; goto _test_eof; 
	_test_eof455: cs = 455; goto _test_eof; 
	_test_eof456: cs = 456; goto _test_eof; 
	_test_eof457: cs = 457; goto _test_eof; 
	_test_eof458: cs = 458; goto _test_eof; 
	_test_eof459: cs = 459; goto _test_eof; 
	_test_eof1104: cs = 1104; goto _test_eof; 
	_test_eof460: cs = 460; goto _test_eof; 
	_test_eof461: cs = 461; goto _test_eof; 
	_test_eof462: cs = 462; goto _test_eof; 
	_test_eof463: cs = 463; goto _test_eof; 
	_test_eof1105: cs = 1105; goto _test_eof; 
	_test_eof1106: cs = 1106; goto _test_eof; 
	_test_eof1107: cs = 1107; goto _test_eof; 
	_test_eof464: cs = 464; goto _test_eof; 
	_test_eof465: cs = 465; goto _test_eof; 
	_test_eof1108: cs = 1108; goto _test_eof; 
	_test_eof466: cs = 466; goto _test_eof; 
	_test_eof467: cs = 467; goto _test_eof; 
	_test_eof468: cs = 468; goto _test_eof; 
	_test_eof1109: cs = 1109; goto _test_eof; 
	_test_eof469: cs = 469; goto _test_eof; 
	_test_eof470: cs = 470; goto _test_eof; 
	_test_eof471: cs = 471; goto _test_eof; 
	_test_eof472: cs = 472; goto _test_eof; 
	_test_eof473: cs = 473; goto _test_eof; 
	_test_eof474: cs = 474; goto _test_eof; 
	_test_eof475: cs = 475; goto _test_eof; 
	_test_eof476: cs = 476; goto _test_eof; 
	_test_eof477: cs = 477; goto _test_eof; 
	_test_eof478: cs = 478; goto _test_eof; 
	_test_eof479: cs = 479; goto _test_eof; 
	_test_eof480: cs = 480; goto _test_eof; 
	_test_eof481: cs = 481; goto _test_eof; 
	_test_eof482: cs = 482; goto _test_eof; 
	_test_eof483: cs = 483; goto _test_eof; 
	_test_eof484: cs = 484; goto _test_eof; 
	_test_eof485: cs = 485; goto _test_eof; 
	_test_eof486: cs = 486; goto _test_eof; 
	_test_eof487: cs = 487; goto _test_eof; 
	_test_eof488: cs = 488; goto _test_eof; 
	_test_eof489: cs = 489; goto _test_eof; 
	_test_eof490: cs = 490; goto _test_eof; 
	_test_eof491: cs = 491; goto _test_eof; 
	_test_eof492: cs = 492; goto _test_eof; 
	_test_eof493: cs = 493; goto _test_eof; 
	_test_eof494: cs = 494; goto _test_eof; 
	_test_eof495: cs = 495; goto _test_eof; 
	_test_eof496: cs = 496; goto _test_eof; 
	_test_eof497: cs = 497; goto _test_eof; 
	_test_eof498: cs = 498; goto _test_eof; 
	_test_eof499: cs = 499; goto _test_eof; 
	_test_eof500: cs = 500; goto _test_eof; 
	_test_eof501: cs = 501; goto _test_eof; 
	_test_eof502: cs = 502; goto _test_eof; 
	_test_eof503: cs = 503; goto _test_eof; 
	_test_eof504: cs = 504; goto _test_eof; 
	_test_eof505: cs = 505; goto _test_eof; 
	_test_eof506: cs = 506; goto _test_eof; 
	_test_eof507: cs = 507; goto _test_eof; 
	_test_eof508: cs = 508; goto _test_eof; 
	_test_eof509: cs = 509; goto _test_eof; 
	_test_eof510: cs = 510; goto _test_eof; 
	_test_eof511: cs = 511; goto _test_eof; 
	_test_eof512: cs = 512; goto _test_eof; 
	_test_eof513: cs = 513; goto _test_eof; 
	_test_eof514: cs = 514; goto _test_eof; 
	_test_eof515: cs = 515; goto _test_eof; 
	_test_eof516: cs = 516; goto _test_eof; 
	_test_eof517: cs = 517; goto _test_eof; 
	_test_eof518: cs = 518; goto _test_eof; 
	_test_eof519: cs = 519; goto _test_eof; 
	_test_eof520: cs = 520; goto _test_eof; 
	_test_eof521: cs = 521; goto _test_eof; 
	_test_eof522: cs = 522; goto _test_eof; 
	_test_eof523: cs = 523; goto _test_eof; 
	_test_eof524: cs = 524; goto _test_eof; 
	_test_eof525: cs = 525; goto _test_eof; 
	_test_eof526: cs = 526; goto _test_eof; 
	_test_eof527: cs = 527; goto _test_eof; 
	_test_eof528: cs = 528; goto _test_eof; 
	_test_eof529: cs = 529; goto _test_eof; 
	_test_eof530: cs = 530; goto _test_eof; 
	_test_eof531: cs = 531; goto _test_eof; 
	_test_eof532: cs = 532; goto _test_eof; 
	_test_eof533: cs = 533; goto _test_eof; 
	_test_eof534: cs = 534; goto _test_eof; 
	_test_eof535: cs = 535; goto _test_eof; 
	_test_eof536: cs = 536; goto _test_eof; 
	_test_eof537: cs = 537; goto _test_eof; 
	_test_eof538: cs = 538; goto _test_eof; 
	_test_eof539: cs = 539; goto _test_eof; 
	_test_eof540: cs = 540; goto _test_eof; 
	_test_eof541: cs = 541; goto _test_eof; 
	_test_eof542: cs = 542; goto _test_eof; 
	_test_eof543: cs = 543; goto _test_eof; 
	_test_eof544: cs = 544; goto _test_eof; 
	_test_eof545: cs = 545; goto _test_eof; 
	_test_eof546: cs = 546; goto _test_eof; 
	_test_eof547: cs = 547; goto _test_eof; 
	_test_eof548: cs = 548; goto _test_eof; 
	_test_eof549: cs = 549; goto _test_eof; 
	_test_eof550: cs = 550; goto _test_eof; 
	_test_eof551: cs = 551; goto _test_eof; 
	_test_eof552: cs = 552; goto _test_eof; 
	_test_eof553: cs = 553; goto _test_eof; 
	_test_eof554: cs = 554; goto _test_eof; 
	_test_eof555: cs = 555; goto _test_eof; 
	_test_eof556: cs = 556; goto _test_eof; 
	_test_eof557: cs = 557; goto _test_eof; 
	_test_eof558: cs = 558; goto _test_eof; 
	_test_eof559: cs = 559; goto _test_eof; 
	_test_eof560: cs = 560; goto _test_eof; 
	_test_eof561: cs = 561; goto _test_eof; 
	_test_eof562: cs = 562; goto _test_eof; 
	_test_eof563: cs = 563; goto _test_eof; 
	_test_eof1110: cs = 1110; goto _test_eof; 
	_test_eof564: cs = 564; goto _test_eof; 
	_test_eof565: cs = 565; goto _test_eof; 
	_test_eof566: cs = 566; goto _test_eof; 
	_test_eof567: cs = 567; goto _test_eof; 
	_test_eof568: cs = 568; goto _test_eof; 
	_test_eof569: cs = 569; goto _test_eof; 
	_test_eof570: cs = 570; goto _test_eof; 
	_test_eof571: cs = 571; goto _test_eof; 
	_test_eof572: cs = 572; goto _test_eof; 
	_test_eof573: cs = 573; goto _test_eof; 
	_test_eof574: cs = 574; goto _test_eof; 
	_test_eof575: cs = 575; goto _test_eof; 
	_test_eof576: cs = 576; goto _test_eof; 
	_test_eof577: cs = 577; goto _test_eof; 
	_test_eof578: cs = 578; goto _test_eof; 
	_test_eof579: cs = 579; goto _test_eof; 
	_test_eof580: cs = 580; goto _test_eof; 
	_test_eof581: cs = 581; goto _test_eof; 
	_test_eof582: cs = 582; goto _test_eof; 
	_test_eof583: cs = 583; goto _test_eof; 
	_test_eof584: cs = 584; goto _test_eof; 
	_test_eof585: cs = 585; goto _test_eof; 
	_test_eof586: cs = 586; goto _test_eof; 
	_test_eof587: cs = 587; goto _test_eof; 
	_test_eof588: cs = 588; goto _test_eof; 
	_test_eof589: cs = 589; goto _test_eof; 
	_test_eof590: cs = 590; goto _test_eof; 
	_test_eof591: cs = 591; goto _test_eof; 
	_test_eof592: cs = 592; goto _test_eof; 
	_test_eof593: cs = 593; goto _test_eof; 
	_test_eof594: cs = 594; goto _test_eof; 
	_test_eof595: cs = 595; goto _test_eof; 
	_test_eof596: cs = 596; goto _test_eof; 
	_test_eof597: cs = 597; goto _test_eof; 
	_test_eof598: cs = 598; goto _test_eof; 
	_test_eof599: cs = 599; goto _test_eof; 
	_test_eof600: cs = 600; goto _test_eof; 
	_test_eof601: cs = 601; goto _test_eof; 
	_test_eof602: cs = 602; goto _test_eof; 
	_test_eof603: cs = 603; goto _test_eof; 
	_test_eof604: cs = 604; goto _test_eof; 
	_test_eof1111: cs = 1111; goto _test_eof; 
	_test_eof605: cs = 605; goto _test_eof; 
	_test_eof606: cs = 606; goto _test_eof; 
	_test_eof1112: cs = 1112; goto _test_eof; 
	_test_eof607: cs = 607; goto _test_eof; 
	_test_eof608: cs = 608; goto _test_eof; 
	_test_eof609: cs = 609; goto _test_eof; 
	_test_eof610: cs = 610; goto _test_eof; 
	_test_eof611: cs = 611; goto _test_eof; 
	_test_eof612: cs = 612; goto _test_eof; 
	_test_eof613: cs = 613; goto _test_eof; 
	_test_eof614: cs = 614; goto _test_eof; 
	_test_eof615: cs = 615; goto _test_eof; 
	_test_eof616: cs = 616; goto _test_eof; 
	_test_eof617: cs = 617; goto _test_eof; 
	_test_eof618: cs = 618; goto _test_eof; 
	_test_eof619: cs = 619; goto _test_eof; 
	_test_eof620: cs = 620; goto _test_eof; 
	_test_eof1113: cs = 1113; goto _test_eof; 
	_test_eof621: cs = 621; goto _test_eof; 
	_test_eof622: cs = 622; goto _test_eof; 
	_test_eof623: cs = 623; goto _test_eof; 
	_test_eof624: cs = 624; goto _test_eof; 
	_test_eof625: cs = 625; goto _test_eof; 
	_test_eof626: cs = 626; goto _test_eof; 
	_test_eof627: cs = 627; goto _test_eof; 
	_test_eof628: cs = 628; goto _test_eof; 
	_test_eof629: cs = 629; goto _test_eof; 
	_test_eof630: cs = 630; goto _test_eof; 
	_test_eof631: cs = 631; goto _test_eof; 
	_test_eof632: cs = 632; goto _test_eof; 
	_test_eof633: cs = 633; goto _test_eof; 
	_test_eof634: cs = 634; goto _test_eof; 
	_test_eof635: cs = 635; goto _test_eof; 
	_test_eof636: cs = 636; goto _test_eof; 
	_test_eof637: cs = 637; goto _test_eof; 
	_test_eof638: cs = 638; goto _test_eof; 
	_test_eof639: cs = 639; goto _test_eof; 
	_test_eof640: cs = 640; goto _test_eof; 
	_test_eof641: cs = 641; goto _test_eof; 
	_test_eof642: cs = 642; goto _test_eof; 
	_test_eof1114: cs = 1114; goto _test_eof; 
	_test_eof643: cs = 643; goto _test_eof; 
	_test_eof644: cs = 644; goto _test_eof; 
	_test_eof645: cs = 645; goto _test_eof; 
	_test_eof646: cs = 646; goto _test_eof; 
	_test_eof647: cs = 647; goto _test_eof; 
	_test_eof1115: cs = 1115; goto _test_eof; 
	_test_eof648: cs = 648; goto _test_eof; 
	_test_eof649: cs = 649; goto _test_eof; 
	_test_eof650: cs = 650; goto _test_eof; 
	_test_eof651: cs = 651; goto _test_eof; 
	_test_eof652: cs = 652; goto _test_eof; 
	_test_eof1116: cs = 1116; goto _test_eof; 
	_test_eof653: cs = 653; goto _test_eof; 
	_test_eof654: cs = 654; goto _test_eof; 
	_test_eof655: cs = 655; goto _test_eof; 
	_test_eof656: cs = 656; goto _test_eof; 
	_test_eof657: cs = 657; goto _test_eof; 
	_test_eof1117: cs = 1117; goto _test_eof; 
	_test_eof1118: cs = 1118; goto _test_eof; 
	_test_eof1119: cs = 1119; goto _test_eof; 
	_test_eof658: cs = 658; goto _test_eof; 
	_test_eof659: cs = 659; goto _test_eof; 
	_test_eof1120: cs = 1120; goto _test_eof; 
	_test_eof660: cs = 660; goto _test_eof; 
	_test_eof661: cs = 661; goto _test_eof; 
	_test_eof662: cs = 662; goto _test_eof; 
	_test_eof663: cs = 663; goto _test_eof; 
	_test_eof664: cs = 664; goto _test_eof; 
	_test_eof665: cs = 665; goto _test_eof; 
	_test_eof666: cs = 666; goto _test_eof; 
	_test_eof667: cs = 667; goto _test_eof; 
	_test_eof668: cs = 668; goto _test_eof; 
	_test_eof669: cs = 669; goto _test_eof; 
	_test_eof670: cs = 670; goto _test_eof; 
	_test_eof671: cs = 671; goto _test_eof; 
	_test_eof672: cs = 672; goto _test_eof; 
	_test_eof673: cs = 673; goto _test_eof; 
	_test_eof674: cs = 674; goto _test_eof; 
	_test_eof675: cs = 675; goto _test_eof; 
	_test_eof676: cs = 676; goto _test_eof; 
	_test_eof677: cs = 677; goto _test_eof; 
	_test_eof678: cs = 678; goto _test_eof; 
	_test_eof679: cs = 679; goto _test_eof; 
	_test_eof680: cs = 680; goto _test_eof; 
	_test_eof681: cs = 681; goto _test_eof; 
	_test_eof682: cs = 682; goto _test_eof; 
	_test_eof683: cs = 683; goto _test_eof; 
	_test_eof684: cs = 684; goto _test_eof; 
	_test_eof685: cs = 685; goto _test_eof; 
	_test_eof686: cs = 686; goto _test_eof; 
	_test_eof1121: cs = 1121; goto _test_eof; 
	_test_eof1122: cs = 1122; goto _test_eof; 
	_test_eof1123: cs = 1123; goto _test_eof; 
	_test_eof687: cs = 687; goto _test_eof; 
	_test_eof688: cs = 688; goto _test_eof; 
	_test_eof689: cs = 689; goto _test_eof; 
	_test_eof1124: cs = 1124; goto _test_eof; 
	_test_eof1125: cs = 1125; goto _test_eof; 
	_test_eof690: cs = 690; goto _test_eof; 
	_test_eof691: cs = 691; goto _test_eof; 
	_test_eof692: cs = 692; goto _test_eof; 
	_test_eof693: cs = 693; goto _test_eof; 
	_test_eof1126: cs = 1126; goto _test_eof; 
	_test_eof1127: cs = 1127; goto _test_eof; 
	_test_eof694: cs = 694; goto _test_eof; 
	_test_eof695: cs = 695; goto _test_eof; 
	_test_eof696: cs = 696; goto _test_eof; 
	_test_eof697: cs = 697; goto _test_eof; 
	_test_eof1128: cs = 1128; goto _test_eof; 
	_test_eof1129: cs = 1129; goto _test_eof; 
	_test_eof698: cs = 698; goto _test_eof; 
	_test_eof699: cs = 699; goto _test_eof; 
	_test_eof700: cs = 700; goto _test_eof; 
	_test_eof701: cs = 701; goto _test_eof; 
	_test_eof702: cs = 702; goto _test_eof; 
	_test_eof703: cs = 703; goto _test_eof; 
	_test_eof704: cs = 704; goto _test_eof; 
	_test_eof705: cs = 705; goto _test_eof; 
	_test_eof706: cs = 706; goto _test_eof; 
	_test_eof707: cs = 707; goto _test_eof; 
	_test_eof708: cs = 708; goto _test_eof; 
	_test_eof709: cs = 709; goto _test_eof; 
	_test_eof710: cs = 710; goto _test_eof; 
	_test_eof711: cs = 711; goto _test_eof; 
	_test_eof712: cs = 712; goto _test_eof; 
	_test_eof713: cs = 713; goto _test_eof; 
	_test_eof714: cs = 714; goto _test_eof; 
	_test_eof715: cs = 715; goto _test_eof; 
	_test_eof716: cs = 716; goto _test_eof; 
	_test_eof717: cs = 717; goto _test_eof; 
	_test_eof718: cs = 718; goto _test_eof; 
	_test_eof719: cs = 719; goto _test_eof; 
	_test_eof720: cs = 720; goto _test_eof; 
	_test_eof721: cs = 721; goto _test_eof; 
	_test_eof722: cs = 722; goto _test_eof; 
	_test_eof1130: cs = 1130; goto _test_eof; 
	_test_eof723: cs = 723; goto _test_eof; 
	_test_eof724: cs = 724; goto _test_eof; 
	_test_eof725: cs = 725; goto _test_eof; 
	_test_eof726: cs = 726; goto _test_eof; 
	_test_eof727: cs = 727; goto _test_eof; 
	_test_eof728: cs = 728; goto _test_eof; 
	_test_eof729: cs = 729; goto _test_eof; 
	_test_eof730: cs = 730; goto _test_eof; 
	_test_eof731: cs = 731; goto _test_eof; 
	_test_eof732: cs = 732; goto _test_eof; 
	_test_eof733: cs = 733; goto _test_eof; 
	_test_eof734: cs = 734; goto _test_eof; 
	_test_eof735: cs = 735; goto _test_eof; 
	_test_eof736: cs = 736; goto _test_eof; 
	_test_eof737: cs = 737; goto _test_eof; 
	_test_eof1131: cs = 1131; goto _test_eof; 
	_test_eof738: cs = 738; goto _test_eof; 
	_test_eof739: cs = 739; goto _test_eof; 
	_test_eof740: cs = 740; goto _test_eof; 
	_test_eof741: cs = 741; goto _test_eof; 
	_test_eof742: cs = 742; goto _test_eof; 
	_test_eof743: cs = 743; goto _test_eof; 
	_test_eof744: cs = 744; goto _test_eof; 
	_test_eof745: cs = 745; goto _test_eof; 
	_test_eof746: cs = 746; goto _test_eof; 
	_test_eof747: cs = 747; goto _test_eof; 
	_test_eof748: cs = 748; goto _test_eof; 
	_test_eof749: cs = 749; goto _test_eof; 
	_test_eof750: cs = 750; goto _test_eof; 
	_test_eof1132: cs = 1132; goto _test_eof; 
	_test_eof751: cs = 751; goto _test_eof; 
	_test_eof752: cs = 752; goto _test_eof; 
	_test_eof753: cs = 753; goto _test_eof; 
	_test_eof754: cs = 754; goto _test_eof; 
	_test_eof755: cs = 755; goto _test_eof; 
	_test_eof756: cs = 756; goto _test_eof; 
	_test_eof757: cs = 757; goto _test_eof; 
	_test_eof758: cs = 758; goto _test_eof; 
	_test_eof759: cs = 759; goto _test_eof; 
	_test_eof760: cs = 760; goto _test_eof; 
	_test_eof761: cs = 761; goto _test_eof; 
	_test_eof1133: cs = 1133; goto _test_eof; 
	_test_eof1134: cs = 1134; goto _test_eof; 
	_test_eof762: cs = 762; goto _test_eof; 
	_test_eof763: cs = 763; goto _test_eof; 
	_test_eof764: cs = 764; goto _test_eof; 
	_test_eof1135: cs = 1135; goto _test_eof; 
	_test_eof765: cs = 765; goto _test_eof; 
	_test_eof766: cs = 766; goto _test_eof; 
	_test_eof767: cs = 767; goto _test_eof; 
	_test_eof768: cs = 768; goto _test_eof; 
	_test_eof769: cs = 769; goto _test_eof; 
	_test_eof770: cs = 770; goto _test_eof; 
	_test_eof771: cs = 771; goto _test_eof; 
	_test_eof772: cs = 772; goto _test_eof; 
	_test_eof773: cs = 773; goto _test_eof; 
	_test_eof774: cs = 774; goto _test_eof; 
	_test_eof1136: cs = 1136; goto _test_eof; 
	_test_eof1137: cs = 1137; goto _test_eof; 
	_test_eof1138: cs = 1138; goto _test_eof; 
	_test_eof775: cs = 775; goto _test_eof; 
	_test_eof776: cs = 776; goto _test_eof; 
	_test_eof777: cs = 777; goto _test_eof; 
	_test_eof778: cs = 778; goto _test_eof; 
	_test_eof779: cs = 779; goto _test_eof; 
	_test_eof780: cs = 780; goto _test_eof; 
	_test_eof781: cs = 781; goto _test_eof; 
	_test_eof782: cs = 782; goto _test_eof; 
	_test_eof783: cs = 783; goto _test_eof; 
	_test_eof784: cs = 784; goto _test_eof; 
	_test_eof785: cs = 785; goto _test_eof; 
	_test_eof1139: cs = 1139; goto _test_eof; 
	_test_eof1140: cs = 1140; goto _test_eof; 
	_test_eof1141: cs = 1141; goto _test_eof; 
	_test_eof786: cs = 786; goto _test_eof; 
	_test_eof787: cs = 787; goto _test_eof; 
	_test_eof788: cs = 788; goto _test_eof; 
	_test_eof789: cs = 789; goto _test_eof; 
	_test_eof790: cs = 790; goto _test_eof; 
	_test_eof791: cs = 791; goto _test_eof; 
	_test_eof792: cs = 792; goto _test_eof; 
	_test_eof793: cs = 793; goto _test_eof; 
	_test_eof794: cs = 794; goto _test_eof; 
	_test_eof795: cs = 795; goto _test_eof; 
	_test_eof796: cs = 796; goto _test_eof; 
	_test_eof797: cs = 797; goto _test_eof; 
	_test_eof1142: cs = 1142; goto _test_eof; 
	_test_eof798: cs = 798; goto _test_eof; 
	_test_eof799: cs = 799; goto _test_eof; 
	_test_eof800: cs = 800; goto _test_eof; 
	_test_eof1143: cs = 1143; goto _test_eof; 
	_test_eof1144: cs = 1144; goto _test_eof; 
	_test_eof801: cs = 801; goto _test_eof; 
	_test_eof1145: cs = 1145; goto _test_eof; 
	_test_eof1146: cs = 1146; goto _test_eof; 
	_test_eof802: cs = 802; goto _test_eof; 
	_test_eof1147: cs = 1147; goto _test_eof; 
	_test_eof1148: cs = 1148; goto _test_eof; 
	_test_eof803: cs = 803; goto _test_eof; 
	_test_eof804: cs = 804; goto _test_eof; 
	_test_eof805: cs = 805; goto _test_eof; 
	_test_eof806: cs = 806; goto _test_eof; 
	_test_eof807: cs = 807; goto _test_eof; 
	_test_eof808: cs = 808; goto _test_eof; 
	_test_eof809: cs = 809; goto _test_eof; 
	_test_eof810: cs = 810; goto _test_eof; 
	_test_eof811: cs = 811; goto _test_eof; 
	_test_eof812: cs = 812; goto _test_eof; 
	_test_eof813: cs = 813; goto _test_eof; 
	_test_eof814: cs = 814; goto _test_eof; 
	_test_eof815: cs = 815; goto _test_eof; 
	_test_eof816: cs = 816; goto _test_eof; 
	_test_eof817: cs = 817; goto _test_eof; 
	_test_eof818: cs = 818; goto _test_eof; 
	_test_eof819: cs = 819; goto _test_eof; 
	_test_eof820: cs = 820; goto _test_eof; 
	_test_eof821: cs = 821; goto _test_eof; 
	_test_eof822: cs = 822; goto _test_eof; 
	_test_eof823: cs = 823; goto _test_eof; 
	_test_eof824: cs = 824; goto _test_eof; 
	_test_eof825: cs = 825; goto _test_eof; 
	_test_eof826: cs = 826; goto _test_eof; 
	_test_eof827: cs = 827; goto _test_eof; 
	_test_eof828: cs = 828; goto _test_eof; 
	_test_eof829: cs = 829; goto _test_eof; 
	_test_eof830: cs = 830; goto _test_eof; 
	_test_eof831: cs = 831; goto _test_eof; 
	_test_eof832: cs = 832; goto _test_eof; 
	_test_eof833: cs = 833; goto _test_eof; 
	_test_eof834: cs = 834; goto _test_eof; 
	_test_eof835: cs = 835; goto _test_eof; 
	_test_eof836: cs = 836; goto _test_eof; 
	_test_eof837: cs = 837; goto _test_eof; 
	_test_eof838: cs = 838; goto _test_eof; 
	_test_eof839: cs = 839; goto _test_eof; 
	_test_eof840: cs = 840; goto _test_eof; 
	_test_eof841: cs = 841; goto _test_eof; 
	_test_eof842: cs = 842; goto _test_eof; 
	_test_eof843: cs = 843; goto _test_eof; 
	_test_eof844: cs = 844; goto _test_eof; 
	_test_eof1149: cs = 1149; goto _test_eof; 
	_test_eof845: cs = 845; goto _test_eof; 
	_test_eof846: cs = 846; goto _test_eof; 
	_test_eof847: cs = 847; goto _test_eof; 
	_test_eof848: cs = 848; goto _test_eof; 
	_test_eof849: cs = 849; goto _test_eof; 
	_test_eof850: cs = 850; goto _test_eof; 
	_test_eof851: cs = 851; goto _test_eof; 
	_test_eof852: cs = 852; goto _test_eof; 
	_test_eof853: cs = 853; goto _test_eof; 
	_test_eof854: cs = 854; goto _test_eof; 
	_test_eof855: cs = 855; goto _test_eof; 
	_test_eof856: cs = 856; goto _test_eof; 
	_test_eof857: cs = 857; goto _test_eof; 
	_test_eof858: cs = 858; goto _test_eof; 
	_test_eof859: cs = 859; goto _test_eof; 
	_test_eof860: cs = 860; goto _test_eof; 
	_test_eof861: cs = 861; goto _test_eof; 
	_test_eof862: cs = 862; goto _test_eof; 
	_test_eof863: cs = 863; goto _test_eof; 
	_test_eof864: cs = 864; goto _test_eof; 
	_test_eof865: cs = 865; goto _test_eof; 
	_test_eof866: cs = 866; goto _test_eof; 
	_test_eof867: cs = 867; goto _test_eof; 
	_test_eof868: cs = 868; goto _test_eof; 
	_test_eof869: cs = 869; goto _test_eof; 
	_test_eof870: cs = 870; goto _test_eof; 
	_test_eof871: cs = 871; goto _test_eof; 
	_test_eof872: cs = 872; goto _test_eof; 
	_test_eof873: cs = 873; goto _test_eof; 
	_test_eof874: cs = 874; goto _test_eof; 
	_test_eof875: cs = 875; goto _test_eof; 
	_test_eof876: cs = 876; goto _test_eof; 
	_test_eof877: cs = 877; goto _test_eof; 
	_test_eof878: cs = 878; goto _test_eof; 
	_test_eof879: cs = 879; goto _test_eof; 
	_test_eof880: cs = 880; goto _test_eof; 
	_test_eof881: cs = 881; goto _test_eof; 
	_test_eof882: cs = 882; goto _test_eof; 
	_test_eof883: cs = 883; goto _test_eof; 
	_test_eof884: cs = 884; goto _test_eof; 
	_test_eof885: cs = 885; goto _test_eof; 
	_test_eof886: cs = 886; goto _test_eof; 
	_test_eof887: cs = 887; goto _test_eof; 
	_test_eof888: cs = 888; goto _test_eof; 
	_test_eof889: cs = 889; goto _test_eof; 
	_test_eof890: cs = 890; goto _test_eof; 
	_test_eof891: cs = 891; goto _test_eof; 
	_test_eof892: cs = 892; goto _test_eof; 
	_test_eof893: cs = 893; goto _test_eof; 
	_test_eof894: cs = 894; goto _test_eof; 
	_test_eof895: cs = 895; goto _test_eof; 
	_test_eof896: cs = 896; goto _test_eof; 
	_test_eof897: cs = 897; goto _test_eof; 
	_test_eof898: cs = 898; goto _test_eof; 
	_test_eof899: cs = 899; goto _test_eof; 
	_test_eof900: cs = 900; goto _test_eof; 
	_test_eof901: cs = 901; goto _test_eof; 
	_test_eof902: cs = 902; goto _test_eof; 
	_test_eof903: cs = 903; goto _test_eof; 
	_test_eof904: cs = 904; goto _test_eof; 
	_test_eof905: cs = 905; goto _test_eof; 
	_test_eof906: cs = 906; goto _test_eof; 
	_test_eof907: cs = 907; goto _test_eof; 
	_test_eof908: cs = 908; goto _test_eof; 
	_test_eof909: cs = 909; goto _test_eof; 
	_test_eof910: cs = 910; goto _test_eof; 
	_test_eof911: cs = 911; goto _test_eof; 
	_test_eof912: cs = 912; goto _test_eof; 
	_test_eof913: cs = 913; goto _test_eof; 
	_test_eof914: cs = 914; goto _test_eof; 
	_test_eof915: cs = 915; goto _test_eof; 
	_test_eof916: cs = 916; goto _test_eof; 
	_test_eof917: cs = 917; goto _test_eof; 
	_test_eof918: cs = 918; goto _test_eof; 
	_test_eof919: cs = 919; goto _test_eof; 
	_test_eof920: cs = 920; goto _test_eof; 
	_test_eof921: cs = 921; goto _test_eof; 
	_test_eof922: cs = 922; goto _test_eof; 
	_test_eof923: cs = 923; goto _test_eof; 
	_test_eof924: cs = 924; goto _test_eof; 
	_test_eof925: cs = 925; goto _test_eof; 
	_test_eof926: cs = 926; goto _test_eof; 
	_test_eof927: cs = 927; goto _test_eof; 
	_test_eof928: cs = 928; goto _test_eof; 
	_test_eof929: cs = 929; goto _test_eof; 
	_test_eof930: cs = 930; goto _test_eof; 
	_test_eof931: cs = 931; goto _test_eof; 
	_test_eof932: cs = 932; goto _test_eof; 
	_test_eof933: cs = 933; goto _test_eof; 
	_test_eof934: cs = 934; goto _test_eof; 
	_test_eof935: cs = 935; goto _test_eof; 
	_test_eof936: cs = 936; goto _test_eof; 
	_test_eof937: cs = 937; goto _test_eof; 
	_test_eof938: cs = 938; goto _test_eof; 
	_test_eof939: cs = 939; goto _test_eof; 
	_test_eof940: cs = 940; goto _test_eof; 
	_test_eof941: cs = 941; goto _test_eof; 
	_test_eof942: cs = 942; goto _test_eof; 
	_test_eof943: cs = 943; goto _test_eof; 
	_test_eof944: cs = 944; goto _test_eof; 
	_test_eof945: cs = 945; goto _test_eof; 
	_test_eof946: cs = 946; goto _test_eof; 
	_test_eof947: cs = 947; goto _test_eof; 
	_test_eof948: cs = 948; goto _test_eof; 
	_test_eof949: cs = 949; goto _test_eof; 
	_test_eof950: cs = 950; goto _test_eof; 
	_test_eof951: cs = 951; goto _test_eof; 
	_test_eof952: cs = 952; goto _test_eof; 
	_test_eof953: cs = 953; goto _test_eof; 
	_test_eof954: cs = 954; goto _test_eof; 
	_test_eof955: cs = 955; goto _test_eof; 
	_test_eof956: cs = 956; goto _test_eof; 
	_test_eof957: cs = 957; goto _test_eof; 
	_test_eof958: cs = 958; goto _test_eof; 
	_test_eof959: cs = 959; goto _test_eof; 
	_test_eof960: cs = 960; goto _test_eof; 
	_test_eof961: cs = 961; goto _test_eof; 
	_test_eof962: cs = 962; goto _test_eof; 
	_test_eof963: cs = 963; goto _test_eof; 
	_test_eof964: cs = 964; goto _test_eof; 
	_test_eof965: cs = 965; goto _test_eof; 
	_test_eof966: cs = 966; goto _test_eof; 
	_test_eof967: cs = 967; goto _test_eof; 
	_test_eof968: cs = 968; goto _test_eof; 
	_test_eof969: cs = 969; goto _test_eof; 
	_test_eof970: cs = 970; goto _test_eof; 
	_test_eof971: cs = 971; goto _test_eof; 
	_test_eof972: cs = 972; goto _test_eof; 
	_test_eof973: cs = 973; goto _test_eof; 
	_test_eof974: cs = 974; goto _test_eof; 
	_test_eof975: cs = 975; goto _test_eof; 
	_test_eof976: cs = 976; goto _test_eof; 
	_test_eof977: cs = 977; goto _test_eof; 
	_test_eof978: cs = 978; goto _test_eof; 
	_test_eof1150: cs = 1150; goto _test_eof; 
	_test_eof979: cs = 979; goto _test_eof; 
	_test_eof980: cs = 980; goto _test_eof; 
	_test_eof981: cs = 981; goto _test_eof; 
	_test_eof982: cs = 982; goto _test_eof; 
	_test_eof983: cs = 983; goto _test_eof; 
	_test_eof984: cs = 984; goto _test_eof; 
	_test_eof985: cs = 985; goto _test_eof; 
	_test_eof986: cs = 986; goto _test_eof; 
	_test_eof1151: cs = 1151; goto _test_eof; 
	_test_eof987: cs = 987; goto _test_eof; 
	_test_eof988: cs = 988; goto _test_eof; 
	_test_eof989: cs = 989; goto _test_eof; 
	_test_eof990: cs = 990; goto _test_eof; 
	_test_eof991: cs = 991; goto _test_eof; 
	_test_eof1152: cs = 1152; goto _test_eof; 
	_test_eof992: cs = 992; goto _test_eof; 
	_test_eof993: cs = 993; goto _test_eof; 
	_test_eof994: cs = 994; goto _test_eof; 
	_test_eof995: cs = 995; goto _test_eof; 
	_test_eof996: cs = 996; goto _test_eof; 
	_test_eof997: cs = 997; goto _test_eof; 
	_test_eof998: cs = 998; goto _test_eof; 
	_test_eof999: cs = 999; goto _test_eof; 
	_test_eof1000: cs = 1000; goto _test_eof; 
	_test_eof1001: cs = 1001; goto _test_eof; 
	_test_eof1002: cs = 1002; goto _test_eof; 
	_test_eof1003: cs = 1003; goto _test_eof; 
	_test_eof1004: cs = 1004; goto _test_eof; 
	_test_eof1005: cs = 1005; goto _test_eof; 
	_test_eof1006: cs = 1006; goto _test_eof; 
	_test_eof1007: cs = 1007; goto _test_eof; 
	_test_eof1008: cs = 1008; goto _test_eof; 
	_test_eof1009: cs = 1009; goto _test_eof; 
	_test_eof1153: cs = 1153; goto _test_eof; 
	_test_eof1010: cs = 1010; goto _test_eof; 
	_test_eof1011: cs = 1011; goto _test_eof; 
	_test_eof1012: cs = 1012; goto _test_eof; 
	_test_eof1013: cs = 1013; goto _test_eof; 
	_test_eof1014: cs = 1014; goto _test_eof; 
	_test_eof1015: cs = 1015; goto _test_eof; 
	_test_eof1016: cs = 1016; goto _test_eof; 
	_test_eof1017: cs = 1017; goto _test_eof; 
	_test_eof1018: cs = 1018; goto _test_eof; 
	_test_eof1019: cs = 1019; goto _test_eof; 
	_test_eof1020: cs = 1020; goto _test_eof; 
	_test_eof1021: cs = 1021; goto _test_eof; 
	_test_eof1022: cs = 1022; goto _test_eof; 
	_test_eof1023: cs = 1023; goto _test_eof; 
	_test_eof1024: cs = 1024; goto _test_eof; 
	_test_eof1025: cs = 1025; goto _test_eof; 
	_test_eof1026: cs = 1026; goto _test_eof; 
	_test_eof1027: cs = 1027; goto _test_eof; 
	_test_eof1028: cs = 1028; goto _test_eof; 
	_test_eof1154: cs = 1154; goto _test_eof; 
	_test_eof1029: cs = 1029; goto _test_eof; 
	_test_eof1030: cs = 1030; goto _test_eof; 
	_test_eof1031: cs = 1031; goto _test_eof; 
	_test_eof1032: cs = 1032; goto _test_eof; 
	_test_eof1033: cs = 1033; goto _test_eof; 
	_test_eof1034: cs = 1034; goto _test_eof; 
	_test_eof1035: cs = 1035; goto _test_eof; 
	_test_eof1036: cs = 1036; goto _test_eof; 
	_test_eof1037: cs = 1037; goto _test_eof; 
	_test_eof1038: cs = 1038; goto _test_eof; 
	_test_eof1039: cs = 1039; goto _test_eof; 
	_test_eof1040: cs = 1040; goto _test_eof; 
	_test_eof1041: cs = 1041; goto _test_eof; 
	_test_eof1042: cs = 1042; goto _test_eof; 
	_test_eof1043: cs = 1043; goto _test_eof; 
	_test_eof1155: cs = 1155; goto _test_eof; 
	_test_eof1156: cs = 1156; goto _test_eof; 
	_test_eof1157: cs = 1157; goto _test_eof; 
	_test_eof1044: cs = 1044; goto _test_eof; 
	_test_eof1045: cs = 1045; goto _test_eof; 
	_test_eof1046: cs = 1046; goto _test_eof; 
	_test_eof1047: cs = 1047; goto _test_eof; 
	_test_eof1048: cs = 1048; goto _test_eof; 
	_test_eof1049: cs = 1049; goto _test_eof; 
	_test_eof1050: cs = 1050; goto _test_eof; 
	_test_eof1158: cs = 1158; goto _test_eof; 
	_test_eof1051: cs = 1051; goto _test_eof; 
	_test_eof1052: cs = 1052; goto _test_eof; 
	_test_eof1053: cs = 1053; goto _test_eof; 
	_test_eof1054: cs = 1054; goto _test_eof; 
	_test_eof1055: cs = 1055; goto _test_eof; 
	_test_eof1056: cs = 1056; goto _test_eof; 
	_test_eof1057: cs = 1057; goto _test_eof; 
	_test_eof1058: cs = 1058; goto _test_eof; 
	_test_eof1059: cs = 1059; goto _test_eof; 
	_test_eof1060: cs = 1060; goto _test_eof; 
	_test_eof1061: cs = 1061; goto _test_eof; 
	_test_eof1062: cs = 1062; goto _test_eof; 
	_test_eof1063: cs = 1063; goto _test_eof; 
	_test_eof1159: cs = 1159; goto _test_eof; 
	_test_eof1064: cs = 1064; goto _test_eof; 
	_test_eof1065: cs = 1065; goto _test_eof; 
	_test_eof1066: cs = 1066; goto _test_eof; 
	_test_eof1067: cs = 1067; goto _test_eof; 
	_test_eof1068: cs = 1068; goto _test_eof; 
	_test_eof1069: cs = 1069; goto _test_eof; 
	_test_eof1070: cs = 1070; goto _test_eof; 
	_test_eof1160: cs = 1160; goto _test_eof; 
	_test_eof1071: cs = 1071; goto _test_eof; 
	_test_eof1072: cs = 1072; goto _test_eof; 
	_test_eof1073: cs = 1073; goto _test_eof; 
	_test_eof1074: cs = 1074; goto _test_eof; 
	_test_eof1075: cs = 1075; goto _test_eof; 
	_test_eof1076: cs = 1076; goto _test_eof; 
	_test_eof1161: cs = 1161; goto _test_eof; 

	_test_eof: {}
	if ( p == eof )
	{
	switch ( cs ) {
	case 7: 
	case 8: 
	case 144: 
	case 168: 
	case 170: 
	case 218: 
	case 231: 
	case 249: 
	case 303: 
	case 304: 
	case 310: 
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	break;
	case 255: 
	case 256: 
	case 257: 
	case 262: 
	case 263: 
	{
		WARN(ZS_BAD_DNAME_CHAR);
		p--; {goto st253;}
	}
	break;
	case 153: 
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st253;}
	}
	break;
	case 271: 
	{
		WARN(ZS_BAD_TEXT);
		p--; {goto st253;}
	}
	break;
	case 276: 
	case 284: 
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st253;}
	}
	break;
	case 285: 
	case 286: 
	case 287: 
	case 296: 
	{
		ERR(ZS_BAD_ORIGIN);
		p--; {goto st253;}
	}
	break;
	case 298: 
	{
		ERR(ZS_BAD_INCLUDE_FILENAME);
		p--; {goto st253;}
	}
	break;
	case 301: 
	case 311: 
	{
		ERR(ZS_BAD_INCLUDE_ORIGIN);
		p--; {goto st253;}
	}
	break;
	case 1081: 
	{
		s->stop = false;
	}
	break;
	case 145: 
	case 146: 
	case 147: 
	case 148: 
	case 149: 
	case 150: 
	case 151: 
	case 152: 
	case 187: 
	case 188: 
	case 189: 
	case 190: 
	case 191: 
	case 192: 
	case 193: 
	case 194: 
	case 195: 
	{
		ERR(ZS_BAD_DIRECTIVE);
		p--; {goto st253;}
	}
	break;
	case 314: 
	case 315: 
	case 316: 
	case 317: 
	case 318: 
	case 319: 
	case 320: 
	{
		WARN(ZS_BAD_BASE64_CHAR);
		p--; {goto st253;}
	}
	break;
	case 321: 
	case 322: 
	case 323: 
	case 324: 
	case 325: 
	case 326: 
	case 327: 
	case 328: 
	case 329: 
	case 330: 
	case 331: 
	case 332: 
	case 333: 
	case 334: 
	case 335: 
	case 336: 
	case 337: 
	case 338: 
	case 339: 
	case 340: 
	case 341: 
	case 342: 
	case 343: 
	case 344: 
	case 345: 
	case 346: 
	case 347: 
	case 348: 
	case 349: 
	case 350: 
	case 351: 
	case 352: 
	case 353: 
	case 354: 
	case 355: 
	case 356: 
	case 357: 
	case 358: 
	case 359: 
	case 360: 
	case 361: 
	case 362: 
	case 363: 
	case 364: 
	case 365: 
	case 366: 
	case 367: 
	case 368: 
	case 369: 
	case 370: 
	case 371: 
	case 372: 
	case 373: 
	case 374: 
	case 375: 
	case 376: 
	case 377: 
	case 378: 
	case 379: 
	case 380: 
	case 381: 
	case 382: 
	case 383: 
	case 384: 
	case 385: 
	case 386: 
	case 387: 
	case 388: 
	case 389: 
	case 390: 
	case 391: 
	case 392: 
	case 393: 
	case 394: 
	case 395: 
	case 396: 
	case 397: 
	case 398: 
	case 399: 
	case 400: 
	case 401: 
	case 402: 
	case 403: 
	case 404: 
	case 405: 
	case 406: 
	case 407: 
	case 408: 
	case 409: 
	case 410: 
	case 411: 
	case 412: 
	case 413: 
	case 414: 
	case 415: 
	case 416: 
	case 417: 
	case 418: 
	case 419: 
	case 420: 
	case 421: 
	case 422: 
	case 423: 
	case 424: 
	case 425: 
	case 426: 
	case 427: 
	case 428: 
	case 429: 
	case 430: 
	case 431: 
	case 432: 
	case 433: 
	case 434: 
	case 435: 
	case 436: 
	case 437: 
	case 438: 
	case 439: 
	case 440: 
	case 441: 
	case 442: 
	case 443: 
	case 444: 
	case 445: 
	case 446: 
	case 447: 
	{
		WARN(ZS_BAD_BITMAP);
		p--; {goto st253;}
	}
	break;
	case 448: 
	case 455: 
	case 456: 
	case 457: 
	case 464: 
	case 466: 
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {goto st253;}
	}
	break;
	case 467: 
	case 468: 
	case 469: 
	case 470: 
	case 471: 
	case 472: 
	case 473: 
	case 474: 
	case 475: 
	case 476: 
	case 477: 
	case 478: 
	case 479: 
	case 480: 
	case 481: 
	case 482: 
	case 483: 
	case 484: 
	case 485: 
	case 486: 
	case 487: 
	case 488: 
	case 489: 
	case 490: 
	case 491: 
	case 492: 
	case 493: 
	case 494: 
	case 495: 
	case 496: 
	case 497: 
	case 498: 
	case 499: 
	case 500: 
	case 501: 
	case 502: 
	case 503: 
	case 504: 
	case 505: 
	case 506: 
	case 507: 
	case 508: 
	case 509: 
	case 510: 
	case 511: 
	case 512: 
	case 513: 
	case 514: 
	case 515: 
	case 516: 
	case 517: 
	case 518: 
	case 519: 
	case 520: 
	case 521: 
	case 522: 
	case 523: 
	case 524: 
	case 525: 
	case 526: 
	case 527: 
	case 528: 
	case 529: 
	case 530: 
	case 531: 
	case 532: 
	case 533: 
	case 534: 
	case 535: 
	case 536: 
	case 537: 
	case 538: 
	case 539: 
	case 540: 
	case 541: 
	case 542: 
	case 543: 
	case 544: 
	case 545: 
	case 546: 
	case 547: 
	case 548: 
	case 549: 
	case 550: 
	case 551: 
	case 552: 
	case 553: 
	case 554: 
	case 555: 
	case 556: 
	case 557: 
	case 558: 
	case 559: 
	case 560: 
	case 561: 
	{
		WARN(ZS_BAD_ALGORITHM);
		p--; {goto st253;}
	}
	break;
	case 562: 
	case 563: 
	case 564: 
	case 565: 
	case 566: 
	case 567: 
	case 568: 
	case 569: 
	case 570: 
	case 571: 
	case 572: 
	case 573: 
	case 574: 
	case 575: 
	case 576: 
	case 577: 
	case 578: 
	case 579: 
	case 580: 
	case 581: 
	case 582: 
	case 583: 
	case 584: 
	case 585: 
	case 586: 
	case 587: 
	case 588: 
	case 589: 
	case 590: 
	case 591: 
	case 592: 
	case 593: 
	case 594: 
	case 595: 
	case 596: 
	case 597: 
	case 598: 
	case 599: 
	case 600: 
	case 601: 
	case 602: 
	{
		WARN(ZS_BAD_CERT_TYPE);
		p--; {goto st253;}
	}
	break;
	case 5: 
	case 9: 
	case 10: 
	case 234: 
	case 605: 
	case 606: 
	case 607: 
	case 608: 
	case 609: 
	case 610: 
	case 624: 
	case 628: 
	case 632: 
	case 636: 
	case 637: 
	case 638: 
	case 639: 
	case 640: 
	case 641: 
	case 642: 
	case 643: 
	case 644: 
	case 645: 
	case 646: 
	case 647: 
	case 648: 
	case 651: 
	case 652: 
	case 653: 
	case 721: 
	case 722: 
	case 723: 
	case 724: 
	case 725: 
	case 730: 
	case 731: 
	case 732: 
	case 733: 
	case 734: 
	case 735: 
	case 736: 
	case 737: 
	case 738: 
	case 739: 
	case 740: 
	case 741: 
	case 742: 
	case 743: 
	case 744: 
	case 747: 
	case 748: 
	case 749: 
	case 750: 
	case 751: 
	case 752: 
	case 753: 
	case 761: 
	case 767: 
	case 768: 
	case 775: 
	case 776: 
	case 777: 
	case 786: 
	case 787: 
	case 826: 
	case 829: 
	case 830: 
	case 841: 
	case 842: 
	case 843: 
	case 844: 
	case 845: 
	case 846: 
	case 847: 
	case 848: 
	case 849: 
	case 850: 
	case 851: 
	case 852: 
	case 976: 
	case 977: 
	case 978: 
	case 983: 
	case 984: 
	case 985: 
	case 986: 
	case 987: 
	case 988: 
	case 989: 
	case 990: 
	case 991: 
	case 1009: 
	case 1015: 
	case 1018: 
	case 1019: 
	case 1020: 
	case 1031: 
	case 1032: 
	case 1033: 
	case 1044: 
	case 1045: 
	case 1046: 
	case 1051: 
	case 1064: 
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 3: 
	case 16: 
	case 17: 
	case 18: 
	case 19: 
	case 20: 
	case 21: 
	case 25: 
	case 26: 
	case 28: 
	case 29: 
	case 30: 
	case 32: 
	case 33: 
	case 34: 
	case 35: 
	case 37: 
	case 38: 
	case 39: 
	case 41: 
	case 42: 
	case 43: 
	case 46: 
	case 47: 
	case 48: 
	case 49: 
	case 51: 
	case 53: 
	case 54: 
	case 55: 
	case 56: 
	case 58: 
	case 59: 
	case 60: 
	case 61: 
	case 62: 
	case 63: 
	case 64: 
	case 65: 
	case 66: 
	case 67: 
	case 69: 
	case 70: 
	case 73: 
	case 74: 
	case 76: 
	case 78: 
	case 81: 
	case 82: 
	case 83: 
	case 84: 
	case 87: 
	case 88: 
	case 89: 
	case 90: 
	case 92: 
	case 95: 
	case 98: 
	case 99: 
	case 100: 
	case 101: 
	case 103: 
	case 104: 
	case 106: 
	case 108: 
	case 109: 
	case 110: 
	case 113: 
	case 114: 
	case 116: 
	case 118: 
	case 120: 
	case 121: 
	case 122: 
	case 124: 
	case 125: 
	case 126: 
	case 128: 
	case 130: 
	case 131: 
	case 134: 
	case 136: 
	case 137: 
	case 143: 
	case 158: 
	case 162: 
	case 220: 
	case 221: 
	case 223: 
	case 224: 
	case 225: 
	case 227: 
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	break;
	case 279: 
	case 280: 
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st253;}
	}
	break;
	case 288: 
	case 289: 
	case 290: 
	case 295: 
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	{
		ERR(ZS_BAD_ORIGIN);
		p--; {goto st253;}
	}
	break;
	case 258: 
	case 259: 
	case 260: 
	case 261: 
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_DNAME_CHAR);
		p--; {goto st253;}
	}
	break;
	case 291: 
	case 292: 
	case 293: 
	case 294: 
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		ERR(ZS_BAD_ORIGIN);
		p--; {goto st253;}
	}
	break;
	case 305: 
	case 306: 
	case 307: 
	case 308: 
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		ERR(ZS_BAD_INCLUDE_ORIGIN);
		p--; {goto st253;}
	}
	break;
	case 215: 
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 236: 
	case 237: 
	case 238: 
	case 239: 
	case 240: 
	case 241: 
	case 242: 
	case 243: 
	case 244: 
	case 245: 
	case 246: 
	case 247: 
	case 248: 
	case 251: 
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	break;
	case 277: 
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st253;}
	}
	break;
	case 449: 
	case 450: 
	case 458: 
	case 459: 
	case 465: 
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {goto st253;}
	}
	break;
	case 611: 
	case 612: 
	case 613: 
	case 615: 
	case 617: 
	case 619: 
	case 649: 
	case 650: 
	case 715: 
	case 716: 
	case 717: 
	case 718: 
	case 719: 
	case 720: 
	case 726: 
	case 727: 
	case 728: 
	case 729: 
	case 745: 
	case 746: 
	case 765: 
	case 766: 
	case 769: 
	case 770: 
	case 778: 
	case 779: 
	case 780: 
	case 781: 
	case 788: 
	case 789: 
	case 831: 
	case 832: 
	case 833: 
	case 834: 
	case 839: 
	case 840: 
	case 979: 
	case 980: 
	case 981: 
	case 982: 
	case 992: 
	case 993: 
	case 994: 
	case 995: 
	case 996: 
	case 997: 
	case 1021: 
	case 1022: 
	case 1023: 
	case 1024: 
	case 1025: 
	case 1026: 
	case 1034: 
	case 1035: 
	case 1036: 
	case 1037: 
	case 1038: 
	case 1039: 
	case 1047: 
	case 1048: 
	case 1052: 
	case 1053: 
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 132: 
	case 138: 
	case 154: 
	case 157: 
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	break;
	case 835: 
	case 836: 
	case 837: 
	case 838: 
	{
		WARN(ZS_BAD_TIMESTAMP_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 264: 
	case 265: 
	case 270: 
	{
		WARN(ZS_BAD_TEXT_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_TEXT);
		p--; {goto st253;}
	}
	break;
	case 654: 
	case 655: 
	case 656: 
	case 657: 
	{
		s->long_string = false;
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 299: 
	{
		ERR(ZS_BAD_INCLUDE_FILENAME);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	break;
	case 300: 
	case 302: 
	case 309: 
	{
		ERR(ZS_BAD_INCLUDE_ORIGIN);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	break;
	case 603: 
	case 604: 
	case 658: 
	case 659: 
	case 1049: 
	case 1050: 
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 754: 
	case 755: 
	case 756: 
	case 759: 
	case 760: 
	case 762: 
	{
		WARN(ZS_BAD_APL);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 771: 
	case 772: 
	case 773: 
	case 774: 
	case 782: 
	case 783: 
	case 784: 
	case 785: 
	case 998: 
	case 999: 
	case 1016: 
	case 1017: 
	case 1027: 
	case 1028: 
	case 1029: 
	case 1030: 
	case 1040: 
	case 1041: 
	case 1042: 
	case 1043: 
	case 1054: 
	case 1055: 
	case 1056: 
	case 1057: 
	case 1059: 
	case 1060: 
	case 1061: 
	case 1062: 
	case 1065: 
	case 1066: 
	case 1068: 
	case 1069: 
	case 1071: 
	case 1072: 
	case 1074: 
	case 1075: 
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 1000: 
	case 1001: 
	case 1002: 
	case 1003: 
	case 1004: 
	case 1005: 
	case 1006: 
	case 1007: 
	case 1008: 
	case 1010: 
	case 1011: 
	case 1012: 
	case 1013: 
	case 1014: 
	{
		WARN(ZS_BAD_BASE32HEX_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 790: 
	case 791: 
	case 794: 
	case 803: 
	case 804: 
	case 805: 
	case 810: 
	case 811: 
	case 812: 
	case 817: 
	case 818: 
	case 819: 
	case 822: 
	case 824: 
	case 825: 
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 796: 
	case 797: 
	case 798: 
	case 799: 
	case 800: 
	case 801: 
	case 802: 
	{
		WARN(ZS_BAD_GATEWAY_KEY);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 827: 
	case 828: 
	case 853: 
	case 854: 
	case 855: 
	case 856: 
	case 857: 
	case 858: 
	case 859: 
	case 860: 
	case 861: 
	case 862: 
	case 863: 
	case 864: 
	case 865: 
	case 866: 
	case 867: 
	case 868: 
	case 869: 
	case 870: 
	case 871: 
	case 872: 
	case 873: 
	case 874: 
	case 875: 
	case 876: 
	case 877: 
	case 878: 
	case 879: 
	case 880: 
	case 881: 
	case 882: 
	case 883: 
	case 884: 
	case 885: 
	case 886: 
	case 887: 
	case 888: 
	case 889: 
	case 890: 
	case 891: 
	case 892: 
	case 893: 
	case 894: 
	case 895: 
	case 896: 
	case 897: 
	case 898: 
	case 899: 
	case 900: 
	case 901: 
	case 902: 
	case 903: 
	case 904: 
	case 905: 
	case 906: 
	case 907: 
	case 908: 
	case 909: 
	case 910: 
	case 911: 
	case 912: 
	case 913: 
	case 914: 
	case 915: 
	case 916: 
	case 917: 
	case 918: 
	case 919: 
	case 920: 
	case 921: 
	case 922: 
	case 923: 
	case 924: 
	case 925: 
	case 926: 
	case 927: 
	case 928: 
	case 929: 
	case 930: 
	case 931: 
	case 932: 
	case 933: 
	case 934: 
	case 935: 
	case 936: 
	case 937: 
	case 938: 
	case 939: 
	case 940: 
	case 941: 
	case 942: 
	case 943: 
	case 944: 
	case 945: 
	case 946: 
	case 947: 
	case 948: 
	case 949: 
	case 950: 
	case 951: 
	case 952: 
	case 953: 
	case 954: 
	case 955: 
	case 956: 
	case 957: 
	case 958: 
	case 959: 
	case 960: 
	case 961: 
	case 962: 
	case 963: 
	case 964: 
	case 965: 
	case 966: 
	case 967: 
	case 968: 
	case 969: 
	case 970: 
	case 971: 
	case 972: 
	case 973: 
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 660: 
	case 661: 
	case 662: 
	case 663: 
	case 664: 
	case 665: 
	case 666: 
	case 667: 
	case 668: 
	case 669: 
	case 670: 
	case 671: 
	case 672: 
	case 673: 
	case 674: 
	case 675: 
	case 676: 
	case 677: 
	case 678: 
	case 679: 
	case 680: 
	case 681: 
	case 682: 
	case 683: 
	case 684: 
	case 685: 
	case 686: 
	case 687: 
	case 688: 
	case 689: 
	case 690: 
	case 691: 
	case 692: 
	case 693: 
	case 694: 
	case 695: 
	case 696: 
	case 697: 
	case 698: 
	case 699: 
	case 700: 
	case 701: 
	case 702: 
	case 703: 
	case 704: 
	case 705: 
	case 706: 
	case 707: 
	case 708: 
	case 709: 
	case 710: 
	case 711: 
	case 712: 
	case 713: 
	case 714: 
	{
		WARN(ZS_BAD_LOC_DATA);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 6: 
	case 11: 
	case 12: 
	case 23: 
	case 166: 
	case 169: 
	case 185: 
	case 217: 
	case 219: 
	case 229: 
	case 230: 
	case 232: 
	case 233: 
	case 235: 
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	break;
	case 15: 
	case 199: 
	case 250: 
	case 252: 
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	break;
	case 4: 
	case 22: 
	case 24: 
	case 27: 
	case 31: 
	case 36: 
	case 40: 
	case 44: 
	case 45: 
	case 50: 
	case 52: 
	case 57: 
	case 68: 
	case 71: 
	case 72: 
	case 75: 
	case 77: 
	case 79: 
	case 80: 
	case 85: 
	case 86: 
	case 91: 
	case 93: 
	case 94: 
	case 96: 
	case 97: 
	case 102: 
	case 105: 
	case 107: 
	case 111: 
	case 112: 
	case 115: 
	case 117: 
	case 119: 
	case 123: 
	case 127: 
	case 129: 
	case 222: 
	case 226: 
	case 228: 
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 196: 
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	break;
	case 163: 
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 792: 
	case 793: 
	case 806: 
	case 807: 
	case 813: 
	case 814: 
	case 820: 
	case 821: 
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 974: 
	case 975: 
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 1: 
	case 135: 
	case 167: 
	case 198: 
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	break;
	case 133: 
	case 164: 
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 282: 
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st253;}
	}
	break;
	case 614: 
	case 616: 
	case 618: 
	case 620: 
	case 621: 
	case 622: 
	case 623: 
	case 625: 
	case 626: 
	case 627: 
	case 629: 
	case 630: 
	case 631: 
	case 633: 
	case 634: 
	case 635: 
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 2: 
	case 139: 
	case 140: 
	case 141: 
	case 142: 
	case 159: 
	case 160: 
	case 161: 
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	break;
	case 266: 
	case 267: 
	case 268: 
	case 269: 
	case 272: 
	case 273: 
	case 274: 
	case 275: 
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_TEXT_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_TEXT);
		p--; {goto st253;}
	}
	break;
	case 757: 
	case 758: 
	case 763: 
	case 764: 
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_APL);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 808: 
	case 815: 
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 451: 
	case 452: 
	case 453: 
	case 454: 
	case 460: 
	case 461: 
	case 462: 
	case 463: 
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_HEX_RDATA);
		p--; {goto st253;}
	}
	break;
	case 1058: 
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_CHAR_COLON);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 795: 
	case 823: 
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_GATEWAY_KEY);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 1067: 
	case 1070: 
	case 1073: 
	case 1076: 
	{
		WARN(ZS_BAD_CHAR_DASH);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 1063: 
	{
		WARN(ZS_BAD_CHAR_COLON);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_HEX_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 186: 
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	{
		ERR(ZS_BAD_DIRECTIVE);
		p--; {goto st253;}
	}
	break;
	case 165: 
	case 172: 
	case 173: 
	case 174: 
	case 175: 
	case 176: 
	case 177: 
	case 178: 
	case 179: 
	case 180: 
	case 181: 
	case 182: 
	case 183: 
	case 184: 
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	break;
	case 156: 
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	break;
	case 200: 
	case 202: 
	case 203: 
	case 204: 
	case 205: 
	case 206: 
	case 207: 
	case 208: 
	case 209: 
	case 210: 
	case 211: 
	case 212: 
	case 213: 
	case 214: 
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	break;
	case 13: 
	case 171: 
	case 201: 
	case 216: 
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	break;
	case 278: 
	case 281: 
	case 283: 
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	{
		ERR(ZS_BAD_TTL);
		p--; {goto st253;}
	}
	break;
	case 809: 
	case 816: 
	{
		WARN(ZS_BAD_ADDRESS_CHAR);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_GATEWAY);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_GATEWAY_KEY);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	break;
	case 14: 
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	break;
	case 197: 
	{
		s->r_owner_length = 0;
		WARN(ZS_BAD_OWNER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_TIME_UNIT);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_NUMBER);
		p--; {goto st253;}
	}
	{
		WARN(ZS_UNSUPPORTED_TYPE);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_RDATA);
		p--; {goto st253;}
	}
	{
		WARN(ZS_BAD_REST);
		p--; {goto st253;}
	}
	break;
	}
	}

	_out: {}
	}


	// Check if scanner state machine is in uncovered state.
	if (cs == zone_scanner_error) {
		ERR(ZS_UNCOVERED_STATE);
		s->error_counter++;

		// Fill error context data.
		for (s->buffer_length = 0;
		     ((p + s->buffer_length) < pe) &&
		     (s->buffer_length < sizeof(s->buffer) - 1);
		     s->buffer_length++)
		{
			// Only rest of the current line.
			if (*(p + s->buffer_length) == '\n') {
				break;
			}
			s->buffer[s->buffer_length] = *(p + s->buffer_length);
		}

		// Ending string in buffer.
		s->buffer[s->buffer_length++] = 0;

		// Processing error.
		s->process_error(s);

		return;
	}

	// Check unclosed multiline record.
	if (is_eof && s->multiline) {
		ERR(ZS_UNCLOSED_MULTILINE);
		s->error_counter++;
		s->process_error(s);
	}

	// Storing scanner states.
	s->cs  = cs;
	s->top = top;
	memcpy(s->stack, stack, sizeof(stack));

	// Storing r_data pointer.
	s->r_data_tail = rdata_tail - s->r_data;
}

int zs_scanner_parse(zs_scanner_t *s,
                     const char   *start,
                     const char   *end,
                     const bool   final_block)
{
	if (s == NULL || start == NULL || end == NULL) {
		return -1;
	}

	// Parse input block.
	parse_block(s, start, end, false);

	// Parse trailing artificial block (newline char) if not stop.
	if (final_block && !s->stop) {
		parse_block(s, NEWLINE_BLOCK, NEWLINE_BLOCK + 1, true);
	}

	// Check if any errors has occured.
	if (s->error_counter > 0) {
		return -1;
	}

	return 0;
}

int zs_scanner_parse_file(zs_scanner_t *s,
                          const char   *file_name)
{
	long		page_size;
	uint64_t	n_blocks;
	uint64_t	block_id;
	uint64_t	default_block_size;
	struct stat	file_stat;

	if (s == NULL || file_name == NULL) {
		return -1;
	}

	// Getting OS page size.
	page_size = sysconf(_SC_PAGESIZE);
	if (page_size <= 0) {
		ERR(ZS_SYSTEM);
		return -1;
	}

	// Copying file name.
	s->file.name = strdup(file_name);

	// Opening the zone file.
	s->file.descriptor = open(file_name, O_RDONLY);
	if (s->file.descriptor == -1) {
		ERR(ZS_FILE_OPEN);
		free(s->file.name);
		return -1;
	}

	// Get absolute path of the zone file.
	char *full_name = realpath(file_name, NULL);
	if (full_name != NULL) {
		free(s->path);
		s->path = strdup(dirname(full_name));
		free(full_name);
	} else {
		ERR(ZS_FILE_PATH);
		close(s->file.descriptor);
		free(s->file.name);
		return -1;
	}

	// Getting file information.
	if (fstat(s->file.descriptor, &file_stat) == -1) {
		ERR(ZS_FILE_FSTAT);
		close(s->file.descriptor);
		free(s->file.name);
		return -1;
	}

	// Check for directory.
	if (S_ISDIR(file_stat.st_mode)) {
		ERR(ZS_FILE_DIR);
		close(s->file.descriptor);
		free(s->file.name);
		return -1;
	}

	// Check for empty file.
	if (file_stat.st_size == 0) {
		close(s->file.descriptor);
		free(s->file.name);
		return 0;
	}

	// Block size adjustment to multiple of page size.
	default_block_size = (BLOCK_SIZE / page_size) * page_size;

	// Number of blocks which cover the whole file (ceiling operation).
	n_blocks = 1 + ((file_stat.st_size - 1) / default_block_size);

	// Loop over zone file blocks.
	for (block_id = 0; block_id < n_blocks; block_id++) {
		// Current block start to scan.
		uint64_t scanner_start = block_id * default_block_size;
		// Current block size to scan.
		uint64_t block_size = default_block_size;
		// Last data block mark.
		bool final_block = false;
		// Mmapped data.
		char *data;

		// The last block is probably shorter.
		if (block_id == (n_blocks - 1)) {
			block_size = file_stat.st_size - scanner_start;
			final_block = true;
		}

		// Zone file block mapping.
		data = mmap(0, block_size, PROT_READ, MAP_SHARED,
		            s->file.descriptor, scanner_start);
		if (data == MAP_FAILED) {
			ERR(ZS_FILE_MMAP);
			close(s->file.descriptor);
			free(s->file.name);
			return -1;
		}

		// Scan zone file block.
		(void)zs_scanner_parse(s, data, data + block_size, final_block);

		// Zone file block unmapping.
		if (munmap(data, block_size) == -1) {
			ERR(ZS_FILE_MMAP);
			close(s->file.descriptor);
			free(s->file.name);
			return -1;
		}

		// Stop parsing if required.
		if (s->stop) {
			break;
		}
	}

	close(s->file.descriptor);
	free(s->file.name);

	// Check for scanner return.
	if (s->error_counter > 0) {
		return -1;
	}

	return 0;
}
